/*
 * lib-1.0.js:
 */

var DataToJs, JsUtil, Log, XLib, XApplication, XIconLink, XPathResult, XRadioButton;
var XCheckBox, XDate, XFlow, XHorFlow, XVerFlow, XTable, XTree, XTreeRow;
var MWForm, MWPlugin, UILib, XUILib;
var Intl, InstallTrigger, WebSocket, MW, HtmlWriter;

var MagicW = function() {

	/* ================================================================	*
	 * XElement.js 														*
	 * ================================================================	*/

	var XElement = function(data) {

		// addChild ---------------------------------------------------
		XElement.prototype.addChild = function(tag) {
			var child;

			if (this.data._children === undefined)
				this.data._children = [];

			child = {tag: tag};

			this.data._children.push(child);

			return new XElement(child);
		};
		
		// getAttrString ----------------------------------------------
		XElement.prototype.getAttrString = function(attr, def) {
			var s;

			if ((s = this.data[attr]) === undefined)
				return def;

			return s;
		};

		// getChildren ------------------------------------------------
		XElement.prototype.getChildren = function(attr) {
			var i, a, list;

			list = [];

			if ((a = this.data._children) === undefined)
				return list;

			for(i = 0; i < a.length; ++i)
				if (a[i].tag === attr)
					list.push(new XWData(a[i]));

			return list;
		};
	/*
		// getString --------------------------------------------------
		XElement.prototype.getString = function(attr) {
			var v;

			if ((v = this.data[attr]) !== undefined)
				return v;

			throw new Error('attribute \'' + attr + '\' not found');
		};

		// setIntAttr -------------------------------------------------
		XElement.prototype.setIntAttr = function(attr, value) {

			this.data[attr] = value.toString();
		};
	*/
		// setStringAttr ----------------------------------------------
		XElement.prototype.setStringAttr = function(attr, value) {

			this.data[attr] = value;
		};

		// toUIXWData -------------------------------------------------
		XElement.prototype.toUIXWData = function(xappl) {

			return DataToJs.transformFromXElement(this.data, false);
		};

		// Constructor ------------------------------------------------
		{
			this.data = data;
		}
	};

	// getInstanceFromXData -----------------------------------------------
	XElement.getInstanceFromXData = function(xd) {

		return new XElement(xd.data);
	};

	// getInstanceFromXWData ----------------------------------------------
	XElement.getInstanceFromXWData = function(xwd) {

		return new XElement(xwd.data);
	};

	// newInstance --------------------------------------------------------
	XElement.newInstance = function(tag) {

		return new XElement({tag: tag});
	};

	/* ================================================================	*
	 * XHMap.js 														*
	 * ================================================================	*/

	var XHMap = function() {

		// exists -----------------------------------------------------
		XHMap.prototype.exists = function(key) {

			return this.hmap[key] !== undefined;
		};

		// get --------------------------------------------------------
		XHMap.prototype.get = function(key) {
			var v;

			if ((v = this.hmap[key]) !== undefined)
				return v;

			return null;
		};
	/*
		// keys -------------------------------------------------------
		XMap.prototype.keys = function(a) {
			var i;

			i = 0;
			for(var key in this.hmap)
				a[i++] = key;

			return a;
		};
	*/
		// put --------------------------------------------------------
		XHMap.prototype.put = function(key, value) {

			this.hmap[key] = value;
		};
	/*
		// remove -----------------------------------------------------
		XMap.prototype.remove = function(id) {

			delete this.hmap[id];
		};

		// size -------------------------------------------------------
		XMap.prototype.size = function() {

			return this.hmap.length;
		};

		// values -----------------------------------------------------
		XMap.prototype.values = function(a) {
			var i;

			i = 0;
			for(var key in this.hmap)
				a[i++] = this.hmap[key];

			return a;
		};
	*/
		// Constructor ------------------------------------------------
		{
			this.hmap = {};
		}
	};

	// createXActionArray -------------------------------------------------
	XHMap.createXActionArray = function() {

		return new XHMap();
	};


	/* ================================================================	*
	 * XInteger.js 														*
	 * ================================================================	*/

	var XInteger = function() {
	};

	// floor --------------------------------------------------------------
	XInteger.floor = function(v) {

		return Math.floor(v);
	};

	// parseInt -----------------------------------------------------------
	XInteger.parseInt = function(s) {

		return parseInt(s);
	};

	// round --------------------------------------------------------------
	XInteger.round = function(v) {

		return Math.round(v);
	};

	// toPadding ----------------------------------------------------------
	XInteger.toPadding = function(n, padding) {
		var i, s, len;

		s = n.toString();
		len = s.length;

		for(i = len; i < padding; ++i)
			s = '0' + s;

		return s;
	};


	/* ================================================================	*
	 * XPluginLib.js 													*
	 * ================================================================	*/

	var XPluginLib = function() {
	};

	XPluginLib.fonts = [];
	XPluginLib.html = [];

	// create -------------------------------------------------------------
	XPluginLib.create = function(tag, data) {

		switch(tag) {
		case 'plugin-display':
			return new XDisplay(data);
		case 'plugin-slide':
			return new XSlide(data);
		case 'plugin-toc-edit':
			return new XTocEdit(data);
		default:
			return null;
		}
	};

	// addFont ------------------------------------------------------------
	XPluginLib.addFont = function(font) {
	};

	// setFont ------------------------------------------------------------
	XPluginLib.setFont = function(xd, alias) {
		var id;

		if ((id = XPluginLib.fonts[alias]) === undefined)
			throw new Error('font not found: ' + alias);

		xd.setStringAttr('fontid', id);
	};

	// setFonts -----------------------------------------------------------
	XPluginLib.setFonts = function(data) {

		if (data.tag !== 'fonts')
			throw new Error('invalid data: ' + data.tag);

		for(var entry of data._children)
			XPluginLib.fonts[entry.alias] = entry.fontid;
	};

	// setHtml ------------------------------------------------------------
	XPluginLib.setHtml = function(data) {

		for(var entry of data)
			XPluginLib.html[entry.id] = entry.html;
	};

	/* ================================================================	*
	 * XString.js 														*
	 * ================================================================	*/

	var XString = function() {
	};

	// equals -------------------------------------------------------------
	XString.equals = function(s1, s2) {

		return s1 === s2;
	};

	// size ---------------------------------------------------------------
	XString.size = function(s) {

		return s.length;
	};

	/* ================================================================	*
	 * XStringBuilder.js 												*
	 * ================================================================	*/

	var XStringBuilder = function() {

		// addChild ---------------------------------------------------
		XStringBuilder.prototype.append = function(text) {

			this.list.push(text);
		};

		// getText ----------------------------------------------------
		XStringBuilder.prototype.getText = function() {

			return this.list.join('');
		};

		// Constructor ------------------------------------------------
		{
			this.list = []; 
		}
	};

	/* ================================================================	*
	 * XUtil.js 														*
	 * ================================================================	*/

	var XUtil = function() {
	};

	// convertTextToLabel -------------------------------------------------
	XUtil.convertTextToLabel = function(paragraphs) {
		var sb;

		sb = [];
		for(var p of paragraphs) {
			if (p !== paragraphs[0])
				sb.push('<br>');

			sb.push(XUtil.escapeXml(p.getTextContent()));
		}

		return sb.join('');
	};

	// escapeXml ----------------------------------------------------------
	XUtil.escapeXml = function(text) {

		return text.replace(/[<>&'"]/g,
			function (c) {
				switch (c) {
	            case '<':
					return '&lt;';
	            case '>':
					return '&gt;';
	            case '&':
					return '&amp;';
	            case '\'':
					return '&apos;';
	            case '"':
					return '&quot;';
	        	}
	    	}
		);
	};

	// getLang ------------------------------------------------------------
	XUtil.getLang = function() {
		var n1, n2, filename, lang, params, pathname, prefix;

		//lang = navigator.language;
		lang = 'fr';

		pathname = window.location.pathname;

		if ((n1 = pathname.lastIndexOf('/')) === -1)
			return 'fr';
		prefix = pathname.substring(0, n1 + 1);

		filename = pathname.substring(n1);

		if (filename === '/course') {
			params = new URLSearchParams(window.location.search);
			if ((lang = params.get('lang')) !== null)
				return lang;

			return 'fr';
		}

		if ((n1 = filename.indexOf('-')) === -1)
			return lang;

		if ((n2 = filename.lastIndexOf('.')) === -1)
			return lang;

		if (n2 - n1 !== 3)
			return lang;

		return filename.substring(n1 + 1, n2);
	};

	// getLocalTimeZone ---------------------------------------------------
	XUtil.getLocalTimeZone = function() {

	 	return Intl.DateTimeFormat().resolvedOptions().timeZone;
	};

	// removeAllChildren --------------------------------------------------
	XUtil.removeAllChildren = function(parent) {

		while(parent.firstChild !== null)
			parent.removeChild(parent.firstChild);
	};

	// getUserAgent -------------------------------------------------------
	XUtil.getUserAgent = function() {

		return navigator.userAgent;
	};

	// getVersion ---------------------------------------------------------
	XUtil.getVersion = function() {

		return '1.5.5';
	};

	/* ================================================================	*
	 * DocLib.js 														*
	 * ================================================================	*/

	var DocLib = new function() {
	};

	// formGetValue ---------------------------------------------------
	DocLib.formGetValue = function(type, formId, fieldId, itemId) {
		var k, el, wt, exp, root;

		k = formId + '_' + fieldId;

		switch(type) {
		case 'date':
			root = MW.getCard(formId);

			if ((wt = XLib.search(root, k)) === null)
				throw new Error('widget not found: ' + k);

			return wt.uiwt.uidate.getValue();
		case 'explorer':
			root = MW.getCard(formId);

			switch(itemId) {
			case 'tree':
				k += '_tree_view';

				if ((wt = XLib.search(root, k)) === null)
					throw new Error('widget not found: ' + k);

				return wt.getSelectedValue();
			}

			throw new Error('not yet implemented: ' + itemId);

		
		case 'tree':
			root = MW.getCard(formId);

			if ((wt = XLib.search(root, k)) === null)
				throw new Error('widget not found: ' + k);

			return wt.view.getSelectedValue();
		case 'radiobutton':
			exp = 'input[type=radio][name="' + k + '"]:checked';

			if ((el = document.querySelector(exp)) === null)
				return '';

			return el.value;
		}

		el = document.getElementById(k + '_input');

		switch(type) {
		case 'combobox':
			return el.options[el.selectedIndex].value;
		case 'password':
		case 'textarea':
		case 'textfield':
			return el.value;
		default:
			throw new Error('not yet implemented: ' + type + ' (' + k + ')');
		}
	};

	// formGetValue ---------------------------------------------------
	DocLib.formGetValues = function(type, id, fieldId) {
		var i, k, l, r, exp;

		k = id + '_' + fieldId;

		switch(type) {
		case 'checkbox':
			exp = 'input[type=checkbox][name="' + k + '"]:checked';
			l = document.querySelectorAll(exp);

			r = [];
			for(i = 0; i < l.length; ++i)
				r.push(l[i].value);
			
			return r;
		default:
			throw new Error('not yet implemented: ' + type);
		}
	};

	// formSetValue ---------------------------------------------------
	DocLib.formSetValue = function(id, type, value) {
		var a, el;

		if ((el = document.getElementById(id)) === null)
			throw new Error('element not found: ' + id);

		switch(type) {
		case 'checkbox':
			el.checked = true;
			break;
		case 'combobox':
			el.value = value;
			break;
		case 'date':
			a = value.split(':');
			DocLib.formSetValue(id + '_year_input', "combobox", a[0]);
			DocLib.formSetValue(id + '_month_input', "combobox", a[1]);
			DocLib.formSetValue(id + '_day_input', "combobox", a[2]);
			DocLib.formSetValue(id + '_hour_input', "combobox", a[3]);
			DocLib.formSetValue(id + '_min_input', "combobox", a[4]);
			DocLib.formSetValue(id + '_tz_input', "combobox", a[5]);
			break;
		case 'label':
			el.innerHTML = value;
			break;
		case 'password':
		case 'textarea':
		case 'textfield':
			el.value = value;
			break;
		case 'radiobutton':
			el.checked = true;
			break;
		default:
			throw new Error('not yet supported: ' + type);
		}
	};

	// getCookie ------------------------------------------------------
	DocLib.getCookie = function(name) {
		var i, n1, n2, n3, n4, value;
		var nloop;

		value = document.cookie;

		nloop = 0;
		i = 0;
		for(;;) {
			if (++nloop === 200)
				throw 'too many cookies not found...';

			n1 = i;
			for(; i < value.length; ++i)
				if (value.charAt(i) === '=')
					break;
			n2 = i;

			if (++i >= value.length)
				return '';

			if (value.charAt(i) === '"') {
				n3 = ++i;
				for(; i < value.length; ++i)
					if (value.charAt(i) === '"')
						break;
				n4 = i;
				++i;
			} else {
				n3 = i;
				for(; i < value.length; ++i)
					if (value.charAt(i) === ';')
						break;
				n4 = i;
			}

			if (value.substring(n1, n2) === name)
				return value.substring(n3, n4);

			for(++i; i < value.length; ++i)
				if (value.charAt(i) !== ' ')
					break;
		}
	};

	// getElement -----------------------------------------------------
	DocLib.getElement = function(id) {

		return document.getElementById(id);
	};

	// hide -----------------------------------------------------------
	DocLib.hide = function(id) {
		var el;

		el = document.getElementById(id);
		el.style.display = 'none';
	};

	// show -----------------------------------------------------------
	DocLib.show = function(id) {
		var el;

		if ((el = document.getElementById(id)) === null)
			throw new Error('card not found: ' + id);

		el.style.display = 'block';
	};

	/* ================================================================	*
	 * HttpClient.js 													*
	 * ================================================================	*/

	var HttpClient = function(url, appl) {

		// uploadFiles ------------------------------------------------
		HttpClient.prototype.uploadFiles = function(requestId, nextTask, data) {
			var r, res, root, xhttp, xml;

			appl.taskCardShow('lecture-upload');
			appl.waitingStart();

			XApplication.logMsg('http-post: uploadFiles [requestId=' +
										requestId + ',nextTask=' + nextTask + ']');

			xhttp = new XMLHttpRequest();

			xhttp.onreadystatechange = function() {

				if (xhttp.readyState !== 4)				// 4 DONE ...
					return;

				if (xhttp.status !== 200) {
					XApplication.logMsg('http-post: uploadFiles [staus=' + xhttp.status + ']');
					appl.taskNotifyHttpUploadError(requestId + '#http-error');
					return;
				}

				xml = xhttp.responseXML;
				root = xml.documentElement;
				data = new XmlData(xml, root);

				appl.taskNotifyHttpUploadDone(nextTask, data);
			};

			xhttp.open('POST', this.url, true);

			res = {
				bar: DocLib.getElement('uploading-bar'),
				me: this
			};

			this.repaintProgressBar(res.bar, 8, 16);

			xhttp.upload.addEventListener("progress", function(e) {

				res.me.repaintProgressBar(res.bar, e.loaded, e.total);
			});

			xhttp.send(data);
		};

		// post -----------------------------------------------------------
		HttpClient.prototype.post = function(xdoc, callback) {
			var data, id, xhttp, xml;

			id = xdoc.getString('id');
			XApplication.logXml('http-post: ' + id, xdoc);

			xhttp = new XMLHttpRequest();

			xhttp.onreadystatechange = function() {

				if (xhttp.readyState !== 4)				// 4 DONE ...
					return;

				if (xhttp.status !== 200) {
					Log.info('HttpClient.error.requestId: ' + id);
					Log.info(xhttp.responseText, false);
					return;
				}

				xml = xhttp.responseXML;
				data = new XmlData(xml, xml.documentElement);
				XApplication.logXml("http-response: " + id, data);

				if (callback !== null)
					callback(data);
				return;
			};

			xhttp.open('POST', this.url, true);
			xhttp.setRequestHeader('Content-Type', 'application/xml');
			xhttp.send(xdoc.getXml());
		};

		// repaintProgressBar -----------------------------------------
		HttpClient.prototype.repaintProgressBar = function(canvas, nleft, total) {
			var w, h, ctx, left, right;

			ctx = canvas.getContext('2d');
			w = ctx.canvas.width;
			h = ctx.canvas.height;

			left = w * nleft / total;
			right = w - left;

			if (left > 0) {
				ctx.fillStyle = 'teal';
				ctx.fillRect(0, 0, left, h);
			}

			if (right > 0) {
				ctx.fillStyle = '#000000';
				ctx.fillRect(w - right, 0, right, h);
			}
		};

		// xmlPost ----------------------------------------------------
		HttpClient.prototype.xmlPost = function(req, xdoc, reqs) {
			var doc, task, text, xhttp, xml;

			xhttp = new XMLHttpRequest();

			xhttp.onreadystatechange = function() {

				if (xhttp.readyState !== 4)				// 4 DONE ...
					return;

				if (xhttp.status !== 200) {
					text = xhttp.responseText;

					Log.info('HttpClient.error.requestId: ' + req.requestId);
					Log.info(text, false);

					if (xhttp.status === 401 && text.search('<magicw msg="session timeout" ') !== -1) {
						appl.taskNotifyHttpSessionTimeout();
						return;
					}

					appl.taskNotifyHttpRequestError(req.requestId, reqs);
					return;
				}

				doc = xhttp.responseXML;
				xml = new XmlData(doc, doc.documentElement);

				task = (req.taskId !== null) ? req.taskId : req.requestId;
				appl.taskNotifyHttpRequestDone(task, xml, reqs);
			};

			xhttp.open('POST', this.url, true);
			xhttp.setRequestHeader('Content-Type', 'application/xml');
			xhttp.send(xdoc.getXml());
		};

		// Constructor ------------------------------------------------
		{
			this.url = url;
			this.uiappl = appl;
		}
	};

	/* ================================================================	*
	 * HttpError.js 													*
	 * ================================================================	*/

	var HttpError = function() {
	};

	// send -----------------------------------------------------------
	HttpError.send = function(error) {
		var id, msg, url, xdoc, xhttp, xml;

		id = 'javascript-error';
		if ((url = MW.xappl.getUrl()) === null) {
			console.log(error);
			return;
		}
		
		url += '?magicw-xml';

		xdoc = UILib.newXmlData('magicw');
		xdoc.setStringAttr('id', id);

		if (MW.isFirefox)
			msg = "Error: " + error.message + '\n' + error.stack;
		else
		if (MW.isChrome)
			msg = (error.stack === undefined) ? error : error.stack.toString();
		else
			msg = error.toString();

		xdoc.addChildWithContent('msg', msg);
		XApplication.logXml('http-request: ' + id, xdoc);

		xhttp = new XMLHttpRequest();

		xhttp.onreadystatechange = function() {

			if (xhttp.readyState !== 4)				// 4 DONE ...
				return;

			if (xhttp.status !== 200) {
				Log.info('HttpClient.error.requestId: ' + id);
				Log.info(xhttp.responseText, false);
				return;
			}

			xml = xhttp.responseXML;
			xdoc = new XmlData(xml, xml.documentElement);
			XApplication.logXml("http-response: " + id, xdoc);
			return;
		};

		xhttp.open('POST', url, true);
		xhttp.setRequestHeader('Content-Type', 'application/xml');
		xhttp.send(xdoc.getXml());
	};

	/* ================================================================	*
	 * Log.js 															*
	 * ================================================================	*/

	var Log = function() {
	};

	Log.levels = [];

	// info -----------------------------------------------------------
	Log.info = function(msg, ctxt) {

		// display --------------------------------------------------------
		function display(msg, ctxt) {

			function extractWhereFromChrome(line) {
				var n1, n2, method, where;

				if ((n1 = line.indexOf(' at ')) === -1)
					return null;
				n1 += 4;

				if ((n2 = line.indexOf(' ', n1)) === -1)
					return null;

				method = line.substring(n1, n2);

				if ((n1 = line.lastIndexOf('/')) === -1)
					return null;
				++n1;

				if ((n2 = line.lastIndexOf(':')) === -1)
					return null;

				where = line.substring(n1, n2);

				return '[' + where + ':' + method + '()]';
			}

			function extractWhereFromMozilla(line) {
				var n1, n2, method, where;

				if ((n2 = line.lastIndexOf(':')) === -1)
					return null;

				if ((n1 = line.lastIndexOf('/', n2)) === -1)
					return null;

				where = line.substring(n1 + 1, n2);

				if ((n1 = line.indexOf('/')) === -1)
					return null;

				if ((n1 = line.indexOf('/', n1 + 1)) === -1)
					return null;

				if ((n2 = line.indexOf('@', ++n1)) === -1)
					return null;

				method = line.substring(n1, n2);

				return '[ ' + where + ':' + method + '()]';
			}

			// display ---
			var a, e, s;

			if (!ctxt) {
				console.log(msg);
				return;
			}

			e = new Error();

			if (navigator.userAgent.indexOf('Chrome') !== -1) {
				a = e.stack.toString().split("\n");
				if ((s = extractWhereFromChrome(a[3])) !== null)
					msg += ' ' + s;
			} else
			if (navigator.userAgent.indexOf('Mozilla') !== -1) {
				a = e.stack.toString().split("\n");
				if ((s = extractWhereFromMozilla(a[2])) !== null)
					msg += ' ' + s;
			}

			console.log(msg);
		};

		if (ctxt === undefined)
			ctxt = true;

		if (Log.checkLevel('info'))
			display(msg, ctxt);
	};
		
	// info -----------------------------------------------------------
	Log.checkLevel = function(level) {
		var levels;

		if ((levels = Log.levels) === null)
			return false;

		for(var i = 0; i < levels.length; ++i)
			if (level === levels[i])
				return true;

		return false;
	};

	// setLevels ------------------------------------------------------
	Log.setLevels = function() {

		Log.levels = [];

		for(var i = 0; i < arguments.length; ++i)
			Log.levels.push(arguments[i]);
	};

	/* ================================================================	*
	 * MColor.js 														*
	 * ================================================================	*/

	var MColor = new function() {
	};

	// MColor.getActiveColor ------------------------------------------
	MColor.getActiveColor = function(v) {
		var r, g, b;

		r = (v >> 16) & 0xFF;
		g = (v >> 8) & 0xFF;
		b = v & 0xFF;

		r = Math.round(r * 8 / 10);
		g = Math.round(g * 8 / 10);
		b = Math.round(b * 8 / 10);

		return MColor.toJS((r << 16) + (g << 8) + b);
	};

	// MColor.toRGB ---------------------------------------------------
	MColor.getOverColor = function(v) {
		var r, g, b;

		r = (v >> 16) & 0xFF;
		g = (v >> 8) & 0xFF;
		b = v & 0xFF;

		if (r === 0)
			r = 40;
		else
		if ((r = Math.round(r * 10 / 8)) > 255)
			r = 255;

		if (g === 0)
			g = 40;
		else
		if ((g = Math.round(g * 10 / 8)) > 255)
			g = 255;

		if (b === 0)
			b = 40;
		else
		if ((b = Math.round(b * 10 / 8)) > 255)
			b = 255;

		return MColor.toJS((r << 16) + (g << 8) + b);
	};

	// MColor.toJS ----------------------------------------------------
	MColor.toJS = function(v) {
		var s, len;

		s = v.toString(16);
		len = s.length;

		switch(len) {
		case 6:
			return '#' + s;
		case 5:
			return '#0' + s;
		case 4:
			return '#00' + s;
		case 3:
			return '#000' + s;
		case 2:
			return '#0000' + s;
		case 1:
			return '#00000' + s;
		default:
			throw new Error('invalid value: ' + v);
		}
	};

	// MColor.toRGB ---------------------------------------------------
	MColor.toRGB = function(colorname) {

		if (colorname.charAt(0) === '#')
			return(parseInt('0x' + colorname.substring(1)));

		switch(colorname) {
		case 'aliceblue':
			return 0xF0F8FF;
		case 'antiquewhite':
			return 0xFAEBD7;
		case 'aqua':
			return 0xFFFF;
		case 'aquamarine':
			return 0x7FFFD4;
		case 'azure':
			return 0xF0FFFF;
		case 'beige':
			return 0xF5F5DC;
		case 'bisque':
			return 0xFFE4C4;
		case 'black':
			return 0x0;
		case 'blanchedalmond':
			return 0xFFEBCD;
		case 'blue':
			return 0xFF;
		case 'blueviolet':
			return 0x8A2BE2;
		case 'brown':
			return 0xA52A2A;
		case 'burlywood':
			return 0xDEB887;
		case 'cadetblue':
			return 0x5F9EA0;
		case 'chartreuse':
			return 0x7FFF00;
		case 'chocolate':
			return 0xD2691E;
		case 'coral':
			return 0xFF7F50;
		case 'cornflowerblue':
			return 0x6495ED;
		case 'cornsilk':
			return 0xFFF8DC;
		case 'crimson':
			return 0xDC143C;
		case 'cyan':
			return 0xFFFF;
		case 'darkblue':
			return 0x8B;
		case 'darkcyan':
			return 0x8B8B;
		case 'darkgoldenrod':
			return 0xB8860B;
		case 'darkgray':
			return 0xA9A9A9;
		case 'darkgreen':
			return 0x6400;
		case 'darkkhaki':
			return 0xBDB76B;
		case 'darkmagenta':
			return 0x8B008B;
		case 'darkolivegreen':
			return 0x556B2F;
		case 'darkorange':
			return 0xFF8C00;
		case 'darkorchid':
			return 0x9932CC;
		case 'darkred':
			return 0x8B0000;
		case 'darksalmon':
			return 0xE9967A;
		case 'darkseagreen':
			return 0x8FBC8F;
		case 'darkslateblue':
			return 0x483D8B;
		case 'darkslategray':
			return 0x2F4F4F;
		case 'darkturquoise':
			return 0xCED1;
		case 'darkviolet':
			return 0x9400D3;
		case 'deeppink':
			return 0xFF1493;
		case 'deepskyblue':
			return 0xBFFF;
		case 'dimgray':
			return 0x696969;
		case 'dodgerblue':
			return 0x1E90FF;
		case 'firebrick':
			return 0xB22222;
		case 'floralwhite':
			return 0xFFFAF0;
		case 'forestgreen':
			return 0x228B22;
		case 'fuchsia':
			return 0xFF00FF;
		case 'gainsboro':
			return 0xDCDCDC;
		case 'ghostwhite':
			return 0xF8F8FF;
		case 'gold':
			return 0xFFD700;
		case 'goldenrod':
			return 0xDAA520;
		case 'gray':
			return 0x808080;
		case 'green':
			return 0x8000;
		case 'greenyellow':
			return 0xADFF2F;
		case 'honeydew':
			return 0xF0FFF0;
		case 'hotpink':
			return 0xFF69B4;
		case 'indianred':
			return 0xCD5C5C;
		case 'indigo':
			return 0x4B0082;
		case 'ivory':
			return 0xFFFFF0;
		case 'khaki':
			return 0xF0E68C;
		case 'lavender':
			return 0xE6E6FA;
		case 'lavenderblush':
			return 0xFFF0F5;
		case 'lawngreen':
			return 0x7CFC00;
		case 'lemonchiffon':
			return 0xFFFACD;
		case 'lightblue':
			return 0xADD8E6;
		case 'lightcoral':
			return 0xF08080;
		case 'lightcyan':
			return 0xE0FFFF;
		case 'lightgoldenrodyellow':
			return 0xFAFAD2;
		case 'lightgray':
			return 0xD3D3D3;
		case 'lightgreen':
			return 0x90EE90;
		case 'lightpink':
			return 0xFFB6C1;
		case 'lightsalmon':
			return 0xFFA07A;
		case 'lightseagreen':
			return 0x20B2AA;
		case 'lightskyblue':
			return 0x87CEFA;
		case 'lightslategray':
			return 0x778899;
		case 'lightsteelblue':
			return 0xB0C4DE;
		case 'lightyellow':
			return 0xFFFFE0;
		case 'lime':
			return 0xFF00;
		case 'limegreen':
			return 0x32CD32;
		case 'linen':
			return 0xFAF0E6;
		case 'magenta':
			return 0xFF00FF;
		case 'maroon':
			return 0x800000;
		case 'mediumaquamarine':
			return 0x66CDAA;
		case 'mediumblue':
			return 0xCD;
		case 'mediumorchid':
			return 0xBA55D3;
		case 'mediumpurple':
			return 0x9370DB;
		case 'mediumseagreen':
			return 0x3CB371;
		case 'mediumslateblue':
			return 0x7B68EE;
		case 'mediumspringgreen':
			return 0xFA9A;
		case 'mediumturquoise':
			return 0x48D1CC;
		case 'mediumvioletred':
			return 0xC71585;
		case 'midnightblue':
			return 0x191970;
		case 'mintcream':
			return 0xF5FFFA;
		case 'mistyrose':
			return 0xFFE4E1;
		case 'moccasin':
			return 0xFFE4B5;
		case 'navajowhite':
			return 0xFFDEAD;
		case 'navy':
			return 0x80;
		case 'oldlace':
			return 0xFDF5E6;
		case 'olive':
			return 0x808000;
		case 'olivedrab':
			return 0x6B8E23;
		case 'orange':
			return 0xFFA500;
		case 'orangered':
			return 0xFF4500;
		case 'orchid':
			return 0xDA70D6;
		case 'palegoldenrod':
			return 0xEEE8AA;
		case 'palegreen':
			return 0x98FB98;
		case 'paleturquoise':
			return 0xAFEEEE;
		case 'palevioletred':
			return 0xDB7093;
		case 'papayawhip':
			return 0xFFEFD5;
		case 'peachpuff':
			return 0xFFDAB9;
		case 'peru':
			return 0xCD853F;
		case 'pink':
			return 0xFFC0CB;
		case 'plum':
			return 0xDDA0DD;
		case 'powderblue':
			return 0xB0E0E6;
		case 'purple':
			return 0x800080;
		case 'red':
			return 0xFF0000;
		case 'rosybrown':
			return 0xBC8F8F;
		case 'royalblue':
			return 0x4169E1;
		case 'saddlebrown':
			return 0x8B4513;
		case 'salmon':
			return 0xFA8072;
		case 'sandybrown':
			return 0xF4A460;
		case 'seagreen':
			return 0x2E8B57;
		case 'seashell':
			return 0xFFF5EE;
		case 'sienna':
			return 0xA0522D;
		case 'silver':
			return 0xC0C0C0;
		case 'skyblue':
			return 0x87CEEB;
		case 'slateblue':
			return 0x6A5ACD;
		case 'slategray':
			return 0x708090;
		case 'snow':
			return 0xFFFAFA;
		case 'springgreen':
			return 0xFF7F;
		case 'steelblue':
			return 0x4682B4;
		case 'tan':
			return 0xD2B48C;
		case 'teal':
			return 0x8080;
		case 'thistle':
			return 0xD8BFD8;
		case 'tomato':
			return 0xFF6347;
		case 'turquoise':
			return 0x40E0D0;
		case 'violet':
			return 0xEE82EE;
		case 'wheat':
			return 0xF5DEB3;
		case 'white':
			return 0xFFFFFF;
		case 'whitesmoke':
			return 0xF5F5F5;
		case 'yellow':
			return 0xFFFF00;
		case 'yellowgreen':
			return 0x9ACD32;
		default:
			throw new Error('unknown colorname: ' + colorname);
		}
	};

	/* ================================================================	*
	 * XmlData.js 														*
	 * ================================================================	*/

	var XmlData = function(doc, cur) {

		// addChild ---------------------------------------------------
		XmlData.prototype.addChild = function(tag) {
			var el;

			el = this.doc.createElement(tag);

			return new XmlData(this.doc, el);
		};

		// addChildWithContent ----------------------------------------
		XmlData.prototype.addChildWithContent = function(tag, text) {
			var el, node;

			el = this.doc.createElement(tag);
			this.cur.appendChild(el);

			node = this.doc.createTextNode(text);
			el.appendChild(node);
		};

		// getAllChildren ---------------------------------------------
		XmlData.prototype.getAllChildren = function() {
			var i, a, children;

			a = [];

			children = this.cur.children;
			for(i = 0; i < children.length; ++i)
				a.push(new XmlData(this.doc, children[i]));

			return(a);
		};

		// getAttrBool ------------------------------------------------
		XmlData.prototype.getAttrBool = function(attr, def) {
			var v;

			if ((v = this.cur.getAttribute(attr)) === null)
				return def;

			return v === 'true';
		};

		// getAttrInt -------------------------------------------------
		XmlData.prototype.getAttrInt = function(attr, def) {
			var v;

			if ((v = this.cur.getAttribute(attr)) === null)
				return def;

			return parseInt(v);
		};

		// getAttrString ----------------------------------------------
		XmlData.prototype.getAttrString = function(attr, def) {
			var v;

			if ((v = this.cur.getAttribute(attr)) === null)
				return def;

			return v;
		};

		// getBool ----------------------------------------------------
		XmlData.prototype.getBool = function(attr) {
			var v;

			if ((v = this.getAttrBool(attr, null)) === null)
				throw new Error('attribute not found: ' + attr);

			return v;
		};

		// getDouble --------------------------------------------------
		XmlData.prototype.getDouble = function(attr) {
			var v;

			if ((v = this.getString(attr)) === null)
				throw new Error('attribute not found: ' + attr);

			return parseFloat(v);
		};

		// getChildren ------------------------------------------------
		XmlData.prototype.getChildren = function(tag) {
			var i, a, children;

			a = [];

			children = this.cur.children;
			for(i = 0; i < children.length; ++i) {
				if (tag !== children[i].tagName)
					continue;

				a.push(new XmlData(this.doc, children[i]));
			}
			
			return(a);
		};

		// getFirstChild ----------------------------------------------
		XmlData.prototype.getFirstChild = function() {
			var children;
			
			children = this.cur.children;

			if (children.length === 0)
				return null;

			return new XmlData(this.doc, children[0]);
		};

		// getFirstChildWithTag ---------------------------------------
		XmlData.prototype.getFirstChildWithTag = function(tag) {
			var i, children;

			children = this.cur.children;
			for(i = 0; i < children.length; ++i)
				if (tag === children[i].tagName)
					return new XmlData(this.doc, children[i]);
		
			return null;
		};

		// getInt -----------------------------------------------------
		XmlData.prototype.getInt = function(attr) {

			if ((v = this.getString(attr)) === null)
				throw new Error('attribute not found: ' + attr);

			return parseInt(v);
		};

		// getTag -----------------------------------------------------
		XmlData.prototype.getTag = function() {

			return this.cur.tagName;
		};
		
		// getTextContent ---------------------------------------------
		XmlData.prototype.getTextContent = function() {

			return this.cur.textContent;
		};

		// getString --------------------------------------------------
		XmlData.prototype.getString = function(attr) {

			return this.cur.getAttribute(attr);
		};

		// getXml -----------------------------------------------------
		XmlData.prototype.getXml = function() {

			return new XMLSerializer().serializeToString(this.doc);
		};

		// importAsFirstChild -----------------------------------------
		XmlData.prototype.importAsFirstChild = function(src) {
			var children, first, node;

			node = this.doc.importNode(src.cur, true);

			children = this.cur.children;

			if (children.length === 0)
				this.cur.appendChild(node);
			else {
				first = children[0];
				this.cur.insertBefore(node, first);
			}

			return node;
		};

		// importElement ----------------------------------------------
		XmlData.prototype.importElement = function(src) {
			var node;

			node = this.doc.importNode(src.cur, true);
			this.cur.appendChild(node);
		};

		// removeAttr -------------------------------------------------
		XmlData.prototype.removeAttr = function(attr) {

			this.cur.removeAttribute(attr);
		};
		
		// setStringAttr ----------------------------------------------
		XmlData.prototype.setStringAttr = function(attr, value) {

			this.cur.setAttribute(attr, value);
		};

		// searchElement ----------------------------------------------
		XmlData.prototype.searchElement = function(exp) {
			var result;

			result = this.doc.evaluate(exp, this.cur, null, XPathResult.FIRST_ORDERED_NODE_TYPE);

			if (result.singleNodeValue === null)
				return null;

			return new XmlData(this.doc, result.singleNodeValue);
		};

		// searchElements ---------------------------------------------
		XmlData.prototype.searchElements = function(exp) {
			var r, el, result;

			//result = this.doc.evaluate(exp, this.cur, null, XPathResult.UNORDERED_NODE_ITERATOR_TYPE);
			result = this.doc.evaluate(exp, this.cur, null, XPathResult.ORDERED_NODE_ITERATOR_TYPE);

			r = [];

			while((el = result.iterateNext()) !== null)
				r.push(new XmlData(this.doc, el));

			return r;
		};

		// searchTextValue --------------------------------------------
		XmlData.prototype.searchTextValue = function(exp, src, tmpl) {
			var id, text;

			text = this.searchValue(exp);
			id = src.getString("id");
			
			return JsUtil.transformText(tmpl.fonts, id, text);
		};

		// searchValue ------------------------------------------------
		XmlData.prototype.searchValue = function(exp) {
			var result;

			result = this.doc.evaluate(exp, this.cur, null, XPathResult.STRING_TYPE);

			return result.stringValue;
		};

		// write ------------------------------------------------------
		XmlData.prototype.write = function() {

			console.log(this.cur);
		};
		
		// Constructor ------------------------------------------------
		{
			this.doc = doc;
			this.cur = cur;
		}
	};

	// createDocument -----------------------------------------------------
	XmlData.createDocument = function(tag) {
		var doc;

		doc = document.implementation.createDocument(null, tag);

		return new XmlData(doc, doc.documentElement);
	};

	/* ================================================================	*
	 * UILib.js 														*
	 * ================================================================	*/

	var UILib = function() {
	};

	UILib.NO_COLOR	= 'no-color';

	// cloneIntArray ------------------------------------------------------
	UILib.cloneIntArray = function(a) {
		var i, aa;

		aa = [];
		for(i = 0; i < a.length; ++i)
			aa[i] = a[i];

		return aa;
	};

	// getActiveColor -------------------------------------------------------
	UILib.getActiveColor = function(colorname) {
		var c;

		c = MColor.toRGB(colorname);
		return MColor.getActiveColor(c);
	};

	// getOverColor -------------------------------------------------------
	UILib.getOverColor = function(colorname) {
		var c;

		c = MColor.toRGB(colorname);
		return MColor.getOverColor(c);
	};

	// getStringMap -------------------------------------------------------
	UILib.getStringMap = function() {

		return new XMap();
	};

	// getVerticalScrollbarWidth ------------------------------------------
	UILib.getVerticalScrollbarWidth = function(scp) {

		return scp.uiwt.scrollw;
	};

	// getXDataListMap ----------------------------------------------------
	UILib.getXDataListMap = function() {

		return new XMap();
	};

	// getXDataMap --------------------------------------------------------
	UILib.getXDataMap = function() {

		return new XMap();
	};

	// getXFormMap --------------------------------------------------------
	UILib.getXFormMap = function() {

		return new XMap();
	};

	// getXTaskMap --------------------------------------------------------
	UILib.getXTaskMap = function() {

		return new XMap();
	};

	// intToString --------------------------------------------------------
	UILib.intToString = function(v) {

		return v.toString();
	};

	// isCharAt -----------------------------------------------------------
	UILib.isCharAt = function(s, idx, c) {

		if (s.charAt(idx) === c)
			return true;

		return false;
	};

	// mathFloor ----------------------------------------------------------
	UILib.mathFloor = function(v) {

		return Math.floor(v);
	};

	// mathRound ----------------------------------------------------------
	UILib.mathRound = function(v) {

		return Math.round(v);
	};

	// newArray -----------------------------------------------------------
	UILib.newArray = function() {
		var a, i, dim1, dim2, value;

		switch(arguments.length) {
		case 2:
			dim1 = arguments[0];
			value = arguments[1];

			a = [];
			for(i = 0; i < dim1; ++i)
				a[i] = value;
			return a;
		case 3:
			dim1 = arguments[0];
			dim2 = arguments[1];
			value = arguments[2];

			a = [];
			for(i = 0; i < dim1; ++i)
				a[i] = UILib.newArray(dim2, value);
			return a;
		default:
			throw new Error('not yet implemented: ' + arguments.length);
		}
	};

	// newXData -----------------------------------------------------------
	UILib.newXData = function(tag) {

		return new XData({tag:tag,_children:[]});
	};

	// newXmlData ---------------------------------------------------------
	UILib.newXmlData = function(tag) {

		return XmlData.createDocument(tag);
	};

	// parseInt -----------------------------------------------------------
	UILib.parseInt = function(s) {

		return parseInt(s);
	};

	// stringEmpty --------------------------------------------------------
	UILib.stringEmpty = function(s) {

		return s === '';
	};

	// stringLength -------------------------------------------------------
	UILib.stringLength = function(s) {

		return s.length;
	};

	// stringsEquals ------------------------------------------------------
	UILib.stringsEquals = function(s1, s2) {

		return s1 === s2;
	};

	// stringsMatch -------------------------------------------------------
	UILib.stringsMatch = function(s1, s2) {

		return s1.startsWith(s2);
	};

	// warning ------------------------------------------------------------
	UILib.warning = function(msg) {

		Log.info(msg);
	};
/* ====================================================================	*
 * HtmlTag 																*
 * ====================================================================	*/

var HtmlTag = function(text) {

	// Constructor ----------------------------------------------------
	{
		var k, len, n, n1, n2, n3, off, v;

		if (text.charAt(0) !== '/') {
			off = 0;
			this.open = true;
		} else {
			off = 1;
			this.open = false;
		}
		this.attrs = UILib.getStringMap();
		if ((n = text.indexOf(' ', off)) === -1) {
			this.tag = text.substring(off);
			return;
		}
		this.tag = text.substring(off, n);
		off = n + 1;
		len = UILib.stringLength(text);
		for(;;) {
			for(;off < len;++off)
				if (text.charAt(off) !== ' ')
					break;
			if (off === len)
				break;
			if ((n1 = text.indexOf('=', off)) === -1)
				throw new Error("invalid tag: " + text);
			k = text.substring(off, n1);
			if ((n2 = n1 + 1) === len)
				throw new Error("invalid tag: " + text);
			if (text.charAt(n2) !== '"')
				throw new Error("invalid tag: " + text);
			++n2;
			if ((n3 = text.indexOf('"', n2)) === -1)
				throw new Error("invalid tag: " + text);
			v = text.substring(n2, n3);
			this.attrs.put(k, v);
			off = n3 + 1;
		}
	}
};
/* ====================================================================	*
 * JsUtil 																*
 * ====================================================================	*/

var JsUtil = function() {
};

// transformText ------------------------------------------------------
JsUtil.transformText = function(fonts, id, html) {
	var align, first, font, ht, n1, n2, off, sb, v1, v2;

	sb = new XStringBuilder();
	sb.append("<div id=\"");
	sb.append(id);
	sb.append("\">");
	first = true;
	off = 0;
	for(;;) {
		if ((n1 = html.indexOf('<', off)) === -1)
			break;
		sb.append(html.substring(off, n1 ++));
		if ((n2 = html.indexOf('>', n1)) === -1)
			throw new Error("invalid html: " + html);
		ht = new HtmlTag(html.substring(n1, n2));
		if (!ht.open) {
			sb.append("</");
			sb.append(ht.tag);
			sb.append(">");
		} else {
			sb.append("<");
			sb.append(ht.tag);
			switch(ht.tag) {
			case "p":
				sb.append(" class=\"");
				if (first) {
					sb.append("mw-top");
					first = false;
				} else
					sb.append("mw");
				font = fonts.get("default");
				if ((v1 = ht.attrs.get("font")) !== null)
					if ((v2 = fonts.get(v1)) !== null)
						font = v2;
				sb.append(" mwf-");
				sb.append(font);
				if ((align = ht.attrs.get("align")) !== null) {
					switch(align) {
					case "center":
					case "right":
						sb.append(" mw-");
						sb.append(align);
						break;
					}
				}
				sb.append("\"");
				break;
			}
			sb.append(">");
		}
		off = n2 + 1;
	}
	sb.append("</div>");
	return(sb.getText());
};
/* ====================================================================	*
 * XAttrs 																*
 * ====================================================================	*/

var XAttrs = function(list) {

	// check ----------------------------------------------------------
	XAttrs.prototype.check = function(list) {
		var i, names;

		names = list.split(",");
		for(i = 0;i < names.length;++i)
			if (this.search(names[i]) === null)
				throw new Error("attribute not found: " + names[i]);
	};

	// getAttrItems ---------------------------------------------------
	XAttrs.prototype.getAttrItems = function(name) {
		var a, i, k, l, n, s, v, value;

		if ((value = this.search(name)) === null) {
			l = UILib.newArray(0, null);
			return(new XAttrs(l));
		}
		a = value.split(",");
		l = UILib.newArray(a.length, null);
		for(i = 0;i < a.length;++i) {
			s = a[i];
			if ((n = s.indexOf('=')) === -1)
				throw new Error("invalid attributes: " + value);
			k = s.substring(0, n);
			v = s.substring(n + 1);
			l[i] = new XAttr(k, v);
		}
		return(new XAttrs(l));
	};

	// getInt ---------------------------------------------------------
	XAttrs.prototype.getInt = function(name, def) {
		var v;

		if ((v = this.search(name)) !== null)
			return(XInteger.parseInt(v));
		return(def);
	};

	// getList --------------------------------------------------------
	XAttrs.prototype.getList = function() {
		return(this.list);
	};

	// getString ------------------------------------------------------
	XAttrs.prototype.getString = function(name) {
		var v;

		if ((v = this.search(name)) !== null)
			return(v);
		throw new Error("attribute not found: " + name);
	};

	// search ---------------------------------------------------------
	XAttrs.prototype.search = function(name) {
		var i;

		for(i = 0;i < this.list.length;++i)
			if (UILib.stringsEquals(name, this.list[i].name))
				return(this.list[i].value);
		return(null);
	};

	// Constructor ----------------------------------------------------
	{
		this.list = list;
	}
};

// createFromXData ----------------------------------------------------
XAttrs.createFromXData = function(data, names) {
	var i, k, kv, kvs, l, list;

	if (names === null) {
		kvs = data.getAttributes();
		list = UILib.newArray(kvs.length, null);
		for(i = 0;i < list.length;++i) {
			kv = kvs[i];
			list[i] = new XAttr(kv.key, kv.value);
		}
	} else {
		l = names.split(",");
		list = UILib.newArray(l.length, null);
		for(i = 0;i < list.length;++i) {
			k = l[i];
			list[i] = new XAttr(k, data.getAttrString(k, null));
		}
	}
	return(new XAttrs(list));
};
/* ====================================================================	*
 * XAction 																*
 * ====================================================================	*/

var XAction = function(data) {

	this.requestId = null;

	// run ------------------------------------------------------------
	XAction.prototype.run = function(taskid, appl) {
		XApplication.logMsg("XAction.run[" + taskid + "]: " + this.tag);
		switch(this.tag) {
		case "card-update":
		case "form-set-explorer":
		case "upload-files":
			appl.execTask(this);
			return(true);
		}
		switch(this.tag) {
		case "cache-clear":
			appl.taskCacheClear();
			break;
		case "cache-download":
			appl.taskCacheDownload(this.ref, this.nextTask);
			break;
		case "card-load":
			appl.taskCardLoad(this.id);
			break;
		case "card-show":
			appl.taskCardShow(this.id);
			break;
		case "cards-hide":
			appl.taskCardsHide();
			break;
		case "debug-vars":
			appl.taskDebugVars(this.xref);
			break;
		case "echo":
			appl.taskEcho(this.value);
			break;
		case "exec-task":
			appl.taskExecTask(this.id);
			break;
		case "exit":
			appl.taskExit(this.ref);
			break;
		case "form-clean":
			appl.taskFormClean(this.formId, this.fields);
			break;
		case "form-set":
			appl.taskFormSet(this.cardId, this.formId, this.ref, this.values);
			break;
		case "form-set-tree":
			appl.taskFormSetTree(this.fst);
			break;
		case "html5-action":
			appl.taskHtml5Action(this.ref);
			break;
		case "http-parallel":
			appl.taskHttpParallel(this.requests);
			break;
		case "http-request":
			appl.taskHttpRequest(this.hreq, null);
			break;
		case "if":
			return(appl.taskIf(taskid, this.xif));
		case "plugin-action":
			appl.taskPluginAction(this.widgetId, this.ref, this.value);
			break;
		case "plugin-object":
			appl.taskPluginObject(this.id, this.ref, this.value);
			break;
		case "popup":
			appl.taskPopupShow(this);
			break;
		case "popup-info":
			appl.taskPopupInfo(this.id, this.value);
			break;
		case "popup-error":
			appl.taskPopupError(this.id, this.value);
			break;
		case "return":
			return(false);
		case "set-var":
			appl.taskSetVar(this.id, this.ref);
			break;
		case "switch":
			return(appl.taskSwitch(taskid, this.xswitch));
		case "template-update":
			appl.taskTemplateUpdate(this.id, this.ref, this.templateId, this.visible);
			break;
		case "widget":
			appl.taskWidget(this.xwa);
			break;
		case "widget-clean":
			appl.taskWidgetClean(this.cardId, this.widgetId);
			break;
		case "widget-config":
			appl.taskWidgetConfig(this.id, this.ref, this.value);
			break;
		case "widget-hide":
			appl.taskWidgetShow(this.cardId, this.widgetId, false);
			break;
		case "widget-set":
			appl.taskWidgetSet(this.cardId, this.widgetId, this.value);
			break;
		case "widget-show":
			appl.taskWidgetShow(this.cardId, this.widgetId, true);
			break;
		default:
			throw new Error("not yet implemented: " + this.tag);
		}
		return(true);
	};

	// Constructor ----------------------------------------------------
	{
		var children, i;

		this.tag = data.getTag();
		switch(this.tag) {
		case "cache-clear":
			break;
		case "cache-download":
			this.ref = data.getString("ref");
			this.nextTask = data.getString("nextTask");
			break;
		case "card-load":
		case "card-show":
			this.id = data.getString("id");
			break;
		case "card-update":
			this.cardId = data.getString("cardId");
			this.widgetId = data.getString("widgetId");
			this.value = data.getAttrString("value", null);
			this.xref = data.getAttrString("xref", null);
			break;
		case "cards-hide":
			break;
		case "debug-vars":
			this.xref = data.getAttrString("xref", null);
			break;
		case "echo":
			this.value = data.getString("value");
			break;
		case "exec-task":
			this.id = data.getString("id");
			break;
		case "exit":
			this.ref = data.getAttrString("msg", "exit");
			break;
		case "form-clean":
			this.formId = data.getString("formId");
			this.fields = data.getAttrString("fields", null);
			break;
		case "form-set":
			this.cardId = data.getAttrString("cardId", null);
			this.formId = data.getString("formId");
			this.ref = data.getAttrString("ref", null);
			this.values = XAction.getValues("root", data);
			break;
		case "form-set-explorer":
			this.attrs = XAttrs.createFromXData(data, null);
			this.attrs.check("cardId,formId,inputId,tref,vref");
			break;
		case "form-set-tree":
			this.fst = new XFormSetTree(data);
			break;
		case "html5-action":
			this.ref = data.getString("ref");
			break;
		case "http-parallel":
			this.requests = new XHttpRequests(data);
			break;
		case "http-request":
			this.hreq = new XHttpRequest(data);
			break;
		case "if":
			this.xif = new XIfAction(data);
			break;
		case "plugin-action":
			this.widgetId = data.getString("widgetId");
			this.ref = data.getString("ref");
			this.value = data.getAttrString("value", null);
			break;
		case "plugin-object":
			this.id = data.getString("id");
			this.ref = data.getString("ref");
			this.value = data.getAttrString("value", null);
			break;
		case "popup":
			this.id = data.getString("id");
			this.templateId = data.getString("templateId");
			this.values = XAction.getValues("popup", data);
			children = data.getChildren("task");
			if (children.length === 0)
				throw new Error("no task defined for popup: " + this.id);
			this.tasks = UILib.newArray(children.length, null);
			for(i = 0;i < this.tasks.length;++i) {
				this.tasks[i] = new XTask(children[i]);
				this.values.addChildWithContent(this.tasks[i].id, "#popup#" + this.id + "#" + this.tasks[i].id);
			}
			break;
		case "popup-error":
			this.id = data.getString("id");
			this.ref = data.getString("ref");
			this.value = data.getAttrString("value", null);
			break;
		case "popup-info":
			this.id = data.getString("id");
			this.ref = data.getString("ref");
			this.value = data.getString("value");
			break;
		case "return":
			break;
		case "set-var":
			this.id = data.getString("id");
			this.ref = data.getString("ref");
			this.cardId = data.getAttrString("cardId", null);
			break;
		case "switch":
			this.xswitch = new XSwitchAction(data);
			break;
		case "template-update":
			this.id = data.getString("id");
			this.ref = data.getString("ref");
			this.templateId = data.getString("templateId");
			this.visible = data.getAttrBool("visible", false);
			break;
		case "upload-files":
			this.attrs = XAttrs.createFromXData(data, null);
			this.attrs.check("requestId,nextTask");
			break;
		case "widget":
			this.xwa = new XWidgetAction(data);
			break;
		case "widget-clean":
			this.cardId = data.getString("cardId");
			this.widgetId = data.getString("widgetId");
			break;
		case "widget-config":
			this.id = data.getString("id");
			this.ref = data.getString("ref");
			this.value = data.getString("value");
			break;
		case "widget-hide":
			this.cardId = data.getString("cardId");
			this.widgetId = data.getString("widgetId");
			break;
		case "widget-show":
			this.cardId = data.getString("cardId");
			this.widgetId = data.getString("widgetId");
			break;
		case "widget-set":
			this.cardId = data.getString("cardId");
			this.widgetId = data.getString("widgetId");
			this.value = data.getTextContent();
			break;
		default:
			throw new Error("invalid tag: " + this.tag);
		}
	}
};

// getValues ----------------------------------------------------------
XAction.getValues = function(tag, parent) {
	var child, children, i, k, r;

	r = UILib.newXData(tag);
	children = parent.getChildren("value");
	for(i = 0;i < children.length;++i) {
		child = children[i];
		if ((k = child.getAttrString("id", null)) === null)
			throw new Error("attribute 'id' not found");
		r.addChildWithContent(k, child.getTextContent());
	}
	return(r);
};
/* ====================================================================	*
 * XApplConfig 															*
 * ====================================================================	*/

var XApplConfig = function(data, vars) {

	// Constructor ----------------------------------------------------
	{
		var children, d, i, k, v;

		if (data === null)
			return;
		this.applname = data.getAttrString("applname", null);
		if ((d = data.getFirstChildWithTag("servlet")) !== null) {
			this.server = d.getAttrString("server", null);
			this.uri = d.getAttrString("uri", null);
			this.endpoint = d.getAttrString("endpoint", null);
			this.mode = d.getAttrString("mode", "prod");
		}
		if ((d = data.getFirstChildWithTag("layout")) !== null)
			this.bg = d.getAttrString("bg", null);
		children = data.getChildren("var");
		for(i = 0;i < children.length;++i) {
			d = children[i];
			k = d.getString("id");
			v = d.getTextContent();
			vars.putVar(k, v);
		}
		children = data.getChildren("canvas-img");
		this.canvasImgs = UILib.newArray(children.length, null);
		for(i = 0;i < children.length;++i)
			this.canvasImgs[i] = children[i].getTextContent();
	}
};
/* ====================================================================	*
 * XFile 																*
 * ====================================================================	*/

var XFile = function(xd) {

	// Constructor ----------------------------------------------------
	{
		this.tag = xd.getTag();
		this.path = xd.getTextContent();
		switch(this.tag) {
		case "img":
		case "migration-img":
			this.width = xd.getInt("width");
			this.height = xd.getInt("height");
			break;
		}
	}
};

// getList ------------------------------------------------------------
XFile.getList = function(xd, exp) {
	var el, files, i, list;

	if ((el = xd.searchElement(exp)) === null)
		return(null);
	list = el.getAllChildren();
	files = UILib.newArray(list.length, null);
	for(i = 0;i < files.length;++i)
		files[i] = new XFile(list[i]);
	return(files);
};
/* ====================================================================	*
 * XTemplate 															*
 * ====================================================================	*/

var XTemplate = function(appl, tmpl) {

	// copyChildren ---------------------------------------------------
	XTemplate.prototype.copyChildren = function(src, dst) {
		var children, i;

		children = src.getAllChildren();
		for(i = 0;i < children.length;++i)
			dst.importElement(children[i]);
	};

	// eval -----------------------------------------------------------
	XTemplate.prototype.eval = function(exp, cdata) {
		var last, v, vp, xd;

		if (UILib.stringsEquals(exp, "$content"))
			return(cdata.getTextContent());
		if (UILib.isCharAt(exp, 0, '@'))
			return(cdata.getAttrString(exp.substring(1), ""));
		if (UILib.isCharAt(exp, 0, '$')) {
			vp = new XVarPath(exp);
			if ((xd = this.appl.getData(vp.dataId)) === null)
				throw new Error("data not found: " + exp);
			if ((v = xd.searchValue(vp.xpath)) === null)
				throw new Error("xpath not found: " + vp.xpath);
			return(v);
		}
		if (UILib.isCharAt(exp, 0, '/')) {
			if ((xd = this.appl.getData("popup")) === null)
				throw new Error("data not found: popup");
			if ((v = xd.searchValue(exp)) === null)
				throw new Error("expression not found: " + exp);
			return(this.appl.parseText(v));
		}
		if (UILib.stringsMatch(exp, "./")) {
			if ((v = cdata.searchValue(exp.substring(2))) === null)
				throw new Error("expression not found: " + exp);
			return(v);
		}
		switch(exp) {
		case "#tag":
			return(cdata.getTag());
		}
		last = UILib.stringLength(exp) -1;
		if (exp.charAt(last) === ')') {
			if (UILib.stringsMatch(exp, "#calc("))
				return(this.evalCalc(exp.substring(6, last)));
		}
		throw new Error("invalid expression: " + exp);
	};

	// evalCalc -------------------------------------------------------
	XTemplate.prototype.evalCalc = function(line) {
		var arg, args, i, idx, key, l, len, op, r, v, values;

		args = line.split(" ");
		len = args.length;
		if ((len % 2) === 0)
			throw new Error("invalid number of arguments: " + line);
		idx = 0;
		op = UILib.newArray(len / 2, 0);
		for(i = 1;i < len;i += 2) {
			switch(args[i]) {
			case "+":
				op[idx ++] = 1;
				break;
			case "-":
				op[idx ++] = 2;
				break;
			case "*":
				op[idx ++] = 3;
				break;
			default:
				throw new Error("invalid operator: " + line);
			}
		}
		r = 0;
		idx = 0;
		values = UILib.newArray(op.length, 0);
		for(i = 0;i < len;i += 2) {
			arg = args[i];
			if (UILib.stringsMatch(arg, "$#")) {
				key = arg.substring(2);
				if ((l = this.index.get(key)) === null)
					throw new Error("invalid key: " + key);
				v = l.length;
			} else
				v = UILib.parseInt(arg);
			if (i === 0)
				r = v;
			else
				values[idx ++] = v;
		}
		for(i = 0;i < op.length;++i) {
			switch(op[i]) {
			case 1:
				r += values[i];
				break;
			case 2:
				r -= values[i];
				break;
			case 3:
				r *= values[i];
				break;
			}
		}
		return(UILib.intToString(r));
	};

	// parse ----------------------------------------------------------
	XTemplate.prototype.parse = function(src, dst, cdata) {
		var a, attrs, bg, c, child, children, exp, i, k, l, tag, vp, xd, xml;

		attrs = src.getAttributes();
		for(i = 0;i < attrs.length;++i) {
			a = attrs[i];
			if (!UILib.isCharAt(a.key, 0, '_'))
				dst.setStringAttr(a.key, a.value);
			else
				dst.setStringAttr(a.key.substring(1), this.eval(a.value, cdata));
		}
		switch(src.getTag()) {
		case "labelbutton":
			bg = dst.getAttrString("bg", null);
			if (dst.getAttrString("over", null) === null)
				dst.setStringAttr("over", UILib.getOverColor(bg));
			if (dst.getAttrString("active", null) === null)
				dst.setStringAttr("active", UILib.getActiveColor(bg));
			return;
		case "text":
			if ((exp = src.getAttrString("_content", null)) !== null) {
				vp = new XVarPath(exp);
				if ((xd = this.appl.getData(vp.dataId)) === null)
					throw new Error("data not found: " + exp);
				xml = xd.searchTextValue(vp.xpath, src, this);
				dst.addXText(xml);
			}
			this.copyChildren(src, dst);
			return;
		}
		children = src.getAllChildren();
		for(i = 0;i < children.length;++i) {
			c = children[i];
			tag = c.getTag();
			switch(tag) {
			case "template-foreach":
				if ((k = c.getAttrString("key", null)) === null)
					throw new Error("template-foreach - attribute 'key' not found");
				l = this.index.get(k);
				this.parseForEach(c.getAllChildren(), l);
				break;
			default:
				child = dst.addChild(tag);
				this.parse(c, child, cdata);
			}
		}
	};

	// parseForEach ---------------------------------------------------
	XTemplate.prototype.parseForEach = function(children, cdata) {
		var c, child, i, j;

		for(i = 0;i < cdata.length;++i) {
			for(j = 0;j < children.length;++j) {
				c = children[j];
				child = this.res.addChild(c.getTag());
				this.parse(c, child, cdata[i]);
			}
		}
	};

	// Constructor ----------------------------------------------------
	{
		var child, children, fontid, i, id, key, path, rpath, tag, vp, widget, xd;

		this.appl = appl;
		this.index = UILib.getXDataListMap();
		if (tmpl.getAttrString("ref", null) === null)
			throw new Error("attribute 'ref' not found");
		if ((tag = tmpl.getAttrString("widget", null)) === null)
			throw new Error("attribute 'widget' not found");
		children = tmpl.getChildren("template-font");
		this.fonts = UILib.getStringMap();
		for(i = 0;i < children.length;++i) {
			child = children[i];
			if ((id = child.getAttrString("id", null)) === null)
				throw new Error("template-font - attribute 'id' not found");
			fontid = child.getAttrString("fontid", null);
			this.fonts.put(id, fontid);
		}
		children = tmpl.getChildren("template-index");
		for(i = 0;i < children.length;++i) {
			child = children[i];
			if ((key = child.getAttrString("key", null)) === null)
				throw new Error("template-index - attribute 'key' not found");
			if ((path = child.getAttrString("path", null)) === null)
				throw new Error("template-index - attribute 'path' not found");
			vp = new XVarPath(path);
			if ((xd = appl.getData(vp.dataId)) === null)
				throw new Error("template-index - data not found: " + path);
			rpath = appl.parseText(vp.xpath);
			this.index.put(key, xd.searchElements(rpath));
		}
		if ((widget = tmpl.getFirstChildWithTag(tag)) === null)
			throw new Error("tag not found: " + tag);
		this.res = UILib.newXData(tag);
		this.parse(widget, this.res, null);
	}
};

// getWidget ----------------------------------------------------------
XTemplate.getWidget = function(appl, tmpl) {
	var xt;

	xt = new XTemplate(appl, tmpl);
	return(xt.res);
};
/* ====================================================================	*
 * XApplication 														*
 * ====================================================================	*/

var XApplication = function(uiappl) {

	this.config = null;
	this.tasks = null;

	// execTask -------------------------------------------------------
	XApplication.prototype.execTask = function(action) {
		switch(action.tag) {
		case "card-update":
			this.uiappl.cardUpdate(action, this.vars);
			break;
		case "form-set-explorer":
			this.uiappl.setFormExplorer(action.attrs);
			break;
		case "upload-files":
			this.uiappl.uploadFiles(action.attrs);
			break;
		default:
			throw new Error("invalid action: " + action.tag);
		}
	};

	// getAllData -----------------------------------------------------
	XApplication.prototype.getAllData = function() {
		return(this.vars.getAllData());
	};

	// getApplName ----------------------------------------------------
	XApplication.prototype.getApplName = function() {
		if (this.config === null)
			return(null);
		return(this.config.applname);
	};

	// getBackground --------------------------------------------------
	XApplication.prototype.getBackground = function() {
		if (this.config === null)
			return(null);
		return(this.config.bg);
	};

	// getConfig ------------------------------------------------------
	XApplication.prototype.getConfig = function() {
		return(this.config);
	};

	// getData --------------------------------------------------------
	XApplication.prototype.getData = function(id) {
		return(this.vars.getData(id));
	};

	// getEndpoint ----------------------------------------------------
	XApplication.prototype.getEndpoint = function() {
		switch(this.config.server) {
		case "localhost:8084":
		case "192.168.1.123:8084":
			return("ws://" + this.config.server + this.config.endpoint);
		default:
			return("wss://" + this.config.server + this.config.endpoint);
		}
	};

	// getFormInputValue ----------------------------------------------
	XApplication.prototype.getFormInputValue = function(id) {
		var entries, fieldId, formId, i, input, itemId, n, s, xf;

		if ((n = id.indexOf('#')) === -1)
			throw new Error("invalid id: " + id);
		formId = id.substring(0, n);
		s = id.substring(n + 1);
		if ((n = s.indexOf('-')) === -1) {
			fieldId = s;
			itemId = null;
		} else {
			fieldId = s.substring(0, n);
			itemId = s.substring(n + 1);
		}
		xf = this.getXForm(formId);
		entries = xf.entries;
		for(i = 0;i < entries.length;++i) {
			input = entries[i];
			if (XString.equals(fieldId, input.id))
				return(this.uiappl.getFormField(input.type, formId, fieldId, itemId));
		}
		throw new Error("form not found: " + id);
	};

	// getWidgetInputValue --------------------------------------------
	XApplication.prototype.getWidgetInputValue = function(id) {
		var cardId, n, widgetId;

		if ((n = id.indexOf('#')) === -1)
			throw new Error("invalid id: " + id);
		cardId = id.substring(0, n);
		widgetId = id.substring(n + 1);
		return(this.uiappl.widgetGet(cardId, widgetId));
	};

	// getTasks -------------------------------------------------------
	XApplication.prototype.getTasks = function() {
		return(this.tasks);
	};

	// getUrl ---------------------------------------------------------
	XApplication.prototype.getUrl = function() {
		if (this.config.server === null)
			return(null);
		switch(this.config.server) {
		case "localhost:8084":
		case "192.168.1.123:8084":
			return("http://" + this.config.server + this.config.uri);
		default:
			return("https://" + this.config.server + this.config.uri);
		}
	};

	// getVar ---------------------------------------------------------
	XApplication.prototype.getVar = function(id) {
		var dataId, n, path, xd;

		if (!id.startsWith("#"))
			return(this.vars.getVar(id));
		if ((n = id.indexOf("/", 1)) === -1)
			throw new Error("invalid syntax: " + id);
		dataId = id.substring(1, n);
		path = id.substring(n + 1);
		if ((xd = this.getData(dataId)) === null)
			throw new Error("data not found: " + id);
		return(xd.searchValue(path));
	};

	// getXForm -------------------------------------------------------
	XApplication.prototype.getXForm = function(ref) {
		var form;

		if ((form = this.forms.get(ref)) !== null)
			return(form);
		form = this.uiappl.getXForm(ref);
		this.forms.put(ref, form);
		return(form);
	};

	// init -----------------------------------------------------------
	XApplication.prototype.init = function(data) {
		this.config = new XApplConfig(data.getFirstChildWithTag("config"), this.vars);
		this.tasks = new XTasks(this, data.getChildren("task"));
		XApplication.mode = this.config.mode;
	};

	// listVars -------------------------------------------------------
	XApplication.prototype.listVars = function() {
		return(this.vars.listVars());
	};

	// parseText ------------------------------------------------------
	XApplication.prototype.parseText = function(text) {
		return(this.vars.parse(text));
	};

	// putData --------------------------------------------------------
	XApplication.prototype.putData = function(id, data) {
		this.vars.putData(id, data);
	};

	// putVar ---------------------------------------------------------
	XApplication.prototype.putVar = function(id, ref) {
		this.vars.putVar(id, ref);
	};

	// runTask --------------------------------------------------------
	XApplication.prototype.runTask = function(taskId) {
		XApplication.logMsg("=====> runTask: " + taskId);
		if (this.tasks !== null)
			this.tasks.run(taskId);
		else
			UILib.warning("no task defined: " + taskId);
	};

	// runWizardTask --------------------------------------------------
	XApplication.prototype.runWizardTask = function(wizardId, taskId) {
		this.uiappl.wizardTask(wizardId, taskId);
	};

	// searchAction ---------------------------------------------------
	XApplication.prototype.searchAction = function(actions, tag, id) {
		var a, i, j, r, t;

		for(i = 0;i < actions.length;++i) {
			a = actions[i];
			if (UILib.stringsEquals(a.tag, tag))
				if (UILib.stringsEquals(a.id, id))
					return(a);
			switch(a.tag) {
			case "if":
				if ((r = this.searchAction(a.xif.ifblock, tag, id)) !== null)
					return(r);
				if ((r = this.searchAction(a.xif.ifblock, tag, id)) !== null)
					return(r);
				break;
			case "popup":
				t = a.tasks;
				for(j = 0;j < t.length;++j)
					if ((r = this.searchAction(t[j].actions, tag, id)) !== null)
						return(r);
				break;
			}
		}
		return(null);
	};

	// taskCacheClear -------------------------------------------------
	XApplication.prototype.taskCacheClear = function() {
		this.uiappl.cacheClear();
	};

	// taskCacheDownload ----------------------------------------------
	XApplication.prototype.taskCacheDownload = function(ref, nextTask) {
		var dataId, exp, files, n, xd;

		if ((n = ref.indexOf(':')) === -1)
			throw new Error("invalid ref: " + ref);
		dataId = ref.substring(0, n);
		exp = ref.substring(n + 1);
		xd = this.getData(dataId);
		if ((files = XFile.getList(xd, exp)) === null)
			throw new Error("element not found: " + dataId + ":" + exp);
		this.uiappl.cacheDownload(files, nextTask);
	};

	// taskCardLoad ---------------------------------------------------
	XApplication.prototype.taskCardLoad = function(id) {
		this.uiappl.cardLoad(id);
	};

	// taskCardShow ---------------------------------------------------
	XApplication.prototype.taskCardShow = function(id) {
		this.uiappl.cardShow(id);
	};

	// taskCardsHide --------------------------------------------------
	XApplication.prototype.taskCardsHide = function() {
		this.uiappl.cardsHide();
	};

	// taskDebugVars --------------------------------------------------
	XApplication.prototype.taskDebugVars = function(xref) {
		var data, i, keys, kv, list, v, xd;

		if (xref !== null) {
			if ((xd = this.vars.getData(xref)) !== null) {
				XApplication.logXml("taskDebugVar: " + xref, xd);
				return;
			}
			if ((v = this.vars.getVar(xref)) !== null) {
				XApplication.logMsg("----taskDebugVars----");
				XApplication.logMsg(xref + ": " + v);
				return;
			}
		}
		XApplication.logMsg("----taskDebugVars----");
		list = this.vars.listVars();
		for(i = 0;i < list.length;++i) {
			kv = list[i];
			XApplication.logMsg(kv.key + ": " + kv.value);
		}
		XApplication.logMsg("--");
		data = this.vars.getAllData();
		keys = UILib.newArray(data.size(), null);
		data.keys(keys);
		for(i = 0;i < keys.length;++i)
			XApplication.logMsg(keys[i]);
		XApplication.logMsg("---------------------");
	};

	// taskEcho -------------------------------------------------------
	XApplication.prototype.taskEcho = function(value) {
		XApplication.logMsg("=====> echo: " + value);
	};

	// taskExecTask ---------------------------------------------------
	XApplication.prototype.taskExecTask = function(id) {
		this.runTask(id);
	};

	// taskExist ------------------------------------------------------
	XApplication.prototype.taskExist = function(taskid) {
		return(this.tasks.exists(taskid));
	};

	// taskExit -------------------------------------------------------
	XApplication.prototype.taskExit = function(msg) {
		this.uiappl.exit(msg);
	};

	// taskFormClean --------------------------------------------------
	XApplication.prototype.taskFormClean = function(id, fields) {
		var i, input, l, lfields, xform;

		if ((xform = this.getXForm(id)) === null)
			throw new Error("form not found: " + id);
		if (fields === null) {
			l = xform.entries;
			for(i = 0;i < l.length;++i) {
				input = l[i];
				this.uiappl.cleanFormField(input.type, id, input.id);
			}
			return;
		}
		lfields = fields.split(",");
		for(i = 0;i < lfields.length;++i) {
			input = xform.getInput(lfields[i]);
			this.uiappl.cleanFormField(input.type, id, input.id);
		}
	};

	// taskFormSet ----------------------------------------------------
	XApplication.prototype.taskFormSet = function(cardId, id, ref, data) {
		var child, children, dataId, dref, fieldId, i, input, l, n, options, path, text, type, value, xform;

		if ((xform = this.getXForm(id)) === null)
			throw new Error("form not found: " + id);
		children = data.getAllChildren();
		for(i = 0;i < children.length;++i) {
			child = children[i];
			fieldId = child.getTag();
			type = xform.getType(fieldId);
			if ((text = child.getTextContent()) === null)
				text = "";
			switch(type) {
			case "checkbox":
			case "radiobutton":
				if (UILib.stringsMatch(text, "${#")) {
					if ((n = text.indexOf('/')) === -1)
						throw new Error("invalid var: " + text);
					dataId = text.substring(3, n);
					dref = this.getData(dataId);
					path = text.substring(n + 1, UILib.stringLength(text) -1);
					options = dref.getChildren(path);
					for(n = 0;n < options.length;++n) {
						value = options[n].getTextContent();
						this.uiappl.setFormField(cardId, type, id, fieldId, value);
					}
					continue;
				}
			}
			value = this.parseText(text);
			this.uiappl.setFormField(cardId, type, id, fieldId, value);
		}
		if (ref === null)
			return;
		if (!UILib.isCharAt(ref, 0, '#'))
			throw new Error("invalid ref: " + ref);
		if ((n = ref.indexOf('/')) === -1)
			throw new Error("invalid ref: " + ref);
		dataId = ref.substring(1, n);
		path = "/magicw/" + ref.substring(n + 1) + "/";
		dref = this.getData(dataId);
		l = xform.entries;
		for(i = 0;i < l.length;++i) {
			input = l[i];
			if ((value = dref.searchValue(path + input.id)) !== null)
				this.uiappl.setFormField(cardId, input.type, id, input.id, value);
		}
	};

	// taskFormSetTree ------------------------------------------------
	XApplication.prototype.taskFormSetTree = function(fst) {
		this.uiappl.setFormTree(fst);
	};

	// taskHtml5Action ------------------------------------------------
	XApplication.prototype.taskHtml5Action = function(ref) {
		this.uiappl.html5Action(ref);
	};

	// taskHttpParallel -----------------------------------------------
	XApplication.prototype.taskHttpParallel = function(reqs) {
		var i, list, results;

		list = reqs.list;
		results = reqs.results;
		for(i = 0;i < list.length;++i) {
			results[i] = null;
			this.taskHttpRequest(list[i], reqs);
		}
	};

	// taskHttpRequest ------------------------------------------------
	XApplication.prototype.taskHttpRequest = function(hreq, reqs) {
		var data, i, tag, value, values, xdoc, xform;

		this.waitingStart();
		xdoc = UILib.newXmlData("magicw");
		xdoc.setStringAttr("id", hreq.requestId);
		if (hreq.formId !== null) {
			if ((xform = this.getXForm(hreq.formId)) === null)
				throw new Error("form not found: " + hreq.formId);
			if (!xform.read(xdoc, this.uiappl)) {
				XApplication.logXml("form-check", xdoc);
				this.runTask(hreq.formId + "#http-check");
				this.waitingStop();
				return;
			}
		}
		values = hreq.values.getAllChildren();
		for(i = 0;i < values.length;++i) {
			tag = values[i].getTag();
			value = this.parseText(values[i].getTextContent());
			xdoc.addChildWithContent(tag, value);
		}
		if (hreq.dataId !== null)
			if ((data = this.getData(hreq.dataId)) !== null)
				xdoc.importElement(data);
		XApplication.logXml("http-request: " + hreq.requestId, xdoc);
		this.uiappl.httpRequest(hreq, xdoc, reqs);
	};

	// taskIf ---------------------------------------------------------
	XApplication.prototype.taskIf = function(taskid, xif) {
		var block, i, v1, v2;

		if ((v1 = this.vars.getVar(xif.var)) === null)
			throw new Error("invalid var: " + xif.var);
		v2 = this.vars.parse(xif.value);
		if (XString.equals(v1, v2))
			block = xif.ifblock;
		else
			block = xif.elseblock;
		for(i = 0;i < block.length;++i)
			if (!block[i].run(taskid, this))
				return(false);
		return(true);
	};

	// taskNotifyEvent ------------------------------------------------
	XApplication.prototype.taskNotifyEvent = function(taskid, data) {
		var n, popup, task, wizard;

		if (data !== null)
			this.putVar("action-data", data);
		if (UILib.stringsMatch(taskid, "#popup#")) {
			if ((n = taskid.indexOf('#', 7)) === -1)
				throw new Error("invalid taskid: " + taskid);
			popup = taskid.substring(7, n);
			task = taskid.substring(n + 1);
			if (this.tasks !== null)
				this.taskNotifyPopupTask(popup, task);
			else
				UILib.warning("no task defined: " + task);
			return;
		}
		if (UILib.stringsMatch(taskid, "#wizard#")) {
			if ((n = taskid.indexOf('#', 8)) === -1)
				throw new Error("invalid taskid: " + taskid);
			wizard = taskid.substring(8, n);
			task = taskid.substring(n + 1);
			this.runWizardTask(wizard, task);
			return;
		}
		this.runTask(taskid);
	};

	// taskNotifyHttpRequestDone --------------------------------------
	XApplication.prototype.taskNotifyHttpRequestDone = function(requestId, data, reqs) {
		var id, result;

		XApplication.logXml("http-response: " + requestId, data);
		if ((id = data.getAttrString("dataId", null)) === null)
			id = data.getString("id");
		this.putData(id, data);
		result = data.getString("result");
		if (reqs !== null) {
			reqs.notifyRequest(this, requestId, result);
			return;
		}
		this.waitingStop();
		this.runTask(requestId + "#" + result);
	};

	// taskNotifyHttpRequestError -------------------------------------
	XApplication.prototype.taskNotifyHttpRequestError = function(requestId, reqs) {
		var taskid;

		if (reqs !== null) {
			reqs.notifyRequest(this, requestId, "http-error");
			return;
		}
		this.waitingStop();
		taskid = requestId + "#http-error";
		if (this.tasks.exists(taskid))
			this.runTask(taskid);
		else
			this.runTask("#http-error");
	};

	// taskNotifyHttpSessionTimeout -----------------------------------
	XApplication.prototype.taskNotifyHttpSessionTimeout = function() {
		this.waitingStop();
		this.runTask("#http-timeout");
	};

	// taskNotifyHttpUploadDone ---------------------------------------
	XApplication.prototype.taskNotifyHttpUploadDone = function(nextTask, data) {
		var id;

		this.waitingStop();
		id = data.getString("id");
		XApplication.logXml("http-response: " + id, data);
		this.putData(id, data);
		this.runTask(nextTask);
	};

	// taskNotifyHttpUploadError --------------------------------------
	XApplication.prototype.taskNotifyHttpUploadError = function(nextTask) {
		this.waitingStop();
		if (this.tasks.exists(nextTask))
			this.runTask(nextTask);
		else
			this.runTask("#http-error");
	};

	// taskNotifyPopupTask --------------------------------------------
	XApplication.prototype.taskNotifyPopupTask = function(popupId, taskId) {
		var action, i, id, l, t, tag;

		if (UILib.stringsMatch(popupId, "_popup-error-")) {
			tag = "popup-error";
			id = popupId.substring(13);
		} else
			if (UILib.stringsMatch(popupId, "_popup-info-")) {
				tag = "popup-info";
				id = popupId.substring(12);
			} else {
				tag = "popup";
				id = popupId;
			}
		action = null;
		l = this.tasks.list();
		for(i = 0;i < l.length;++i)
			if ((action = this.searchAction(l[i].actions, tag, id)) !== null)
				break;
		if (action === null)
			throw new Error("action not found: <" + tag + " id=\"" + id + "\">");
		if (UILib.stringsEquals(action.tag, "popup-error")) {
			XApplication.logMsg("cardShow: " + action.ref);
			this.taskCardShow(action.ref);
			return;
		}
		if (UILib.stringsEquals(action.tag, "popup-info")) {
			XApplication.logMsg("cardShow: " + action.ref);
			this.taskCardShow(action.ref);
			return;
		}
		if (UILib.stringsEquals(action.tag, "popup")) {
			l = action.tasks;
			for(i = 0;i < l.length;++i) {
				t = l[i];
				if (XString.equals(t.id, taskId)) {
					t.run(this);
					return;
				}
			}
			throw new Error("task not found: popupId=" + id + " taskId=" + taskId);
		}
		throw new Error("not yet implemented: " + tag);
	};

	// taskPopupInfo --------------------------------------------------
	XApplication.prototype.taskPopupInfo = function(id, value) {
		var a, c1, d;

		d = UILib.newXData("popup");
		d.setStringAttr("id", "_popup-info-" + id);
		d.setStringAttr("templateId", "popup-info");
		c1 = d.addChild("value");
		c1.setStringAttr("id", "text");
		c1.setContent(this.parseText(value));
		c1 = d.addChild("task");
		c1.setStringAttr("id", "continue");
		a = new XAction(d);
		this.taskPopupShow(a);
	};

	// taskPluginAction -----------------------------------------------
	XApplication.prototype.taskPluginAction = function(widgetId, ref, value) {
		if (value !== null)
			value = this.vars.parse(value);
		this.uiappl.pluginAction(widgetId, ref, value);
	};

	// taskPluginObject -----------------------------------------------
	XApplication.prototype.taskPluginObject = function(id, ref, value) {
		this.uiappl.pluginObject(id, ref, value);
	};

	// taskPopupError -------------------------------------------------
	XApplication.prototype.taskPopupError = function(id, value) {
		var a, c1, d;

		d = UILib.newXData("popup");
		d.setStringAttr("id", "_popup-error-" + id);
		d.setStringAttr("templateId", "popup-error");
		c1 = d.addChild("value");
		c1.setStringAttr("id", "text");
		if (value !== null)
			c1.setContent(this.parseText(value));
		else
			c1.setContent(this.vars.getVar("popup-error-default"));
		c1 = d.addChild("task");
		c1.setStringAttr("id", "continue");
		a = new XAction(d);
		this.taskPopupShow(a);
	};

	// taskPopupShow --------------------------------------------------
	XApplication.prototype.taskPopupShow = function(a) {
		var res, tmpl, xwdata;

		this.putData("popup", a.values);
		tmpl = this.uiappl.getTemplate(a.templateId);
		res = XTemplate.getWidget(this, tmpl);
		xwdata = this.uiappl.getXWData(res);
		this.uiappl.popupShow(a.id, XLib.createWidget(xwdata));
	};

	// taskSetVar -----------------------------------------------------
	XApplication.prototype.taskSetVar = function(id, ref) {
		this.putVar(id, ref);
	};

	// taskSwitch -----------------------------------------------------
	XApplication.prototype.taskSwitch = function(taskid, xswitch) {
		var block, i, v;

		if ((v = this.vars.getVar(xswitch.var)) === null)
			throw new Error("invalid var: " + xswitch.var);
		if ((block = xswitch.blocks.get(v)) === null)
			if ((block = xswitch.blocks.get("")) === null)
				return(true);
		for(i = 0;i < block.length;++i)
			if (!block[i].run(taskid, this))
				return(false);
		return(true);
	};

	// taskTemplateUpdate ---------------------------------------------
	XApplication.prototype.taskTemplateUpdate = function(id, ref, templateId, visible) {
		this.uiappl.templateUpdate(id, ref, templateId, visible);
	};

	// taskWidget -----------------------------------------------------
	XApplication.prototype.taskWidget = function(xwa) {
		this.uiappl.widgetExec(xwa);
	};

	// taskWidgetClean ------------------------------------------------
	XApplication.prototype.taskWidgetClean = function(cardId, widgetId) {
		this.uiappl.widgetClean(cardId, widgetId);
	};

	// taskWidgetConfig -----------------------------------------------
	XApplication.prototype.taskWidgetConfig = function(id, ref, text) {
		this.uiappl.widgetConfig(id, ref, this.parseText(text));
	};

	// taskWidgetSet --------------------------------------------------
	XApplication.prototype.taskWidgetSet = function(cardId, widgetId, value) {
		this.uiappl.widgetSet(cardId, widgetId, value);
	};

	// taskWidgetShow -------------------------------------------------
	XApplication.prototype.taskWidgetShow = function(cardId, widgetId, visible) {
		this.uiappl.widgetShow(cardId, widgetId, visible);
	};

	// waitingStart ---------------------------------------------------
	XApplication.prototype.waitingStart = function() {
		this.uiappl.waitingStart();
	};

	// waitingStop ----------------------------------------------------
	XApplication.prototype.waitingStop = function() {
		this.uiappl.waitingStop();
	};

	// Constructor ----------------------------------------------------
	{
		this.uiappl = uiappl;
		this.vars = new XVars(this);
		this.forms = UILib.getXFormMap();
	}
};

// isModeDebug --------------------------------------------------------
XApplication.isModeDebug = function() {
	if (XApplication.mode === null)
		return(true);
	return(UILib.stringsEquals(XApplication.mode, "debug"));
};

// logMsg -------------------------------------------------------------
XApplication.logMsg = function(msg) {
	console.log(msg);
};

// logWarning ---------------------------------------------------------
XApplication.logWarning = function(msg) {
	console.log(msg);
};

// logXml -------------------------------------------------------------
XApplication.logXml = function(type, xdoc) {
	var c, msg, n, text;

	if (XString.equals(type, "http-response: javascript-error"))
		return;
	XApplication.logMsg("-----" + type + "-----");
	if (XString.equals(type, "http-request: javascript-error")) {
		if ((c = xdoc.getFirstChildWithTag("msg")) !== null) {
			text = c.getTextContent();
			if ((n = text.indexOf('\n')) === -1)
				msg = text;
			else
				msg = text.substring(0, n);
			XApplication.logMsg("  " + msg);
		}
	}
	xdoc.write();
};
/* ====================================================================	*
 * XAttr 																*
 * ====================================================================	*/

var XAttr = function(name, value) {

	// Constructor ----------------------------------------------------
	{
		this.name = name;
		this.value = value;
	}
};
/* ====================================================================	*
 * XBorder 																*
 * ====================================================================	*/

var XBorder = function(color, thickness) {

	// getWidth -------------------------------------------------------
	XBorder.prototype.getWidth = function() {
		return(this.thickness.getWidth());
	};

	// getHeight ------------------------------------------------------
	XBorder.prototype.getHeight = function() {
		return(this.thickness.getHeight());
	};

	// Constructor ----------------------------------------------------
	{
		this.color = color;
		this.thickness = thickness;
	}
};
/* ====================================================================	*
 * XWidget 																*
 * ====================================================================	*/

var XWidget = function(data) {

	this.prefWidth = 0;
	this.children = null;
	this.uiwt = null;
	this.cwidth = 0;
	this.cheight = 0;
	this.vwidth = 0;
	this.vheight = 0;

	// getEdgeWidth ---------------------------------------------------
	XWidget.prototype.getEdgeWidth = function() {
		return(this.margin.getWidth() + this.border.getWidth() + this.padding.getWidth());
	};

	// getEdgeHeight --------------------------------------------------
	XWidget.prototype.getEdgeHeight = function() {
		return(this.margin.getHeight() + this.border.getHeight() + this.padding.getHeight());
	};

	// getMaxX --------------------------------------------------------
	XWidget.prototype.getMaxX = function() {
		return(this.margin.right + this.border.thickness.right + this.padding.right);
	};

	// getX -----------------------------------------------------------
	XWidget.prototype.getX = function() {
		return(this.margin.left + this.border.thickness.left + this.padding.left);
	};

	// getY -----------------------------------------------------------
	XWidget.prototype.getY = function() {
		return(this.margin.top + this.border.thickness.top + this.padding.top);
	};

	// hasChildren ----------------------------------------------------
	XWidget.prototype.hasChildren = function() {
		return(this.children !== null);
	};

	// layoutInit -----------------------------------------------------
	XWidget.prototype.layoutInit = function() {
		this.prefWidth = this.minWidth;
	};

	// layoutSetLocation ----------------------------------------------
	XWidget.prototype.layoutSetLocation = function(vx, vy, anchor) {
		var dh, dw;

		if (anchor === null)
			anchor = this.anchor;
		if ((dw = this.vwidth - this.cwidth) < 0)
			dw = 0;
		if ((dh = this.vheight - this.cheight) < 0)
			dh = 0;
		switch(anchor) {
		case "northwest":
			break;
		case "north":
			vx += dw / 2;
			break;
		case "northeast":
			vx += dw;
			break;
		case "west":
			vy += dh / 2;
			break;
		case "center":
			vx += dw / 2;
			vy += dh / 2;
			break;
		case "east":
			vx += dw;
			vy += dh / 2;
			break;
		case "southwest":
			vy += dh;
			break;
		case "south":
			vx += dw / 2;
			vy += dh;
			break;
		case "southeast":
			vx += dw;
			vy += dh;
			break;
		case "golden":
			vx += dw / 2;
			vy += dh /(XLib.GoldenRatio + 1);
			break;
		default:
			throw new Error("invalid anchor: " + anchor);
		}
		this.uiwt.layoutSetLocation(vx, vy, this.cwidth, this.cheight);
	};

	// layoutSetMinimumWidth ------------------------------------------
	XWidget.prototype.layoutSetMinimumWidth = function() {
		if (this.hidden) {
			this.cwidth = 0;
			return;
		}
		this.cwidth = this.minWidth + this.getEdgeWidth();
	};

	// layoutSetMinimumHeight -----------------------------------------
	XWidget.prototype.layoutSetMinimumHeight = function() {
		if (this.hidden) {
			this.cheight = 0;
			return;
		}
		this.cheight = this.minHeight + this.getEdgeHeight();
	};

	// layoutSetResponsive --------------------------------------------
	XWidget.prototype.layoutSetResponsive = function(key) {
	};

	// layoutSetViewportWidth -----------------------------------------
	XWidget.prototype.layoutSetViewportWidth = function(vwidth) {
		var w;

		if (this.hidden) {
			this.cwidth = 0;
			return;
		}
		if ((this.vwidth = vwidth) < this.cwidth)
			throw new Error("invalid vwidth: [" + this.id + "] " + vwidth + " < " + this.cwidth);
		if (this.expandHor === 0) {
			if ((this.cwidth = this.prefWidth + this.getEdgeWidth()) > vwidth)
				this.cwidth = vwidth;
			return;
		}
		if ((w = UILib.mathFloor(vwidth * this.expandHor / 100)) > this.cwidth)
			this.cwidth = w;
	};

	// layoutSetViewportHeight ----------------------------------------
	XWidget.prototype.layoutSetViewportHeight = function(vheight) {
		var h;

		if (this.hidden) {
			this.cheight = 0;
			return;
		}
		if ((this.vheight = vheight) < this.cheight)
			throw new Error("invalid vheight: " + vheight + " < " + this.cheight);
		if ((h = UILib.mathFloor(vheight * this.expandVer / 100)) > this.cheight)
			this.cheight = h;
	};

	// Constructor ----------------------------------------------------
	{
		var sz, v;

		if (data === null)
			return;
		this.id = data.getAttrString("id", null);
		this.type = data.getTag();
		this.props = new XProps(data);
		this.margin = data.getEdge("margin");
		this.border = data.getBorder("border");
		this.padding = data.getEdge("padding");
		sz = data.getSize("size");
		this.minWidth = sz.width;
		this.minHeight = sz.height;
		if ((v = data.getPercent("expand")) > 0)
			this.expandHor = this.expandVer = v;
		else {
			this.expandHor = data.getPercent("expandh");
			this.expandVer = data.getPercent("expandv");
		}
		this.anchor = data.getAnchor("anchor");
		this.fg = data.getColor("fg");
		this.bg = data.getColor("bg");
		this.responsive = data.getStringArray("responsive");
		this.visibility = true;
		this.hidden = false;
	}
};
/* ====================================================================	*
 * XButton 																*
 * ====================================================================	*/

var XButton = function(xdata) {

	XWidget.call(this, xdata);

	// Constructor ----------------------------------------------------
	{
		if ((this.text = xdata.getAttrString("text", null)) === null)
			throw new Error("attr text not found...");
		this.action = xdata.getAttrString("action", null);
		this.data = xdata.getAttrString("data", null);
	}
};
/* ====================================================================	*
 * XCheckBox 															*
 * ====================================================================	*/

var XCheckBox = function(data) {

	XHorFlow.call(this, data);

	// Constructor ----------------------------------------------------
	{
		this.text = data.getAttrString("text", null);
		this.key = data.getAttrString("key", null);
		this.group = data.getAttrString("group", null);
	}
};

// create -------------------------------------------------------------
XCheckBox.create = function(data) {
	var attrs, bg, c1, c2, fg, font, group, i, id, key, name, text, value, xd;

	xd = UILib.newXData("checkbox");
	bg = fg = font = group = id = key = text = null;
	attrs = data.getAttributes();
	for(i = 0;i < attrs.length;++i) {
		name = attrs[i].key;
		value = attrs[i].value;
		switch(name) {
		case "id":
			id = value;
			xd.setStringAttr(name, value);
			break;
		case "margin":
		case "border":
		case "padding":
		case "size":
		case "expand":
		case "expandh":
		case "expandv":
		case "anchor":
			xd.setStringAttr(name, value);
			break;
		case "bg":
			bg = value;
			xd.setStringAttr(name, value);
			break;
		case "fg":
			fg = value;
			xd.setStringAttr(name, value);
			break;
		case "align":
		case "gap":
			xd.setStringAttr(name, value);
			break;
		case "text":
			text = value;
			break;
		case "font":
			font = value;
			break;
		case "key":
			key = value;
			break;
		case "group":
			group = value;
			break;
		}
	}
	if (id === null)
		throw new Error("attribute id not found");
	if (text === null)
		throw new Error("attribute text not found");
	if (key === null)
		throw new Error("attribute key not found");
	if (group === null)
		throw new Error("attribute group not found");
	if (font === null)
		font = data.getFont();
	xd.setStringAttr("text", text);
	xd.setStringAttr("key", key);
	xd.setStringAttr("group", group);
	c1 = xd.addChild("checkbox-item");
	c1.setStringAttr("id", id + "_checkbox");
	c1.setStringAttr("key", key);
	c1.setStringAttr("group", group);
	c2 = xd.addChild("label");
	c2.setStringAttr("id", id + "_label");
	c2.setStringAttr("text", text);
	if (bg !== null) {
		c1.setStringAttr("bg", bg);
		c2.setStringAttr("bg", bg);
	}
	if (fg !== null)
		c2.setStringAttr("fg", fg);
	if (font !== null)
		c2.setStringAttr("font", font);
	return(new XCheckBox(data.toUIXWData(xd, true)));
};
/* ====================================================================	*
 * XCheckBoxItem 														*
 * ====================================================================	*/

var XCheckBoxItem = function(data) {

	XWidget.call(this, data);

	// Constructor ----------------------------------------------------
	{
		if ((this.key = data.getAttrString("key", null)) === null)
			throw new Error("attribute 'key' not found");
		if ((this.group = data.getAttrString("group", null)) === null)
			throw new Error("attribute 'group' not found");
	}
};
/* ====================================================================	*
 * XComboBox 															*
 * ====================================================================	*/

var XComboBox = function(data) {

	XWidget.call(this, data);

	this.options = null;

	// initOptions ----------------------------------------------------
	XComboBox.prototype.initOptions = function(children) {
		var d, i, key, value;

		this.options = UILib.newArray(children.length, null);
		for(i = 0;i < children.length;++i) {
			d = children[i];
			if ((key = d.getAttrString("key", null)) === null)
				throw new Error("attribute 'key' not found");
			if ((value = d.getAttrString("value", null)) === null)
				throw new Error("attribute 'value' not found");
			this.options[i] = new XKeyValue(key, value);
		}
	};

	// Constructor ----------------------------------------------------
	{
		this.initOptions(data.getAllChildElements("option"));
	}
};
/* ====================================================================	*
 * XDate 																*
 * ====================================================================	*/

var XDate = function(data) {

	XFlow.call(this, data);

	// Constructor ----------------------------------------------------
	{
		this.yearOffset = 2006;
	}
};

// create -------------------------------------------------------------
XDate.create = function(data) {
	var attrs, bg, cb, i, id, months, name, opt, value, xd, year;

	xd = UILib.newXData("date");
	id = bg = null;
	attrs = data.getAttributes();
	for(i = 0;i < attrs.length;++i) {
		name = attrs[i].key;
		value = attrs[i].value;
		switch(name) {
		case "id":
			id = value;
			xd.setStringAttr("id", id);
			break;
		case "margin":
		case "border":
		case "padding":
		case "size":
		case "expand":
		case "expandh":
		case "expandv":
		case "anchor":
			xd.setStringAttr(name, value);
			break;
		case "bg":
			bg = value;
			xd.setStringAttr(name, value);
			break;
		case "align":
		case "gap":
		case "interline":
			xd.setStringAttr(name, value);
			break;
		}
	}
	if (id === null)
		throw new Error("attribute id not found");
	cb = xd.addChild("combobox");
	if (bg !== null)
		cb.setStringAttr("bg", bg);
	cb.setStringAttr("id", id + "_day");
	for(i = 1;i <= 31;++i) {
		opt = cb.addChild("option");
		opt.setIntAttr("key", i);
		opt.setIntAttr("value", i);
	}
	months = UILib.newArray(12, null);
	months[0] = "Janvier";
	months[1] = "Février";
	months[2] = "Mars";
	months[3] = "Avril";
	months[4] = "Mai";
	months[5] = "Juin";
	months[6] = "Juillet";
	months[7] = "Août";
	months[8] = "Septembre";
	months[9] = "Octobre";
	months[10] = "Novembre";
	months[11] = "Décembre";
	cb = xd.addChild("combobox");
	if (bg !== null)
		cb.setStringAttr("bg", bg);
	if (bg !== null)
		cb.setStringAttr("bg", bg);
	cb.setStringAttr("id", id + "_month");
	for(i = 1;i <= 12;++i) {
		opt = cb.addChild("option");
		opt.setIntAttr("key", i);
		opt.setStringAttr("value", months[i -1]);
	}
	cb = xd.addChild("combobox");
	if (bg !== null)
		cb.setStringAttr("bg", bg);
	cb.setStringAttr("id", id + "_year");
	for(i = 0;i < 23;++i) {
		year = 2006 + i;
		opt = cb.addChild("option");
		opt.setIntAttr("key", year);
		opt.setIntAttr("value", year);
	}
	cb = xd.addChild("combobox");
	if (bg !== null)
		cb.setStringAttr("bg", bg);
	cb.setStringAttr("id", id + "_hour");
	for(i = 0;i <= 23;++i) {
		opt = cb.addChild("option");
		opt.setIntAttr("key", i);
		opt.setStringAttr("value", i + " h");
	}
	cb = xd.addChild("combobox");
	if (bg !== null)
		cb.setStringAttr("bg", bg);
	cb.setStringAttr("id", id + "_min");
	for(i = 0;i < 60;i += 5) {
		opt = cb.addChild("option");
		opt.setIntAttr("key", i);
		opt.setStringAttr("value", i + " min");
	}
	cb = xd.addChild("combobox");
	if (bg !== null)
		cb.setStringAttr("bg", bg);
	cb.setStringAttr("id", id + "_tz");
	opt = cb.addChild("option");
	opt.setStringAttr("key", "local");
	opt.setStringAttr("value", "zzz");
	opt = cb.addChild("option");
	opt.setStringAttr("key", "utc");
	opt.setStringAttr("value", "UTC");
	return(new XDate(data.toUIXWData(xd, true)));
};
/* ====================================================================	*
 * XDisplay 															*
 * ====================================================================	*/

var XDisplay = function(data) {

	XWidget.call(this, data);

	this.appl = null;
	this.index = 0;
	this.files = null;
	this.plugin = null;

	// doAction -------------------------------------------------------
	XDisplay.prototype.doAction = function(ref, value) {
		var i, path, xd;

		switch(ref) {
		case "init":
			xd = this.appl.getData("lecture-info");
			if ((this.files = XFile.getList(xd, "toc")) === null)
				throw new Error("element not found: lecture-info:toc");
			this.index = 0;
			this.plugin.start();
			if (this.files.length > 0)
				this.plugin.setImage(this.files[this.index].path);
			break;
		case "set-slide":
			if (!UILib.stringsMatch(value, "cache:"))
				throw new Error("invalid value: " + value);
			path = value.substring(6);
			for(i = 0;i < this.files.length;++i) {
				if (XString.equals(path, this.files[i].path)) {
					if (this.index === i)
						return;
					this.index = i;
					this.plugin.setImage(path);
					this.plugin.repaint();
					return;
				}
			}
			throw new Error("invalid value: " + path);
		default:
			throw new Error("not yet implemented: " + ref);
		}
	};

	// getCurrentSlide ------------------------------------------------
	XDisplay.prototype.getCurrentSlide = function() {
		var n, path;

		path = this.files[this.index].path;
		if ((n = path.indexOf('/')) === -1)
			throw new Error("invalid path: " + path);
		return(path.substring(n + 1));
	};

	// sendPointerPositions -------------------------------------------
	XDisplay.prototype.sendPointerPositions = function(value) {
		this.appl.taskPluginObject("transmitter", "pointer", value);
	};

	// setCanvasSize --------------------------------------------------
	XDisplay.prototype.setCanvasSize = function(canvasw, canvash) {
	};

	// startRecording -------------------------------------------------
	XDisplay.prototype.startRecording = function() {
		this.plugin.startRecording();
	};

	// stopPointerDragging --------------------------------------------
	XDisplay.prototype.stopPointerDragging = function() {
		this.plugin.stopPointerDragging(true);
	};

	// stopRecording --------------------------------------------------
	XDisplay.prototype.stopRecording = function() {
		this.plugin.stopRecording();
	};

	// Constructor ----------------------------------------------------
	{
		if ((this.actionLabel1 = data.getAttrString("action1", null)) === null)
			throw new Error("attribute 'action1' not found...");
		if ((this.actionLabel2 = data.getAttrString("action2", null)) === null)
			throw new Error("attribute 'action2' not found...");
	}
};
/* ====================================================================	*
 * XDropDown 															*
 * ====================================================================	*/

var XDropDown = function(xdata) {

	XWidget.call(this, xdata);

	// layoutInit -----------------------------------------------------
	XDropDown.prototype.layoutInit = function() {
		this.minWidth = this.uilabel.getMinimumWidth();
		this.prefWidth = this.uilabel.getPrefWidth();
		this.popup.layoutInit();
	};

	// layoutSetViewportWidth -----------------------------------------
	XDropDown.prototype.layoutSetViewportWidth = function(vwidth) {
		var sz;

		XWidget.prototype.layoutSetViewportWidth.call(this, vwidth);
		sz = this.uilabel.format(this.cwidth - this.getEdgeWidth());
		if (this.expandHor === 0)
			this.cwidth = sz.width + this.getEdgeWidth();
		this.minHeight = sz.height;
	};

	// Constructor ----------------------------------------------------
	{
		this.uilabel = xdata.getLabel(false);
		this.action = xdata.getAttrString("action", null);
		this.active = xdata.getAttrString("active", null);
		this.over = xdata.getAttrString("over", null);
		this.data = xdata.getAttrString("data", null);
		this.popup = XLib.createWidget(xdata.getFirstChild());
	}
};
/* ====================================================================	*
 * XEdge 																*
 * ====================================================================	*/

var XEdge = function(top, right, bottom, left) {

	// isEmpty --------------------------------------------------------
	XEdge.prototype.isEmpty = function() {
		return(this.top === 0 && this.right === 0 && this.bottom === 0 && this.left === 0);
	};

	// getWidth -------------------------------------------------------
	XEdge.prototype.getWidth = function() {
		return(this.left + this.right);
	};

	// getHeight ------------------------------------------------------
	XEdge.prototype.getHeight = function() {
		return(this.top + this.bottom);
	};

	// Constructor ----------------------------------------------------
	{
		this.top = top;
		this.right = right;
		this.bottom = bottom;
		this.left = left;
	}
};
/* ====================================================================	*
 * XElementAttrs 														*
 * ====================================================================	*/

var XElementAttrs = function(attrs) {

	// copyWidgetAttrs ------------------------------------------------
	XElementAttrs.prototype.copyWidgetAttrs = function(dst) {
		var a, i;

		for(i = 0;i < this.attrs.length;++i) {
			a = this.attrs[i];
			switch(a.key) {
			case "id":
			case "margin":
			case "border":
			case "padding":
			case "size":
			case "expand":
			case "expandh":
			case "expandv":
			case "anchor":
			case "bg":
				dst.setStringAttr(a.key, a.value);
				break;
			}
		}
	};

	// getString ------------------------------------------------------
	XElementAttrs.prototype.getString = function(name) {
		var s;

		if ((s = this.searchString(name, null)) === null)
			throw new Error("attribute not found: " + name);
		return(s);
	};

	// getValue -------------------------------------------------------
	XElementAttrs.prototype.getValue = function(name, data) {
		var s;

		if ((s = this.searchString(name, null)) !== null)
			return(s);
		return(data.getDefault(name));
	};

	// searchString ---------------------------------------------------
	XElementAttrs.prototype.searchString = function(name, def) {
		var i;

		for(i = 0;i < this.attrs.length;++i)
			if (XString.equals(this.attrs[i].key, name))
				return(this.attrs[i].value);
		return(def);
	};

	// Constructor ----------------------------------------------------
	{
		this.attrs = attrs;
	}
};

// create -------------------------------------------------------------
XElementAttrs.create = function(attrs) {
	return(new XElementAttrs(attrs));
};
/* ====================================================================	*
 * XFlow 																*
 * ====================================================================	*/

var XFlow = function(data) {

	XWidget.call(this, data);

	this.nleft = 0;

	// init -----------------------------------------------------------
	XFlow.prototype.init = function() {
		var i;

		for(i = 0;i < this.children.length;++i)
			if (this.children[i].expandHor === 0)
				this.children[i].expandHor = 100;
	};

	// layoutInit -----------------------------------------------------
	XFlow.prototype.layoutInit = function() {
		var i;

		for(i = 0;i < this.children.length;++i)
			this.children[i].layoutInit();
	};

	// layoutSetLocation ----------------------------------------------
	XFlow.prototype.layoutSetLocation = function(vx, vy, anchor) {
		var i, wt, x, y;

		XWidget.prototype.layoutSetLocation.call(this, vx, vy, anchor);
		x = this.getX();
		y = this.getY();
		switch(this.align) {
		case "center":
			x += this.nleft / 2;
			break;
		case "right":
			x += this.nleft;
			break;
		}
		for(i = 0;i < this.children.length;++i) {
			wt = this.children[i];
			wt.layoutSetLocation(x + this.cellx[i], y + this.celly[i], wt.anchor);
		}
	};

	// layoutSetMinimumWidth ------------------------------------------
	XFlow.prototype.layoutSetMinimumWidth = function() {
		var i, w, wt;

		XWidget.prototype.layoutSetMinimumWidth.call(this);
		w = 0;
		for(i = 0;i < this.children.length;++i) {
			wt = this.children[i];
			wt.layoutSetMinimumWidth();
			w += wt.prefWidth + wt.getEdgeWidth();
		}
		w += this.gap *(this.children.length -1);
		if (this.prefWidth < w)
			this.prefWidth = w;
		w = 0;
		for(i = 0;i < this.children.length;++i) {
			wt = this.children[i];
			if (w < wt.cwidth)
				w = wt.cwidth;
		}
		w += this.getEdgeWidth();
		if (this.cwidth < w)
			this.cwidth = w;
	};

	// layoutSetMinimumHeight -----------------------------------------
	XFlow.prototype.layoutSetMinimumHeight = function() {
		var h, l;

		XWidget.prototype.layoutSetMinimumHeight.call(this);
		l = this.children.length -1;
		h = this.celly[l] + this.cellh[l];
		h += this.getEdgeHeight();
		if (this.cheight < h)
			this.cheight = h;
	};

	// layoutSetViewportWidth -----------------------------------------
	XFlow.prototype.layoutSetViewportWidth = function(vwidth) {
		var dw, h, i, j, off, rowh, w, wmin, wt, x, y;

		XWidget.prototype.layoutSetViewportWidth.call(this, vwidth);
		wmin = this.nleft = this.cwidth - this.getEdgeWidth();
		off = rowh = 0;
		x = y = 0;
		for(i = 0;i < this.children.length;++i) {
			wt = this.children[i];
			w = wt.prefWidth + wt.getEdgeWidth();
			if (i === off)
				dw = w;
			else
				dw = w + this.gap;
			if (dw > wmin) {
				if (i > 0) {
					if (this.nleft > wmin)
						this.nleft = wmin > 0?wmin:0;
					for(j = off;j < i;++j)
						this.cellh[j] = rowh;
					x = 0;
					y += rowh + this.interline;
					off = i;
					wmin = this.cwidth - this.getEdgeWidth();
					rowh = 0;
				}
				if (w > wmin) {
					w = wmin;
					this.nleft = 0;
				}
			}
			this.cellx[i] = x;
			this.celly[i] = y;
			this.cellw[i] = w;
			wt.layoutSetViewportWidth(w);
			wt.layoutSetMinimumHeight();
			wmin -= dw;
			x += w + this.gap;
			if ((h = wt.cheight) > rowh)
				rowh = h;
		}
		if (this.nleft > wmin)
			this.nleft = wmin > 0?wmin:0;
		for(i = off;i < this.children.length;++i)
			this.cellh[i] = rowh;
	};

	// layoutSetViewportHeight ----------------------------------------
	XFlow.prototype.layoutSetViewportHeight = function(vheight) {
		var i;

		XWidget.prototype.layoutSetViewportHeight.call(this, vheight);
		for(i = 0;i < this.children.length;++i)
			this.children[i].layoutSetViewportHeight(this.cellh[i]);
	};

	// Constructor ----------------------------------------------------
	{
		this.children = XLib.createChildrenWidget(data);
		this.align = data.getAlign("align");
		this.gap = data.getInt("gap");
		this.interline = data.getInt("interline");
		this.cellx = UILib.newArray(this.children.length, 0);
		this.celly = UILib.newArray(this.children.length, 0);
		this.cellw = UILib.newArray(this.children.length, 0);
		this.cellh = UILib.newArray(this.children.length, 0);
		this.init();
	}
};
/* ====================================================================	*
 * XFont 																*
 * ====================================================================	*/

var XFont = function(alias, fg, fontname) {

	// Constructor ----------------------------------------------------
	{
		this.alias = alias;
		this.fg = fg;
		this.fontname = fontname;
	}
};
/* ====================================================================	*
 * XForm 																*
 * ====================================================================	*/

var XForm = function(data) {

	// getInput -------------------------------------------------------
	XForm.prototype.getInput = function(fieldId) {
		var i;

		for(i = 0;i < this.entries.length;++i)
			if (UILib.stringsEquals(fieldId, this.entries[i].id))
				return(this.entries[i]);
		throw new Error("invalid input id: " + fieldId);
	};

	// getType --------------------------------------------------------
	XForm.prototype.getType = function(fieldId) {
		var i;

		for(i = 0;i < this.entries.length;++i)
			if (UILib.stringsEquals(fieldId, this.entries[i].id))
				return(this.entries[i].type);
		for(i = 0;i < this.labels.length;++i)
			if (UILib.stringsEquals(fieldId, this.labels[i].id))
				return("label");
		for(i = 0;i < this.links.length;++i)
			if (UILib.stringsEquals(fieldId, this.links[i].id))
				return("link");
		throw new Error("no fieldId: " + fieldId);
	};

	// read -----------------------------------------------------------
	XForm.prototype.read = function(data, uiappl) {
		var i, input, ok;

		ok = true;
		for(i = 0;i < this.entries.length;++i) {
			input = this.entries[i];
			if (!input.read(this.id, data, uiappl))
				ok = false;
		}
		if (!ok)
			data.setStringAttr("status", "error");
		return(ok);
	};

	// Constructor ----------------------------------------------------
	{
		var i, j, l, n1, n2, n3, off, part, parts;

		this.id = data.getString("id");
		parts = data.getChildren("part");
		if (parts.length === 0) {
			part = new XFormPart(data, -1);
			this.entries = part.entries;
			this.labels = part.labels;
			this.links = part.links;
			return;
		}
		n1 = n2 = n3 = 0;
		l = UILib.newArray(parts.length, null);
		for(i = 0;i < l.length;++i) {
			l[i] = new XFormPart(parts[i], i + 1);
			n1 += l[i].entries.length;
			n2 += l[i].labels.length;
			n3 += l[i].links.length;
		}
		off = 0;
		this.entries = UILib.newArray(n1, null);
		for(i = 0;i < l.length;++i) {
			part = l[i];
			for(j = 0;j < part.entries.length;++j)
				this.entries[off ++] = l[i].entries[j];
		}
		off = 0;
		this.labels = UILib.newArray(n2, null);
		for(i = 0;i < l.length;++i) {
			part = l[i];
			for(j = 0;j < part.labels.length;++j)
				this.labels[off ++] = l[i].labels[j];
		}
		off = 0;
		this.links = UILib.newArray(n3, null);
		for(i = 0;i < l.length;++i) {
			part = l[i];
			for(j = 0;j < part.links.length;++j)
				this.links[off ++] = l[i].links[j];
		}
	}
};
/* ====================================================================	*
 * XFormItem 															*
 * ====================================================================	*/

var XFormItem = function(type, id) {

	// Constructor ----------------------------------------------------
	{
		this.type = type;
		this.id = id;
	}
};
/* ====================================================================	*
 * XFormPart 															*
 * ====================================================================	*/

var XFormPart = function(data, partId) {

	// Constructor ----------------------------------------------------
	{
		var children, i;

		children = data.getChildren("input");
		this.entries = UILib.newArray(children.length, null);
		for(i = 0;i < this.entries.length;++i)
			this.entries[i] = new XInput(children[i], partId);
		children = data.getChildren("label");
		this.labels = UILib.newArray(children.length, null);
		for(i = 0;i < children.length;++i)
			this.labels[i] = new XFormItem("label", children[i].getString("id"));
		children = data.getChildren("link");
		this.links = UILib.newArray(children.length, null);
		for(i = 0;i < children.length;++i)
			this.links[i] = new XFormItem("link", children[i].getString("id"));
	}
};
/* ====================================================================	*
 * XFormSetTree 														*
 * ====================================================================	*/

var XFormSetTree = function(data) {

	// Constructor ----------------------------------------------------
	{
		this.cardId = data.getString("cardId");
		this.formId = data.getString("formId");
		this.inputId = data.getString("inputId");
		this.partId = data.getAttrInt("partId", -1);
		this.xref = data.getString("xref");
	}
};
/* ====================================================================	*
 * XGrid 																*
 * ====================================================================	*/

var XGrid = function(data) {

	XWidget.call(this, data);

	this.ncols = 0;
	this.nrows = 0;
	this.gapw = 0;
	this.gaph = 0;
	this.colsmin = null;
	this.rowsmin = null;
	this.colseq = null;
	this.rowseq = null;
	this.weightm = null;
	this.weighth = null;
	this.weightv = null;
	this.cells = null;
	this.tcells = null;
	this.colsw = null;
	this.rowsh = null;

	// getGridShrinkList ----------------------------------------------
	XGrid.prototype.getGridShrinkList = function(data) {
		var i, l, list;

		l = data.getAllChildElements("shrink");
		if (l.length === 0)
			return(null);
		list = UILib.newArray(l.length, null);
		for(i = 0;i < list.length;++i)
			list[i] = new XGridShrink(l[i], this.layout.gapw, this.layout.gaph);
		return(list);
	};

	// insertExtra ----------------------------------------------------
	XGrid.prototype.insertExtra = function(tab, xtra) {
		var i;

		for(i = 0;i < tab.length;++i)
			tab[i] += xtra[i];
	};

	// layoutInit -----------------------------------------------------
	XGrid.prototype.layoutInit = function() {
		var i;

		for(i = 0;i < this.children.length;++i)
			this.children[i].layoutInit();
	};

	// layoutSetLocation ----------------------------------------------
	XGrid.prototype.layoutSetLocation = function(vx, vy, anchor) {
		var c, tx, ty, x, y;

		XWidget.prototype.layoutSetLocation.call(this, vx, vy, anchor);
		ty = this.getY();
		for(y = 0;y < this.nrows;++y) {
			tx = this.getX();
			for(x = 0;x < this.ncols;++x) {
				if ((c = this.cells[y][x]) !== null && c.wt !== null)
					c.wt.layoutSetLocation(tx, ty, c.anchor);
				tx += this.colsw[x] + this.gapw;
			}
			ty += this.rowsh[y] + this.gaph;
		}
	};

	// layoutSetMinimumWidth ------------------------------------------
	XGrid.prototype.layoutSetMinimumWidth = function() {
		var a, i, w;

		XWidget.prototype.layoutSetMinimumWidth.call(this);
		for(i = 0;i < this.children.length;++i)
			this.children[i].layoutSetMinimumWidth();
		this.colsw = UILib.cloneIntArray(this.colsmin);
		if ((w = this.packCols(this.colsw, false) + this.getEdgeWidth()) < this.cwidth) {
			a = this.splitExtra(this.weighth, this.cwidth - w);
			this.insertExtra(this.colsw, a);
		} else
			this.cwidth = w;
		this.prefWidth = this.cwidth - this.getEdgeWidth();
		if (this.minWidth < this.cwidth)
			this.minWidth = this.cwidth - this.getEdgeWidth();
	};

	// layoutSetMinimumHeight -----------------------------------------
	XGrid.prototype.layoutSetMinimumHeight = function() {
		var a, h, i;

		XWidget.prototype.layoutSetMinimumHeight.call(this);
		for(i = 0;i < this.children.length;++i)
			this.children[i].layoutSetMinimumHeight();
		this.rowsh = UILib.cloneIntArray(this.rowsmin);
		if ((h = this.packRows() + this.getEdgeHeight()) < this.cheight) {
			a = this.splitExtra(this.weightv, this.cheight - h);
			this.insertExtra(this.rowsh, a);
		} else
			this.cheight = h;
	};

	// layoutSetResponsive --------------------------------------------
	XGrid.prototype.layoutSetResponsive = function(key) {
		var c, i, x, y;

		for(y = 0;y < this.nrows;++y)
			for(x = 0;x < this.ncols;++x)
				if ((c = this.cells[y][x]) !== null && c.wt !== null)
					c.wt.layoutSetResponsive(key);
		if (key !== null) {
			for(i = 0;i < this.shrinkl.length;++i) {
				if (UILib.stringsEquals(this.shrinkl[i].key, key)) {
					this.setLayout(i);
					return;
				}
			}
		}
		this.setLayout( -1);
	};

	// layoutSetViewportWidth -----------------------------------------
	XGrid.prototype.layoutSetViewportWidth = function(vwidth) {
		var a, c, delta, i, n, w, wmax, wpref, x, y;

		if ((this.vwidth = vwidth) < this.cwidth)
			throw new Error("invalid vwidth:[" + this.type + "] " + vwidth + " < " + this.cwidth);
		if (vwidth > this.cwidth) {
			if (this.expandHor === 0)
				wmax = vwidth;
			else
				if ((wmax = vwidth * this.expandHor / 100) < this.cwidth)
					wmax = this.cwidth;
			if (UILib.stringsEquals(this.weightm, "percent") && wmax > this.cwidth) {
				this.setColsWidth(wmax - this.cwidth);
				this.cwidth = wmax;
			}
			if (wmax > this.cwidth) {
				delta = UILib.newArray(this.ncols, 0);
				wpref = this.packPrefCols(delta);
				if ((w = wpref + this.getEdgeWidth()) > wmax)
					w = wmax;
				if ((n = w - this.cwidth) > 0) {
					a = this.splitPrefWidth(delta, n);
					this.insertExtra(this.colsw, a);
					this.cwidth = w;
				}
			}
			if (wmax > this.cwidth && this.expandHor > 0) {
				a = this.splitExtra(this.weighth, wmax - this.cwidth);
				this.insertExtra(this.colsw, a);
				this.cwidth = wmax;
			}
		}
		this.prefWidth = this.cwidth - this.getEdgeWidth();
		for(y = 0;y < this.nrows;++y) {
			for(x = 0;x < this.ncols;++x) {
				if ((c = this.cells[y][x]) === null || c.wt === null)
					continue;
				w = this.gapw *(c.cellw -1);
				for(i = 0;i < c.cellw;++i)
					w += this.colsw[x + i];
				c.wt.layoutSetViewportWidth(w);
			}
		}
	};

	// layoutSetViewportHeight ----------------------------------------
	XGrid.prototype.layoutSetViewportHeight = function(vheight) {
		var a, c, h, i, x, y;

		if ((this.vheight = vheight) < this.cheight)
			throw new Error("invalid vheight: " + vheight + " < " + this.cheight);
		if ((h = vheight * this.expandVer / 100) > this.cheight) {
			a = this.splitExtra(this.weightv, h - this.cheight);
			this.insertExtra(this.rowsh, a);
			this.cheight = h;
		}
		for(y = 0;y < this.nrows;++y) {
			for(x = 0;x < this.ncols;++x) {
				if ((c = this.cells[y][x]) === null || c.wt === null)
					continue;
				h = this.gaph *(c.cellh -1);
				for(i = 0;i < c.cellh;++i)
					h += this.rowsh[y + i];
				c.wt.layoutSetViewportHeight(h);
			}
		}
	};

	// packCols -------------------------------------------------------
	XGrid.prototype.packCols = function(widths, preferred) {
		var a, c, delta, i, j, max, off, v, w, wh, x, y;

		for(x = 0;x < this.ncols;++x) {
			max = 0;
			for(y = 0;y < this.nrows;++y) {
				if ((c = this.tcells[y][x]) === null || c.wt === null)
					continue;
				if (c.cellw > 1)
					continue;
				if (preferred)
					v = c.wt.prefWidth + c.wt.getEdgeWidth();
				else
					v = c.wt.cwidth;
				if (v > max)
					max = v;
			}
			if (widths[x] < max)
				widths[x] = max;
		}
		for(x = 0;x < this.ncols;++x) {
			for(y = 0;y < this.nrows;++y) {
				if ((c = this.tcells[y][x]) === null)
					continue;
				if (c.wt === null || c.cellw === 1)
					continue;
				if (preferred)
					v = c.wt.prefWidth + c.wt.getEdgeWidth();
				else
					v = c.wt.cwidth;
				off = x - c.cellw + 1;
				for(i = off;i < x;++i)
					v -= widths[i] + this.gapw;
				if (widths[x] >= v)
					continue;
				delta = v - widths[x];
				wh = UILib.newArray(c.cellw, 0);
				for(i = 0;i < c.cellw;++i)
					wh[i] = this.weighth[off + i];
				a = this.splitExtra(wh, delta);
				for(i = 0;i < c.cellw;++i)
					widths[off + i] += a[i];
			}
		}
		if (this.colseq !== null) {
			for(i = 0;i < this.colseq.length;++i) {
				max = 0;
				a = this.colseq[i];
				for(j = 0;j < a.length;++j)
					if ((v = widths[a[j]]) > max)
						max = v;
				for(j = 0;j < a.length;++j)
					widths[a[j]] = max;
			}
		}
		w = 0;
		for(x = 0;x < this.ncols;++x)
			w += widths[x];
		w += this.gapw *(this.ncols -1);
		return(w);
	};

	// packPrefCols ---------------------------------------------------
	XGrid.prototype.packPrefCols = function(delta) {
		var i, v, w, widths;

		widths = UILib.newArray(this.ncols, 0);
		w = this.packCols(widths, true);
		for(i = 0;i < this.ncols;++i)
			if ((v = widths[i] - this.colsw[i]) > 0)
				delta[i] = v;
		return(w);
	};

	// packRows -------------------------------------------------------
	XGrid.prototype.packRows = function() {
		var a, c, delta, h, i, j, max, off, v, wv, x, y;

		for(y = 0;y < this.nrows;++y) {
			max = 0;
			for(x = 0;x < this.ncols;++x) {
				if ((c = this.tcells[y][x]) === null || c.wt === null)
					continue;
				if (c.cellh > 1)
					continue;
				if ((v = c.wt.cheight) > max)
					max = v;
			}
			if (this.rowsh[y] < max)
				this.rowsh[y] = max;
		}
		for(y = 0;y < this.nrows;++y) {
			for(x = 0;x < this.ncols;++x) {
				if ((c = this.tcells[y][x]) === null || c.wt === null)
					continue;
				if (c.cellh === 1)
					continue;
				v = c.wt.cheight;
				off = y - c.cellh + 1;
				for(i = off;i < y;++i)
					v -= this.rowsh[i] + this.gaph;
				if (this.rowsh[y] >= v)
					continue;
				delta = v - this.rowsh[y];
				wv = UILib.newArray(c.cellh, 0);
				for(i = 0;i < c.cellh;++i)
					wv[i] = this.weightv[off + i];
				a = this.splitExtra(wv, delta);
				for(i = 0;i < c.cellh;++i)
					this.rowsh[off + i] += a[i];
			}
		}
		if (this.rowseq !== null) {
			for(i = 0;i < this.rowseq.length;++i) {
				max = 0;
				a = this.rowseq[i];
				for(j = 0;j < a.length;++j)
					if ((v = this.rowsh[a[j]]) > max)
						max = v;
				for(j = 0;j < a.length;++j)
					this.rowsh[a[i]] = max;
			}
		}
		h = 0;
		for(i = 0;i < this.nrows;++i)
			h += this.rowsh[i];
		return(h + this.gaph *(this.nrows -1));
	};

	// reset ----------------------------------------------------------
	XGrid.prototype.reset = function() {
		var c, x, y;

		this.tcells = UILib.newArray(this.nrows, this.ncols, null);
		for(y = 0;y < this.nrows;++y)
			for(x = 0;x < this.ncols;++x)
				if ((c = this.cells[y][x]) !== null)
					this.tcells[y + c.cellh -1][x + c.cellw -1] = c;
	};

	// setColsWidth ---------------------------------------------------
	XGrid.prototype.setColsWidth = function(delta) {
		var i, ncolsw, w;

		w = 0;
		for(i = 0;i < this.ncols;++i)
			w += this.colsw[i];
		ncolsw = this.splitExtra(this.weighth, w + delta);
		XLib.adjustMinimumWidth(this.colsw, ncolsw);
		for(i = 0;i < this.ncols;++i)
			this.colsw[i] = ncolsw[i];
	};

	// setDefaultLayout -----------------------------------------------
	XGrid.prototype.setDefaultLayout = function() {
		this.ncols = this.layout.ncols;
		this.nrows = this.layout.nrows;
		this.gapw = this.layout.gapw;
		this.gaph = this.layout.gaph;
		this.colsmin = this.layout.colsmin;
		this.rowsmin = this.layout.rowsmin;
		this.colseq = this.layout.colseq;
		this.rowseq = this.layout.rowseq;
		this.weightm = this.layout.weightm;
		this.weighth = this.layout.weighth;
		this.weightv = this.layout.weightv;
		this.cells = this.layout.cells;
		this.children = this.layout.children;
		this.colsw = UILib.newArray(this.ncols, 0);
		this.rowsh = UILib.newArray(this.nrows, 0);
		this.reset();
	};

	// setLayout ------------------------------------------------------
	XGrid.prototype.setLayout = function(v) {
		var c, nparts, offy, r, ry, wt, x, y;

		if (this.ilayout === v)
			return;
		if ((this.ilayout = v) === -1) {
			this.setDefaultLayout();
			return;
		}
		r = this.shrinkl[this.ilayout];
		if (this.layout.nrows % r.nlines !== 0)
			throw new Error("can't split: " + this.layout.nrows + " % " + r.nlines + " != 0");
		nparts = this.layout.nrows / r.nlines;
		this.ncols = r.ncols;
		this.nrows = nparts * r.nrows;
		this.gapw = r.gapw;
		this.gaph = r.gaph;
		this.colsmin = r.colsmin;
		this.rowsmin = r.getRowsMin(nparts);
		this.colseq = r.colseq;
		this.rowseq = r.getRowsEq(nparts);
		this.weighth = r.weighth;
		this.weightv = r.getWeightv(nparts);
		this.cells = UILib.newArray(this.nrows, this.ncols, null);
		offy = ry = 0;
		for(y = 0;y < this.nrows;++y) {
			for(x = 0;x < this.ncols;++x) {
				if ((c = r.cells[ry][x]) === null)
					continue;
				wt = this.layout.search(offy, c.offset);
				this.cells[y][x] = new XGridCell(c.cellw, c.cellh, wt, c.anchor);
			}
			if (++ry === r.nrows) {
				ry = 0;
				offy += r.nlines;
			}
		}
		this.reset();
	};

	// splitExtra -----------------------------------------------------
	XGrid.prototype.splitExtra = function(weight, delta) {
		var i, n, nleft, nw, off, xtra;

		xtra = UILib.newArray(weight.length, 0);
		nleft = UILib.mathFloor(delta);
		nw = 0;
		for(i = 0;i < weight.length;++i) {
			if (weight[i] === 0.0)
				continue;
			n = UILib.mathFloor(weight[i] * delta);
			xtra[i] = n;
			nleft -= n;
			++nw;
		}
		if (nleft === 0)
			return(xtra);
		if ((off = UILib.mathFloor((nw - nleft) / 2)) < 0)
			off = 0;
		for(;;) {
			for(i = 0;i < weight.length;++i) {
				if (weight[i] === 0.0)
					continue;
				if (off > 0) {
					--off;
					continue;
				}
				++xtra[i];
				if (--nleft === 0)
					return(xtra);
			}
		}
	};

	// splitPrefWidth -------------------------------------------------
	XGrid.prototype.splitPrefWidth = function(delta, dx) {
		var i, total, w, weight;

		w = 0;
		for(i = 0;i < delta.length;++i)
			w += delta[i];
		if (w === dx)
			return(delta);
		total = w;
		weight = UILib.newArray(delta.length, 0);
		for(i = 0;i < delta.length;++i)
			weight[i] = delta[i];
		for(i = 0;i < delta.length;++i)
			weight[i] /= total;
		return(this.splitExtra(weight, dx));
	};

	// Constructor ----------------------------------------------------
	{
		this.layout = new XGridLayout(data);
		this.shrinkl = this.getGridShrinkList(data);
		this.ilayout = -1;
		this.setDefaultLayout();
	}
};
/* ====================================================================	*
 * XGridCell 															*
 * ====================================================================	*/

var XGridCell = function(cellw, cellh, wt, anchor) {

	// Constructor ----------------------------------------------------
	{
		this.cellw = cellw;
		this.cellh = cellh;
		this.wt = wt;
		this.anchor = anchor;
	}
};
/* ====================================================================	*
 * XGridLayout 															*
 * ====================================================================	*/

var XGridLayout = function(data) {

	// getGridCells ---------------------------------------------------
	XGridLayout.prototype.getGridCells = function(data) {
		var cdata, i, ldata, list, sz, wdata, wt;

		ldata = data.getAllChildElements("cell");
		list = UILib.newArray(ldata.length, null);
		for(i = 0;i < list.length;++i) {
			cdata = ldata[i];
			sz = cdata.getSize("size");
			wt = null;
			if ((wdata = cdata.getFirstChild()) !== null)
				wt = XLib.createWidget(wdata);
			list[i] = new XGridCell(sz.width, sz.height, wt, null);
		}
		return(list);
	};

	// search ---------------------------------------------------------
	XGridLayout.prototype.search = function(offy, offset) {
		var c, n, x, y;

		if (offset === -1)
			return(null);
		n = 0;
		for(y = offy;y < this.nrows;++y) {
			for(x = 0;x < this.ncols;++x) {
				if ((c = this.cells[y][x]) === null)
					continue;
				if (c.wt === null)
					continue;
				if (offset === n)
					return(c.wt);
				++n;
			}
		}
		throw new Error("rows=" + offy + ",offset=" + offset);
	};

	// Constructor ----------------------------------------------------
	{
		var c, i, list, map, nchildren, offx, offy, wt, x, y;

		if ((this.ncols = data.getInt("ncols")) === 0)
			throw new Error("invalid ncols: 0");
		if ((this.nrows = data.getInt("nrows")) === 0)
			throw new Error("invalid nrows: 0");
		if ((this.gapw = data.getAttrInt("gap", -1)) !== -1)
			this.gaph = this.gapw;
		else {
			this.gapw = data.getInt("gapw");
			this.gaph = data.getInt("gaph");
		}
		this.caption = data.getColor("caption");
		this.even = data.getColor("even");
		this.odd = data.getColor("odd");
		this.colsmin = data.getIntArray("colsmin", this.ncols);
		this.rowsmin = data.getIntArray("rowsmin", this.nrows);
		this.colseq = data.getIntArrayList("colseq", "=", this.ncols);
		this.rowseq = data.getIntArrayList("rowseq", "=", this.nrows);
		this.lweighth = data.getAttrString("weighth", null);
		this.lweightv = data.getAttrString("weightv", null);
		this.weightm = data.getString("weightm");
		this.weighth = data.getPercentArray("weighth", this.lweighth, this.ncols);
		this.weightv = data.getPercentArray("weightv", this.lweightv, this.nrows);
		this.cells = UILib.newArray(this.nrows, this.ncols, null);
		list = this.getGridCells(data);
		nchildren = 0;
		offx = offy = 0;
		map = UILib.newArray(this.nrows, this.ncols, false);
		for(i = 0;i < list.length;++i) {
			if (offy === this.nrows)
				throw new Error("too many cells in grid [nrows=" + this.nrows + ",ncols=" + this.ncols + "]");
			c = list[i];
			this.cells[offy][offx] = c;
			if (offx + c.cellw > this.ncols)
				c.cellw = this.ncols - offx;
			if (offy + c.cellh > this.nrows)
				c.cellh = this.nrows - offy;
			for(y = 0;y < c.cellh;++y)
				for(x = 0;x < c.cellw;++x)
					map[offy + y][offx + x] = true;
			for(;;) {
				if (++offx === this.ncols) {
					offx = 0;
					if (++offy === this.nrows)
						break;
				}
				if (map[offy][offx] === false)
					break;
			}
			if (c.wt !== null)
				++nchildren;
		}
		this.children = UILib.newArray(nchildren, null);
		nchildren = 0;
		for(i = 0;i < list.length;++i)
			if ((wt = list[i].wt) !== null)
				this.children[nchildren ++] = wt;
		if (this.caption !== UILib.NO_COLOR) {
			for(x = 0;x < this.ncols;++x)
				if ((c = this.cells[0][x]) !== null)
					if ((wt = c.wt) !== null)
						wt.bg = this.caption;
			offy = 1;
		} else
			offy = 0;
		if (this.even !== UILib.NO_COLOR)
			for(y = offy;y < this.nrows;y += 2)
				for(x = 0;x < this.ncols;++x)
					if ((c = this.cells[y][x]) !== null)
						if ((wt = c.wt) !== null)
							wt.bg = this.even;
		if (this.odd !== UILib.NO_COLOR)
			for(y = offy + 1;y < this.nrows;y += 2)
				for(x = 0;x < this.ncols;++x)
					if ((c = this.cells[y][x]) !== null)
						if ((wt = c.wt) !== null)
							wt.bg = this.odd;
	}
};
/* ====================================================================	*
 * XGridShrink 															*
 * ====================================================================	*/

var XGridShrink = function(data, gapw, gaph) {

	// getGridShrinkCells ---------------------------------------------
	XGridShrink.prototype.getGridShrinkCells = function(data) {
		var anchor, i, l, list, offset, sz;

		l = data.getAllChildElements("cell");
		list = UILib.newArray(l.length, null);
		for(i = 0;i < list.length;++i) {
			offset = l[i].getInt("offset");
			sz = l[i].getSize("size");
			anchor = l[i].getAnchor("anchor");
			list[i] = new XGridShrinkCell(sz.width, sz.height, offset, anchor);
		}
		return(list);
	};

	// getRowsEq ------------------------------------------------------
	XGridShrink.prototype.getRowsEq = function(nparts) {
		var a, dst, i, j, k, n, off, src;

		if (this.rowseq === null)
			return(null);
		a = UILib.newArray(this.rowseq.length, null);
		for(i = 0;i < this.rowseq.length;++i) {
			src = this.rowseq[i];
			a[i] = dst = UILib.newArray(src.length * nparts, 0);
			n = off = 0;
			for(j = 0;j < nparts;++j) {
				for(k = 0;k < src.length;++k)
					dst[off ++] = n + src[k];
				n += this.nrows;
			}
		}
		return(a);
	};

	// getRowsMin -----------------------------------------------------
	XGridShrink.prototype.getRowsMin = function(nparts) {
		var a, i, j, off;

		a = UILib.newArray(this.nrows * nparts, 0);
		off = 0;
		for(i = 0;i < nparts;++i)
			for(j = 0;j < this.nrows;++j)
				a[off ++] = this.rowsmin[j];
		return(a);
	};

	// getWeightv -----------------------------------------------------
	XGridShrink.prototype.getWeightv = function(nparts) {
		var a, i, j, off;

		a = UILib.newArray(this.nrows * nparts, 0);
		off = 0;
		for(i = 0;i < nparts;++i)
			for(j = 0;j < this.nrows;++j)
				a[off ++] = this.weightv[j];
		for(i = 0;i < a.length;++i)
			a[i] /= nparts;
		return(a);
	};

	// Constructor ----------------------------------------------------
	{
		var c, i, list, map, offx, offy, x, y;

		if ((this.key = data.getAttrString("key", null)) === null)
			throw new Error("attr key not found...");
		if ((this.nlines = data.getAttrInt("nlines", -1)) === -1)
			throw new Error("attr nlines not found...");
		if ((this.ncols = data.getAttrInt("ncols", -1)) === -1)
			throw new Error("attr ncols not found...");
		if ((this.nrows = data.getAttrInt("nrows", -1)) === -1)
			throw new Error("attr nrows not found...");
		if ((this.gapw = data.getAttrInt("gap", -1)) !== -1)
			this.gaph = this.gapw;
		else {
			this.gapw = data.getAttrInt("gapw", gapw);
			this.gaph = data.getAttrInt("gaph", gaph);
		}
		this.colsmin = data.getIntArray("colsmin", this.ncols);
		this.rowsmin = data.getIntArray("rowsmin", this.nrows);
		this.colseq = data.getIntArrayList("colseq", "=", this.ncols);
		this.rowseq = data.getIntArrayList("rowseq", "=", this.nrows);
		this.lweighth = data.getAttrString("weighth", null);
		this.lweightv = data.getAttrString("weightv", null);
		this.weighth = data.getPercentArray("weighth", this.lweighth, this.ncols);
		this.weightv = data.getPercentArray("weightv", this.lweightv, this.nrows);
		this.cells = UILib.newArray(this.nrows, this.ncols, null);
		list = this.getGridShrinkCells(data);
		offx = offy = 0;
		map = UILib.newArray(this.nrows, this.ncols, false);
		for(i = 0;i < list.length;++i) {
			if (offy === this.nrows)
				throw new Error("too many cells in grid");
			c = list[i];
			this.cells[offy][offx] = c;
			if (offx + c.cellw > this.ncols)
				c.cellw = this.ncols - offx;
			if (offy + c.cellh > this.nrows)
				c.cellh = this.nrows - offy;
			for(y = 0;y < c.cellh;++y)
				for(x = 0;x < c.cellw;++x)
					map[offy + y][offx + x] = true;
			for(;;) {
				if (++offx === this.ncols) {
					offx = 0;
					if (++offy === this.nrows)
						break;
				}
				if (map[offy][offx] === false)
					break;
			}
		}
	}
};
/* ====================================================================	*
 * XGridShrinkCell 														*
 * ====================================================================	*/

var XGridShrinkCell = function(cellw, cellh, offset, anchor) {

	// Constructor ----------------------------------------------------
	{
		this.cellw = cellw;
		this.cellh = cellh;
		this.offset = offset;
		this.anchor = anchor;
	}
};
/* ====================================================================	*
 * XHorFlow 															*
 * ====================================================================	*/

var XHorFlow = function(data) {

	XWidget.call(this, data);

	this.nleft = 0;
	this.rowh = 0;

	// layoutInit -----------------------------------------------------
	XHorFlow.prototype.layoutInit = function() {
		var dw, i, mw, pw, w, wt;

		mw = pw = 0;
		for(i = 0;i < this.children.length;++i) {
			wt = this.children[i];
			dw = wt.getEdgeWidth();
			wt.layoutInit();
			wt.layoutSetMinimumWidth();
			mw += wt.minWidth + dw;
			this.prefw[i] = wt.prefWidth + dw;
			pw += this.prefw[i];
		}
		w = this.gap *(this.children.length -1);
		mw += w;
		pw += w;
		if (this.minWidth < mw)
			this.minWidth = mw;
		if (this.prefWidth < pw)
			this.prefWidth = pw;
	};

	// layoutSetLocation ----------------------------------------------
	XHorFlow.prototype.layoutSetLocation = function(vx, vy, anchor) {
		var i, wt, x, y;

		XWidget.prototype.layoutSetLocation.call(this, vx, vy, anchor);
		x = this.getX();
		y = this.getY();
		switch(this.align) {
		case "center":
			x += this.nleft / 2;
			break;
		case "right":
			x += this.nleft;
			break;
		}
		for(i = 0;i < this.children.length;++i) {
			wt = this.children[i];
			wt.layoutSetLocation(x, y, wt.anchor);
			x += this.colsw[i] + this.gap;
		}
	};

	// layoutSetMinimumWidth ------------------------------------------
	XHorFlow.prototype.layoutSetMinimumWidth = function() {
		var i;

		XWidget.prototype.layoutSetMinimumWidth.call(this);
		for(i = 0;i < this.children.length;++i)
			this.children[i].layoutSetMinimumWidth();
	};

	// layoutSetMinimumHeight -----------------------------------------
	XHorFlow.prototype.layoutSetMinimumHeight = function() {
		var i;

		XWidget.prototype.layoutSetMinimumHeight.call(this);
		for(i = 0;i < this.children.length;++i)
			this.children[i].layoutSetMinimumHeight();
	};

	// layoutSetViewportWidth -----------------------------------------
	XHorFlow.prototype.layoutSetViewportWidth = function(vwidth) {
		var dw, h, i, minw, n, wt, xtra;

		XWidget.prototype.layoutSetViewportWidth.call(this, vwidth);
		xtra = UILib.newArray(this.children.length, 0);
		for(i = 0;i < this.children.length;++i)
			xtra[i] = this.prefw[i] - this.children[i].cwidth;
		this.nleft = this.cwidth - this.getEdgeWidth() -(this.children.length -1) * this.gap;
		for(i = 0;i < this.children.length;++i) {
			this.colsw[i] = this.children[i].cwidth;
			this.nleft -= this.colsw[i];
		}
		while(this.nleft > 0) {
			n = 0;
			minw = 0;
			for(i = 0;i < this.children.length;++i) {
				if (xtra[i] > 0) {
					++n;
					if (minw === 0)
						minw = xtra[i];
					else
						if (xtra[i] < minw)
							minw = xtra[i];
				}
			}
			if (n === 0)
				break;
			if ((dw = UILib.mathFloor(this.nleft / n)) > minw)
				dw = minw;
			if (dw === 0)
				dw = 1;
			for(i = 0;i < this.children.length;++i) {
				if (xtra[i] > 0) {
					xtra[i] -= dw;
					this.colsw[i] += dw;
					if ((this.nleft -= dw) === 0)
						break;
				}
			}
		}
		if (this.nleft > 0 && this.weight !== -1) {
			this.colsw[this.weight] += this.nleft;
			this.nleft = 0;
		}
		this.rowh = 0;
		for(i = 0;i < this.children.length;++i) {
			wt = this.children[i];
			wt.layoutSetViewportWidth(this.colsw[i]);
			if ((h = wt.minHeight + wt.getEdgeHeight()) > this.rowh)
				this.rowh = h;
		}
		this.minHeight = this.rowh;
	};

	// layoutSetViewportHeight ----------------------------------------
	XHorFlow.prototype.layoutSetViewportHeight = function(vheight) {
		var i;

		XWidget.prototype.layoutSetViewportHeight.call(this, vheight);
		for(i = 0;i < this.children.length;++i)
			this.children[i].layoutSetViewportHeight(this.rowh);
	};

	// Constructor ----------------------------------------------------
	{
		this.children = XLib.createChildrenWidget(data);
		this.weight = data.getAttrInt("weight", -1);
		this.gap = data.getInt("gap");
		this.align = data.getAlign("align");
		this.prefw = UILib.newArray(this.children.length, 0);
		this.colsw = UILib.newArray(this.children.length, 0);
	}
};
/* ====================================================================	*
 * XHttpRequest 														*
 * ====================================================================	*/

var XHttpRequest = function(data) {

	// Constructor ----------------------------------------------------
	{
		this.requestId = data.getString("requestId");
		this.formId = data.getAttrString("formId", null);
		this.taskId = data.getAttrString("taskId", null);
		this.dataId = data.getAttrString("dataId", null);
		this.values = XAction.getValues("root", data);
	}
};
/* ====================================================================	*
 * XHttpRequests 														*
 * ====================================================================	*/

var XHttpRequests = function(data) {

	// notifyRequest --------------------------------------------------
	XHttpRequests.prototype.notifyRequest = function(appl, requestId, res) {
		var i, taskid;

		for(i = 0;i < this.list.length;++i)
			if (UILib.stringsEquals(requestId, this.list[i].requestId))
				break;
		if (i === this.list.length)
			throw new Error("requestId not found: " + requestId);
		if (UILib.stringsEquals(res, "http-error"))
			this.error = true;
		this.results[i] = res;
		for(i = 0;i < this.results.length;++i)
			if (this.results[i] === null)
				return;
		appl.waitingStop();
		if (!this.error) {
			appl.taskExecTask(this.id + "#http-ok");
			return;
		}
		taskid = this.id + "#http-error";
		if (!appl.taskExist(taskid))
			taskid = "#http-error";
		appl.taskExecTask(taskid);
	};

	// Constructor ----------------------------------------------------
	{
		var children, i;

		this.id = data.getString("id");
		children = data.getChildren("http-request");
		this.list = UILib.newArray(children.length, null);
		this.results = UILib.newArray(children.length, null);
		for(i = 0;i < this.list.length;++i)
			this.list[i] = new XHttpRequest(children[i]);
		this.error = false;
	}
};
/* ====================================================================	*
 * XIconLink 															*
 * ====================================================================	*/

var XIconLink = function(data) {

	XHorFlow.call(this, data);

	// Constructor ----------------------------------------------------
	{
		this.ref = data.getAttrString("ref", null);
		this.iconsize = data.getAttrString("iconsize", null);
		this.text = data.getAttrString("text", null);
	}
};

// create -------------------------------------------------------------
XIconLink.create = function(data) {
	var attrs, bg, c1, c2, fg, font, i, iconsz, id, name, ref, text, value, xd;

	xd = UILib.newXData("iconlink");
	bg = fg = font = iconsz = id = ref = text = null;
	attrs = data.getAttributes();
	for(i = 0;i < attrs.length;++i) {
		name = attrs[i].key;
		value = attrs[i].value;
		switch(name) {
		case "id":
			id = value;
			xd.setStringAttr(name, value);
			break;
		case "margin":
		case "border":
		case "padding":
		case "size":
		case "expand":
		case "expandh":
		case "expandv":
		case "anchor":
			xd.setStringAttr(name, value);
			break;
		case "bg":
			bg = value;
			xd.setStringAttr(name, value);
			break;
		case "fg":
			fg = value;
			xd.setStringAttr(name, value);
			break;
		case "align":
		case "gap":
			xd.setStringAttr(name, value);
			break;
		case "ref":
			ref = value;
			break;
		case "iconsize":
			iconsz = value;
			break;
		case "text":
			text = value;
			break;
		case "font":
			font = value;
			break;
		}
	}
	if (id === null)
		throw new Error("attribute id not found");
	if (ref === null)
		throw new Error("attribute ref not found");
	if (text === null)
		throw new Error("attribute text not found");
	if (font === null)
		font = data.getFont();
	xd.setStringAttr("ref", ref);
	xd.setStringAttr("text", text);
	if (iconsz !== null)
		xd.setStringAttr("iconsize", iconsz);
	c1 = xd.addChild("picture");
	c1.setStringAttr("id", id + "_picture");
	c1.setStringAttr("ref", ref);
	if (iconsz !== null)
		c1.setStringAttr("size", iconsz);
	c2 = xd.addChild("link");
	c2.setStringAttr("id", id + "_link");
	c2.setStringAttr("text", text);
	if (bg !== null) {
		c1.setStringAttr("bg", bg);
		c2.setStringAttr("bg", bg);
	}
	if (fg !== null)
		c2.setStringAttr("fg", fg);
	if (font !== null)
		c2.setStringAttr("font", font);
	return(new XIconLink(data.toUIXWData(xd, true)));
};
/* ====================================================================	*
 * XIfAction 															*
 * ====================================================================	*/

var XIfAction = function(data) {

	// Constructor ----------------------------------------------------
	{
		var children, i, n1, n2;

		this.var = data.getString("var");
		this.value = data.getAttrString("value", "");
		children = data.getAllChildren();
		for(n1 = 0;n1 < children.length;++n1)
			if (XString.equals(children[n1].getTag(), "else"))
				break;
		this.ifblock = UILib.newArray(n1, null);
		for(i = 0;i < n1;++i)
			this.ifblock[i] = new XAction(children[i]);
		if (n1 === children.length)
			n2 = 0;
		else
			n2 = children.length - ++n1;
		this.elseblock = UILib.newArray(n2, null);
		for(i = 0;i < n2;++i)
			this.elseblock[i] = new XAction(children[n1 ++]);
	}
};
/* ====================================================================	*
 * XTableColumn 														*
 * ====================================================================	*/

var XTableColumn = function(el) {

	// setConfig ------------------------------------------------------
	XTableColumn.prototype.setConfig = function(cfg) {
		var child;

		child = cfg.addChild("column");
		child.setStringAttr("type", this.type);
		child.setStringAttr("ref", this.ref);
		child.setStringAttr("label", this.label);
		child.setStringAttr("imgsz", this.imgsz);
		child.setStringAttr("action", this.action);
		child.setStringAttr("data", this.data);
		child.setStringAttr("align", this.align);
		child.setStringAttr("font", this.font);
		child.setStringAttr("fontid", this.fontid);
		child.setStringAttr("caption", this.caption);
		child.setStringAttr("group", this.group);
	};

	// Constructor ----------------------------------------------------
	{
		this.type = el.getString("type");
		if ((this.ref = el.getAttrString("ref", null)) === null)
			this.label = el.getString("label");
		this.imgsz = el.getAttrString("imgsz", null);
		this.action = el.getAttrString("action", null);
		this.data = el.getAttrString("data", null);
		this.align = el.getAttrString("align", null);
		this.font = el.getAttrString("font", "plain");
		this.fontid = el.getAttrString("fontid", null);
		this.caption = el.getAttrString("caption", null);
		this.group = el.getAttrString("group", null);
	}
};

// getColumns ---------------------------------------------------------
XTableColumn.getColumns = function(parent, tag) {
	var children, cols, i;

	children = parent.getChildren(tag);
	cols = UILib.newArray(children.length, null);
	for(i = 0;i < cols.length;++i)
		cols[i] = new XTableColumn(children[i]);
	return(cols);
};
/* ====================================================================	*
 * XInput 																*
 * ====================================================================	*/

var XInput = function(data, partId) {

	// getOptions -----------------------------------------------------
	XInput.prototype.getOptions = function(children) {
		var d, i, k, kvs, v;

		kvs = UILib.newArray(children.length, null);
		for(i = 0;i < kvs.length;++i) {
			d = children[i];
			if ((k = d.getAttrString("key", null)) === null)
				throw new Error("attribute 'key' not found");
			if ((v = d.getAttrString("value", null)) === null)
				throw new Error("attribute 'value' not found");
			kvs[i] = new XKeyValue(k, v);
		}
		return(kvs);
	};

	// add ------------------------------------------------------------
	XInput.prototype.add = function(root, value, options) {
		if (UILib.stringEmpty(value)) {
			if (!this.required)
				return(true);
			this.error(root, "required");
			return(false);
		}
		if (this.lowercase)
			value = value.toLowerCase();
		if (options !== null)
			if (!this.checkOption(root, value))
				return(false);
		if (!UILib.stringEmpty(value))
			root.addChildWithContent(this.id, value);
		return(true);
	};

	// checkOption ----------------------------------------------------
	XInput.prototype.checkOption = function(root, value) {
		var i;

		for(i = 0;i < this.options.length;++i)
			if (UILib.stringsEquals(value, this.options[i].key))
				return(true);
		this.error(root, "invalid value: " + value);
		return(false);
	};

	// error ----------------------------------------------------------
	XInput.prototype.error = function(root, value) {
		var el;

		el = root.addChild(this.id);
		el.setStringAttr("error", value);
	};

	// read -----------------------------------------------------------
	XInput.prototype.read = function(formId, root, uiappl) {
		var i, l, value;

		switch(this.type) {
		case "hidden":
			value = uiappl.getKey(formId + "#" + this.id);
			if (this.required && value === null) {
				this.error(root, "required");
				return(false);
			}
			if (value === null)
				return(true);
			return(this.add(root, value, null));
		case "checkbox":
			l = uiappl.getFormFields(this.type, formId, this.id);
			if (this.required && l.length === 0) {
				this.error(root, "required");
				return(false);
			}
			for(i = 0;i < l.length;++i) {
				if (!this.checkOption(root, l[i]))
					return(false);
				root.addChildWithContent(this.id, l[i]);
			}
			return(true);
		default:
			value = uiappl.getFormField(this.type, formId, this.id, null);
			return(this.add(root, value, this.options));
		}
	};

	// Constructor ----------------------------------------------------
	{
		var el, v;

		this.partId = partId;
		this.type = data.getString("type");
		this.id = data.getString("id");
		this.text = data.getAttrString("text", null);
		this.required = data.getAttrBool("required", false);
		this.comment = data.getAttrString("comment", null);
		this.options = null;
		switch(this.type) {
		case "checkbox":
		case "radiobutton":
			this.layout = data.getAttrString("layout", "flow");
			this.props = data.getAttrString("props", null);
			if ((v = data.getAttrString("value", null)) === null) {
				this.options = this.getOptions(data.getChildren("option"));
				break;
			}
			this.options = UILib.newArray(1, null);
			this.options[0] = new XKeyValue(this.id, v);
			break;
		case "combobox":
			this.options = this.getOptions(data.getChildren("option"));
			break;
		case "explorer":
			this.attrs = XAttrs.createFromXData(data, "explorer,tree,table");
			el = XElement.getInstanceFromXData(data);
			this.tcols = XTableColumn.getColumns(el, "table-column");
			break;
		case "date":
			break;
		case "hidden":
			break;
		case "textfield":
			this.lowercase = data.getAttrBool("lowercase", false);
			break;
		case "password":
			break;
		case "textarea":
			this.nrows = data.getAttrInt("nrows", 0);
			this.nospellcheck = data.getAttrBool("nospellcheck", false);
			break;
		case "tree":
			this.template = data.getAttrString("template", null);
			this.props = data.getString("props");
			break;
		default:
			throw new Error("invalid type: " + this.type);
		}
	}
};
/* ====================================================================	*
 * XKeyValue 															*
 * ====================================================================	*/

var XKeyValue = function(key, value) {

	// Constructor ----------------------------------------------------
	{
		this.key = key;
		this.value = value;
	}
};
/* ====================================================================	*
 * XLabel 																*
 * ====================================================================	*/

var XLabel = function(data) {

	XWidget.call(this, data);

	// layoutInit -----------------------------------------------------
	XLabel.prototype.layoutInit = function() {
		if (this.nowrap) {
			this.prefWidth = this.uilabel.getPrefWidth();
			if (this.minWidth === 0)
				this.minWidth = this.prefWidth;
			return;
		}
		this.minWidth = this.uilabel.getMinimumWidth();
		this.prefWidth = this.uilabel.getPrefWidth();
	};

	// layoutSetViewportWidth -----------------------------------------
	XLabel.prototype.layoutSetViewportWidth = function(vwidth) {
		var sz;

		XWidget.prototype.layoutSetViewportWidth.call(this, vwidth);
		sz = this.uilabel.format(this.cwidth - this.getEdgeWidth());
		if (this.expandHor === 0)
			this.cwidth = sz.width + this.getEdgeWidth();
		if (this.cwidth > vwidth)
			this.cwidth = vwidth;
		this.minHeight = sz.height;
	};

	// Constructor ----------------------------------------------------
	{
		this.nowrap = data.getBool("nowrap");
		this.uilabel = data.getLabel(this.nowrap);
	}
};
/* ====================================================================	*
 * XLabelButton 														*
 * ====================================================================	*/

var XLabelButton = function(xdata) {

	XWidget.call(this, xdata);

	// layoutInit -----------------------------------------------------
	XLabelButton.prototype.layoutInit = function() {
		if (this.nowrap) {
			this.prefWidth = this.uilabel.getPrefWidth();
			if (this.minWidth === 0)
				this.minWidth = this.prefWidth;
			return;
		}
		this.minWidth = this.uilabel.getMinimumWidth();
		this.prefWidth = this.uilabel.getPrefWidth();
	};

	// layoutSetViewportWidth -----------------------------------------
	XLabelButton.prototype.layoutSetViewportWidth = function(vwidth) {
		var sz;

		if (this.hidden) {
			this.cwidth = 0;
			return;
		}
		XWidget.prototype.layoutSetViewportWidth.call(this, vwidth);
		sz = this.uilabel.format(this.cwidth - this.getEdgeWidth());
		if (this.expandHor === 0)
			this.cwidth = sz.width + this.getEdgeWidth();
		this.minHeight = sz.height;
	};

	// Constructor ----------------------------------------------------
	{
		this.nowrap = xdata.getBool("nowrap");
		this.uilabel = xdata.getLabel(false);
		this.action = xdata.getAttrString("action", null);
		this.active = xdata.getAttrString("active", null);
		this.over = xdata.getAttrString("over", null);
		this.data = xdata.getAttrString("data", null);
		this.disabled = false;
	}
};
/* ====================================================================	*
 * XScrollPane 															*
 * ====================================================================	*/

var XScrollPane = function(data) {

	XWidget.call(this, data);

	this.view = null;
	this.scrollw = 0;
	this.lvx = 0;
	this.lvy = 0;
	this.lanchor = null;

	// initChild ------------------------------------------------------
	XScrollPane.prototype.initChild = function(d) {
		this.children = UILib.newArray(1, null);
		this.children[0] = XLib.createWidget(d);
		this.view = this.children[0];
		if (UILib.stringsEquals(this.view.type, "widget"))
			return;
		if (d.getAttrString("border", null) !== null)
			throw new Error("attr border is not allowed in the scrollpane view tag");
		this.view.border.thickness = new XEdge(0, 0, 0, 0);
		this.view.expandHor = 100;
	};

	// layoutInit -----------------------------------------------------
	XScrollPane.prototype.layoutInit = function() {
		XWidget.prototype.layoutInit.call(this);
		this.view.layoutInit();
		this.scrollw = UILib.getVerticalScrollbarWidth(this);
	};

	// layoutSetLocation ----------------------------------------------
	XScrollPane.prototype.layoutSetLocation = function(vx, vy, anchor) {
		this.lvx = vx;
		this.lvy = vy;
		this.lanchor = anchor;
		XWidget.prototype.layoutSetLocation.call(this, vx, vy, anchor);
	};

	// layoutSetMinimumWidth ------------------------------------------
	XScrollPane.prototype.layoutSetMinimumWidth = function() {
		var w;

		XWidget.prototype.layoutSetMinimumWidth.call(this);
		this.view.layoutSetMinimumWidth();
		if ((w = this.view.cwidth + this.getEdgeWidth() + this.scrollw) > this.cwidth)
			this.cwidth = w;
		if ((w = this.view.prefWidth + this.scrollw) > this.prefWidth)
			this.prefWidth = w;
		if ((w = this.cwidth - this.getEdgeWidth()) > this.prefWidth)
			this.prefWidth = w;
		if (this.minWidth < this.cwidth)
			this.minWidth = this.cwidth - this.getEdgeWidth();
	};

	// layoutSetMinimumHeight -----------------------------------------
	XScrollPane.prototype.layoutSetMinimumHeight = function() {
		XWidget.prototype.layoutSetMinimumHeight.call(this);
		this.view.layoutSetMinimumHeight();
	};

	// layoutSetViewportWidth -----------------------------------------
	XScrollPane.prototype.layoutSetViewportWidth = function(vwidth) {
		XWidget.prototype.layoutSetViewportWidth.call(this, vwidth);
		this.view.layoutSetViewportWidth(this.cwidth - this.getEdgeWidth() - this.scrollw);
	};

	// layoutSetViewportHeight ----------------------------------------
	XScrollPane.prototype.layoutSetViewportHeight = function(vheight) {
		XWidget.prototype.layoutSetViewportHeight.call(this, vheight);
		if (this.expandVer === 0 && this.minHeight === 0)
			this.cheight = vheight;
		this.view.layoutSetViewportHeight(this.view.cheight);
	};

	// layoutUpdateUI -------------------------------------------------
	XScrollPane.prototype.layoutUpdateUI = function(visible) {
		this.view.layoutInit();
		if (!visible)
			return;
		this.view.layoutSetMinimumWidth();
		this.view.layoutSetViewportWidth(this.cwidth - this.getEdgeWidth() - this.scrollw);
		this.view.layoutSetMinimumHeight();
		this.view.layoutSetViewportHeight(this.view.cheight);
		this.layoutSetLocation(this.lvx, this.lvy, this.lanchor);
	};

	// setView --------------------------------------------------------
	XScrollPane.prototype.setView = function(view) {
		if (!view.margin.isEmpty())
			throw new Error("margin is not empty: " + view.id);
		if (!view.border.thickness.isEmpty())
			throw new Error("border is not empty: " + view.id);
		this.view = view;
		this.children[0] = view;
	};

	// Constructor ----------------------------------------------------
	{
		if (data === null)
			return;
		if (data.getAttrString("padding", null) !== null)
			throw new Error("attr padding is not allowed in the scrollpane tag");
		this.padding = new XEdge(0, 0, 0, 0);
		this.initChild(data.getFirstChild());
	}
};
/* ====================================================================	*
 * XTable 																*
 * ====================================================================	*/

var XTable = function(data, cfg) {

	XScrollPane.call(this, data);

	// getGroup -------------------------------------------------------
	XTable.prototype.getGroup = function() {
		var child, i, list;

		list = this.cfg.getChildren("column");
		for(i = 0;i < list.length;++i) {
			child = list[i];
			switch(child.getString("type")) {
			case "checkbox":
				return(child.getAttrString("group", null));
			}
		}
		throw new Error("checkbox not found...");
	};

	// getNewInstance -------------------------------------------------
	XTable.prototype.getNewInstance = function(appl, xref) {
		this.cfg.setStringAttr("props", "xref=" + xref);
		return(XLib.createWidget(this.cfg.toUIXWData(appl)));
	};

	// Constructor ----------------------------------------------------
	{
		this.cfg = cfg;
	}
};

// addCaption ---------------------------------------------------------
XTable.addCaption = function(xgrid, hd, col) {
	var cell, label;

	cell = xgrid.addChild("cell");
	label = cell.addChild("label");
	if (col.align === null)
		label.setStringAttr("align", "left");
	else
		label.setStringAttr("align", col.align);
	label.setStringAttr("expand", "100%");
	label.setStringAttr("padding", "8:5:8:5");
	label.setStringAttr("text", col.caption);
	label.setStringAttr("fg", hd.fg);
	if (col.fontid !== null)
		label.setStringAttr("fontid", hd.fontid);
	else
		label.setStringAttr("font", hd.font);
};

// addCheckbox --------------------------------------------------------
XTable.addCheckbox = function(xgrid, xrow, col) {
	var cell, checkbox, key;

	cell = xgrid.addChild("cell");
	checkbox = cell.addChild("checkbox-item");
	checkbox.setStringAttr("expand", "100%");
	checkbox.setStringAttr("padding", "8:5:8:5");
	checkbox.setStringAttr("group", col.group);
	if ((key = xrow.searchValue(col.ref)) === null)
		throw new Error("ref not found");
	checkbox.setStringAttr("key", key);
};

// addLabel -----------------------------------------------------------
XTable.addLabel = function(xgrid, xrow, col) {
	var cell, label, text;

	cell = xgrid.addChild("cell");
	label = cell.addChild("label");
	if (col.align === null)
		label.setStringAttr("align", "left");
	else
		label.setStringAttr("align", col.align);
	label.setStringAttr("expand", "100%");
	label.setStringAttr("padding", "8:5:8:5");
	if ((text = xrow.searchValue(col.ref)) === null)
		text = "";
	label.setStringAttr("text", text);
	if (col.fontid !== null)
		label.setStringAttr("fontid", col.fontid);
	else
		label.setStringAttr("font", col.font);
};

// addLink ------------------------------------------------------------
XTable.addLink = function(xgrid, xrow, col) {
	var cell, link, s;

	cell = xgrid.addChild("cell");
	link = cell.addChild("link");
	if (col.action !== null) {
		link.setStringAttr("action", col.action);
		if (col.data !== null)
			if ((s = xrow.searchValue(col.data)) !== null)
				link.setStringAttr("data", s);
	}
	if (col.align === null)
		link.setStringAttr("align", "left");
	else
		link.setStringAttr("align", col.align);
	link.setStringAttr("padding", "8:5:8:5");
	link.setStringAttr("expand", "100%");
	if (col.ref === null)
		s = col.label;
	else
		if ((s = xrow.searchValue(col.ref)) === null)
			s = "";
	link.setStringAttr("text", s);
	if (col.fontid !== null)
		link.setStringAttr("fontid", col.fontid);
	else
		link.setStringAttr("font", col.font);
};

// addPictbutton ------------------------------------------------------
XTable.addPictbutton = function(xgrid, xrow, col) {
	var cell, pb, s;

	cell = xgrid.addChild("cell");
	pb = cell.addChild("pictbutton");
	if (col.action !== null) {
		pb.setStringAttr("action", col.action);
		if (col.data !== null)
			if ((s = xrow.searchValue(col.data)) !== null)
				pb.setStringAttr("data", s);
	}
	pb.setStringAttr("expand", "100%");
	pb.setStringAttr("ref", col.ref);
	pb.setStringAttr("size", col.imgsz);
	pb.setStringAttr("naturalsz", col.imgsz);
};

// create -------------------------------------------------------------
XTable.create = function(data) {
	var attrs, cfg, child, col, cols, header, i, id, j, name, nrows, path, props, rows, scroll, v, xdata, xgrid, xroot, xrow;

	if ((id = data.getAttrString("id", null)) === null)
		throw new Error("id not found");
	if ((path = data.getAttrString("path", null)) === null)
		throw new Error("attribute 'path' not found");
	scroll = data.getAttrInt("scroll", 0);
	props = new XProps(data);
	xdata = xgrid = null;
	rows = null;
	if (data.isRawWidget())
		xroot = UILib.newXData("table");
	else
		if ((xdata = data.getXData(props.get("xref"))) === null)
			xroot = UILib.newXData("table");
		else {
			rows = xdata.searchElements("/magicw" + path);
			if (rows.length === 0) {
				xdata = null;
				xroot = UILib.newXData("table");
			} else {
				if (rows.length <= scroll)
					xroot = xgrid = UILib.newXData("grid");
				else {
					xroot = UILib.newXData("table");
					xgrid = xroot.addChild("grid");
				}
			}
		}
	xroot.setStringAttr("id", id);
	xroot.setStringAttr("bg", "#E8E8E8");
	xroot.setStringAttr("props", props.write("caching,xref"));
	attrs = data.getAttributes();
	for(i = 0;i < attrs.length;++i) {
		name = attrs[i].key;
		v = attrs[i].value;
		switch(name) {
		case "anchor":
		case "border":
		case "expand":
		case "expandh":
		case "expandv":
		case "size":
			xroot.setStringAttr(name, v);
			break;
		case "margin":
			xroot.setStringAttr(name, v);
			break;
		}
	}
	if ((child = data.getFirstChildWithTag("header")) !== null)
		header = new XTableHeader(XElement.getInstanceFromXWData(child));
	else
		header = null;
	cols = XTableColumn.getColumns(XElement.getInstanceFromXWData(data), "column");
	cfg = XElement.newInstance("table");
	cfg.setStringAttr("id", id);
	cfg.setStringAttr("path", path);
	if (header !== null)
		header.setConfig(cfg);
	for(i = 0;i < cols.length;++i)
		cols[i].setConfig(cfg);
	if (data.isRawWidget())
		return(new XTableRaw(data.toUIXWData(xroot, false), path, scroll, header, cols));
	if (xdata === null) {
		xroot.addChild("area");
		return(new XTable(data.toUIXWData(xroot, true), cfg));
	}
	nrows = rows.length;
	if (header !== null)
		++nrows;
	xgrid.setStringAttr("id", id + "_view");
	xgrid.setIntAttr("ncols", cols.length);
	xgrid.setIntAttr("nrows", nrows);
	xgrid.setIntAttr("gapw", 1);
	xgrid.setStringAttr("bg", "silver");
	xgrid.setStringAttr("odd", "gainsboro");
	xgrid.setStringAttr("even", "whitesmoke");
	if (header !== null) {
		xgrid.setStringAttr("caption", header.bg);
		for(j = 0;j < cols.length;++j)
			XTable.addCaption(xgrid, header, cols[j]);
	}
	for(i = 0;i < rows.length;++i) {
		xrow = rows[i];
		for(j = 0;j < cols.length;++j) {
			col = cols[j];
			switch(col.type) {
			case "checkbox":
				XTable.addCheckbox(xgrid, xrow, col);
				break;
			case "label":
				XTable.addLabel(xgrid, xrow, col);
				break;
			case "link":
				XTable.addLink(xgrid, xrow, col);
				break;
			case "pictbutton":
				XTable.addPictbutton(xgrid, xrow, col);
				break;
			default:
				throw new Error("invalid type: " + col.type);
			}
		}
	}
	if (xroot === xgrid)
		return(new XGrid(data.toUIXWData(xroot, true)));
	return(new XTable(data.toUIXWData(xroot, true), cfg));
};
/* ====================================================================	*
 * XMessagesList 														*
 * ====================================================================	*/

var XMessagesList = function() {
};

// create -------------------------------------------------------------
XMessagesList.create = function(data) {
	var attrs, bg, gap, h5w, paddingv, props, root;

	XPluginLib.addFont(new XFont("XMessagesList#title", "white", "small:bold"));
	XPluginLib.addFont(new XFont("XMessagesList#sender", "white", "small:bold"));
	XPluginLib.addFont(new XFont("XMessagesList#date", "whitesmoke", "small:plain:italic"));
	XPluginLib.addFont(new XFont("XMessagesList#body", "white", "small:plain"));
	if (data.isRawWidget()) {
		h5w = new H5Widget(data);
		return(h5w);
	}
	attrs = XElementAttrs.create(data.getAttributes());
	bg = attrs.getValue("bg", data);
	paddingv = attrs.searchString("paddingv", "5");
	gap = attrs.searchString("gap", "10");
	root = UILib.newXData("scroll");
	attrs.copyWidgetAttrs(root);
	props = "bg=" + bg + ",gap=" + gap + ",paddingv=" + paddingv;
	root.setStringAttr("props", props);
	root.addChild("area");
	return(new XScrollPane(data.toUIXWData(root, true)));
};

// getNewView ---------------------------------------------------------
XMessagesList.getNewView = function(cardId, scroll, data, label) {
	var bg, body, entries, entry, field, flow, i, ip, msg, paragraphs, props, sender, view;

	props = scroll.props;
	bg = props.get("bg");
	view = UILib.newXData("vflow");
	view.setStringAttr("bg", bg);
	view.setStringAttr("padding", props.get("paddingv"));
	view.setStringAttr("gap", props.get("gap"));
	view.setStringAttr("expandh", "100%");
	flow = view.addChild("hflow");
	flow.setStringAttr("bg", bg);
	flow.setStringAttr("expandh", "100%");
	flow.setStringAttr("weight", "0");
	field = flow.addChild("label");
	field.setStringAttr("text", label);
	field.setStringAttr("bg", bg);
	XPluginLib.setFont(field, "XMessagesList#title");
	field = flow.addChild("pictbutton");
	field.setStringAttr("ref", "refresh-16.png");
	field.setStringAttr("bg", bg);
	field.setStringAttr("margin", "0:10:0:0");
	field.setStringAttr("action", cardId + "-messages-refresh");
	field.setStringAttr("size", "16x16");
	entries = data.getChildren("entry");
	for(i = 0;i < entries.length;++i) {
		entry = entries[i];
		if (entry.getAttrString("hidden", null) !== null)
			continue;
		msg = view.addChild("vflow");
		msg.setStringAttr("expandh", "100%");
		msg.setStringAttr("bg", "black");
		msg.setStringAttr("padding", "10");
		flow = msg.addChild("hflow");
		flow.setStringAttr("expandh", "100%");
		flow.setStringAttr("bg", "black");
		flow.setStringAttr("weight", "0");
		sender = entry.searchValue("sender");
		ip = entry.searchValue("ip");
		if (!XString.equals(ip, ""))
			sender += " (" + ip + ")";
		field = flow.addChild("label");
		field.setStringAttr("text", sender);
		field.setStringAttr("bg", "black");
		field.setStringAttr("anchor", "west");
		field.setStringAttr("align", "left");
		XPluginLib.setFont(field, "XMessagesList#sender");
		field = flow.addChild("pictbutton");
		field.setStringAttr("ref", "list-remove-16.png");
		field.setStringAttr("bg", "black");
		field.setStringAttr("action", cardId + "-messages-hidden");
		field.setStringAttr("data", entry.getString("id"));
		field.setStringAttr("size", "16x16");
		field = msg.addChild("label");
		field.setStringAttr("text", entry.searchValue("date"));
		field.setStringAttr("anchor", "west");
		field.setStringAttr("align", "left");
		field.setStringAttr("bg", "black");
		XPluginLib.setFont(field, "XMessagesList#date");
		body = entry.getFirstChildWithTag("body");
		paragraphs = body.getAllChildren();
		field = msg.addChild("label");
		field.setStringAttr("text", XUtil.convertTextToLabel(paragraphs));
		field.setStringAttr("anchor", "west");
		field.setStringAttr("align", "left");
		field.setStringAttr("bg", "black");
		XPluginLib.setFont(field, "XMessagesList#body");
	}
	return(view);
};
/* ====================================================================	*
 * XThumbnails 															*
 * ====================================================================	*/

var XThumbnails = function() {
};

// create -------------------------------------------------------------
XThumbnails.create = function(data) {
	var attrs, bg, gapv, imgp, imgw, paddingv, path, props, root, xref;

	if (data.isRawWidget())
		return(new H5Widget(data));
	attrs = XElementAttrs.create(data.getAttributes());
	bg = attrs.getValue("bg", data);
	xref = attrs.getString("xref");
	path = attrs.getString("path");
	gapv = attrs.searchString("gapv", null);
	imgw = attrs.searchString("imgw", null);
	paddingv = attrs.searchString("paddingv", null);
	imgp = attrs.searchString("imgp", null);
	root = UILib.newXData("scroll");
	attrs.copyWidgetAttrs(root);
	props = "bg=" + bg + ",path=" + path + ",xref=" + xref;
	if (gapv !== null)
		props += ",gapv=" + gapv;
	if (imgw !== null)
		props += ",imgw=" + imgw;
	if (paddingv !== null)
		props += ",paddingv=" + paddingv;
	if (imgp !== null)
		props += ",imgp=" + imgp;
	root.setStringAttr("props", props);
	root.addChild("area");
	return(new XScrollPane(data.toUIXWData(root, true)));
};

// getNewView ---------------------------------------------------------
XThumbnails.getNewView = function(cardId, scroll, appl) {
	var bg, d1, d2, gapv, i, id, img, imgp, imgs, imgw, n1, n2, paddingv, path, props, r, r1, r2, ratio, view, xd, ximg, xref;

	props = scroll.props;
	xref = props.get("xref");
	path = props.get("path");
	if ((xd = appl.getData(xref)) === null)
		throw new Error("element not found: " + xref + ":" + path);
	if ((imgs = XFile.getList(xd, path)) === null)
		throw new Error("element not found: " + xref + ":" + path);
	if (imgs.length === 0)
		return(UILib.newXData("area"));
	bg = props.get("bg");
	gapv = props.lookup("gapv");
	imgw = props.lookup("imgw");
	paddingv = props.lookup("paddingv");
	imgp = props.lookup("imgp");
	view = UILib.newXData("vflow");
	view.setStringAttr("id", scroll.id + "_vflow");
	view.setStringAttr("expandh", "100%");
	view.setStringAttr("bg", bg);
	if (gapv !== null)
		view.setStringAttr("gap", gapv);
	if (paddingv !== null)
		view.setStringAttr("padding", paddingv);
	r1 = 16.0 / 9;
	r2 = 4.0 / 3;
	n1 = n2 = 0;
	for(i = 0;i < imgs.length;++i) {
		img = imgs[i];
		r = img.width / img.height;
		if ((d1 = r - r1) < 0)
			d1 = - d1;
		if ((d2 = r - r2) < 0)
			d2 = - d2;
		if (d1 < d2)
			n1 ++;
		else
			n2 ++;
	}
	ratio =(n1 > n2)?"16/9":"4/3";
	for(i = 0;i < imgs.length;++i) {
		img = imgs[i];
		id = scroll.id + "_button" + XInteger.toPadding(i, 3);
		ximg = view.addChild("pictbutton");
		ximg.setStringAttr("id", id);
		path = img.path;
		if (path.startsWith("file:") || path.startsWith("cache:"))
			ximg.setStringAttr("ref", path);
		else
			ximg.setStringAttr("ref", "cache:" + img.path);
		ximg.setStringAttr("action", cardId + "-slide-update");
		ximg.setStringAttr("data", "cache:" + img.path);
		ximg.setStringAttr("bg", bg);
		ximg.setStringAttr("expandh", "100%");
		ximg.setStringAttr("ratio", ratio);
		ximg.setStringAttr("naturalsz", img.width + "x" + img.height);
		if (imgw !== null)
			ximg.setStringAttr("size", imgw + "x0");
		if (imgp !== null)
			ximg.setStringAttr("padding", imgp);
	}
	return(view);
};
/* ====================================================================	*
 * XLib 																*
 * ====================================================================	*/

var XLib = function() {
};

// adjustMinimumWidth -------------------------------------------------
XLib.adjustMinimumWidth = function(minw, colsw) {
	var i, j, n, ncols, nleft;

	ncols = minw.length;
	for(i = 0;i < ncols;++i) {
		if ((nleft = minw[i] - colsw[i]) <= 0)
			continue;
		for(j = 0;j < ncols;++j) {
			if ((n = colsw[j] - minw[j]) <= 0)
				continue;
			if (n >= nleft) {
				colsw[j] -= nleft;
				colsw[i] += nleft;
				nleft = 0;
				break;
			}
			colsw[j] -= n;
			colsw[i] += n;
			nleft -= n;
		}
		if (nleft !== 0)
			throw new Error("invalid value for nleft: " + nleft);
	}
};

// createChildrenWidget -----------------------------------------------
XLib.createChildrenWidget = function(data) {
	var children, i, list;

	children = data.getAllChildrenWidgets();
	list = UILib.newArray(children.length, null);
	for(i = 0;i < children.length;++i)
		list[i] = XLib.createWidget(children[i]);
	return(list);
};

// createWidget -------------------------------------------------------
XLib.createWidget = function(data) {
	var tag, wt;

	tag = data.getTag();
	switch(tag) {
	case "area":
		return(new XWidget(data));
	case "button":
		return(new XButton(data));
	case "checkbox":
		return(XCheckBox.create(data));
	case "checkbox-item":
		return(new XCheckBoxItem(data));
	case "combobox":
		return(new XComboBox(data));
	case "date":
		return(XDate.create(data));
	case "dropdown":
		return(new XDropDown(data));
	case "flow":
		return(new XFlow(data));
	case "form":
		return(MWForm.create(data));
	case "grid":
		return(new XGrid(data));
	case "hflow":
		return(new XHorFlow(data));
	case "iconlink":
		return(XIconLink.create(data));
	case "label":
		return(new XLabel(data));
	case "labelbutton":
		return(new XLabelButton(data));
	case "link":
		return(new XLink(data));
	case "menubar":
		return(new XMenuBar(data));
	case "messages":
		return(XMessagesList.create(data));
	case "password":
		return(new XPassword(data));
	case "pictbutton":
		return(new XPictButton(data));
	case "picture":
		return(new XPicture(data));
	case "plugin":
		return(new XPlugin(data));
	case "radiobutton":
		return(XRadioButton.create(data));
	case "radiobutton-item":
		return(new XRadioButtonItem(data));
	case "scroll":
		return(new XScrollPane(data));
	case "simplegrid":
		return(new XSimpleGrid(data));
	case "splitpane":
		return(new XSplitPane(data));
	case "stack":
		return(new XStack(data));
	case "table":
		return(XTable.create(data));
	case "text":
		return(new XText(data));
	case "textarea":
		return(new XTextArea(data));
	case "textfield":
		return(new XTextField(data));
	case "thumbnails":
		return(XThumbnails.create(data));
	case "tree":
		return(XTree.create(data));
	case "treearea":
		return(new XTreeArea(data));
	case "treelabel":
		return(new XTreeLabel(data));
	case "treerow":
		return(XTreeRow.create(data));
	case "treetoggle":
		return(new XTreeToggle(data));
	case "treeview":
		return(new XTreeView(data));
	case "vflow":
		return(new XVerFlow(data));
	case "widget":
		return(data.getRefWidget());
	}
	if ((wt = XPluginLib.create(tag, data)) !== null)
		return(wt);
	throw new Error("invalid widgetname: " + tag);
};

// doLayout -----------------------------------------------------------
XLib.doLayout = function(wt, prefw) {
	var height, width;

	wt.layoutSetMinimumWidth();
	if ((width = wt.prefWidth) < prefw)
		width = prefw;
	wt.layoutSetViewportWidth(width);
	wt.layoutSetMinimumHeight();
	height = wt.cheight;
	wt.layoutSetViewportHeight(height);
};

// log ----------------------------------------------------------------
XLib.log = function(root, indent) {
	var i;

	XApplication.logMsg(indent + root.type + ": " + root.id);
	if (!root.hasChildren())
		return;
	indent += "  ";
	for(i = 0;i < root.children.length;++i)
		XLib.log(root.children[i], indent);
};

// search -------------------------------------------------------------
XLib.search = function(wt, id) {
	var children, i, r;

	if (wt.id !== null && UILib.stringsEquals(wt.id, id))
		return(wt);
	if (wt.children === null)
		return(null);
	children = wt.children;
	for(i = 0;i < children.length;++i)
		if ((r = XLib.search(children[i], id)) !== null)
			return(r);
	return(null);
};

// searchParent -------------------------------------------------------
XLib.searchParent = function(wt, id) {
	return(XLib.searchParentW(wt, id, null));
};

// searchParentW ------------------------------------------------------
XLib.searchParentW = function(wt, id, parent) {
	var children, i, r;

	if (wt.id !== null && UILib.stringsEquals(wt.id, id))
		return(parent);
	if (wt.children === null)
		return(null);
	children = wt.children;
	for(i = 0;i < children.length;++i)
		if ((r = XLib.searchParentW(children[i], id, wt)) !== null)
			return(r);
	return(null);
};
/* ====================================================================	*
 * XLink 																*
 * ====================================================================	*/

var XLink = function(xdata) {

	XWidget.call(this, xdata);

	// layoutInit -----------------------------------------------------
	XLink.prototype.layoutInit = function() {
		this.minWidth = this.uilabel.getMinimumWidth();
		this.prefWidth = this.uilabel.getPrefWidth();
	};

	// layoutSetViewportWidth -----------------------------------------
	XLink.prototype.layoutSetViewportWidth = function(vwidth) {
		var sz;

		XWidget.prototype.layoutSetViewportWidth.call(this, vwidth);
		sz = this.uilabel.format(this.cwidth - this.getEdgeWidth());
		if (this.expandHor === 0)
			this.cwidth = sz.width + this.getEdgeWidth();
		this.minHeight = sz.height;
	};

	// Constructor ----------------------------------------------------
	{
		this.uilabel = xdata.getLabel(false);
		this.uilabel.setUnderline();
		this.action = xdata.getAttrString("action", null);
		this.data = xdata.getAttrString("data", null);
	}
};
/* ====================================================================	*
 * XMenuBar 															*
 * ====================================================================	*/

var XMenuBar = function(data) {

	XWidget.call(this, data);

	this.left = null;
	this.right = null;
	this.rowh = 0;
	this.nrows = 0;

	// init -----------------------------------------------------------
	XMenuBar.prototype.init = function(data) {
		var cdata, i, off;

		if ((cdata = data.getFirstChildWithTag("left")) === null)
			this.left = UILib.newArray(0, null);
		else
			this.left = XLib.createChildrenWidget(cdata);
		if ((cdata = data.getFirstChildWithTag("right")) === null)
			this.right = UILib.newArray(0, null);
		else
			this.right = XLib.createChildrenWidget(cdata);
		this.children = UILib.newArray(this.left.length + this.right.length, null);
		for(i = 0;i < this.left.length;++i)
			this.children[i] = this.left[i];
		off = this.left.length;
		for(i = 0;i < this.right.length;++i)
			this.children[off + i] = this.right[i];
	};

	// layoutInit -----------------------------------------------------
	XMenuBar.prototype.layoutInit = function() {
		var i, w, wt;

		for(i = 0;i < this.children.length;++i) {
			wt = this.children[i];
			wt.layoutInit();
			if ((w = wt.prefWidth + wt.getEdgeWidth()) > this.minWidth)
				this.minWidth = w;
			this.prefWidth += w;
		}
		this.prefWidth +=(this.children.length -1) * this.gap;
	};

	// layoutSetLocation ----------------------------------------------
	XMenuBar.prototype.layoutSetLocation = function(vx, vy, anchor) {
		var dy, i, wt, x, xmax, y;

		XWidget.prototype.layoutSetLocation.call(this, vx, vy, anchor);
		x = this.getX();
		y = this.getY();
		xmax = this.cwidth - this.getMaxX();
		for(i = 0;i < this.left.length;++i) {
			wt = this.left[i];
			if (x + wt.cwidth > xmax) {
				y += this.rowh + this.interline;
				x = this.getX();
			}
			dy =(this.rowh - wt.cheight) / 2;
			wt.layoutSetLocation(x, y + dy, "center");
			x += wt.cwidth + this.gap;
		}
		y =(this.nrows -1) *(this.rowh + this.interline);
		x = xmax;
		for(i = this.right.length -1;i >= 0;--i) {
			wt = this.right[i];
			if ((x -= wt.cwidth) < 0) {
				y -= this.rowh + this.interline;
				x = xmax - wt.cwidth;
			}
			dy =(this.rowh - wt.cheight) / 2;
			wt.layoutSetLocation(x, y + dy, "center");
			x -= this.gap;
		}
	};

	// layoutSetMinimumWidth ------------------------------------------
	XMenuBar.prototype.layoutSetMinimumWidth = function() {
		var i;

		XWidget.prototype.layoutSetMinimumWidth.call(this);
		for(i = 0;i < this.children.length;++i)
			this.children[i].layoutSetMinimumWidth();
	};

	// layoutSetMinimumHeight -----------------------------------------
	XMenuBar.prototype.layoutSetMinimumHeight = function() {
		var i;

		XWidget.prototype.layoutSetMinimumHeight.call(this);
		for(i = 0;i < this.children.length;++i)
			this.children[i].layoutSetMinimumHeight();
	};

	// layoutSetViewportWidth -----------------------------------------
	XMenuBar.prototype.layoutSetViewportWidth = function(vwidth) {
		var h, i, nleft, w, wmax, wt;

		XWidget.prototype.layoutSetViewportWidth.call(this, vwidth);
		this.rowh = 0;
		wmax = nleft = this.cwidth - this.getEdgeWidth();
		this.nrows = 1;
		for(i = 0;i < this.children.length;++i) {
			wt = this.children[i];
			w = wt.prefWidth + wt.getEdgeWidth();
			wt.layoutSetViewportWidth(w);
			if ((h = wt.minHeight + wt.getEdgeHeight()) > this.rowh)
				this.rowh = h;
			if ((nleft -= w) < 0) {
				nleft = wmax - w;
				++this.nrows;
			}
			nleft -= this.gap;
		}
		this.minHeight = this.nrows * this.rowh +(this.nrows -1) * this.interline;
	};

	// layoutSetViewportHeight ----------------------------------------
	XMenuBar.prototype.layoutSetViewportHeight = function(vheight) {
		var i;

		XWidget.prototype.layoutSetViewportHeight.call(this, vheight);
		for(i = 0;i < this.children.length;++i)
			this.children[i].layoutSetViewportHeight(this.rowh);
	};

	// Constructor ----------------------------------------------------
	{
		this.gap = data.getInt("gap");
		this.interline = data.getInt("interline");
		this.init(data);
	}
};
/* ====================================================================	*
 * XPassword 															*
 * ====================================================================	*/

var XPassword = function(data) {

	XWidget.call(this, data);

	// Constructor ----------------------------------------------------
	{
		this.ncols = data.getInt("ncols");
	}
};
/* ====================================================================	*
 * XPictButton 															*
 * ====================================================================	*/

var XPictButton = function(xd) {

	XWidget.call(this, xd);

	this.prefh = 0;

	// layoutInit -----------------------------------------------------
	XPictButton.prototype.layoutInit = function() {
		var h;

		XWidget.prototype.layoutInit.call(this);
		if (this.ratio === null || this.vratio === 0.0)
			return;
		if ((h = XInteger.round(this.minWidth / this.vratio)) > this.minHeight)
			this.minHeight = h;
	};

	// layoutSetMinimumHeight -----------------------------------------
	XPictButton.prototype.layoutSetMinimumHeight = function() {
		if (this.ratio === null || this.vratio === 0.0) {
			XWidget.prototype.layoutSetMinimumHeight.call(this);
			return;
		}
		this.cheight =(this.prefh > this.minHeight)?this.prefh:this.minHeight;
		this.cheight += this.getEdgeHeight();
	};

	// layoutSetViewportWidth -----------------------------------------
	XPictButton.prototype.layoutSetViewportWidth = function(vwidth) {
		var w;

		XWidget.prototype.layoutSetViewportWidth.call(this, vwidth);
		if (this.ratio === null || this.vratio === 0.0)
			return;
		w = this.cwidth - this.getEdgeWidth();
		this.prefh = XInteger.round(w / this.vratio);
	};

	// Constructor ----------------------------------------------------
	{
		if ((this.ref = xd.getAttrString("ref", null)) === null)
			throw new Error("attribute ref not found");
		this.action = xd.getAttrString("action", null);
		this.active = xd.getAttrString("active", null);
		this.over = xd.getAttrString("over", null);
		this.data = xd.getAttrString("data", null);
		this.naturalsz = xd.getAttrSize("naturalsz", null);
		if ((this.ratio = xd.getAttrString("ratio", null)) === null)
			return;
		switch(this.ratio) {
		case "16/9":
			this.vratio = 16.0 / 9;
			break;
		case "4/3":
			this.vratio = 4.0 / 3;
			break;
		case "true":
			this.vratio = 0.0;
			break;
		default:
			throw new Error("invalid ratio: " + this.vratio);
		}
	}
};
/* ====================================================================	*
 * XPicture 															*
 * ====================================================================	*/

var XPicture = function(data) {

	XWidget.call(this, data);

	this.prefh = 0;

	// layoutInit -----------------------------------------------------
	XPicture.prototype.layoutInit = function() {
		var h;

		XWidget.prototype.layoutInit.call(this);
		if (this.ratio === null || this.vratio === 0.0)
			return;
		if ((h = XInteger.round(this.minWidth / this.vratio)) > this.minHeight)
			this.minHeight = h;
	};

	// layoutSetMinimumHeight -----------------------------------------
	XPicture.prototype.layoutSetMinimumHeight = function() {
		if (this.ratio === null || this.vratio === 0.0) {
			XWidget.prototype.layoutSetMinimumHeight.call(this);
			return;
		}
		this.cheight =(this.prefh > this.minHeight)?this.prefh:this.minHeight;
		this.cheight += this.getEdgeHeight();
	};

	// layoutSetViewportWidth -----------------------------------------
	XPicture.prototype.layoutSetViewportWidth = function(vwidth) {
		var w;

		XWidget.prototype.layoutSetViewportWidth.call(this, vwidth);
		if (this.ratio === null || this.vratio === 0.0)
			return;
		w = this.cwidth - this.getEdgeWidth();
		this.prefh = XInteger.round(w / this.vratio);
	};

	// Constructor ----------------------------------------------------
	{
		if ((this.ref = data.getAttrString("ref", null)) === null)
			if (this.minWidth === 0 && this.minHeight === 0)
				throw new Error("attribute 'ref' or 'size' not found");
		this.naturalsz = data.getAttrSize("naturalsz", null);
		if ((this.ratio = data.getAttrString("ratio", null)) === null)
			return;
		switch(this.ratio) {
		case "16/9":
			this.vratio = 16.0 / 9;
			break;
		case "4/3":
			this.vratio = 4.0 / 3;
			break;
		case "true":
			this.vratio = 0.0;
			break;
		default:
			throw new Error("invalid ratio: " + this.vratio);
		}
	}
};
/* ====================================================================	*
 * XPlugin 																*
 * ====================================================================	*/

var XPlugin = function(data) {

	XWidget.call(this, data);
};
/* ====================================================================	*
 * XPoint 																*
 * ====================================================================	*/

var XPoint = function(x, y) {

	// Constructor ----------------------------------------------------
	{
		this.x = x;
		this.y = y;
	}
};
/* ====================================================================	*
 * XProps 																*
 * ====================================================================	*/

var XProps = function(data) {

	// compare --------------------------------------------------------
	XProps.prototype.compare = function(key, value) {
		var kv;

		if ((kv = this.search(key)) === null)
			return(false);
		return(UILib.stringsEquals(kv.value, value));
	};

	// exists ---------------------------------------------------------
	XProps.prototype.exists = function(key) {
		return(this.search(key) !== null);
	};

	// get ------------------------------------------------------------
	XProps.prototype.get = function(key) {
		var kv;

		if ((kv = this.search(key)) === null)
			throw new Error("attribut not found: " + key);
		return(kv.value);
	};

	// lookup ---------------------------------------------------------
	XProps.prototype.lookup = function(key) {
		var kv;

		if ((kv = this.search(key)) === null)
			return(null);
		return(kv.value);
	};

	// search ---------------------------------------------------------
	XProps.prototype.search = function(key) {
		var i;

		if (this.values === null)
			return(null);
		for(i = 0;i < this.values.length;++i)
			if (UILib.stringsEquals(key, this.values[i].key))
				return(this.values[i]);
		return(null);
	};

	// set ------------------------------------------------------------
	XProps.prototype.set = function(key, value) {
		var i, kv, tmp;

		if (this.values === null) {
			this.values = UILib.newArray(1, null);
			this.values[0] = new XKeyValue(key, value);
			return;
		}
		for(i = 0;i < this.values.length;++i) {
			kv = this.values[i];
			if (UILib.stringsEquals(kv.key, key)) {
				kv.value = value;
				return;
			}
		}
		tmp = UILib.newArray(this.values.length + 1, null);
		for(i = 0;i < this.values.length;++i)
			tmp[i] = this.values[i];
		tmp[i] = new XKeyValue(key, value);
		this.values = tmp;
	};

	// write ----------------------------------------------------------
	XProps.prototype.write = function(list) {
		var first, i, kv, l, sb;

		l = list.split(",");
		first = true;
		sb = new XStringBuilder();
		for(i = 0;i < l.length;++i) {
			if ((kv = this.search(l[i])) === null)
				continue;
			if (first)
				first = false;
			else
				sb.append(",");
			sb.append(kv.key + "=" + kv.value);
		}
		return(sb.getText());
	};

	// writeAll -------------------------------------------------------
	XProps.prototype.writeAll = function() {
		var i, kv, sb;

		if (this.values === null)
			return("");
		sb = new XStringBuilder();
		for(i = 0;i < this.values.length;++i) {
			if (i > 0)
				sb.append(",");
			kv = this.values[i];
			sb.append(kv.key + "=" + kv.value);
		}
		return(sb.getText());
	};

	// Constructor ----------------------------------------------------
	{
		var attr, i, k, l, n, v;

		this.values = null;
		if ((attr = data.getAttrString("props", null)) === null)
			return;
		l = attr.split(",");
		this.values = UILib.newArray(l.length, null);
		for(i = 0;i < l.length;++i) {
			if ((n = l[i].indexOf('=')) === -1)
				throw new Error("invalid properties: " + attr);
			k = l[i].substring(0, n);
			v = l[i].substring(n + 1);
			this.values[i] = new XKeyValue(k, v);
		}
	}
};
/* ====================================================================	*
 * XRadioButton 														*
 * ====================================================================	*/

var XRadioButton = function(data) {

	XHorFlow.call(this, data);

	// Constructor ----------------------------------------------------
	{
		this.text = data.getAttrString("text", null);
		this.key = data.getAttrString("key", null);
		this.group = data.getAttrString("group", null);
	}
};

// create -------------------------------------------------------------
XRadioButton.create = function(data) {
	var attrs, bg, c1, c2, fg, font, group, i, id, key, name, text, value, xd;

	xd = UILib.newXData("radiobutton");
	bg = fg = font = group = id = key = text = null;
	attrs = data.getAttributes();
	for(i = 0;i < attrs.length;++i) {
		name = attrs[i].key;
		value = attrs[i].value;
		switch(name) {
		case "id":
			id = value;
			xd.setStringAttr(name, value);
			break;
		case "margin":
		case "border":
		case "padding":
		case "size":
		case "expand":
		case "expandh":
		case "expandv":
		case "anchor":
			xd.setStringAttr(name, value);
			break;
		case "bg":
			bg = value;
			xd.setStringAttr(name, value);
			break;
		case "fg":
			fg = value;
			xd.setStringAttr(name, value);
			break;
		case "align":
		case "gap":
			xd.setStringAttr(name, value);
			break;
		case "text":
			text = value;
			break;
		case "font":
			font = value;
			break;
		case "key":
			key = value;
			break;
		case "group":
			group = value;
			break;
		}
	}
	if (id === null)
		throw new Error("attribute id not found");
	if (text === null)
		throw new Error("attribute text not found");
	if (key === null)
		throw new Error("attribute key not found");
	if (group === null)
		throw new Error("attribute group not found");
	if (font === null)
		font = data.getFont();
	xd.setStringAttr("text", text);
	xd.setStringAttr("key", key);
	xd.setStringAttr("group", group);
	c1 = xd.addChild("radiobutton-item");
	c1.setStringAttr("id", id + "_radiobutton");
	c1.setStringAttr("key", key);
	c1.setStringAttr("group", group);
	c2 = xd.addChild("label");
	c2.setStringAttr("id", id + "_label");
	c2.setStringAttr("text", text);
	if (bg !== null) {
		c1.setStringAttr("bg", bg);
		c2.setStringAttr("bg", bg);
	}
	if (fg !== null)
		c2.setStringAttr("fg", fg);
	if (font !== null)
		c2.setStringAttr("font", font);
	return(new XRadioButton(data.toUIXWData(xd, true)));
};
/* ====================================================================	*
 * XRadioButtonItem 													*
 * ====================================================================	*/

var XRadioButtonItem = function(data) {

	XWidget.call(this, data);

	// Constructor ----------------------------------------------------
	{
		if ((this.key = data.getAttrString("key", null)) === null)
			throw new Error("attribute 'key' not found");
		if ((this.group = data.getAttrString("group", null)) === null)
			throw new Error("attribute 'group' not found");
	}
};
/* ====================================================================	*
 * XRectangle 															*
 * ====================================================================	*/

var XRectangle = function(x, y, w, h) {

	// isInside -------------------------------------------------------
	XRectangle.prototype.isInside = function(px, py) {
		var x1, y1;

		x1 = this.x + this.w;
		y1 = this.y + this.h;
		if (px < this.x || x1 <= px)
			return(false);
		if (py < this.y || y1 <= py)
			return(false);
		return(true);
	};

	// Constructor ----------------------------------------------------
	{
		this.x = x;
		this.y = y;
		this.w = w;
		this.h = h;
	}
};
/* ====================================================================	*
 * XRoot 																*
 * ====================================================================	*/

var XRoot = function(widget) {

	// doLayout -------------------------------------------------------
	XRoot.prototype.doLayout = function(x, y, width, height) {
		if (this.keys !== null)
			this.setResponsiveLayout(width);
		this.widget.layoutSetMinimumWidth();
		if (width < this.widget.cwidth)
			width = this.widget.cwidth;
		this.widget.layoutSetViewportWidth(width);
		this.widget.layoutSetMinimumHeight();
		if (height < this.widget.cheight)
			height = this.widget.cheight;
		this.widget.layoutSetViewportHeight(height);
		this.widget.layoutSetLocation(x, y, null);
	};

	// doUpdateUI -----------------------------------------------------
	XRoot.prototype.doUpdateUI = function() {
		if (!this.updateUI)
			return;
		this.widget.layoutInit();
		this.updateUI = false;
	};

	// setResponsiveLayout --------------------------------------------
	XRoot.prototype.setResponsiveLayout = function(width) {
		var i;

		if (this.minWidth === -1) {
			this.widget.layoutSetResponsive(null);
			this.widget.layoutSetMinimumWidth();
			this.minWidth = this.widget.cwidth;
			this.idx = -1;
			for(i = 0;i < this.widths.length;++i)
				this.widths[i] = -1;
		}
		if (width >= this.minWidth) {
			if (this.idx === -1)
				return;
			this.widget.layoutSetResponsive(null);
			this.idx = -1;
			return;
		}
		for(i = 0;;++i) {
			if (this.widths[i] === -1) {
				this.widget.layoutSetResponsive(this.keys[i]);
				this.widget.layoutSetMinimumWidth();
				this.widths[i] = this.widget.cwidth;
			}
			if (i + 1 === this.widths.length || width >= this.widths[i])
				break;
		}
		if (this.idx === i)
			return;
		this.idx = i;
		this.widget.layoutSetResponsive(this.keys[this.idx]);
	};

	// setUpdateUI ----------------------------------------------------
	XRoot.prototype.setUpdateUI = function(updateUI) {
		this.updateUI = updateUI;
	};

	// Constructor ----------------------------------------------------
	{
		this.widget = widget;
		if ((this.keys = widget.responsive) !== null) {
			this.minWidth = -1;
			this.widths = UILib.newArray(this.keys.length, 0);
			this.idx = -1;
		}
		widget.layoutInit();
		this.updateUI = false;
	}
};
/* ====================================================================	*
 * XSimpleGrid 															*
 * ====================================================================	*/

var XSimpleGrid = function(data) {

	XWidget.call(this, data);

	this.cminw = 0;
	this.cellw = 0;
	this.cellh = 0;
	this.ncols = 0;
	this.nrows = 0;

	// getWidth -------------------------------------------------------
	XSimpleGrid.prototype.getWidth = function() {
		return(this.cellw * this.ncols + this.gapw *(this.ncols -1));
	};

	// getHeight ------------------------------------------------------
	XSimpleGrid.prototype.getHeight = function() {
		return(this.cellh * this.nrows + this.gaph *(this.nrows -1));
	};

	// layoutInit -----------------------------------------------------
	XSimpleGrid.prototype.layoutInit = function() {
		var i, wt;

		for(i = 0;i < this.children.length;++i) {
			wt = this.children[i];
			if (wt.expandHor === 0)
				wt.expandHor = 100;
			if (wt.expandVer === 0)
				wt.expandVer = 100;
			wt.layoutInit();
		}
	};

	// layoutSetLocation ----------------------------------------------
	XSimpleGrid.prototype.layoutSetLocation = function(vx, vy, anchor) {
		var dh, dw, h, i, icol, w, wt, x, x0, y;

		XWidget.prototype.layoutSetLocation.call(this, vx, vy, anchor);
		w = this.cwidth - this.getEdgeWidth();
		h = this.cheight - this.getEdgeHeight();
		if (this.compact) {
			dh = UILib.mathFloor((h - this.getHeight()) / 2);
			switch(this.align) {
			case "left":
				x0 = this.getX();
				dw = 0;
				break;
			case "center":
				dw = UILib.mathFloor((w - this.getWidth()) / 2);
				x0 = this.getX() + dw;
				break;
			case "right":
				x0 = this.cwidth - this.getMaxX() - this.ncols * this.cellw -(this.ncols -1) * this.gapw;
				dw = 0;
				break;
			default:
				throw new Error("invalid value: " + this.align);
			}
		} else {
			dw = UILib.mathFloor((w - this.getWidth()) /(this.ncols + 1));
			dh = UILib.mathFloor((h - this.getHeight()) /(this.nrows + 1));
			x0 = this.getX() + dw;
		}
		x = x0;
		y = this.getY() + dh;
		icol = 0;
		if (this.compact)
			dw = dh = 0;
		for(i = 0;i < this.children.length;++i) {
			wt = this.children[i];
			wt.layoutSetLocation(x, y, "center");
			if (++icol < this.ncols)
				x += this.cellw + this.gapw + dw;
			else {
				x = x0;
				icol = 0;
				y += this.cellh + this.gaph + dh;
			}
		}
	};

	// layoutSetMinimumWidth ------------------------------------------
	XSimpleGrid.prototype.layoutSetMinimumWidth = function() {
		var i, wt;

		XWidget.prototype.layoutSetMinimumWidth.call(this);
		this.cminw = this.prefWidth = 0;
		for(i = 0;i < this.children.length;++i) {
			wt = this.children[i];
			wt.layoutSetMinimumWidth();
			if (this.cminw < wt.cwidth)
				this.cminw = wt.cwidth;
			if (this.prefWidth < wt.prefWidth)
				this.prefWidth = wt.prefWidth;
		}
		if (this.prefWidth < this.cminw)
			this.prefWidth = this.cminw;
		this.cwidth = this.cminw * this.mincols;
		this.cwidth +=(this.mincols -1) * this.gapw;
		this.cwidth += this.getEdgeWidth();
		this.prefWidth *= this.mincols;
		this.prefWidth +=(this.mincols -1) * this.gapw;
		this.prefWidth += this.getEdgeWidth();
		if (this.minWidth < this.cwidth)
			this.minWidth = this.cwidth - this.getEdgeWidth();
	};

	// layoutSetMinimumHeight -----------------------------------------
	XSimpleGrid.prototype.layoutSetMinimumHeight = function() {
		var hmax, i, wt;

		XWidget.prototype.layoutSetMinimumHeight.call(this);
		hmax = 0;
		for(i = 0;i < this.children.length;++i) {
			wt = this.children[i];
			wt.layoutSetMinimumHeight();
			if (hmax < wt.cheight)
				hmax = wt.cheight;
		}
		this.cheight =(hmax * this.nrows) +(this.nrows -1) * this.gaph + this.getEdgeHeight();
	};

	// layoutSetViewportWidth -----------------------------------------
	XSimpleGrid.prototype.layoutSetViewportWidth = function(vwidth) {
		var i, w, wt;

		XWidget.prototype.layoutSetViewportWidth.call(this, vwidth);
		this.cellw = 0;
		for(i = 0;i < this.children.length;++i) {
			wt = this.children[i];
			if ((w = wt.prefWidth + wt.getEdgeWidth()) > this.cellw)
				this.cellw = w;
		}
		w = this.cwidth - this.getEdgeWidth() -(this.mincols -1) * this.gapw;
		if (this.cellw * this.mincols > w)
			if ((this.cellw = UILib.mathFloor(w / this.mincols)) < this.cminw)
				this.cellw = this.cminw;
		for(i = 0;i < this.children.length;++i)
			this.children[i].layoutSetViewportWidth(this.cellw);
		this.ncols = this.mincols;
		w = this.getWidth() + this.getEdgeWidth();
		for(;this.ncols < this.children.length && this.ncols < this.maxcols;++this.ncols)
			if ((w += this.cellw + this.gapw) > this.cwidth)
				break;
		this.nrows = UILib.mathFloor((this.children.length + this.ncols -1) / this.ncols);
		this.ncols = UILib.mathFloor((this.children.length + this.nrows -1) / this.nrows);
	};

	// layoutSetViewportHeight ----------------------------------------
	XSimpleGrid.prototype.layoutSetViewportHeight = function(vheight) {
		var h, i;

		XWidget.prototype.layoutSetViewportHeight.call(this, vheight);
		this.cellh = 0;
		for(i = 0;i < this.children.length;++i)
			if ((h = this.children[i].cheight) > this.cellh)
				this.cellh = h;
		for(i = 0;i < this.children.length;++i)
			this.children[i].layoutSetViewportHeight(this.cellh);
	};

	// Constructor ----------------------------------------------------
	{
		this.children = XLib.createChildrenWidget(data);
		if ((this.gapw = data.getAttrInt("gap", -1)) !== -1)
			this.gaph = this.gapw;
		else {
			this.gapw = data.getInt("gapw");
			this.gaph = data.getInt("gaph");
		}
		this.maxcols = data.getAttrInt("maxcols", 0);
		if (this.maxcols < 1 || this.maxcols > this.children.length)
			this.maxcols = this.children.length;
		this.mincols = data.getAttrInt("mincols", 0);
		if (this.mincols < 1 || this.mincols > this.maxcols)
			this.mincols = 1;
		this.compact = data.getBool("compact");
		this.align = data.getAlign("align");
	}
};
/* ====================================================================	*
 * XSize 																*
 * ====================================================================	*/

var XSize = function(width, height) {

	// Constructor ----------------------------------------------------
	{
		this.width = width;
		this.height = height;
	}
};
/* ====================================================================	*
 * XSlide 																*
 * ====================================================================	*/

var XSlide = function(data) {

	XWidget.call(this, data);

	this.appl = null;
	this.plugin = null;
	this.index = 0;
	this.files = null;
	this.parts = null;
	this.millis = 0;
	this.duration = 0;

	// doAction -------------------------------------------------------
	XSlide.prototype.doAction = function(ref, value) {
		var i, path, xd;

		switch(ref) {
		case "init":
			this.layout.reset();
			xd = this.appl.getData(this.xref);
			if ((this.files = XFile.getList(xd, "toc")) === null)
				throw new Error("element not found: lecture-info:toc");
			this.index = 0;
			if (this.files.length > 0)
				this.plugin.setImage(this.files[this.index].path);
			break;
		case "load":
			if (!UILib.stringsMatch(value, "cache:"))
				throw new Error("invalid value: " + value);
			path = value.substring(6);
			for(i = 0;i < this.files.length;++i) {
				if (XString.equals(path, this.files[i].path)) {
					if (this.index === i)
						return;
					this.index = i;
					this.plugin.setImage(path);
					this.plugin.repaint();
					return;
				}
			}
			throw new Error("invalid value: " + path);
		case "route":
			this.plugin.setRoute(this.getRoute(value));
			break;
		default:
			throw new Error("not yet implemented: " + ref);
		}
	};

	// getCurrentSlide ------------------------------------------------
	XSlide.prototype.getCurrentSlide = function() {
		return(this.files[this.index]);
	};

	// getCurrentTime -------------------------------------------------
	XSlide.prototype.getCurrentTime = function() {
		return(this.millisToTime(this.millis));
	};

	// getCurrentTimeInMillis -----------------------------------------
	XSlide.prototype.getCurrentTimeInMillis = function() {
		return(this.millis);
	};

	// getDuration ----------------------------------------------------
	XSlide.prototype.getDuration = function() {
		return(this.millisToTime(this.duration));
	};

	// getDurationInMillis --------------------------------------------
	XSlide.prototype.getDurationInMillis = function() {
		return(this.duration);
	};

	// getRoute -------------------------------------------------------
	XSlide.prototype.getRoute = function(value) {
		var i, j, list, n, prevx, prevy, route, s, x, y;

		if (XString.equals(value, "END"))
			return(UILib.newArray(0, null));
		list = value.split(" ");
		route = UILib.newArray(list.length * 2 -1, null);
		j = 0;
		prevx = prevy = -1;
		for(i = 0;i < list.length;++i) {
			s = list[i];
			if (XString.equals(s, "I")) {
				x = prevx;
				y = prevy;
			} else
				if ((n = s.indexOf(',')) === -1) {
					x = prevx;
					y = prevy;
				} else {
					x = UILib.parseInt(s.substring(0, n));
					y = UILib.parseInt(s.substring(n + 1));
				}
			if (i === 0)
				route[j ++] = new XPoint(x, y);
			else
				if (prevx === -1) {
					route[j ++] = new XPoint( -1, -1);
					route[j ++] = new XPoint(x, y);
				} else {
					route[j ++] = new XPoint((x + prevx) / 2,(y + prevy) / 2);
					route[j ++] = new XPoint(x, y);
				}
			prevx = x;
			prevy = y;
		}
		return(route);
	};

	// initEvents -----------------------------------------------------
	XSlide.prototype.initEvents = function(part) {
		var events, evt, i, idx;

		events = part.events;
		for(i = 0;i < events.length;++i) {
			evt = events[i];
			if (!evt.isSlide())
				continue;
			for(idx = 0;idx < this.files.length;++idx) {
				if (XString.equals(this.files[idx].path, evt.link)) {
					evt.idxf = idx;
					break;
				}
			}
		}
	};

	// initVideoParts -------------------------------------------------
	XSlide.prototype.initVideoParts = function(parts) {
		var ipart;

		this.parts = parts;
		for(ipart = 0;ipart < parts.length;++ipart)
			this.initEvents(parts[ipart]);
		this.millis = 0;
	};

	// millisToTime ---------------------------------------------------
	XSlide.prototype.millisToTime = function(millis) {
		var h, m, nsecs, s;

		nsecs = XInteger.round(millis / 1000);
		h = XInteger.floor(nsecs / 3600);
		nsecs -= h * 3600;
		m = XInteger.floor(nsecs / 60);
		nsecs -= m * 60;
		s = XInteger.toPadding(nsecs, 2);
		if (h === 0)
			return(m + ":" + s);
		return(h + ":" + XInteger.toPadding(m, 2) + ":" + s);
	};

	// setCanvasSize --------------------------------------------------
	XSlide.prototype.setCanvasSize = function(canvasw, canvash) {
		this.layout.setCanvasSize(canvasw, canvash);
	};

	// setCurrentTime -------------------------------------------------
	XSlide.prototype.setCurrentTime = function(millis) {
		this.millis = millis;
	};

	// setDuration ----------------------------------------------------
	XSlide.prototype.setDuration = function(duration) {
		this.duration = duration;
	};

	// Constructor ----------------------------------------------------
	{
		this.xref = data.getAttrString("xref", "lecture-info");
		this.layout = new XSlideLayout(this);
	}
};
/* ====================================================================	*
 * XSlideLayout 														*
 * ====================================================================	*/

var XSlideLayout = function(xslide) {

	this.canvasw = 0;
	this.canvash = 0;
	this.xmouse = 0;
	this.ymouse = 0;
	this.timebarVisible = false;
	this.barlineDuration = 0;
	this.frame = null;
	this.buttonArea = null;
	this.buttonImgPos = null;
	this.buttonOver = false;
	this.buttonPressed = false;
	this.barline = null;
	this.timebar = null;
	this.events = null;
	this.cursorx = 0;
	this.cursory = 0;
	this.cursorw = 0;
	this.cursorDragging = false;
	this.cursorDraggingX = 0;
	this.cursorDraggingDelta = 0;
	this.thumbnailZone = null;
	this.thumbnailZoneOver = false;
	this.thumbnailMillis = 0;
	this.videoLoading = false;
	this.videoRecording = false;
	this.recordingOffset = 0;
	this.recordingTime = 0;

	// getBarLine -----------------------------------------------------
	XSlideLayout.prototype.getBarLine = function() {
		return(this.barline);
	};

	// getButtonImagePosition -----------------------------------------
	XSlideLayout.prototype.getButtonImagePosition = function() {
		return(this.buttonImgPos);
	};

	// getButtonPosition ----------------------------------------------
	XSlideLayout.prototype.getButtonPosition = function() {
		return(this.buttonArea);
	};

	// getCursorDraggingPosition --------------------------------------
	XSlideLayout.prototype.getCursorDraggingPosition = function() {
		return(this.cursorDraggingX);
	};

	// getCursorDraggingTime ------------------------------------------
	XSlideLayout.prototype.getCursorDraggingTime = function() {
		return(XInteger.round(this.thumbnailMillis));
	};

	// getCursorWidth -------------------------------------------------
	XSlideLayout.prototype.getCursorWidth = function() {
		return(this.cursorw);
	};

	// getCursorX -----------------------------------------------------
	XSlideLayout.prototype.getCursorX = function() {
		return(this.cursorx);
	};

	// getCursorY -----------------------------------------------------
	XSlideLayout.prototype.getCursorY = function() {
		return(this.cursory);
	};

	// getEventsPositions ---------------------------------------------
	XSlideLayout.prototype.getEventsPositions = function() {
		return(this.events);
	};

	// getFramePosition -----------------------------------------------
	XSlideLayout.prototype.getFramePosition = function() {
		return(this.frame);
	};

	// getRecordingOffset ---------------------------------------------
	XSlideLayout.prototype.getRecordingOffset = function() {
		return(this.recordingOffset);
	};

	// getRecordingTime -----------------------------------------------
	XSlideLayout.prototype.getRecordingTime = function() {
		return(this.recordingTime);
	};

	// getSlideIndex --------------------------------------------------
	XSlideLayout.prototype.getSlideIndex = function() {
		var duration, evt, evts, i, ipart, last, millis, parts;

		duration = this.xslide.duration;
		if ((millis = XInteger.round((this.xmouse - this.timebar.x) * duration / this.timebar.w)) < 0)
			millis = 0;
		else
			if (millis > duration)
				millis = duration;
		this.thumbnailMillis = millis;
		if ((parts = this.xslide.parts) === null)
			return(0);
		evts = parts[0].events;
		if (evts.length === 0)
			return(0);
		last = evts[0].idxf;
		for(ipart = 0;ipart < parts.length;++ipart) {
			evts = parts[ipart].events;
			for(i = 0;i < evts.length;++i) {
				evt = evts[i];
				if (!evt.isSlide())
					continue;
				if (millis < evt.timecode)
					return(last);
				last = evt.idxf;
			}
		}
		return(last);
	};

	// getSlidePosition -----------------------------------------------
	XSlideLayout.prototype.getSlidePosition = function() {
		var f, r;

		f = this.xslide.getCurrentSlide();
		r = XUILib.getImagePosition(this.canvasw, this.canvash, f.width, f.height);
		return(r);
	};

	// getTimebar -----------------------------------------------------
	XSlideLayout.prototype.getTimebar = function() {
		return(this.timebar);
	};

	// getThumbnailPath -----------------------------------------------
	XSlideLayout.prototype.getThumbnailPath = function() {
		if (this.thumbnailFile === null)
			return null;
		return(this.thumbnailFile.path);
	};

	// getThumbnailPosition -------------------------------------------
	XSlideLayout.prototype.getThumbnailPosition = function() {
		var h, r, w;

		w = this.thumbnailFile.width;
		h = this.thumbnailFile.height;
		r = XUILib.getImagePosition(this.canvasw / 3, this.canvash / 3, w, h);
		if ((r.x = this.xmouse - r.w / 2) < 2)
			r.x = 2;
		if (r.x + r.w + 2 > this.canvasw)
			r.x = this.canvasw - r.w -2;
		r.y = this.frame.y - r.h -1;
		return(r);
	};

	// getThumbnailTime -----------------------------------------------
	XSlideLayout.prototype.getThumbnailTime = function() {
		return(this.xslide.millisToTime(this.thumbnailMillis));
	};

	// hasVideoRecording ----------------------------------------------
	XSlideLayout.prototype.hasVideoRecording = function() {
		return(this.videoRecording);
	};

	// isButtonOver ---------------------------------------------------
	XSlideLayout.prototype.isButtonOver = function() {
		return(this.buttonOver);
	};

	// isButtonPressed ------------------------------------------------
	XSlideLayout.prototype.isButtonPressed = function() {
		return(this.buttonPressed);
	};

	// isButtonVisible ------------------------------------------------
	XSlideLayout.prototype.isButtonVisible = function() {
		if (this.cursorDragging || this.videoLoading)
			return(false);
		switch(this.videoMode) {
		case XSlideLayout.LIVE_END:
		case XSlideLayout.VIDEO_STARTING:
		case XSlideLayout.VIDEO_END:
			return(false);
		default:
			return(true);
		}
	};

	// isCursorDragging -----------------------------------------------
	XSlideLayout.prototype.isCursorDragging = function() {
		return(this.cursorDragging && this.thumbnailZoneOver);
	};

	// isLivePlaying --------------------------------------------------
	XSlideLayout.prototype.isLivePlaying = function() {
		switch(this.videoMode) {
		case XSlideLayout.LIVE:
		case XSlideLayout.LIVE_END:
			return(true);
		default:
			return(false);
		}
	};

	// isThumbnailVisible ---------------------------------------------
	XSlideLayout.prototype.isThumbnailVisible = function() {
		return(this.thumbnailZoneOver);
	};

	// isTimebarVisible -----------------------------------------------
	XSlideLayout.prototype.isTimebarVisible = function() {
		return(this.timebarVisible);
	};

	// isVideoAtEnd ---------------------------------------------------
	XSlideLayout.prototype.isVideoAtEnd = function() {
		switch(this.videoMode) {
		case XSlideLayout.LIVE_END:
		case XSlideLayout.VIDEO_END:
			return(true);
		default:
			return(false);
		}
	};

	// isVideoInPause -------------------------------------------------
	XSlideLayout.prototype.isVideoInPause = function() {
		return(this.videoMode === XSlideLayout.VIDEO_PAUSE);
	};

	// isVideoPlaying -------------------------------------------------
	XSlideLayout.prototype.isVideoPlaying = function() {
		switch(this.videoMode) {
		case XSlideLayout.VIDEO_STARTING:
		case XSlideLayout.VIDEO_RUNNING:
			return(true);
		default:
			return(false);
		}
	};

	// reset ----------------------------------------------------------
	XSlideLayout.prototype.reset = function() {
		this.timebarVisible = false;
		this.buttonOver = false;
		this.buttonPressed = false;
		this.cursorDragging = false;
		this.thumbnailZoneOver = false;
		this.videoLoading = false;
		this.videoRecording = false;
		this.videoMode = XSlideLayout.NO_VIDEO;
	};

	// setCanvasSize --------------------------------------------------
	XSlideLayout.prototype.setCanvasSize = function(canvasw, canvash) {
		this.canvasw = canvasw;
		this.canvash = canvash;
	};

	// setEventsPositions ---------------------------------------------
	XSlideLayout.prototype.setEventsPositions = function() {
		var evts, i, ipart, n, nevents, parts, r, x;

		if ((parts = this.xslide.parts) === null) {
			this.events = UILib.newArray(0, 0);
			return;
		}
		r = this.timebar.w / this.xslide.duration;
		nevents = 0;
		for(ipart = 0;ipart < parts.length;++ipart)
			nevents += parts[ipart].getNumberOfSlideEvents();
		n = 0;
		this.events = UILib.newArray(nevents, 0);
		for(ipart = 0;ipart < parts.length;++ipart) {
			evts = parts[ipart].events;
			for(i = 0;i < evts.length;++i) {
				if (!evts[i].isSlide())
					continue;
				x = XInteger.round(r * evts[i].timecode);
				if (x + this.timebar.h > this.timebar.w)
					x = this.timebar.w - this.timebar.h;
				this.events[n ++] = x;
			}
		}
	};

	// setMouseDown ---------------------------------------------------
	XSlideLayout.prototype.setMouseDown = function() {
		var d, r;

		if (this.videoMode === XSlideLayout.VIDEO_STARTING)
			return;
		if (this.buttonArea.isInside(this.xmouse, this.ymouse)) {
			this.buttonPressed = true;
			return;
		}
		d = this.cursorw / 2;
		r = new XRectangle(this.cursorx - d, this.cursory - d, this.cursorw, this.cursorw);
		if (r.isInside(this.xmouse, this.ymouse)) {
			this.cursorDragging = true;
			this.cursorDraggingDelta = this.xmouse - this.cursorx;
			this.cursorDraggingX = this.xmouse - this.cursorDraggingDelta;
		}
	};

	// setMouseEvents -------------------------------------------------
	XSlideLayout.prototype.setMouseEvents = function(x, y, pressed) {
		var idx, max, px;

		this.xmouse = x;
		this.ymouse = y;
		if ((this.thumbnailZoneOver = this.thumbnailZone.isInside(x, y)) === true) {
			idx = this.getSlideIndex();
			if (idx < this.xslide.files.length)
				this.thumbnailFile = this.xslide.files[idx];
		}
		if (this.cursorDragging && this.thumbnailZoneOver) {
			if ((px = x - this.cursorDraggingDelta) < this.timebar.x)
				px = this.timebar.x;
			else
				if ((max = this.timebar.x + this.timebar.w) < px)
					px = max;
			this.cursorDraggingX = px;
			return;
		}
		if (this.buttonArea.isInside(x, y) && this.videoMode !== XSlideLayout.VIDEO_STARTING) {
			this.buttonOver = true;
			this.buttonPressed = pressed;
		} else {
			this.buttonOver = false;
			this.buttonPressed = false;
		}
	};

	// setMouseUp -----------------------------------------------------
	XSlideLayout.prototype.setMouseUp = function() {
		this.buttonPressed = false;
		this.cursorDragging = false;
	};

	// setRecordingStartingAt -----------------------------------------
	XSlideLayout.prototype.setRecordingStartingAt = function(offset, time) {
		this.recordingOffset = offset;
		this.recordingTime = time;
	};

	// setTimebarVisible ----------------------------------------------
	XSlideLayout.prototype.setTimebarVisible = function(visible) {
		var t;

		this.timebarVisible = visible;
		t = UILib.mathRound(this.xslide.getDurationInMillis() / 1000);
		this.barlineDuration =(t + this.barlineDelta) * 1000;
	};

	// update ---------------------------------------------------------
	XSlideLayout.prototype.update = function() {
		var d, frameh, gap, h, marg, pad, tbarh, tbarw, w, x, y;

		marg = 20;
		frameh = 48;
		gap = 10;
		tbarh = 5;
		pad = 5;
		this.cursorw = 16;
		x = marg;
		y = this.canvash - marg - frameh;
		w = this.canvasw -2 * marg;
		h = frameh;
		this.frame = new XRectangle(x, y, w, h);
		d = 5 * pad;
		this.thumbnailZone = new XRectangle(x, y - d, w, h + d);
		w = this.buttonImgSz.width + 6 * pad;
		h = this.buttonImgSz.height + pad;
		x = this.frame.x + gap;
		y = this.frame.y +(this.frame.h - h) / 2;
		this.buttonArea = new XRectangle(x, y, w, h);
		this.buttonImgPos = new XPoint(x + 3 * pad, y + pad / 2);
		x = this.buttonArea.x + this.buttonArea.w + gap + this.cursorw / 2;
		y = this.frame.y + 3 * pad;
		w = this.frame.w - x;
		if (this.xslide.duration > this.barlineDuration)
			this.barlineDuration += this.barlineDelta * 1000;
		tbarw = w * this.xslide.duration / this.barlineDuration;
		this.timebar = new XRectangle(x, y, tbarw, tbarh);
		if ((w -= tbarw) === 0)
			this.barline = null;
		else {
			x += tbarw;
			y += tbarh / 2;
			this.barline = new XRectangle(x, y, w, 1);
		}
		this.setEventsPositions();
		this.cursorx = this.timebar.x + XInteger.round(this.xslide.millis * tbarw / this.xslide.duration);
		this.cursory = this.timebar.y + this.timebar.h / 2;
	};

	// setVideoLoading ------------------------------------------------
	XSlideLayout.prototype.setVideoLoading = function(loading) {
		this.videoLoading = loading;
	};

	// setVideoMode ---------------------------------------------------
	XSlideLayout.prototype.setVideoMode = function(mode) {
		switch(mode) {
		case "no-video":
			this.videoMode = XSlideLayout.NO_VIDEO;
			return;
		case "live":
			this.videoMode = XSlideLayout.LIVE;
			return;
		case "live-end":
			this.videoMode = XSlideLayout.LIVE_END;
			return;
		case "video-starting":
			this.videoMode = XSlideLayout.VIDEO_STARTING;
			return;
		case "video-running":
			this.videoMode = XSlideLayout.VIDEO_RUNNING;
			return;
		case "video-end":
			this.videoMode = XSlideLayout.VIDEO_END;
			return;
		case "video-pause":
			this.videoMode = XSlideLayout.VIDEO_PAUSE;
			return;
		default:
			throw new Error("invalid mode: " + mode);
		}
	};

	// setVideoRecording ----------------------------------------------
	XSlideLayout.prototype.setVideoRecording = function(recording) {
		this.videoRecording = recording;
	};

	// Constructor ----------------------------------------------------
	{
		this.xslide = xslide;
		this.buttonImgSz = new XSize(24, 24);
		this.videoMode = XSlideLayout.NO_VIDEO;
		this.barlineDelta = 60;
		this.thumbnailFile = null;
	}
};
/* ====================================================================	*
 * XSplitPane 															*
 * ====================================================================	*/

var XSplitPane = function(data) {

	XWidget.call(this, data);

	this.dividers = null;
	this.minv = 0;
	this.pwidth = 0;
	this.idivider = 0;
	this.percent = null;
	this.first = 0;
	this.last = 0;
	this.ori = null;

	// copy -----------------------------------------------------------
	XSplitPane.prototype.copy = function(a) {
		var i, r;

		r = UILib.newArray(a.length, 0);
		for(i = 0;i < a.length;++i)
			r[i] = a[i];
		return(r);
	};

	// getDivider -----------------------------------------------------
	XSplitPane.prototype.getDivider = function(n) {
		var h, i, maxh, maxw, w, x, y;

		switch(this.orientation) {
		case XLib.Horizontal:
			x = this.getX() + this.values[0] + 1;
			y = this.getY();
			w = this.gap -2;
			maxh = this.cheight - this.getEdgeHeight();
			if ((h = maxh / 2) > 50)
				h =(maxh < 50)?maxh:50;
			y +=(maxh - h) / 2;
			for(i = 1;i <= n;++i)
				x += this.gap + this.values[i];
			break;
		case XLib.Vertical:
			x = this.getX();
			y = this.getY() + this.values[0] + 1;
			maxw = this.cwidth - this.getEdgeWidth();
			h = this.gap -2;
			if ((w = maxw / 2) > 50)
				w =(maxw < 50)?maxw:50;
			x +=(maxw - w) / 2;
			for(i = 1;i <= n;++i)
				y += this.gap + this.values[i];
			break;
		default:
			throw new Error("internal error: " + this.orientation);
		}
		return(new XRectangle(x, y, w, h));
	};

	// getDividerZone -------------------------------------------------
	XSplitPane.prototype.getDividerZone = function(px, py) {
		var h, i, w, x, y;

		switch(this.orientation) {
		case XLib.Horizontal:
			x = this.getX() + this.values[0];
			y = this.getY();
			h = this.cheight - this.getEdgeHeight();
			for(i = 0;i < this.dividers.length;++i) {
				if (x <= px && px < x + this.gap && y <= py && py < y + h)
					return(i);
				x += this.gap + this.values[i + 1];
			}
			return( -1);
		case XLib.Vertical:
			x = this.getX();
			w = this.cwidth - this.getEdgeWidth();
			y = this.getY() + this.values[0];
			for(i = 0;i < this.dividers.length;++i) {
				if (x <= px && px < x + w && y <= py && py < y + this.gap)
					return(i);
				y += this.gap + this.values[i + 1];
			}
			return( -1);
		default:
			throw new Error("internal error:" + this.orientation);
		}
	};

	// initDividers ---------------------------------------------------
	XSplitPane.prototype.initDividers = function(data) {
		var i, r, v, wt;

		v = data.getString("buttonc");
		r = UILib.newXData("area");
		r.setStringAttr("bg", v);
		r.setStringAttr("expand", "100%");
		this.dividers = UILib.newArray(this.panes.length -1, null);
		for(i = 0;i < this.dividers.length;++i) {
			wt = XLib.createWidget(data.toUIXWData(r, true));
			wt.layoutInit();
			this.dividers[i] = wt;
		}
		this.idivider = -1;
		this.children = UILib.newArray(this.panes.length + this.dividers.length, null);
		for(i = 0;i < this.children.length;++i)
			this.children[i] =(i < this.panes.length)?this.panes[i]:this.dividers[i - this.panes.length];
	};

	// initHorizontal -------------------------------------------------
	XSplitPane.prototype.initHorizontal = function() {
		var i, n, nleft, w, widths;

		this.percent = UILib.newArray(this.panes.length, 0);
		nleft = this.cwidth - this.getEdgeWidth() - this.dividers.length * this.gap;
		if (this.weight === null) {
			if (this.pwidth <= nleft) {
				w = this.pwidth;
				widths = this.prefw;
			} else {
				w = this.minv;
				widths = this.min;
			}
			for(i = 0;i < this.panes.length;++i)
				this.percent[i] = widths[i] / w;
			return;
		}
		widths = UILib.newArray(this.panes.length, 0);
		n = nleft;
		for(i = 0;i < this.panes.length;++i) {
			this.percent[i] = this.weight[i];
			widths[i] = UILib.mathFloor(nleft * this.percent[i]);
			n -= widths[i];
		}
		for(i = 0;i < n;++i)
			widths[i] ++;
		XLib.adjustMinimumWidth(this.min, widths);
		w = nleft;
		for(i = 0;i < this.panes.length;++i)
			this.percent[i] = widths[i] / w;
	};

	// initVertical ---------------------------------------------------
	XSplitPane.prototype.initVertical = function() {
		var h, heights, i, n, nleft;

		this.percent = UILib.newArray(this.panes.length, 0);
		if (this.weight === null) {
			for(i = 0;i < this.panes.length;++i)
				this.percent[i] = this.min[i] / this.minv;
			return;
		}
		heights = UILib.newArray(this.panes.length, 0);
		nleft = this.cheight - this.getEdgeHeight() - this.dividers.length * this.gap;
		n = nleft;
		for(i = 0;i < this.panes.length;++i) {
			this.percent[i] = this.weight[i];
			heights[i] = UILib.mathFloor(nleft * this.percent[i]);
			n -= heights[i];
		}
		for(i = 0;i < n;++i)
			heights[i] ++;
		XLib.adjustMinimumWidth(this.min, heights);
		h = nleft;
		for(i = 0;i < this.panes.length;++i)
			this.percent[i] = heights[i] / h;
	};

	// layoutInit -----------------------------------------------------
	XSplitPane.prototype.layoutInit = function() {
		var dw, i, mw, pw, w, wt;

		for(i = 0;i < this.panes.length;++i) {
			wt = this.panes[i];
			if (wt.expandHor === 0)
				wt.expandHor = 100;
			if (wt.expandVer === 0)
				wt.expandVer = 100;
			wt.layoutInit();
			wt.layoutSetMinimumWidth();
		}
		mw = pw = 0;
		switch(this.orientation) {
		case XLib.Horizontal:
			for(i = 0;i < this.panes.length;++i) {
				wt = this.panes[i];
				dw = wt.getEdgeWidth();
				this.min[i] = wt.minWidth + dw;
				mw += this.min[i];
				this.prefw[i] = wt.prefWidth + dw;
				pw += this.prefw[i];
			}
			this.minv = mw;
			this.pwidth = pw;
			mw += this.gap * this.dividers.length;
			pw += this.gap * this.dividers.length;
			break;
		case XLib.Vertical:
			for(i = 0;i < this.panes.length;++i) {
				wt = this.panes[i];
				dw = wt.getEdgeWidth();
				if ((w = wt.minWidth + dw) > mw)
					mw = w;
				if ((w = wt.prefWidth + dw) > pw)
					pw = w;
			}
			break;
		}
		if (this.minWidth < mw)
			this.minWidth = mw;
		if (this.prefWidth < pw)
			this.prefWidth = pw;
	};

	// layoutSetLocation ----------------------------------------------
	XSplitPane.prototype.layoutSetLocation = function(vx, vy, anchor) {
		var i, r, wt, x, y;

		XWidget.prototype.layoutSetLocation.call(this, vx, vy, anchor);
		x = this.getX();
		y = this.getY();
		switch(this.orientation) {
		case XLib.Horizontal:
			for(i = 0;i < this.panes.length;++i) {
				wt = this.panes[i];
				wt.layoutSetLocation(x, y, wt.anchor);
				x += this.values[i] + this.gap;
			}
			break;
		case XLib.Vertical:
			for(i = 0;i < this.panes.length;++i) {
				wt = this.panes[i];
				wt.layoutSetLocation(x, y, wt.anchor);
				y += this.values[i] + this.gap;
			}
			break;
		}
		for(i = 0;i < this.dividers.length;++i) {
			r = this.getDivider(i);
			wt = this.dividers[i];
			wt.layoutSetMinimumWidth();
			wt.layoutSetViewportWidth(r.w);
			wt.layoutSetMinimumHeight();
			wt.layoutSetViewportHeight(r.h);
			wt.layoutSetLocation(r.x, r.y, "north");
		}
	};

	// layoutSetMinimumWidth ------------------------------------------
	XSplitPane.prototype.layoutSetMinimumWidth = function() {
		var i;

		XWidget.prototype.layoutSetMinimumWidth.call(this);
		for(i = 0;i < this.panes.length;++i)
			this.panes[i].layoutSetMinimumWidth();
	};

	// layoutSetMinimumHeight -----------------------------------------
	XSplitPane.prototype.layoutSetMinimumHeight = function() {
		var h, i, wt;

		XWidget.prototype.layoutSetMinimumHeight.call(this);
		switch(this.orientation) {
		case XLib.Horizontal:
			h = 0;
			for(i = 0;i < this.panes.length;++i) {
				wt = this.panes[i];
				wt.layoutSetMinimumHeight();
				if (h < wt.cheight)
					h = wt.cheight;
			}
			break;
		case XLib.Vertical:
			h = 0;
			for(i = 0;i < this.panes.length;++i) {
				wt = this.panes[i];
				wt.layoutSetMinimumHeight();
				this.min[i] = this.values[i] = wt.cheight;
				h += wt.cheight;
			}
			this.minv = h;
			h += this.dividers.length * this.gap;
			break;
		default:
			throw new Error("internal error:" + this.orientation);
		}
		h += this.getEdgeHeight();
		if (h > this.cheight)
			this.cheight = h;
	};

	// layoutSetViewportWidth -----------------------------------------
	XSplitPane.prototype.layoutSetViewportWidth = function(vwidth) {
		var i, nleft, total, w;

		XWidget.prototype.layoutSetViewportWidth.call(this, vwidth);
		switch(this.orientation) {
		case XLib.Horizontal:
			if (this.percent === null)
				this.initHorizontal();
			total = this.cwidth - this.getEdgeWidth() - this.dividers.length * this.gap;
			nleft = total;
			for(i = 0;i < this.panes.length;++i) {
				w = this.panes[i].cwidth;
				this.values[i] = w;
				nleft -= w;
			}
			if (nleft < 0)
				throw new Error("vwidth too small: " + vwidth + "/" + nleft);
			for(i = 0;i < this.panes.length;++i) {
				if ((w = UILib.mathRound(total * this.percent[i]) - this.values[i]) < 0)
					w = 0;
				else
					if (w > nleft)
						w = nleft;
				this.values[i] += w;
				nleft -= w;
			}
			for(i = 0;i < nleft;++i)
				this.values[i] += 1;
			for(i = 0;i < this.panes.length;++i)
				this.panes[i].layoutSetViewportWidth(this.values[i]);
			break;
		case XLib.Vertical:
			w = this.cwidth - this.getEdgeWidth();
			for(i = 0;i < this.panes.length;++i)
				this.panes[i].layoutSetViewportWidth(w);
			break;
		}
	};

	// layoutSetViewportHeight ----------------------------------------
	XSplitPane.prototype.layoutSetViewportHeight = function(vheight) {
		var h, i, nleft, total;

		XWidget.prototype.layoutSetViewportHeight.call(this, vheight);
		switch(this.orientation) {
		case XLib.Horizontal:
			h = this.cheight - this.getEdgeHeight();
			for(i = 0;i < this.panes.length;++i)
				this.panes[i].layoutSetViewportHeight(h);
			break;
		case XLib.Vertical:
			if (this.percent === null)
				this.initVertical();
			total = this.cheight - this.getEdgeHeight() - this.dividers.length * this.gap;
			nleft = total;
			for(i = 0;i < this.panes.length;++i) {
				h = this.panes[i].cheight;
				this.values[i] = h;
				nleft -= h;
			}
			if (nleft < 0)
				throw new Error("vheight too small: " + vheight + "/" + nleft);
			for(i = 0;i < this.panes.length;++i) {
				if ((h = UILib.mathRound(total * this.percent[i]) - this.values[i]) < 0)
					h = 0;
				else
					if (h > nleft)
						h = nleft;
				this.values[i] += h;
				nleft -= h;
			}
			for(i = 0;i < nleft;++i)
				this.values[i] += 1;
			for(i = 0;i < this.panes.length;++i)
				this.panes[i].layoutSetViewportHeight(this.values[i]);
			break;
		default:
			throw new Error("internal error: " + this.orientation);
		}
	};

	// mouseDragged ---------------------------------------------------
	XSplitPane.prototype.mouseDragged = function(px, py) {
		var d, delta, i, idx1, idx2, p, tmp, total, v;

		if (this.idivider === -1)
			return(false);
		switch(this.orientation) {
		case XLib.Horizontal:
			p = px;
			tmp = this.values;
			total = this.cwidth - this.getEdgeWidth();
			break;
		case XLib.Vertical:
			p = py;
			tmp = this.values;
			total = this.cheight - this.getEdgeHeight();
			break;
		default:
			throw new Error("internal error: " + this.orientation);
		}
		total -= this.dividers.length * this.gap;
		if (p === this.last)
			return(false);
		if ((delta = this.first - p) > 0) {
			idx1 = this.idivider;
			idx2 = this.idivider + 1;
		} else {
			delta = - delta;
			idx1 = this.idivider + 1;
			idx2 = this.idivider;
		}
		if ((v = this.ori[idx1] - delta) < this.min[idx1])
			return(false);
		tmp[idx1] = v;
		tmp[idx2] = v + delta;
		this.percent[idx1] = tmp[idx1] / total;
		d = 1.0;
		for(i = 0;i < this.percent.length;++i)
			if (i !== idx2)
				d -= this.percent[i];
		this.percent[idx2] = d;
		this.last = p;
		return(true);
	};

	// mouseIsInsideDivider -------------------------------------------
	XSplitPane.prototype.mouseIsInsideDivider = function(px, py) {
		return(this.getDividerZone(px, py) !== -1);
	};

	// mousePressed ---------------------------------------------------
	XSplitPane.prototype.mousePressed = function(px, py) {
		if ((this.idivider = this.getDividerZone(px, py)) === -1)
			return(false);
		switch(this.orientation) {
		case XLib.Horizontal:
			this.first = this.last = px;
			this.ori = this.copy(this.values);
			break;
		case XLib.Vertical:
			this.first = this.last = py;
			this.ori = this.copy(this.values);
			break;
		}
		return(true);
	};

	// mouseReset -----------------------------------------------------
	XSplitPane.prototype.mouseReset = function() {
		this.idivider = -1;
	};

	// Constructor ----------------------------------------------------
	{
		this.gap = data.getInt("gap");
		this.orientation = data.getOrientation("orientation");
		this.panes = XLib.createChildrenWidget(data);
		if ((this.lweight = data.getAttrString("weight", null)) !== null)
			this.weight = data.getPercentArray("weight", this.lweight, this.panes.length);
		else
			this.weight = null;
		this.initDividers(data);
		this.min = UILib.newArray(this.panes.length, 0);
		this.values = UILib.newArray(this.panes.length, 0);
		switch(this.orientation) {
		case XLib.Horizontal:
			this.prefw = UILib.newArray(this.panes.length, 0);
			break;
		}
	}
};
/* ====================================================================	*
 * XStack 																*
 * ====================================================================	*/

var XStack = function(data) {

	XWidget.call(this, data);

	// layoutInit -----------------------------------------------------
	XStack.prototype.layoutInit = function() {
		var i, minw, prefw, wt;

		minw = prefw = 0;
		for(i = 0;i < this.children.length;++i) {
			wt = this.children[i];
			wt.layoutInit();
			if (minw < wt.minWidth)
				minw = wt.minWidth;
			if (prefw < wt.prefWidth)
				prefw = wt.prefWidth;
		}
		if (prefw < minw)
			prefw = minw;
		if (minw > this.minWidth)
			this.minWidth = minw;
		if (prefw > this.prefWidth)
			this.prefWidth = prefw;
	};

	// layoutSetLocation ----------------------------------------------
	XStack.prototype.layoutSetLocation = function(vx, vy, anchor) {
		XWidget.prototype.layoutSetLocation.call(this, vx, vy, anchor);
		this.view.layoutSetLocation(this.getX(), this.getY(), "center");
	};

	// layoutSetMinimumWidth ------------------------------------------
	XStack.prototype.layoutSetMinimumWidth = function() {
		var i;

		XWidget.prototype.layoutSetMinimumWidth.call(this);
		this.layoutInit();
		for(i = 0;i < this.children.length;++i)
			this.children[i].layoutSetMinimumWidth();
	};

	// layoutSetMinimumHeight -----------------------------------------
	XStack.prototype.layoutSetMinimumHeight = function() {
		var i, maxh, wt;

		XWidget.prototype.layoutSetMinimumHeight.call(this);
		maxh = 0;
		for(i = 0;i < this.children.length;++i) {
			wt = this.children[i];
			wt.layoutSetMinimumHeight();
			if (maxh < wt.cheight)
				maxh = wt.cheight;
		}
		this.cheight = maxh + this.getEdgeHeight();
	};

	// layoutSetViewportWidth -----------------------------------------
	XStack.prototype.layoutSetViewportWidth = function(vwidth) {
		var i, w;

		XWidget.prototype.layoutSetViewportWidth.call(this, vwidth);
		w = this.cwidth - this.getEdgeWidth();
		for(i = 0;i < this.children.length;++i)
			this.children[i].layoutSetViewportWidth(w);
	};

	// layoutSetViewportHeight ----------------------------------------
	XStack.prototype.layoutSetViewportHeight = function(vheight) {
		var h, i;

		XWidget.prototype.layoutSetViewportHeight.call(this, vheight);
		h = this.cheight - this.getEdgeHeight();
		for(i = 0;i < this.children.length;++i)
			this.children[i].layoutSetViewportHeight(h);
	};

	// reset ----------------------------------------------------------
	XStack.prototype.reset = function() {
		this.idx = 0;
		this.view = this.children[this.idx];
	};

	// task -----------------------------------------------------------
	XStack.prototype.task = function(taskId) {
		var uier, uieu;

		switch(taskId) {
		case "next":
			if (this.idx + 1 === this.children.length)
				throw new Error("invalid task: " + taskId + " [idx=" + this.idx + "]");
			this.view = this.children[ ++this.idx];
			break;
		case "prev":
			if (this.idx === 0)
				throw new Error("invalid task: " + taskId + " [idx=" + this.idx + "]");
			this.view = this.children[ --this.idx];
			break;
		default:
			throw new Error("not yet implemented: " + taskId);
		}
		this.view.uiwt.layoutReset();
		uieu =this.uiwt;
		uieu.eventUpdateUI();
		uier =this.uiwt;
		uier.eventResizeUI();
	};

	// showWidget -----------------------------------------------------
	XStack.prototype.showWidget = function(widgetid) {
		var i, uier, uieu;

		for(i = 0;i < this.children.length;++i)
			if (XString.equals(widgetid, this.children[i].id))
				break;
		if (i === this.children.length)
			throw new Error("invalid widgetId: " + widgetid);
		if (i === this.idx)
			return;
		this.idx = i;
		this.view = this.children[this.idx];
		this.view.uiwt.layoutReset();
		uieu =this.uiwt;
		uieu.eventUpdateUI();
		uier =this.uiwt;
		uier.eventResizeUI();
	};

	// Constructor ----------------------------------------------------
	{
		this.children = XLib.createChildrenWidget(data);
		this.idx = 0;
		this.view = this.children[this.idx];
	}
};
/* ====================================================================	*
 * XSwitchAction 														*
 * ====================================================================	*/

var XSwitchAction = function(data) {

	// add ------------------------------------------------------------
	XSwitchAction.prototype.add = function(key, children, begin, end) {
		var block, i, len;

		if (this.blocks.exists(key))
			throw new Error("duplicate key: [" + key + "]");
		len = end - begin;
		block = UILib.newArray(len, null);
		for(i = 0;i < len;++i)
			block[i] = new XAction(children[begin + i]);
		this.blocks.put(key, block);
	};

	// Constructor ----------------------------------------------------
	{
		var child, children, i, key, off, tag;

		this.var = data.getString("var");
		this.blocks = XHMap.createXActionArray();
		children = data.getAllChildren();
		off = -1;
		key = null;
		for(i = 0;i < children.length;++i) {
			child = children[i];
			tag = child.getTag();
			switch(tag) {
			case "case":
				if (off !== -1)
					this.add(key, children, off, i);
				key = child.getString("value");
				off = i + 1;
				break;
			case "default":
				if (off !== -1)
					this.add(key, children, off, i);
				key = "";
				off = i + 1;
				break;
			default:
				if (off === -1)
					throw new Error("<case> or <default/> not found: " + tag);
			}
		}
		if (key !== null)
			this.add(key, children, off, i);
	}
};
/* ====================================================================	*
 * XTableHeader 														*
 * ====================================================================	*/

var XTableHeader = function(el) {

	// setConfig ------------------------------------------------------
	XTableHeader.prototype.setConfig = function(cfg) {
		var child;

		child = cfg.addChild("header");
		child.setStringAttr("bg", this.bg);
		child.setStringAttr("fg", this.fg);
		child.setStringAttr("font", this.font);
		child.setStringAttr("fontid", this.fontid);
	};

	// Constructor ----------------------------------------------------
	{
		this.bg = el.getAttrString("bg", "white");
		this.fg = el.getAttrString("fg", "black");
		this.font = el.getAttrString("font", "plain");
		this.fontid = el.getAttrString("fontid", null);
	}
};
/* ====================================================================	*
 * XTask 																*
 * ====================================================================	*/

var XTask = function(data) {

	// run ------------------------------------------------------------
	XTask.prototype.run = function(appl) {
		var i;

		for(i = 0;i < this.actions.length;++i)
			if (!this.actions[i].run(this.id, appl))
				return(false);
		return(true);
	};

	// Constructor ----------------------------------------------------
	{
		var children, i;

		this.id = data.getString("id");
		children = data.getAllChildren();
		this.actions = UILib.newArray(children.length, null);
		for(i = 0;i < this.actions.length;++i)
			this.actions[i] = new XAction(children[i]);
	}
};
/* ====================================================================	*
 * XTasks 																*
 * ====================================================================	*/

var XTasks = function(appl, children) {

	// exists ---------------------------------------------------------
	XTasks.prototype.exists = function(id) {
		return(this.map.exists(id));
	};

	// list -----------------------------------------------------------
	XTasks.prototype.list = function() {
		var tasks;

		tasks = UILib.newArray(this.map.size(), null);
		this.map.values(tasks);
		return(tasks);
	};

	// run ------------------------------------------------------------
	XTasks.prototype.run = function(id) {
		var task;

		if ((task = this.map.get(id)) === null)
			throw new Error("invalid task: " + id);
		return(task.run(this.appl));
	};

	// Constructor ----------------------------------------------------
	{
		var i, task;

		this.appl = appl;
		this.map = UILib.getXTaskMap();
		for(i = 0;i < children.length;++i) {
			task = new XTask(children[i]);
			if (task.id === null)
				throw new Error("attribute 'id' for a task is not defined");
			this.map.put(task.id, task);
		}
	}
};
/* ====================================================================	*
 * XText 																*
 * ====================================================================	*/

var XText = function(data) {

	XWidget.call(this, data);

	// layoutInit -----------------------------------------------------
	XText.prototype.layoutInit = function() {
		if ((this.minWidth = this.uitext.getMinimumWidth()) < this.textWidthMin)
			this.minWidth = this.textWidthMin;
		if ((this.prefWidth = this.uitext.getPrefWidth()) < this.minWidth)
			this.prefWidth = this.minWidth;
		if (this.textWidthMax > 0) {
			if (this.textWidthMax < this.minWidth)
				this.textWidthMax = this.minWidth;
			if (this.prefWidth > this.textWidthMax)
				this.prefWidth = this.textWidthMax;
		}
	};

	// layoutSetViewportWidth -----------------------------------------
	XText.prototype.layoutSetViewportWidth = function(vwidth) {
		var fw, sz, w;

		if ((this.vwidth = vwidth) < this.cwidth)
			throw new Error("invalid vwidth: " + vwidth + " < " + this.cwidth);
		if (this.expandHor > 0)
			if ((w = vwidth * this.expandHor / 100) > this.cwidth)
				vwidth = w;
		if ((this.cwidth = this.prefWidth + this.getEdgeWidth()) > vwidth)
			this.cwidth = vwidth;
		fw = this.forcew;
		if (this.cwidth < vwidth && 0 < this.expandHor) {
			this.cwidth = vwidth;
			fw = true;
		}
		w = this.cwidth - this.getEdgeWidth();
		if (this.textWidthMax > 0)
			if (w > this.textWidthMax)
				w = this.textWidthMax;
		sz = this.uitext.format(w, fw);
		if (sz.width < this.minWidth)
			sz.width = this.minWidth;
		this.cwidth = sz.width + this.getEdgeWidth();
		this.minHeight = sz.height;
	};

	// Constructor ----------------------------------------------------
	{
		this.uitext = data.getText(this.fg);
		this.textWidthMin = data.getAttrInt("widthmin", 0);
		this.textWidthMax = data.getAttrInt("widthmax", 0);
		this.forcew = data.getBool("forcew");
		this.minWidth = this.minHeight = 0;
	}
};
/* ====================================================================	*
 * XTextArea 															*
 * ====================================================================	*/

var XTextArea = function(data) {

	XWidget.call(this, data);

	// Constructor ----------------------------------------------------
	{
		this.inputc = data.getColor("inputc");
		this.ncols = data.getInt("ncols");
		this.nrows = data.getInt("nrows");
		this.nospellcheck = data.getBool("nospellcheck");
	}
};
/* ====================================================================	*
 * XTextField 															*
 * ====================================================================	*/

var XTextField = function(data) {

	XWidget.call(this, data);

	// Constructor ----------------------------------------------------
	{
		this.ncols = data.getInt("ncols");
	}
};
/* ====================================================================	*
 * XTocEdit 															*
 * ====================================================================	*/

var XTocEdit = function(data) {

	XWidget.call(this, data);

	this.lectureId = null;
	this.files = null;
	this.cellw = 0;
	this.cellh = 0;
	this.nrows = 0;
	this.left = 0;
	this.appl = null;
	this.plugin = null;

	// addImages ------------------------------------------------------
	XTocEdit.prototype.addImages = function() {
		var i, newf, off, pos, tmp, xd;

		off = this.getOffset();
		xd = this.appl.getData("lecture-edit-add");
		if ((newf = XFile.getList(xd, "images")) === null)
			throw new Error("element not found: lecture-edit-add:images");
		tmp = UILib.newArray(this.files.length + newf.length, null);
		pos = 0;
		for(i = 0;i < off;++i)
			tmp[pos ++] = this.files[i];
		for(i = 0;i < newf.length;++i)
			tmp[pos ++] = newf[i];
		for(i = off;i < this.files.length;++i)
			tmp[pos ++] = this.files[i];
		this.files = tmp;
		this.plugin.init(this.files);
		this.plugin.cardUpdated();
	};

	// doAction -------------------------------------------------------
	XTocEdit.prototype.doAction = function(ref) {
		switch(ref) {
		case "init":
			this.init();
			break;
		case "new-images":
			this.addImages();
			break;
		case "remove":
			this.removeImage();
			break;
		case "save":
			this.save();
			break;
		case "zoom-in":
			if (this.ncols === 1)
				break;
			--this.ncols;
			this.plugin.cardUpdated();
			break;
		case "zoom-out":
			if (this.ncols === 8)
				break;
			++this.ncols;
			this.plugin.cardUpdated();
			break;
		default:
			throw new Error("not yet implemented: " + ref);
		}
	};

	// getImagePosition -----------------------------------------------
	XTocEdit.prototype.getImagePosition = function(idx) {
		var px, py, x, y;

		x = idx % this.ncols;
		y = UILib.mathFloor(idx / this.ncols);
		px = this.getX() + this.left + this.gap;
		px += x *(this.cellw + this.gap);
		py = this.getY() + this.gap;
		py += y *(this.cellh + this.gap);
		return(new XPoint(px, py));
	};

	// getOffset ------------------------------------------------------
	XTocEdit.prototype.getOffset = function() {
		var s;

		s = this.appl.getVar("lectureEditOffset");
		if (!XString.equals(s, "end"))
			return(UILib.parseInt(s));
		return(this.files.length);
	};

	// getPositionInfo ------------------------------------------------
	XTocEdit.prototype.getPositionInfo = function(x, y) {
		var idx, maxx, maxy, px, py;

		px = this.getX();
		py = this.getY();
		maxx = px + this.ncols * this.cellw +(this.ncols + 1) * this.gap;
		maxy = py + this.nrows * this.cellh +(this.nrows + 1) * this.gap;
		if (x < px || y < py || maxx <= x || maxy <= y)
			return(null);
		idx = 0;
		while(idx < this.files.length) {
			if ((py += this.gap) > y)
				return(new XTocEditPosInfo(idx, false));
			if ((py += this.cellh) > y)
				break;
			idx += this.ncols;
		}
		if (idx >= this.files.length)
			return(new XTocEditPosInfo(this.files.length, false));
		px += this.left;
		while(idx < this.files.length) {
			if ((px += this.gap) > x)
				break;
			if ((px += this.cellw) > x)
				return(new XTocEditPosInfo(idx, true));
			++idx;
		}
		return(new XTocEditPosInfo(idx, false));
	};

	// init -----------------------------------------------------------
	XTocEdit.prototype.init = function() {
		var child, xd;

		xd = this.appl.getData("lecture-info");
		child = xd.getFirstChildWithTag("lectureId");
		this.lectureId = child.getTextContent();
		if ((this.files = XFile.getList(xd, "toc")) === null)
			throw new Error("element not found: lecture-info:toc");
		this.plugin.init(this.files);
	};

	// moveImage ------------------------------------------------------
	XTocEdit.prototype.moveImage = function(isrc, idst) {
		var i, tmp;

		tmp = this.files[isrc];
		if (idst < isrc) {
			for(i = isrc;i > idst;--i)
				this.files[i] = this.files[i -1];
		} else {
			for(i = isrc;i < idst;++i)
				this.files[i] = this.files[i + 1];
		}
		this.files[idst] = tmp;
	};

	// removeImage ----------------------------------------------------
	XTocEdit.prototype.removeImage = function() {
		var i, idx, tmp;

		idx = this.plugin.getSelected();
		tmp = UILib.newArray(this.files.length -1, null);
		for(i = 0;i < idx;++i)
			tmp[i] = this.files[i];
		for(i = idx;i < tmp.length;++i)
			tmp[i] = this.files[i + 1];
		this.files = tmp;
		this.plugin.removeImage(this.files, idx);
	};

	// save -----------------------------------------------------------
	XTocEdit.prototype.save = function() {
		var i, n, path, xd;

		xd = UILib.newXmlData("toc");
		for(i = 0;i < this.files.length;++i) {
			path = this.files[i].path;
			if ((n = path.indexOf('/')) === -1)
				throw new Error("invalid path: " + path);
			xd.addChildWithContent("img", path.substring(n + 1));
		}
		this.appl.putData("lecture-edit-toc", xd);
	};

	// layoutSetViewportWidth -----------------------------------------
	XTocEdit.prototype.layoutSetViewportWidth = function(vwidth) {
		var w;

		XWidget.prototype.layoutSetViewportWidth.call(this, vwidth);
		for(;;) {
			w = this.cwidth - this.getEdgeWidth() - this.gap *(this.ncols + 1);
			this.cellw = UILib.mathFloor(w / this.ncols);
			this.left = UILib.mathFloor((w % this.ncols) / 2);
			this.cellh = UILib.mathFloor(this.cellw / 4) * 3;
			if (this.cellw > 96 && this.cellh > 64)
				break;
			--this.ncols;
		}
		this.nrows = UILib.mathFloor((this.files.length +(this.ncols -1)) / this.ncols);
		this.minHeight = this.nrows * this.cellh +(this.nrows + 1) * this.gap;
	};

	// Constructor ----------------------------------------------------
	{
		this.iborder = 2;
		this.gap = 10;
		this.ncols = 4;
	}
};
/* ====================================================================	*
 * XTocEditPosInfo 														*
 * ====================================================================	*/

var XTocEditPosInfo = function(idx, inside) {

	// Constructor ----------------------------------------------------
	{
		this.idx = idx;
		this.inside = inside;
	}
};
/* ====================================================================	*
 * XTree 																*
 * ====================================================================	*/

var XTree = function(data, cfg) {

	XScrollPane.call(this, data);

	// init -----------------------------------------------------------
	XTree.prototype.init = function() {
		var v;

		v =this.view;
		v.set(this);
	};

	// getNewInstance -------------------------------------------------
	XTree.prototype.getNewInstance = function(appl, xref) {
		this.cfg.setStringAttr("props", "xref=" + xref);
		return(XLib.createWidget(this.cfg.toUIXWData(appl)));
	};

	// replace --------------------------------------------------------
	XTree.prototype.replace = function(tree) {
		var v;

		this.setView(tree.view);
		v =tree.view;
		v.tree = this;
	};

	// setPath --------------------------------------------------------
	XTree.prototype.setPath = function(path) {
		var v;

		v =this.view;
		v.setPath(path);
	};

	// Constructor ----------------------------------------------------
	{
		this.cfg = cfg;
		this.init();
	}
};

// create -------------------------------------------------------------
XTree.create = function(data) {
	var action, attrs, bg, cfg, depth, depths, el, entries, entry, fg, font, fontid, hgap, i, id, j, labels, lminw, name, path, props, rpadding, selection, v, vgap, xdata, xtree, xview;

	xtree = UILib.newXData("tree");
	if ((id = data.getAttrString("id", null)) === null)
		throw new Error("id not found");
	xtree.setStringAttr("id", id);
	props = new XProps(data);
	xtree.setStringAttr("props", props.write("caching,xref"));
	if ((vgap = data.getAttrString("vgap", null)) !== null)
		xtree.setStringAttr("vgap", vgap);
	if ((hgap = data.getAttrString("hgap", null)) !== null)
		xtree.setStringAttr("hgap", hgap);
	action = bg = fg = font = fontid = lminw = selection = rpadding = null;
	attrs = data.getAttributes();
	for(i = 0;i < attrs.length;++i) {
		name = attrs[i].key;
		v = attrs[i].value;
		switch(name) {
		case "margin":
		case "border":
		case "size":
		case "expand":
		case "expandh":
		case "expandv":
		case "anchor":
			xtree.setStringAttr(name, v);
			break;
		case "bg":
			bg = v;
			break;
		case "fg":
			fg = v;
			break;
		case "action":
			action = v;
			break;
		case "font":
			font = v;
			break;
		case "fontid":
			fontid = v;
			break;
		case "selection":
			selection = v;
			break;
		case "rpadding":
			rpadding = v;
			break;
		case "labelminw":
			lminw = v;
			break;
		}
	}
	if (fontid !== null)
		xtree.setStringAttr("fontid", fontid);
	if (bg === null)
		bg = data.getDefault("/treeview/bg");
	xtree.setStringAttr("bg", bg);
	cfg = XElement.newInstance("tree");
	cfg.setStringAttr("id", id);
	cfg.setStringAttr("vgap", vgap);
	cfg.setStringAttr("hgap", hgap);
	for(i = 0;i < attrs.length;++i) {
		name = attrs[i].key;
		switch(name) {
		case "margin":
		case "border":
		case "size":
		case "expand":
		case "expandh":
		case "expandv":
		case "anchor":
		case "bg":
		case "fg":
		case "action":
		case "font":
		case "fontid":
		case "selection":
		case "rpadding":
		case "labelminw":
			cfg.setStringAttr(name, attrs[i].value);
			break;
		}
	}
	if (vgap === null)
		vgap = data.getDefault("/tree/vgap");
	if (hgap === null)
		hgap = data.getDefault("/tree/hgap");
	if (fg === null)
		fg = data.getDefault("/treelabel/fg");
	if (selection === null)
		selection = data.getDefault("/treerow/selection");
	if (rpadding === null)
		rpadding = data.getDefault("/treerow/padding");
	if (data.isRawWidget()) {
		xtree.setStringAttr("vgap", vgap);
		xtree.setStringAttr("hgap", hgap);
		xtree.setStringAttr("bg", bg);
		xtree.setStringAttr("fg", fg);
		xtree.setStringAttr("selection", selection);
		xtree.setStringAttr("rpadding", rpadding);
		if (action !== null)
			xtree.setStringAttr("action", action);
		if (font !== null)
			xtree.setStringAttr("font", font);
		if (lminw !== null)
			xtree.setStringAttr("labelminw", lminw);
		return(new XTreeRaw(data.toUIXWData(xtree, false)));
	}
	xview = xtree.addChild("treeview");
	xview.setStringAttr("id", id + "_view");
	xview.setStringAttr("bg", bg);
	if ((xdata = data.getXData(props.get("xref"))) === null)
		return(new XTree(data.toUIXWData(xtree, true), cfg));
	entries = xdata.getAllChildren();
	depths = UILib.newArray(entries.length, 0);
	labels = UILib.newArray(entries.length, null);
	depth = 0;
	for(i = 0;i < entries.length;++i) {
		entry = entries[i];
		depths[i] = entry.getAttrInt("depth", 0);
		labels[i] = entry.getTextContent();
		if (depth < depths[i])
			depth = depths[i];
	}
	if (action !== null)
		xview.setStringAttr("action", action);
	if (vgap !== null)
		xview.setStringAttr("gap", vgap);
	path = UILib.newArray(depth + 1, null);
	for(i = 0;i < entries.length;++i) {
		j = depths[i];
		path[j] = labels[i];
		for(++j;j < path.length;++j)
			path[j] = null;
		el = xview.addChild("treerow");
		if (rpadding !== null)
			el.setStringAttr("padding", rpadding);
		el.setStringAttr("expandh", "100%");
		if (fg !== null)
			el.setStringAttr("fg", fg);
		if (bg !== null)
			el.setStringAttr("bg", bg);
		if (selection !== null)
			el.setStringAttr("selection", selection);
		if (hgap !== null)
			el.setStringAttr("gap", hgap);
		el.setStringAttr("depth", UILib.intToString(depths[i]));
		if (font !== null)
			el.setStringAttr("font", font);
		if (lminw !== null) {
			el.setStringAttr("labelminw", lminw);
			el.setStringAttr("nowrap", "true");
		}
		el.setStringAttr("text", labels[i]);
		el.setStringAttr("value", XTree.getPath(path));
	}
	return(new XTree(data.toUIXWData(xtree, true), cfg));
};

// getPath ------------------------------------------------------------
XTree.getPath = function(path) {
	var i, sb;

	sb = new XStringBuilder();
	sb.append("/");
	for(i = 1;i < path.length;++i) {
		if (path[i] === null)
			break;
		if (i > 1)
			sb.append("/");
		sb.append(path[i]);
	}
	return(sb.getText());
};
/* ====================================================================	*
 * XTreeArea 															*
 * ====================================================================	*/

var XTreeArea = function(data) {

	XWidget.call(this, data);

	this.view = null;
	this.idx = 0;

	// init -----------------------------------------------------------
	XTreeArea.prototype.init = function(view, idx) {
		this.view = view;
		this.idx = idx;
	};

	// Constructor ----------------------------------------------------
	{
		this.selection = data.getColor("selection");
	}
};
/* ====================================================================	*
 * XTreeLabel 															*
 * ====================================================================	*/

var XTreeLabel = function(data) {

	XLabel.call(this, data);

	this.view = null;
	this.idx = 0;

	// init -----------------------------------------------------------
	XTreeLabel.prototype.init = function(view, idx) {
		this.view = view;
		this.idx = idx;
	};

	// Constructor ----------------------------------------------------
	{
		this.selection = data.getColor("selection");
	}
};
/* ====================================================================	*
 * XTreeRow 															*
 * ====================================================================	*/

var XTreeRow = function(data) {

	XHorFlow.call(this, data);

	this.view = null;
	this.idx = 0;
	this.visible = false;

	// init -----------------------------------------------------------
	XTreeRow.prototype.init = function(view, idx) {
		this.view = view;
		this.idx = idx;
		if (this.area !== null)
			this.area.init(view, idx);
		this.toggle.init(view, idx);
		this.label.init(view, idx);
	};

	// layoutSetLocation ----------------------------------------------
	XTreeRow.prototype.layoutSetLocation = function(vx, vy, anchor) {
		XHorFlow.prototype.layoutSetLocation.call(this, vx, vy, anchor);
	};

	// Constructor ----------------------------------------------------
	{
		this.depth = data.getAttrInt("depth", 0);
		this.text = data.getAttrString("text", null);
		this.value = data.getAttrString("value", null);
		this.selection = data.getColor("selection");
		if (this.children.length === 2) {
			this.area = null;
			this.toggle =this.children[0];
			this.label =this.children[1];
		} else {
			this.area =this.children[0];
			this.toggle =this.children[1];
			this.label =this.children[2];
		}
	}
};

// create -------------------------------------------------------------
XTreeRow.create = function(data) {
	var attrs, bg, c1, c2, c3, depth, fg, font, gap, i, id, lminw, mode, name, selection, text, v, value, w, xd;

	xd = UILib.newXData("treerow");
	bg = fg = font = id = lminw = selection = text = value = null;
	mode = "none";
	depth = -1;
	gap = 5;
	attrs = data.getAttributes();
	for(i = 0;i < attrs.length;++i) {
		name = attrs[i].key;
		v = attrs[i].value;
		switch(name) {
		case "id":
			id = v;
			xd.setStringAttr(name, v);
			break;
		case "margin":
		case "border":
		case "padding":
		case "size":
		case "expand":
		case "expandh":
		case "expandv":
		case "anchor":
			xd.setStringAttr(name, v);
			break;
		case "bg":
			bg = v;
			xd.setStringAttr(name, v);
			break;
		case "fg":
			fg = v;
			xd.setStringAttr(name, v);
			break;
		case "align":
		case "gap":
			xd.setStringAttr(name, v);
			break;
		case "text":
			text = v;
			break;
		case "value":
			value = v;
			break;
		case "font":
			font = v;
			break;
		case "mode":
			mode = v;
			break;
		case "depth":
			depth = UILib.parseInt(v);
			break;
		case "selection":
			selection = v;
			xd.setStringAttr("selection", v);
			break;
		case "labelminw":
			lminw = v;
			break;
		}
	}
	if (text === null)
		throw new Error("attribute text not found");
	if (font === null)
		font = data.getFont();
	if (depth === -1)
		throw new Error("attribute depth not found");
	xd.setIntAttr("depth", depth);
	xd.setStringAttr("text", text);
	xd.setStringAttr("value", value);
	if (depth === 0) {
		c1 = null;
		xd.setStringAttr("weight", "1");
	} else {
		w = depth * 10 +(depth -1) * gap;
		xd.setStringAttr("weight", "2");
		c1 = xd.addChild("treearea");
		c1.setStringAttr("size", w + "x10");
	}
	c2 = xd.addChild("treetoggle");
	c2.setStringAttr("mode", mode);
	c2.setStringAttr("size", "12x12");
	c3 = xd.addChild("treelabel");
	c3.setStringAttr("id", id + "_label");
	c3.setStringAttr("text", text);
	c3.setStringAttr("align", "left");
	c3.setStringAttr("expandh", "100%");
	if (lminw !== null) {
		c3.setStringAttr("nowrap", "true");
		c3.setStringAttr("size", lminw + "x0");
	}
	if (bg !== null) {
		if (c1 !== null)
			c1.setStringAttr("bg", bg);
		c2.setStringAttr("bg", bg);
		c3.setStringAttr("bg", bg);
	}
	if (fg !== null) {
		c2.setStringAttr("fg", fg);
		c3.setStringAttr("fg", fg);
	}
	if (font !== null)
		c3.setStringAttr("font", font);
	if (selection !== null) {
		if (c1 !== null)
			c1.setStringAttr("selection", selection);
		c2.setStringAttr("selection", selection);
		c3.setStringAttr("selection", selection);
	}
	return(new XTreeRow(data.toUIXWData(xd, true)));
};
/* ====================================================================	*
 * XTreeToggle 															*
 * ====================================================================	*/

var XTreeToggle = function(data) {

	XWidget.call(this, data);

	this.view = null;
	this.idx = 0;

	// init -----------------------------------------------------------
	XTreeToggle.prototype.init = function(view, idx) {
		this.view = view;
		this.idx = idx;
	};

	// Constructor ----------------------------------------------------
	{
		if (!this.padding.isEmpty())
			throw new Error("padding is not empty (JavaScript)");
		this.mode = data.getAttrString("mode", "none");
		this.selection = data.getColor("selection");
	}
};
/* ====================================================================	*
 * XTreeView 															*
 * ====================================================================	*/

var XTreeView = function(data) {

	XVerFlow.call(this, data);

	this.tree = null;
	this.all = null;
	this.path = null;
	this.first = false;

	// buildTree ------------------------------------------------------
	XTreeView.prototype.buildTree = function() {
		var i, next, tr;

		this.all = UILib.newArray(this.children.length, null);
		for(i = 0;i < this.all.length;++i)
			this.all[i] =this.children[i];
		this.first = true;
		for(i = 0;i < this.children.length;++i) {
			tr =this.children[i];
			tr.init(this, i);
			if (i === 0)
				tr.toggle.mode = "open";
			else
				if (i + 1 === this.children.length)
					tr.toggle.mode = "none";
				else {
					next =this.children[i + 1];
					if (tr.depth + 1 === next.depth)
						tr.toggle.mode = "close";
					else
						tr.toggle.mode = "none";
				}
			tr.visible = false;
		}
	};

	// getSelectedValue -----------------------------------------------
	XTreeView.prototype.getSelectedValue = function() {
		return(this.all[this.selected].value);
	};

	// initTree -------------------------------------------------------
	XTreeView.prototype.initTree = function() {
		var depth, i, j, n, r;

		n = 0;
		for(i = 0;i < this.all.length;++i) {
			r = this.all[i];
			switch(r.depth) {
			case 0:
			case 1:
				r.visible = true;
				++n;
				break;
			}
		}
		if (this.path !== null) {
			for(i = 0;i < this.all.length;++i)
				if (XString.equals(this.path, this.all[i].value))
					break;
			if (i < this.all.length) {
				this.selected = i;
				if ((depth = this.all[i].depth) >= 2) {
					for(j = i + 1;j < this.all.length;++j)
						if (this.all[j].depth < depth)
							break;
					--j;
					while(depth >= 2) {
						while(this.all[j].depth >= depth) {
							if (this.all[j].depth === depth) {
								this.all[j].visible = true;
								++n;
							}
							--j;
						}
						this.all[j].toggle.mode = "open";
						--depth;
					}
				}
			}
		}
		n = 0;
		this.children = UILib.newArray(n, null);
		for(i = 0;i < this.all.length;++i)
			if (this.all[i].visible)
				this.children[n ++] = this.all[i];
	};

	// layoutInit -----------------------------------------------------
	XTreeView.prototype.layoutInit = function() {
		var uieu;

		if (this.first) {
			XVerFlow.prototype.layoutInit.call(this);
			this.initTree();
			this.first = false;
			uieu =this.uiwt;
			uieu.eventUpdateUI();
		}
		XVerFlow.prototype.layoutInit.call(this);
	};

	// set ------------------------------------------------------------
	XTreeView.prototype.set = function(tree) {
		this.tree = tree;
	};

	// setPath --------------------------------------------------------
	XTreeView.prototype.setPath = function(path) {
		this.path = path;
	};

	// taskRowSelected ------------------------------------------------
	XTreeView.prototype.taskRowSelected = function(idx) {
		var cur, old, uiea, uier, uieu;

		if (this.selected === idx)
			return;
		if (this.selected === -1)
			old = null;
		else
			old = this.all[this.selected];
		this.selected = idx;
		cur = this.all[idx];
		if (old !== null) {
			uieu =old.uiwt;
			uieu.eventUpdateUI();
			uier =old.uiwt;
			uier.eventRepaint();
		}
		uieu =cur.uiwt;
		uieu.eventUpdateUI();
		uier =cur.uiwt;
		uier.eventRepaint();
		uiea =this.uiwt;
		uiea.eventActionUI(cur.value);
	};

	// taskTogglePressed ----------------------------------------------
	XTreeView.prototype.taskTogglePressed = function(idx) {
		var cur, i, n, nentries, old, ori, pos, tr, uiea, uieu;

		cur = this.all[idx];
		for(pos = 0;pos < this.children.length;++pos)
			if (this.children[pos] === cur)
				break;
		if (pos === this.children.length)
			throw new Error("invalid index: " + idx);
		if (this.selected !== idx) {
			old = this.all[this.selected];
			this.selected = idx;
			uieu =old.uiwt;
			uieu.eventUpdateUI();
			uieu =cur.uiwt;
			uieu.eventUpdateUI();
			uiea =this.uiwt;
			uiea.eventActionUI(cur.value);
		}
		switch(cur.toggle.mode) {
		case "close":
			nentries = 0;
			for(i = pos + 1;i < this.children.length;++i) {
				tr =this.children[i];
				if (tr.depth <= cur.depth)
					break;
				switch(tr.toggle.mode) {
				case "open":
					tr.toggle.mode = "close";
					break;
				}
				++nentries;
			}
			ori = this.children;
			this.children = UILib.newArray(ori.length - nentries, null);
			for(i = 0;i <= pos;++i)
				this.children[i] = ori[i];
			for(;i < this.children.length;++i)
				this.children[i] = ori[i + nentries];
			break;
		case "open":
			nentries = 0;
			for(i = idx + 1;i < this.all.length;++i) {
				tr = this.all[i];
				if (tr.depth <= cur.depth)
					break;
				if (tr.depth === cur.depth + 1)
					++nentries;
			}
			ori = this.children;
			this.children = UILib.newArray(ori.length + nentries, null);
			for(n = 0;n <= pos;++n)
				this.children[n] = ori[n];
			for(i = idx + 1;i < this.all.length;++i) {
				tr = this.all[i];
				if (tr.depth <= cur.depth)
					break;
				if (tr.depth === cur.depth + 1)
					this.children[n ++] = this.all[i];
			}
			for(i = pos + 1;i < ori.length;++i)
				this.children[n ++] = ori[i];
			break;
		default:
			throw new Error("invalid mode: " + cur.toggle.mode);
		}
		for(i = 0;i < this.all.length;++i) {
			tr = this.all[i];
			tr.visible = false;
		}
		for(i = 0;i < this.children.length;++i) {
			tr =this.children[i];
			tr.visible = true;
		}
		uieu =this.uiwt;
		uieu.eventUpdateUI();
		if (this.tree !== null)
			this.tree.layoutUpdateUI(true);
	};

	// Constructor ----------------------------------------------------
	{
		this.action = data.getAttrString("action", null);
		this.selected = 0;
		this.buildTree();
	}
};
/* ====================================================================	*
 * XUILib 																*
 * ====================================================================	*/

var XUILib = function() {
};

// getImagePosition ---------------------------------------------------
XUILib.getImagePosition = function(width, height, imgw, imgh) {
	var h, ratio, v, w, x, y;

	w = width;
	h = height;
	if (imgw <= w && imgh <= h) {
		x =(w - imgw) / 2;
		y =(h - imgh) / 2;
		return(new XRectangle(x, y, imgw, imgh));
	}
	x = y = 0;
	ratio = imgw / imgh;
	if ((v = XInteger.round(width / ratio)) <= h) {
		y +=(h - v) / 2;
		h = v;
	} else {
		v = XInteger.round(h * ratio);
		x +=(w - v) / 2;
		w = v;
	}
	return(new XRectangle(x, y, w, h));
};
/* ====================================================================	*
 * XVarPath 															*
 * ====================================================================	*/

var XVarPath = function(path) {

	// Constructor ----------------------------------------------------
	{
		var n, s;

		if (!UILib.stringsMatch(path, "${"))
			throw new Error("invalid path: " + path);
		if ((n = path.indexOf('}', 1)) === -1)
			throw new Error("invalid path: " + path);
		this.dataId = path.substring(2, n);
		if (!UILib.isCharAt(path, n + 1, '/'))
			throw new Error("invalid path: " + path);
		s = path.substring(n + 1);
		if (UILib.stringsEquals(s, "/"))
			this.xpath = "/magicw/*";
		else
			this.xpath = "/magicw" + s;
	}
};
/* ====================================================================	*
 * XVars 																*
 * ====================================================================	*/

var XVars = function(appl) {

	// getAllData -----------------------------------------------------
	XVars.prototype.getAllData = function() {
		return(this.hdata);
	};

	// getData --------------------------------------------------------
	XVars.prototype.getData = function(id) {
		return(this.hdata.get(id));
	};

	// getVar ---------------------------------------------------------
	XVars.prototype.getVar = function(id) {
		return(this.hvars.get(id));
	};

	// listVars -------------------------------------------------------
	XVars.prototype.listVars = function() {
		var i, keys, list;

		keys = UILib.newArray(this.hvars.size(), null);
		list = UILib.newArray(keys.length, null);
		this.hvars.keys(keys);
		for(i = 0;i < keys.length;++i)
			list[i] = new XKeyValue(keys[i], this.hvars.get(keys[i]));
		return(list);
	};

	// parse ----------------------------------------------------------
	XVars.prototype.parse = function(text) {
		var data, el, k, len, n1, n2, n3, off, v, value, varname;

		len = UILib.stringLength(text);
		value = "";
		off = 0;
		while(off < len) {
			if ((n1 = text.indexOf("${", off)) === -1)
				return(value + text.substring(off));
			if ((n2 = text.indexOf("}", n1 + 1)) === -1)
				throw new Error("invalid text: " + text);
			varname = text.substring(n1, n2 + 1);
			if (n1 > off)
				value += text.substring(off, n1);
			if (UILib.isCharAt(text, n1 + 2, '#')) {
				if ((n3 = text.indexOf("/", n1 + 2)) === -1)
					throw new Error("invalid text: " + text);
				if (n3 > n2)
					throw new Error("invalid text: " + text);
				v = null;
				if ((data = this.hdata.get(text.substring(n1 + 3, n3))) !== null) {
					k = "/magicw" + text.substring(n3, n2);
					if ((el = data.searchElement(k)) === null)
						v = "";
					else
						if ((v = el.getTextContent()) === null)
							v = "";
				}
				if (v !== null)
					value += v;
				else
					XApplication.logMsg(varname + " not found");
			} else {
				k = text.substring(n1 + 2, n2);
				switch(k) {
				case "LANG":
					value += XUtil.getLang();
					break;
				case "TIME_ZONE":
					value += XUtil.getLocalTimeZone();
					break;
				case "DUDAL-VERSION":
					value += XUtil.getVersion();
					break;
				case "USER-AGENT":
					value += XUtil.getUserAgent();
					break;
				default:
					if ((v = this.hvars.get(k)) !== null)
						value += v;
					else
						XApplication.logMsg(varname + " not found");
				}
			}
			off = n2 + 1;
		}
		return(value);
	};

	// putData --------------------------------------------------------
	XVars.prototype.putData = function(id, data) {
		this.hdata.put(id, data);
	};

	// putVar ---------------------------------------------------------
	XVars.prototype.putVar = function(id, ref) {
		var s, value, vp, xd;

		if (ref === null) {
			this.hvars.put(id, null);
			return;
		}
		if (UILib.stringsMatch(ref, "data:")) {
			vp = new XVarPath(ref.substring(5));
			if ((xd = this.getData(vp.dataId)) === null)
				throw new Error("invalid ref: " + ref);
			s = this.parse(vp.xpath);
			if ((value = xd.searchValue(s)) === null)
				throw new Error("value not found: " + ref);
		} else
			if (UILib.stringsMatch(ref, "input-form:"))
				value = this.appl.getFormInputValue(ref.substring(11));
			else
				if (UILib.stringsMatch(ref, "input-widget:"))
					value = this.appl.getWidgetInputValue(ref.substring(13));
				else
					value = this.parse(ref);
		this.hvars.put(id, value);
	};

	// removeData -----------------------------------------------------
	XVars.prototype.removeData = function(data) {
		var tag;

		tag = data.getTag();
		this.hdata.remove(tag);
	};

	// Constructor ----------------------------------------------------
	{
		this.appl = appl;
		this.hvars = UILib.getStringMap();
		this.hdata = UILib.getXDataMap();
	}
};
/* ====================================================================	*
 * XVerFlow 															*
 * ====================================================================	*/

var XVerFlow = function(data) {

	XWidget.call(this, data);

	// layoutInit -----------------------------------------------------
	XVerFlow.prototype.layoutInit = function() {
		var dw, i, mw, pw, w, wt;

		mw = pw = 0;
		for(i = 0;i < this.children.length;++i) {
			wt = this.children[i];
			dw = wt.getEdgeWidth();
			wt.layoutInit();
			wt.layoutSetMinimumWidth();
			if ((w = wt.minWidth + dw) > mw)
				mw = w;
			if ((w = wt.prefWidth + dw) > pw)
				pw = w;
		}
		if (this.minWidth < mw)
			this.minWidth = mw;
		if (this.prefWidth < pw)
			this.prefWidth = pw;
	};

	// layoutSetLocation ----------------------------------------------
	XVerFlow.prototype.layoutSetLocation = function(vx, vy, anchor) {
		var i, wt, x, y;

		XWidget.prototype.layoutSetLocation.call(this, vx, vy, anchor);
		x = this.getX();
		y = this.getY();
		for(i = 0;i < this.children.length;++i) {
			wt = this.children[i];
			wt.layoutSetLocation(x, y, wt.anchor);
			y += this.rowsh[i] + this.gap;
		}
	};

	// layoutSetMinimumWidth ------------------------------------------
	XVerFlow.prototype.layoutSetMinimumWidth = function() {
		var i;

		XWidget.prototype.layoutSetMinimumWidth.call(this);
		for(i = 0;i < this.children.length;++i)
			this.children[i].layoutSetMinimumWidth();
	};

	// layoutSetMinimumHeight -----------------------------------------
	XVerFlow.prototype.layoutSetMinimumHeight = function() {
		var h, i, wt;

		XWidget.prototype.layoutSetMinimumHeight.call(this);
		h = 0;
		for(i = 0;i < this.children.length;++i) {
			wt = this.children[i];
			wt.layoutSetMinimumHeight();
			this.rowsh[i] = wt.cheight;
			h += this.rowsh[i];
		}
		this.cheight = h +(this.children.length -1) * this.gap + this.getEdgeHeight();
	};

	// layoutSetViewportWidth -----------------------------------------
	XVerFlow.prototype.layoutSetViewportWidth = function(vwidth) {
		var colw, i;

		XWidget.prototype.layoutSetViewportWidth.call(this, vwidth);
		colw = this.cwidth - this.getEdgeWidth();
		for(i = 0;i < this.children.length;++i)
			this.children[i].layoutSetViewportWidth(colw);
	};

	// layoutSetViewportHeight ----------------------------------------
	XVerFlow.prototype.layoutSetViewportHeight = function(vheight) {
		var dh, i, ori;

		ori = this.cheight;
		XWidget.prototype.layoutSetViewportHeight.call(this, vheight);
		if (this.weight !== -1 &&(dh = this.cheight - ori) > 0)
			this.rowsh[this.weight] += dh;
		for(i = 0;i < this.children.length;++i)
			this.children[i].layoutSetViewportHeight(this.rowsh[i]);
	};

	// Constructor ----------------------------------------------------
	{
		this.children = XLib.createChildrenWidget(data);
		this.gap = data.getInt("gap");
		this.weight = data.getAttrInt("weight", -1);
		this.rowsh = UILib.newArray(this.children.length, 0);
	}
};
/* ====================================================================	*
 * XVideoEvent 															*
 * ====================================================================	*/

var XVideoEvent = function(xdata, lectureId) {

	// isSlide --------------------------------------------------------
	XVideoEvent.prototype.isSlide = function() {
		return(XString.equals(this.type, "slide"));
	};

	// Constructor ----------------------------------------------------
	{
		this.type = xdata.getTag();
		this.timecode = xdata.getDouble("timecode");
		switch(this.type) {
		case "slide":
			this.link = lectureId + "/" + xdata.getString("link");
			break;
		case "pointer":
			this.route = xdata.getString("route");
			break;
		}
		this.idxf = -1;
	}
};
/* ====================================================================	*
 * XVideoPart 															*
 * ====================================================================	*/

var XVideoPart = function(xdata, lectureId) {

	// addPointerEvent ------------------------------------------------
	XVideoPart.prototype.addPointerEvent = function(event) {
		var i, tmp;

		tmp = UILib.newArray(this.events.length + 1, null);
		for(i = 0;i < this.events.length;++i)
			tmp[i] = this.events[i];
		tmp[i] = event;
		this.events = tmp;
	};

	// getNumberOfSlideEvents -----------------------------------------
	XVideoPart.prototype.getNumberOfSlideEvents = function() {
		var i, n;

		n = 0;
		for(i = 0;i < this.events.length;++i)
			if (this.events[i].isSlide())
				++n;
		return(n);
	};

	// mergeSlides ----------------------------------------------------
	XVideoPart.prototype.mergeSlides = function(slides) {
		var i, j, k, n, tmp;

		n = slides.length;
		for(i = 0;i < this.events.length;++i)
			if (XString.equals(this.events[i].type, "pointer"))
				++n;
		tmp = UILib.newArray(n, null);
		j = 0;
		k = 0;
		for(i = 0;i < tmp.length;++i) {
			for(;k < this.events.length;++k)
				if (XString.equals(this.events[k].type, "pointer"))
					break;
			if (j === slides.length) {
				tmp[i] = this.events[k ++];
				continue;
			}
			if (k === this.events.length) {
				tmp[i] = slides[j ++];
				continue;
			}
			if (slides[j].timecode < this.events[k].timecode)
				tmp[i] = slides[j ++];
			else
				tmp[i] = this.events[k ++];
		}
		this.events = tmp;
	};

	// Constructor ----------------------------------------------------
	{
		var children, i;

		this.id = xdata.getString("id");
		this.duration = xdata.getInt("duration");
		this.starting = xdata.getInt("starting");
		this.live = xdata.getBool("live");
		children = xdata.getAllChildren();
		this.events = UILib.newArray(children.length, null);
		for(i = 0;i < children.length;++i)
			this.events[i] = new XVideoEvent(children[i], lectureId);
	}
};
/* ====================================================================	*
 * XWidgetAction 														*
 * ====================================================================	*/

var XWidgetAction = function(data) {

	// Constructor ----------------------------------------------------
	{
		this.action = data.getString("action");
		this.ref = data.getString("ref");
		switch(this.action) {
		case "disabled":
		case "enabled":
		case "hide":
		case "show":
		case "visibility-on":
		case "visibility-off":
			break;
		case "background":
			this.value = data.getString("value");
			break;
		case "show-widget":
			this.value = data.getString("value");
			break;
		default:
			throw new Error("invalid action: " + this.action);
		}
	}
};

	// XApplication: static fields ------------------------------------
	XApplication.mode = null;

	// XLib: static fields --------------------------------------------
	XLib.Horizontal = 0;
	XLib.Vertical = 1;
	XLib.GoldenRatio = 1.618;

	// XSlideLayout: static fields ------------------------------------
	XSlideLayout.NO_VIDEO = 0;
	XSlideLayout.LIVE = 1;
	XSlideLayout.LIVE_END = 2;
	XSlideLayout.VIDEO_STARTING = 3;
	XSlideLayout.VIDEO_RUNNING = 4;
	XSlideLayout.VIDEO_END = 5;
	XSlideLayout.VIDEO_PAUSE = 6;

	// XButton: heritage ----------------------------------------------
	XButton.prototype = Object.create(XWidget.prototype);
	XButton.prototype.constructor = XButton;

	// XHorFlow: heritage ---------------------------------------------
	XHorFlow.prototype = Object.create(XWidget.prototype);
	XHorFlow.prototype.constructor = XHorFlow;

	// XCheckBox: heritage --------------------------------------------
	XCheckBox.prototype = Object.create(XHorFlow.prototype);
	XCheckBox.prototype.constructor = XCheckBox;

	// XCheckBoxItem: heritage ----------------------------------------
	XCheckBoxItem.prototype = Object.create(XWidget.prototype);
	XCheckBoxItem.prototype.constructor = XCheckBoxItem;

	// XComboBox: heritage --------------------------------------------
	XComboBox.prototype = Object.create(XWidget.prototype);
	XComboBox.prototype.constructor = XComboBox;

	// XFlow: heritage ------------------------------------------------
	XFlow.prototype = Object.create(XWidget.prototype);
	XFlow.prototype.constructor = XFlow;

	// XDate: heritage ------------------------------------------------
	XDate.prototype = Object.create(XFlow.prototype);
	XDate.prototype.constructor = XDate;

	// XDisplay: heritage ---------------------------------------------
	XDisplay.prototype = Object.create(XWidget.prototype);
	XDisplay.prototype.constructor = XDisplay;

	// XDropDown: heritage --------------------------------------------
	XDropDown.prototype = Object.create(XWidget.prototype);
	XDropDown.prototype.constructor = XDropDown;

	// XGrid: heritage ------------------------------------------------
	XGrid.prototype = Object.create(XWidget.prototype);
	XGrid.prototype.constructor = XGrid;

	// XIconLink: heritage --------------------------------------------
	XIconLink.prototype = Object.create(XHorFlow.prototype);
	XIconLink.prototype.constructor = XIconLink;

	// XLabel: heritage -----------------------------------------------
	XLabel.prototype = Object.create(XWidget.prototype);
	XLabel.prototype.constructor = XLabel;

	// XLabelButton: heritage -----------------------------------------
	XLabelButton.prototype = Object.create(XWidget.prototype);
	XLabelButton.prototype.constructor = XLabelButton;

	// XScrollPane: heritage ------------------------------------------
	XScrollPane.prototype = Object.create(XWidget.prototype);
	XScrollPane.prototype.constructor = XScrollPane;

	// XTable: heritage -----------------------------------------------
	XTable.prototype = Object.create(XScrollPane.prototype);
	XTable.prototype.constructor = XTable;

	// XLink: heritage ------------------------------------------------
	XLink.prototype = Object.create(XWidget.prototype);
	XLink.prototype.constructor = XLink;

	// XMenuBar: heritage ---------------------------------------------
	XMenuBar.prototype = Object.create(XWidget.prototype);
	XMenuBar.prototype.constructor = XMenuBar;

	// XPassword: heritage --------------------------------------------
	XPassword.prototype = Object.create(XWidget.prototype);
	XPassword.prototype.constructor = XPassword;

	// XPictButton: heritage ------------------------------------------
	XPictButton.prototype = Object.create(XWidget.prototype);
	XPictButton.prototype.constructor = XPictButton;

	// XPicture: heritage ---------------------------------------------
	XPicture.prototype = Object.create(XWidget.prototype);
	XPicture.prototype.constructor = XPicture;

	// XPlugin: heritage ----------------------------------------------
	XPlugin.prototype = Object.create(XWidget.prototype);
	XPlugin.prototype.constructor = XPlugin;

	// XRadioButton: heritage -----------------------------------------
	XRadioButton.prototype = Object.create(XHorFlow.prototype);
	XRadioButton.prototype.constructor = XRadioButton;

	// XRadioButtonItem: heritage -------------------------------------
	XRadioButtonItem.prototype = Object.create(XWidget.prototype);
	XRadioButtonItem.prototype.constructor = XRadioButtonItem;

	// XSimpleGrid: heritage ------------------------------------------
	XSimpleGrid.prototype = Object.create(XWidget.prototype);
	XSimpleGrid.prototype.constructor = XSimpleGrid;

	// XSlide: heritage -----------------------------------------------
	XSlide.prototype = Object.create(XWidget.prototype);
	XSlide.prototype.constructor = XSlide;

	// XSplitPane: heritage -------------------------------------------
	XSplitPane.prototype = Object.create(XWidget.prototype);
	XSplitPane.prototype.constructor = XSplitPane;

	// XStack: heritage -----------------------------------------------
	XStack.prototype = Object.create(XWidget.prototype);
	XStack.prototype.constructor = XStack;

	// XText: heritage ------------------------------------------------
	XText.prototype = Object.create(XWidget.prototype);
	XText.prototype.constructor = XText;

	// XTextArea: heritage --------------------------------------------
	XTextArea.prototype = Object.create(XWidget.prototype);
	XTextArea.prototype.constructor = XTextArea;

	// XTextField: heritage -------------------------------------------
	XTextField.prototype = Object.create(XWidget.prototype);
	XTextField.prototype.constructor = XTextField;

	// XTocEdit: heritage ---------------------------------------------
	XTocEdit.prototype = Object.create(XWidget.prototype);
	XTocEdit.prototype.constructor = XTocEdit;

	// XTree: heritage ------------------------------------------------
	XTree.prototype = Object.create(XScrollPane.prototype);
	XTree.prototype.constructor = XTree;

	// XTreeArea: heritage --------------------------------------------
	XTreeArea.prototype = Object.create(XWidget.prototype);
	XTreeArea.prototype.constructor = XTreeArea;

	// XTreeLabel: heritage -------------------------------------------
	XTreeLabel.prototype = Object.create(XLabel.prototype);
	XTreeLabel.prototype.constructor = XTreeLabel;

	// XTreeRow: heritage ---------------------------------------------
	XTreeRow.prototype = Object.create(XHorFlow.prototype);
	XTreeRow.prototype.constructor = XTreeRow;

	// XTreeToggle: heritage ------------------------------------------
	XTreeToggle.prototype = Object.create(XWidget.prototype);
	XTreeToggle.prototype.constructor = XTreeToggle;

	// XVerFlow: heritage ---------------------------------------------
	XVerFlow.prototype = Object.create(XWidget.prototype);
	XVerFlow.prototype.constructor = XVerFlow;

	// XTreeView: heritage --------------------------------------------
	XTreeView.prototype = Object.create(XVerFlow.prototype);
	XTreeView.prototype.constructor = XTreeView;

	/* ================================================================	*
	 * HttpFiles.js 													*
	 * ================================================================	*/

	var HttpFiles = function(appl) {

		// cacheDownload ----------------------------------------------
		HttpFiles.prototype.cacheDownload = function(files, nextTask) {
			var i, f, img, res;

			if (files.length === 0) {
				this.appl.runTask(nextTask);
				return;
			}

			appl.taskCardShow('lecture-download');
			
			res = {
				appl: this.appl,
				nleft: files.length,
				nextTask: nextTask,
				total: files.length,
				bar: null,
				me: this
			};

			switch(nextTask) {
			case 'lecture-edit-ok':
				res.bar = DocLib.getElement('downloading-bar');
				this.repaintProgressBar(res.bar, files.length, files.length);
				break;
			case 'lecture-edit-add-done':
			case 'transmitter-open-ok':
				res.bar = DocLib.getElement('downloading-bar');
				this.repaintProgressBar(res.bar, files.length, files.length);
				break;
			case 'receiver-show':
				res.bar = DocLib.getElement('downloading-bar');
				this.repaintProgressBar(res.bar, files.length, files.length);
				break;
			}

			for(i = 0; i < files.length; ++i) {
				f = files[i];

				switch(f.tag) {
				case 'img':
				case 'migration-img':
					this.images[f.path] = img = new Image;
					img.src = this.getUrl(f.tag, f.path);
					img._path = f.path;

					img.onload = function() {

						try {
							Log.info('-----http-downloaded: ' + this._path + ' -----');
							
							if (--res.nleft > 0) {
								if (res.bar === null)
									return;

								res.me.repaintProgressBar(res.bar, res.nleft, res.total);
								return;
							}

							res.appl.runTask(res.nextTask);
						}
						catch(err) {
							HttpError.send(err);
							MW.xappl.runTask('#http-jserror');
						}
					};
					break;
				default:
					throw new Error('not yet supported: ' + f.tag);
				}
			}
		};

		// cacheOffline -----------------------------------------------
		HttpFiles.prototype.cacheOffline = function(ref, nextTask) {
			var e, i, xo, img, list, path, res, toc;

			if ((xo = this.appl.getData("lecture-info")) === null)
				throw new Error('lecture-info not found');

			toc = xo.searchElement('toc');
			list = toc.getAllChildren();
			
			if (list.length === 0) {
				this.appl.runTask(nextTask);
				return;
			}

			res = {appl:this.appl, nleft:list.length, nextTask:nextTask};

			for(i = 0; i < list.length; ++i) {
				e = list[i];
				path = e.getTextContent();
				this.images[path] = img = new Image;
				img.src = path;

				img.onload = function() {
					
					try {
						if (--res.nleft > 0)
							return;

						res.appl.runTask(nextTask);
					}
					catch(err) {
						HttpError.send(err);
						MW.xappl.runTask('#http-jserror');
					}
				};
			}
		};

		// getImage ---------------------------------------------------
		HttpFiles.prototype.getImage = function(path) {
			var img;

			if ((img = this.images[path]) === undefined)
				throw new Error('invalid image path: ' + path);

			if (!img.complete)
				return null;

			return img;
		};

		// getUrl -----------------------------------------------------
		HttpFiles.prototype.getUrl = function(tag, path) {
			var url;

			url = this.appl.getUrl() + '?magicw-xml';

			switch(tag) {
			case 'img':
				url += '&file-download=';
				break;
			case 'migration-img':
				url += '&migration-download=';
				break;
			default:
				throw new Error('not yet supported: ' + tag);
			}

			url += encodeURIComponent(path);

			if (this.hashSession !== null)
				url += '&' + this.hashSession;

			return url;
		};

		// initCanvasImages -------------------------------------------
		HttpFiles.prototype.initCanvasImages = function() {
			var i, n, cfg, img, list, path, url;

			if ((url = this.appl.getUrl()) === null)
				return;

			if ((n = url.lastIndexOf('/')) === -1)
				throw new Error('invalid url: ' + url);

			url = url.substring(0, n) + '/images/';

			cfg = this.appl.getConfig();
			list = cfg.canvasImgs;

			for(i = 0; i < list.length; ++i) {
				path = 'MW.' + list[i];
				img = new Image;
				img.src = url + list[i];

				this.images[path] = img;
			}
		};

		// repaintProgressBar -----------------------------------------
		HttpFiles.prototype.repaintProgressBar = function(canvas, nleft, total) {
			var w, h, ctx, left, right;

			ctx = canvas.getContext('2d');
			w = ctx.canvas.width;
			h = ctx.canvas.height;

			right = w * nleft / total;
			left = w - right;

			if (left > 0) {
				ctx.fillStyle = 'teal';
				ctx.fillRect(0, 0, left, h);
			}

			if (right > 0) {
				ctx.fillStyle = '#000000';
				ctx.fillRect(w - right, 0, right, h);
			}
		};

		// setHash ----------------------------------------------------
		HttpFiles.prototype.setHash = function(hash) {

			this.hashSession = hash;
		};

		// Constructor ------------------------------------------------
		{
			this.appl = appl;
			this.hashSession = null;
			this.images = [];
		}
	};

	/* ================================================================	*
	 * UILibN.js 														*
	 * ================================================================	*/

	var UILibN = function() {
	};

	// UILibN.drawBorder --------------------------------------------------
	UILibN.drawBorder = function(ctx, x, y, w, h, thickness) {

		if (thickness === 1) {
			ctx.beginPath();
			ctx.rect(x, y, w, h);
			ctx.stroke();
			return;
		}

		ctx.fillRect(x, y, w, thickness);
		ctx.fillRect(x + w - thickness, y + thickness, thickness, h - thickness);
		ctx.fillRect(x, y + h - thickness, w, thickness);
		ctx.fillRect(x, y + thickness, thickness, h - thickness * 2);
	};

	/* ================================================================	*
	 * UploadFilesTask.js 												*
	 * ================================================================	*/

	var UploadFilesTask = function() {

		// chooserDone ------------------------------------------------
		UploadFilesTask.prototype.chooserDone = function() {

			this.exec(this.input.files);
		};

		// exec -------------------------------------------------------
		UploadFilesTask.prototype.exec = function(files) {
			var i, v, appl, attr, cl, data, entry, iterator, list, msg;

			appl = MW.xappl;

			data = new FormData();
			data.append('id', this.requestId);

			list = this.attrs.getList();
			for(i = 0; i < list.length; ++i) {
				attr = list[i];

				switch(attr.name) {
				case "requestId":
				case "nextTask":
					break;
				default:
					v = attr.value;
					if (v.startsWith('${') && v.endsWith('}'))
						v = appl.getVar(v.substring(2, v.length - 1));

					data.append(attr.name, v);
				}
			}
			
			for(i = 0; i < files.length; ++i)
				data.append('file', files[i]);

			XApplication.logMsg("-----http-request: " + this.requestId + "-----");
			iterator = data.entries();
			for(;;) {
				entry = iterator.next();
				if (entry.done)
					break;

				try {
					v = entry.value[1];
					if (v instanceof File)
						msg = v.name + ":" + v.size + " bytes:" + v.type;
					else
						msg = v;

					XApplication.logMsg('\t' + entry.value[0] + '=' + msg);
				}
				catch(err) {
					XApplication.logMsg(err);
				}
			}

			cl = new HttpClient(appl.getUrl() + '?magicw-xml', appl);
			cl.uploadFiles(this.requestId, this.nextTask, data);
		};

		// init -------------------------------------------------------
		UploadFilesTask.prototype.init = function(attrs) {
			var i, attr, list;

			this.attrs = attrs;

			list = attrs.getList();
			for(i = 0; i < list.length; ++i) {
				attr = list[i];

				switch(attr.name) {
				case "requestId":
					this.requestId = attr.value;
					break;
				case "nextTask":
					this.nextTask = attr.value;
					break;
				}
			}

		};

		// openChooser ------------------------------------------------
		UploadFilesTask.prototype.openChooser = function(attrs) {

			this.init(attrs);

			this.input.value = null;
			this.input.click();
		};

		// Constructor ----------------------------------------------------
		{
			this.input = document.createElement('input');
			this.input.type = 'file';
			this.input.multiple = true;

			this.input.onchange = function(evt) {

				MW.uiappl.uploadFilesTask.chooserDone();
			};
		};
	};

	/* ================================================================	*
	 * WLib.js 															*
	 * ================================================================	*/

	var WLib = function() {

		// getWidget --------------------------------------------------
		WLib.prototype.getWidget = function(cardId, widgetId) {
			var card, wt;

			if ((card = MW.roots[cardId]) === undefined)
				throw new Error('cardId not found: ' + cardId);

			if ((wt = XLib.search(card.widget, widgetId)) !== null)
				return wt;

			throw new Error('widgetId not found: ' + cardId + ' (' + cardId + ',' + widgetId + ')');
		};

		// isCurrentCardId ------------------------------------------------
		WLib.prototype.isCurrentCardId = function(cardId) {

			if (MW.xroot === null)
				return false;

			return cardId === MW.xroot.widget.id;
		};

		// updateScrollPaneView ---------------------------------------
		WLib.prototype.updateScrollPaneView = function(scroll, view, visible) {
			var sb, html, owner;

			owner = scroll.uiwt.element;

			if (owner.childElementCount !== 1)
				throw new Error('invalid scroll pane');

			owner.removeChild(owner.firstElementChild);

			sb = [];
			HtmlWriter.createWidget(sb, view);
			html = sb.join('');

			/*
			 * les longueurs des textes sont uniquement calculées si
			 * le fragment est inséré dans body ...
			 */
			range = document.createRange();
			range.selectNode(document.body);
			fragment = range.createContextualFragment(html);
			owner.appendChild(fragment);

			MW.initWidget(view, scroll);
			scroll.setView(view);

			scroll.layoutUpdateUI(visible);
		};


		// widgetUpdate -----------------------------------------------
		WLib.prototype.widgetUpdate = function(owner, wt, parent) {
			var el, fragment, html, range, sb;

			if (owner.childElementCount !== 1)
				throw new Error('not yet implemented');

			owner.removeChild(owner.firstElementChild);

			sb = [];
			HtmlWriter.createWidget(sb, wt);
			html = sb.join('');

			/*
			 * les longueurs des textes sont uniquement calculées si
			 * le fragment est inséré dans body ...
			 */
			range = document.createRange();
			range.selectNode(document.body);
			fragment = range.createContextualFragment(html);
			document.body.appendChild(fragment);

			MW.initWidget(wt, parent);

			el = DocLib.getElement(wt.id);
			owner.appendChild(el);
		};

		// widgetUpdatePicture ----------------------------------------
		WLib.prototype.widgetUpdatePicture = function(pict, ref) {
			var el, id, img, path;

			if (ref.startsWith('${')) {
				 id = MW.xappl.parseText(ref);

				if (!id.startsWith('cache:'))
					throw new Error('not yet implemented: ' +  id);

				path = id.substring(6);
				pict.ref = id;
			} else
			if (ref.startsWith('cache:')) {
				path = ref.substring(6);
				pict.ref = ref;
			} else {
				if (!ref.startsWith('var:'))
					throw new Error('not yet implemented: ' +  ref);

				id = ref.substring(4);
				if ((path = MW.xappl.getVar(id)) === null)
					throw new Error('var not found: ' + id);

				if (path === '')
					return;

				pict.ref = ref;
			}

			img = MW.HttpFiles.getImage(path);
			pict.naturalsz = new XSize(img.naturalWidth, img.naturalHeight);

			el = pict.uiwt.element;
			if (pict.ratio !== null)
				el = el.firstElementChild;

			el.src = img.src;
		};

		// widgetUpdateTable ------------------------------------------
		WLib.prototype.widgetUpdateTable = function(wt, xref, visible) {
			var t;

			t = wt.getNewInstance(MW.xappl, xref);

			this.widgetUpdate(wt.uiwt.element, t.view, wt);

			wt.setView(t.view);
			wt.layoutUpdateUI(visible);
		};

		// widgetUpdateTree -------------------------------------------
		WLib.prototype.widgetUpdateTree = function(wt, xref, path, visible) {
			var t;

			t = wt.getNewInstance(MW.xappl, xref);

			if (path !== null)
				t.setPath(path);

			this.widgetUpdate(wt.uiwt.element, t.view, wt);

			wt.replace(t);

			wt.layoutUpdateUI(visible);
		};

		// Constructor ----------------------------------------------------
		{
		}
	};

	/* ================================================================	*
	 * XWebSocket.js 													*
	 * ================================================================	*/

	var XWebSocket = function() {
	};

	// open ---------------------------------------------------------------
	XWebSocket.open = function(name, url, listener) {
		var label, ws;

		label = '=====> Websocket[' + name + ']: ';

		XApplication.logMsg(label + 'open');
		XApplication.logMsg('websocket: ' + url);

		ws = new WebSocket(url);

		// getBufferedAmount ----------------------------------------------	
		ws.getBufferedAmount = function() {

			return ws.bufferedAmount;
		};

		// onclose --------------------------------------------------------	
		ws.onclose = function(evt) {

			switch(evt.code) {
			case 1000:
				XApplication.logMsg(label + 'onclose [' + evt.code + ']');
				listener.wsClose(this);
				break;
			default:
				XApplication.logMsg(label + 'onclose [' + evt.code + ']: ' + evt.reason);
				listener.wsErrorOnClose(this, evt.code, evt.reason);
			}
		};

		// onmessage ------------------------------------------------------	
		ws.onmessage = function(evt) {
			var doc, parser;

			XApplication.logMsg(label + 'receive');

			try {
				parser = new DOMParser();
				doc = parser.parseFromString(evt.data, 'application/xml');
				XApplication.logMsg(doc);
			}
			catch(err) {
				HttpError.send(err);
			}

			listener.wsMessage(this, doc);
		};

		// onopen ---------------------------------------------------------	
		ws.onopen = function(evt) {

			listener.wsOpen(this);
		};

		// sendData -------------------------------------------------------	
		ws.sendData = function(data) {

			if (ws.readyState !== WebSocket.OPEN)
				return;
			
			ws.send(data);
		};

		// sendMsg --------------------------------------------------------	
		ws.sendMsg = function(msg) {

			if (ws.readyState !== WebSocket.OPEN)
				return;

			ws.send(msg);
		};
	};


	/* ================================================================	*
	 * MDate.js 														*
	 * ================================================================	*/

	var MDate = function(id) {

		// getValue ---------------------------------------------------
		MDate.prototype.getValue = function() {
			var el, y, d, m, h, mm, tz;

			el = DocLib.getElement(this.id + '_day_input');
			d = MDate.toTwoDigit(el.value);

			el = DocLib.getElement(this.id + '_month_input');
			m = MDate.toTwoDigit(el.value);

			el = DocLib.getElement(this.id + '_year_input');
			y = MDate.toTwoDigit(el.value);

			el = DocLib.getElement(this.id + '_hour_input');
			h = MDate.toTwoDigit(el.value);

			el = DocLib.getElement(this.id + '_min_input');
			mm = MDate.toTwoDigit(el.value);

			el = DocLib.getElement(this.id + '_tz_input');
			tz = (el.selectedIndex === 0) ? this.offset : 'UTC';

			return y + '-' + m + '-' + d + ' ' + h + 'h' + mm + 'm ' + tz;
		};

		// reset ------------------------------------------------------
		MDate.prototype.reset = function() {
			var d, el, month, year;

			d = new Date();
			this.offset = d.getTimezoneOffset();

			if (this.offset === 0)
				this.offset = '+00';
			else
			if (this.offset > 0)
				this.offset = '+' + this.offset;

			el = DocLib.getElement(this.id + '_day_input');
			el.selectedIndex = d.getDate() - 1;

			el = DocLib.getElement(this.id + '_month_input');
			month = d.getMonth();
			el.selectedIndex = month;

			el = DocLib.getElement(this.id + '_year_input');
			year = d.getFullYear();
			el.selectedIndex = year - el.options[0].value;

			el = DocLib.getElement(this.id + '_hour_input');
			el.selectedIndex = d.getHours();

			el = DocLib.getElement(this.id + '_min_input');
			el.selectedIndex = d.getMinutes() / 5;

			el = DocLib.getElement(this.id + '_tz_input');
			el.selectedIndex = 0;

			MDate.checkMonth(this.id, month, year);
		};

		// Constructor -------------------------------------------------------
		{
			var el;

			this.id = id;

			this.reset();
			
			el = DocLib.getElement(this.id + '_month_input');
			el.onchange = function(event) {

				try {
					MDate.onChange(id);
				}
				catch(err) {
					HttpError.send(err);
					MW.xappl.runTask('#http-jserror');
				}
			};
			
			el = DocLib.getElement(this.id + '_year_input');
			el.onchange = function(event) {

				try {
					MDate.onChange(id);
				}
				catch(err) {
					HttpError.send(err);
					MW.xappl.runTask('#http-jserror');
				}
			};
			
			el = DocLib.getElement(this.id + '_tz_input');
			el.options[0].text = XUtil.getLocalTimeZone();
		}
	};

	// checkMonth --------------------------------------------------------
	MDate.checkMonth = function(id, month, year) {
		var el, options, ndays;

		el = DocLib.getElement(id + '_day_input');
		options = el.options;

		switch(month) {
		case 0: case 2: case 4: case 6: case 7: case 9: case 11:
			options[28].disabled = false;
			options[29].disabled = false;
			options[30].disabled = false;
			break;
		case 1:
			ndays = new Date(year, month + 1, 0).getDate();
			options[28].disabled = ndays === 28;
			options[29].disabled = true;
			options[30].disabled = true;
			break;
		case 3: case 5: case 8: case 10:
			options[30].disabled = true;
			break;
		}

		return this.minWidth;
	};

	// onChange ----------------------------------------------------------
	MDate.onChange = function(id) {
		var el, month, year;

		el = DocLib.getElement(id + '_month_input');
		month = el.value - 1;

		el = DocLib.getElement(id + '_year_input');
		year = el.value;

		MDate.checkMonth(id, month, year);
	};

	// toTwoDigit --------------------------------------------------------
	MDate.toTwoDigit = function(v) {

		if (v < 10)
			return '0' + v;

		return v;
	};

	/* ================================================================	*
	 * MWLib.js 														*
	 * ================================================================	*/

	var MWLib = function() {
	};

	MWLib.GoldenRatio	= 1.618;
	MWLib.Scrollw		= 0;

	// checkWidgetId ------------------------------------------------------
	MWLib.checkWidgetId = function(wt) {
		var ids;

		ids = [];

		MWLib.checkWidgetIdWalk(ids, wt);
	};

	// checkWidgetIdWalk --------------------------------------------------
	MWLib.checkWidgetIdWalk = function(ids, wt) {
		var i, a, el;

		if ((el = ids[wt.id]) !== undefined)
			throw new Error('duplicate id: ' + wt.id + ": " + wt.type + "/" + el.type);

		ids[wt.id] = wt;

		if ((a = wt.children) === undefined)
			return;

		if (a === null)
			return;
		
		for(i = 0; i < a.length; ++i)
			MWLib.checkWidgetIdWalk(ids, a[i]);
	};

	// getBorderLeft ------------------------------------------------------
	MWLib.getBorderLeft = function(style) {

		return parseFloat(style.getPropertyValue('border-left-width'));
	};

	// getBorderRight -----------------------------------------------------
	MWLib.getBorderRight = function(style) {

		return parseFloat(style.getPropertyValue('border-right-width'));
	};

	// getBordertop -------------------------------------------------------
	MWLib.getBorderTop = function(style) {

		return parseFloat(style.getPropertyValue('border-top-width'));
	};

	// getBorderBottom ----------------------------------------------------
	MWLib.getBorderBottom = function(style) {

		return parseFloat(style.getPropertyValue('border-bottom-width'));
	};

	// getEdgeWidth -------------------------------------------------------
	MWLib.getEdgeWidth = function(style) {
		var w;

		w = parseFloat(style.getPropertyValue('margin-left'));
		w += parseFloat(style.getPropertyValue('margin-right'));
		w += parseFloat(style.getPropertyValue('border-left-width'));
		w += parseFloat(style.getPropertyValue('border-right-width'));
		w += parseFloat(style.getPropertyValue('padding-left'));
		w += parseFloat(style.getPropertyValue('padding-right'));

		return w;
	};

	// getEdgeHeight ------------------------------------------------------
	MWLib.getEdgeHeight = function(style) {
		var h;

		h = parseFloat(style.getPropertyValue('margin-top'));
		h += parseFloat(style.getPropertyValue('margin-bottom'));
		h += parseFloat(style.getPropertyValue('border-top-width'));
		h += parseFloat(style.getPropertyValue('border-bottom-width'));
		h += parseFloat(style.getPropertyValue('padding-top'));
		h += parseFloat(style.getPropertyValue('padding-bottom'));

		return h;
	};

	// getPaddingLeft -----------------------------------------------------
	MWLib.getPaddingLeft = function(style) {

		return parseFloat(style.getPropertyValue('padding-left'));
	};

	// getPaddingRight ----------------------------------------------------
	MWLib.getPaddingRight = function(style) {

		return parseFloat(style.getPropertyValue('padding-right'));
	};

	// getPaddingTop ------------------------------------------------------
	MWLib.getPaddingTop = function(style) {

		return parseFloat(style.getPropertyValue('padding-top'));
	};

	// getPaddingBottom  --------------------------------------------------
	MWLib.getPaddingBottom = function(style) {

		return parseFloat(style.getPropertyValue('padding-bottom'));
	};

	// getWidth -----------------------------------------------------------
	MWLib.getWidth = function(el) {

		return Math.ceil(el.getBoundingClientRect().width);
	};

	// getHeight ----------------------------------------------------------
	MWLib.getHeight = function(el) {

		return Math.ceil(el.getBoundingClientRect().height);
	};

	// getScrollbarWidth --------------------------------------------------
	MWLib.getScrollbarWidth = function() {
		var w1, w2, outer, inner;

		if (MWLib.Scrollw !== 0)
			return MWLib.Scrollw;

		outer = document.createElement("div");
		outer.style.visibility = "hidden";
		outer.style.width = "100px";
		//outer.style.msOverflowStyle = "scrollbar"; // needed for WinJS apps
		document.body.appendChild(outer);
		w1 = outer.offsetWidth;

		// force scrollbars
		outer.style.overflow = "scroll";

		// add innerdiv
		inner = document.createElement("div");
		inner.style.width = "100%";
		outer.appendChild(inner);        
		w2 = inner.offsetWidth;

		// remove divs
		outer.parentNode.removeChild(outer);

		MWLib.Scrollw =  w1 - w2;

		return MWLib.Scrollw;
	};


	/* ================================================================	*
	 * MLabel.js 														*
	 * ================================================================	*/

	var MLabel = function(data) {

		// getMinimumWidth ---------------------------------------------------
		MLabel.prototype.getMinimumWidth = function() {

			return this.minWidth;
		};

		// getPrefWidth ------------------------------------------------------
		MLabel.prototype.getPrefWidth = function() {

			return this.prefWidth;
		};

		// format ------------------------------------------------------------
		MLabel.prototype.format = function(width) {

			this.p.style.width = width + 'px';

			return new XSize(MWLib.getWidth(this.p), MWLib.getHeight(this.p));
		};

		// init --------------------------------------------------------------
		MLabel.prototype.init = function(el) {
			var st, style1, style2;

			this.p = el.firstElementChild;
			st = window.getComputedStyle(this.p);

			if (st.whiteSpace === 'nowrap') {
				this.prefWidth = MWLib.getWidth(this.p);
				return;
			}

			style1 = el.style;
			style2 = this.p.style;

			style2.overflow = 'auto';
			style1.width = '0px';
			style2.width = '0px';
			this.minWidth = this.p.scrollWidth;

			style2.overflow = 'visible';
			style1.width = 'auto';
			style2.width = 'auto';
			style2.whiteSpace = 'nowrap';
			
			this.prefWidth = MWLib.getWidth(this.p);
			style2.whiteSpace = 'normal';
		};

		// setUndefline ------------------------------------------------------
		MLabel.prototype.setUnderline = function() {
		};

		// Constructor -------------------------------------------------------
		{
			this.text = data.text;
		}
	};

	/* ================================================================	*
	 * MText.js 														*
	 * ================================================================	*/

	var MText = function(data) {

		// format ------------------------------------------------------------
		MText.prototype.format = function(width) {
			var w, h;

			this.el.style.width = width + 'px';
			this.el.style.height = 'auto';

			w = MWLib.getWidth(this.el) - this.edgew;
			h = MWLib.getHeight(this.el) - this.edgeh;

			return new XSize(w, h);
		};

		// getMinimumWidth ---------------------------------------------------
		MText.prototype.getMinimumWidth = function() {

			return this.minWidth;
		};

		// getPrefWidth ------------------------------------------------------
		MText.prototype.getPrefWidth = function() {

			return this.prefWidth;
		};

		// init --------------------------------------------------------------
		MText.prototype.init = function(el) {
			var i, l1, l2, bl, br, bt, bb, pl, pr, pt, pb, st;

			this.el = el;

			st = window.getComputedStyle(el);

			bl = MWLib.getBorderLeft(st);
			br = MWLib.getBorderRight(st);
			bt = MWLib.getBorderTop(st);
			bb = MWLib.getBorderBottom(st);
			pl = MWLib.getPaddingLeft(st);
			pr = MWLib.getPaddingRight(st);
			pt = MWLib.getPaddingTop(st);
			pb = MWLib.getPaddingBottom(st);

			// minWidth ...
			st = el.style;
			st.overflow = 'auto';
			st.width = '0px';
			st.borderLeftWidth = '0px';
			st.borderRightWidth = '0px';
			st.paddingLeft = '0px';
			st.paddingRight = '0px';
			this.minWidth = this.el.scrollWidth;

			st.overflow = 'visible';
			st.width = 'auto';

			// prefWidth ...
			l1 = el.getElementsByTagName('p');
			l2 = el.getElementsByTagName('li').length;

			for(i = 0; i < l1.length; ++i)
				l1[i].style.whiteSpace = 'nowrap';

			for(i = 0; i < l2.length; ++i)
				l2[i].style.whiteSpace = 'nowrap';

			this.prefWidth = MWLib.getWidth(this.el);

			for(i = 0; i < l1.length; ++i)
				l1[i].style.whiteSpace = 'normal';

			for(i = 0; i < l2.length; ++i)
				l2[i].style.whiteSpace = 'normal';
			
			// restore ....
			st.borderLeftWidth = bl + 'px';
			st.borderRightWidth = br + 'px';
			st.paddingLeft = pl + 'px';
			st.paddingRight = pr + 'px';

			this.edgew = bl + br + pl + pr;
			this.edgeh = bt + bb + pt + pb;
		};

		// Constructor -------------------------------------------------------
		{
			this.text = data.text;
		}
	};

	/* ================================================================	*
	 * MTreeToggle.js 													*
	 * ================================================================	*/

	var MTreeToggle = function(wt) {

		// onclick -----------------------------------------------------------
		MTreeToggle.prototype.onclick = function(x, y) {
			var r;

			try {
				r = this.canvas.getBoundingClientRect();
				x -= r.left + this.tx;
				y -= r.top + this.ty;

				switch(this.wt.mode) {
				case 'open':
					this.wt.mode = 'close';
					break;
				case 'close':
					this.wt.mode = 'open';
					break;
				}

				this.repaint();
			}
			catch(err) {
				HttpError.send(err);
				MW.xappl.runTask('#http-jserror');
			}
		};

		// repaint -----------------------------------------------------------
		MTreeToggle.prototype.repaint = function() {
			var  bg, ctx;

			bg = this.wt.bg;

			if (this.wt.view !== null)
				if (this.wt.idx === this.wt.view.selected)
					bg = this.wt.selection;

			ctx = this.canvas.getContext('2d');

			ctx.fillStyle = bg;
			ctx.fillRect(0, 0, 100, 100);	// <canvas width="100" height="100"/>

			ctx.beginPath();
			switch(this.wt.mode) {
			case 'open':
				ctx.moveTo(0, 25);
				ctx.lineTo(100, 25);
				ctx.lineTo(50, 88);
				break;
			case 'close':
				ctx.moveTo(25, 0);
				ctx.lineTo(25, 100);
				ctx.lineTo(88, 50);
				break;
			}

			ctx.fillStyle = this.wt.fg;
			ctx.fill();
		};

		// Constructor -------------------------------------------------------
		{
			this.wt = wt;
			this.canvas = wt.uiwt.element;

			if (wt.mode === 'none')
				return;

			this.canvas.onclick = function(evt) {

				try {
					wt.uiwt.uitoggle.onclick(evt.clientX, evt.clientY);

					if (wt.view !== null)
						wt.view.taskTogglePressed(wt.idx);
				}
				catch(err) {
					HttpError.send(err);
					MW.xappl.runTask('#http-jserror');
				}
			};
		}
	};

	/* ================================================================	*
	 * UIWidget.js 														*
	 * ================================================================	*/

	var UIWidget = function(element, widget, parent) {

		// layoutReset ------------------------------------------------
		UIWidget.prototype.layoutReset = function() {

			switch(this.widget.type) {
			case 'stack':
				this.widget.reset();
				this.eventUpdateUI();
				break;
			}
		};

		// layoutSetLocation ------------------------------------------
		UIWidget.prototype.layoutSetLocation = function(vx, vy, vwidth, vheight) {
			var x, y, w, h, p, r, el, tx, ty, th, img, wt, mtop, style;

			wt = this.widget;

			x = vx;
			y = vy;

			if ((p = this.parent) !== null) {
				x -= p.margin.left + p.border.thickness.left;
				y -= p.margin.top + p.border.thickness.right;
			}

			w = Math.floor(vwidth - wt.getEdgeWidth());
			h = Math.floor(vheight - wt.getEdgeHeight());

			style = this.element.style;
			style.left = x + 'px';
			style.top = y + 'px';
			style.width = w + 'px';
			style.height = h + 'px';

			switch(wt.type) {
			case 'plugin':
				switch(wt.id) {
				case 'transmitter-audio':
	//				this.widget.plugin.setSize(w, h);
	//				this.widget.plugin.repaint();
					break;
	//			case 'uploading-bar':
	//				this.widget.setCanvasSize(w, h);
	//				this.widget.plugin.repaint(w, h);
					return;
				}
				return;
			case 'plugin-display':
				this.widget.setCanvasSize(w, h);
				this.widget.plugin.repaint(w, h);
				return;
			case 'plugin-slide':
				this.widget.setCanvasSize(w, h);
				this.widget.plugin.repaint(w, h);
				return;
			case 'plugin-toc-edit':
				this.widget.plugin.repaint(w, h);
				return;
			}

			switch(wt.type) {
			case 'area':
			case 'dropdown':
			case 'flow':
			case 'date':
			case 'grid':
			case 'hflow':
			case 'checkbox':
			case 'iconlink':
			case 'radiobutton':
			case 'link':
			case 'text':
			case 'menubar':
			case 'plugin':
			case 'simplegrid':
			case 'splitpane':
			case 'stack':
			case 'treearea':
			case 'treelabel':
			case 'treetoggle':
			case 'treeview':
			case 'vflow':
				break;
			case 'label':
			case 'labelbutton':
				el = this.widget.uilabel.p;
				th = MWLib.getHeight(el);

				mtop = this.widget.margin.top;

				if (h > th)
					mtop += (h - th) / 2;

				el.style.marginTop = mtop + 'px';
				break;
			case 'button':
			case 'combobox':
				style = this.child.style;
				style.width = w + 'px';
				style.height = h + 'px';
				break;
			case 'checkbox-item':
			case 'password':
			case 'radiobutton-item':
			case 'textarea':
			case 'textfield':
				style = this.child.style;
				style.width = (w - this.edgeWidth) + 'px';
				style.height = (h - this.edgeHeight) + 'px';
				break;
			case 'pictbutton':
			case 'picture':
				if ((img = this.element.firstElementChild) === null)
					break;

				r = XUILib.getImagePosition(w, h, wt.naturalsz.width, wt.naturalsz.height);

				if (wt.ratio !== null) {
					img.width = r.w;
					img.height = r.h;
				}

				style = img.style;
				style.marginLeft = r.x + 'px';
				style.marginTop = r.y + 'px';
				break;
			case 'scroll':
			case 'table':
			case 'tree':
				tx = wt.margin.left + wt.border.thickness.left;
				ty = wt.margin.top + wt.border.thickness.top;
				wt.view.layoutSetLocation(tx, ty, 'north');
				break;
			case 'treerow':
				if (wt.idx === wt.view.selected)
					wt.uiwt.eventUpdateUI();
				break;
			default:
				throw new Error('not yet implemented: ' + wt.type);
			}
		};

		// Constructor -------------------------------------------------------
		{
			var img, style, sz;
			
			this.element = element;
			this.widget = widget;
			this.parent = parent;

			switch(widget.type) {
			case 'area':
			case 'dropdown':
			case 'flow':
			case 'date':
			case 'grid':
			case 'hflow':
			case 'checkbox':
			case 'iconlink':
			case 'radiobutton':
			case 'label':
			case 'labelbutton':
			case 'link':
			case 'plugin':
			case 'plugin-display':
			case 'plugin-slide':
			case 'plugin-toc-edit':
			case 'text':
			case 'menubar':
			case 'tree':
			case 'scroll':
			case 'simplegrid':
			case 'splitpane':
			case 'table':
			case 'treearea':
			case 'treelabel':
			case 'treetoggle':
			case 'vflow':
				break;
			case 'button':
			case 'combobox':
				this.child = element.firstElementChild;
				break;
			case 'checkbox-item':
			case 'password':
			case 'radiobutton-item':
			case 'textarea':
			case 'textfield':
				this.child = element.firstElementChild;
				style = window.getComputedStyle(this.child);
				this.edgeWidth = MWLib.getEdgeWidth(style);
				this.edgeHeight = MWLib.getEdgeHeight(style);
				break;
			case 'pictbutton':
				if (widget.active === undefined)
					widget.active = UILib.getActiveColor(widget.bg);

				if (widget.over === undefined)
					widget.over = UILib.getOverColor(widget.bg);
			case 'picture':
				if (widget.ref === null) {
					if (widget.minWidth === 0 || widget.minHeight === 0)
						throw new Error('invalid picture size: ' + widget.ref);

					widget.naturalsz = new XSize(widget.minWidth,  widget.minHeight);
					break;
				}

				if (widget.ref.startsWith('cache:')) {
					if (widget.ratio === null)
						img = element;
					else
						img = element.firstElementChild;

					if (widget.naturalsz === null) {
						Log.info(widget.ref + ': ' + img.width + 'x' + img.height);
						widget.naturalsz = new XSize(img.width, img.height);
					}
					break;
				}

				if (widget.naturalsz === null) {
					if (widget.minWidth === 0 || widget.minHeight === 0)
						throw new Error('invalid picture size: ' + widget.ref);

					widget.naturalsz = new XSize(widget.minWidth, widget.minHeight);
					return;
				}

				sz = widget.naturalsz;
				if (sz.width === 0 || sz.height === 0)
					throw new Error('invalid picture natural size: ' + widget.ref);
				break;
			case 'stack':
				this.eventResizeUI = function() {

					MW.actionEventResized();
				};

				this.eventUpdateUI = function() {
					var i, l, el, stack;

					stack = this.widget;
					l = stack.children;
					for(i = 0; i < l.length; ++i) {
						el = l[i].uiwt.element;
						if (i === stack.idx)
							el.style.visibility = 'inherit';
						else
							el.style.visibility = 'hidden';
					}
				};
				break;
			case 'treerow':
				this.eventUpdateUI = function() {
					var bg, el, row;

					row = this.widget;

					if (row.idx === row.view.selected)
						bg = row.selection;
					else
						bg = row.bg;

					row.uiwt.element.style.backgroundColor = bg;

					if (row.area !== null) {
						el = row.area.uiwt.element;
						el.style.backgroundColor = bg;
					}

					row.toggle.uiwt.uitoggle.repaint();

					el = row.label.uiwt.element;
					el.style.backgroundColor = bg;
				};

				this.eventRepaint = function() {
				};
				break;
			case 'treeview':
				this.eventUpdateUI = function() {
					var i, all, child, style, row;

					all = this.widget.all;
					for(i = 0; i < all.length; ++i) {
						row = all[i];

						if (row.visible && row.toggle.mode !== 'none')
							row.toggle.uiwt.uitoggle.repaint();
						
						child = row.uiwt.element;
						style = child.style;
						style.display = (row.visible) ? 'block' : 'none';
					}
				};

				this.eventActionUI = function(data) {

					MW.actionTaskEvent(this.widget, data);
				};
				break;
			default:
				throw new Error('not yet implemented: ' + widget.type);
			}
		};
	};

	/* ================================================================	*
	 * XData.js 														*
	 * ================================================================	*/

	var XData = function(data) {

		// addChild ---------------------------------------------------
		XData.prototype.addChild = function(tag) {
			var child;

			if (this.data['_children'] === undefined)
				this.data['_children'] = [];

			child = {tag:tag};

			this.data['_children'].push(child);

			return new XData(child);
		};

		// addChildWithContent ----------------------------------------
		XData.prototype.addChildWithContent = function(tag, text) {

			this.data['_children'].push({tag:tag,_content:text});
		};

		// addXText ------------------------------------------------------
		XData.prototype.addXText = function(value) {

			this.data['text'] = value;
		};

		// getAllChildren ---------------------------------------------
		XData.prototype.getAllChildren = function() {
			var i, a, list;

			list = [];

			if ((a = this.data['_children']) === undefined)
				return list;
			
			for(i = 0; i < a.length; ++i)
				list.push(new XData(a[i]));

			return list;
		};

		// getAttrBool ------------------------------------------------
		XData.prototype.getAttrBool = function(attr, def) {
			var b;

			if ((b = this.data[attr]) !== undefined)
				return b;

			return def;
		};

		// getAttrInt -------------------------------------------------
		XData.prototype.getAttrInt = function(attr, def) {
			var v;

			if ((v = this.data[attr]) === undefined)
				return def;

			return v;
		};

		// getAttrString ----------------------------------------------
		XData.prototype.getAttrString = function(attr, def) {
			var s;

			if ((s = this.data[attr]) === undefined)
				return def;

			return s;
		};

		// getAttributes ----------------------------------------------
		XData.prototype.getAttributes = function() {
			var kvs;

			kvs = [];

			for(var id in this.data) {
				switch(id) {
				case 'tag':
				case '_children':
				case '_content':
					break;
				default:
					kvs.push(new XKeyValue(id, this.data[id]));
				}
			}

			return kvs;
		};

		// getChildren ------------------------------------------------
		XData.prototype.getChildren = function(attr) {
			var i, a, list;

			list = [];

			if ((a = this.data['_children']) === undefined)
				return list;

			for(i = 0; i < a.length; ++i)
				if (a[i].tag === attr)
					list.push(new XData(a[i]));

			return list;
		};

		// getFirstChildWithTag ---------------------------------------
		XData.prototype.getFirstChildWithTag = function(attr) {
			var a, i;

			if ((a = this.data['_children']) === undefined)
				return null;

			for(i = 0; i < a.length; ++i)
				if (a[i].tag === attr)
					return new XData(a[i]);

			return null;
		};

		// getString --------------------------------------------------
		XData.prototype.getString = function(attr) {

			return this.data[attr];
		};

		// getTag -----------------------------------------------------
		XData.prototype.getTag = function() {

			return(this.data['tag']);
		};

		// getTextContent ---------------------------------------------
		XData.prototype.getTextContent = function() {
			var v;

			if ((v = this.data['_content']) === undefined)
				return null;

			return v;
		};

		// searchElement ----------------------------------------------
		XData.prototype.searchElement = function(exp) {
			var i, j, children, cur, tag, path;

			if (exp.charAt(0) !== '/' && !exp.startsWith('./')) {
				children = this.data['_children'];
				for(i = 0; i < children.length; ++i)
					if (exp === children[i].tag)
						return(children[i]);
				
				throw new Error('expression not found: ' + exp);
			}

			path = exp.substring(1).split('/');
			cur = this.data;

			if (path[0] !== '' && path[0] !== cur['tag'])
				throw new Error('expression not found: ' + exp);

			for(i = 1; i < path.length; ++i) {
				tag = path[i];

				children = cur['_children'];
				for(j = 0; j < children.length; ++j)
					if (tag === children[j].tag)
						break;

				if (j === children.length)
					throw new Error('expression not found: ' + exp);

				cur = children[j];
			}

			return cur;
		};

		// searchElements ---------------------------------------------
		XData.prototype.searchElements = function(exp) {
			var i, el, child, children, res, s1, s2, tag, xd;

			i = exp.lastIndexOf('/');
			s1 = exp.substring(0, i);
			s2 = exp.substring(i + 1);
			
			el = this.searchElement(s1);
			children = el._children;

			res = [];
			for(i = 0; i < children.length; ++i) {
				child = children[i];
				
				if (s2 === '*')
					tag = child.tag;
				else {
					if (s2 !== child.tag)
						continue;
					tag = s2;
				}

				xd = new XData({tag:tag});

				if (child._children !== undefined)
					xd.setChildren(child._children);
				else
				if (child._content !== undefined)
					xd.setContent(child._content);
				else
					throw new Error('not data found...');

				res.push(xd);
			}
		
			return res;
		};

		// searchTextValue --------------------------------------------
		XData.prototype.searchTextValue = function(exp, src, tmpl) {
			var id, text;

			text = this.searchValue(exp);
			id = src.getString("id");
			
			return JsUtil.transformText(tmpl.fonts, id, text);
		};

		// searchValue ------------------------------------------------
		XData.prototype.searchValue = function(exp) {
			var el;

			if (exp === '.' || exp === 'text()')
				return this.data._content;

			el = this.searchElement(exp);

			return el._content;
		};

		// setChildren ------------------------------------------------
		XData.prototype.setChildren = function(children) {

			this.data['_children'] = children;
		};

		// setContent -------------------------------------------------
		XData.prototype.setContent = function(value) {

			this.data['_content'] = value;
		};
		
		// setIntAttr -------------------------------------------------
		XData.prototype.setIntAttr = function(attr, value) {

			this.data[attr] = value;
		};
		
		// setStringAttr ----------------------------------------------
		XData.prototype.setStringAttr = function(attr, value) {

			this.data[attr] = value;
		};

		// write ------------------------------------------------------
		XData.prototype.write = function() {

			console.log(this.data);
		};

		// Constructor ------------------------------------------------
		{
			this.data = data;
		}
	};

	/* ================================================================	*
	 * XMap.js 															*
	 * ================================================================	*/

	var XMap = function() {

		// exists -----------------------------------------------------
		XMap.prototype.exists = function(id) {

			return this.hmap[id] !== undefined;
		};
		
		// get --------------------------------------------------------
		XMap.prototype.get = function(id) {
			var v;

			if ((v = this.hmap[id]) !== undefined)
				return v;

			return null;
		};

		// keys -------------------------------------------------------
		XMap.prototype.keys = function(a) {
			var i;

			i = 0;
			for(var key in this.hmap)
				a[i++] = key;

			return a;
		};

		// put --------------------------------------------------------
		XMap.prototype.put = function(id, data) {

			this.hmap[id] = data;
		};

		// remove -----------------------------------------------------
		XMap.prototype.remove = function(id) {

			delete this.hmap[id];
		};

		// size -------------------------------------------------------
		XMap.prototype.size = function() {

			return this.hmap.length;
		};

		// values -----------------------------------------------------
		XMap.prototype.values = function(a) {
			var i;

			i = 0;
			for(var key in this.hmap)
				a[i++] = this.hmap[key];

			return a;
		};

		// Constructor ------------------------------------------------
		{
			this.hmap = {};
		}
	};

	/* ================================================================	*
	 * XOfflineData.js 													*
	 * ================================================================	*/

	var XOfflineData = function(data) {
	/*
		// addChild ---------------------------------------------------
		XOfflineData.prototype.addChild = function(tag) {
			var child;

			if (this.data['_children'] === undefined)
				this.data['_children'] = [];

			child = {tag:tag};

			this.data['_children'].push(child);

			return new XOfflineData(child);
		};

		// addChildWithContent ----------------------------------------
		XOfflineData.prototype.addChildWithContent = function(tag, text) {

			this.data['_children'].push({tag:tag,_content:text});
		};

		// addXText ------------------------------------------------------
		XOfflineData.prototype.addXText = function(value) {

			this.data['text'] = value;
		};
	*/
		// getAllChildren ---------------------------------------------
		XOfflineData.prototype.getAllChildren = function() {
			var i, a, list;

			list = [];

			if ((a = this.data['_children']) === undefined)
				return list;
			
			for(i = 0; i < a.length; ++i)
				list.push(new XOfflineData(a[i]));

			return list;
		};
	/*
		// getAttrBool ------------------------------------------------
		XOfflineData.prototype.getAttrBool = function(attr, def) {
			var b;

			if ((b = this.data[attr]) !== undefined)
				return b;

			return def;
		};

		// getAttrInt -------------------------------------------------
		XOfflineData.prototype.getAttrInt = function(attr, def) {
			var v;

			if ((v = this.data[attr]) === undefined)
				return def;

			return v;
		};
	*/
		// getAttrString ----------------------------------------------
		XOfflineData.prototype.getAttrString = function(attr, def) {
			var s;

			if ((s = this.data[attr]) === undefined)
				return def;

			return s;
		};
	/*
		// getAttributes ----------------------------------------------
		XOfflineData.prototype.getAttributes = function() {
			var kvs;

			kvs = [];

			for(var id in this.data) {
				switch(id) {
				case 'tag':
				case '_children':
				case '_content':
					break;
				default:
					kvs.push(new XKeyValue(id, this.data[id]));
				}
			}

			return kvs;
		};
	*/
		// getBool ----------------------------------------------------
		XOfflineData.prototype.getBool = function(attr) {
			var v;

			if ((v = this.data[attr]) === undefined)
				throw new Error('attribute not found: ' + attr);
			
			return v === 'true';
		};

		// getChildren ------------------------------------------------
		XOfflineData.prototype.getChildren = function(attr) {
			var i, a, list;

			list = [];

			if ((a = this.data['_children']) === undefined)
				return list;

			for(i = 0; i < a.length; ++i)
				if (a[i].tag === attr)
					list.push(new XOfflineData(a[i]));

			return list;
		};

		// getDouble --------------------------------------------------
		XOfflineData.prototype.getDouble = function(attr) {
			var v;

			if ((v = this.data[attr]) === undefined)
				throw new Error('attribute not found: ' + attr);
			
			return parseFloat(v);
		};

		// getFirstChildWithTag ---------------------------------------
		XOfflineData.prototype.getFirstChildWithTag = function(attr) {
			var a, i;

			if ((a = this.data['_children']) === undefined)
				return null;

			for(i = 0; i < a.length; ++i)
				if (a[i].tag === attr)
					return new XOfflineData(a[i]);

			return null;
		};

		// getInt -----------------------------------------------------
		XOfflineData.prototype.getInt = function(attr) {
			var v;

			if ((v = this.data[attr]) === undefined)
				return 0;

			return parseInt(v);
		};

		// getString --------------------------------------------------
		XOfflineData.prototype.getString = function(attr) {

			return this.data[attr];
		};

		// getTag -----------------------------------------------------
		XOfflineData.prototype.getTag = function() {

			return(this.data['tag']);
		};

		// getTextContent ---------------------------------------------
		XOfflineData.prototype.getTextContent = function() {
			var v;

			if ((v = this.data['_content']) === undefined)
				return null;

			return v;
		};

		// removeAttr -------------------------------------------------
		XOfflineData.prototype.removeAttr = function(attr) {

			delete this.data[attr];
		};

		// searchElement ----------------------------------------------
		XOfflineData.prototype.searchElement = function(exp) {
			var i, j, children, cur, path, tag;

			if (exp.charAt(0) !== '/') {
				children = this.data['_children'];
				for(i = 0; i < children.length; ++i)
					if (exp === children[i].tag)
						return(new XOfflineData(children[i]));

				return null;
			}

			path = exp.substring(1).split('/');
			cur = this.data;

			for(i = 1; i < path.length; ++i) {
				tag = path[i];

				children = cur['_children'];
				for(j = 0; j < children.length; ++j)
					if (tag === children[j].tag)
						break;

				if (j === children.length)
					throw new Error('expression not found: ' + exp);

				cur = children[j];
			}

			return new XOfflineData(cur);
		};
	/*
		// searchElements ---------------------------------------------
		XOfflineData.prototype.searchElements = function(exp) {
			var i, child, children, res;

			children = this.getChildren(exp);

			res = [];
			for(i = 0; i < children.length; ++i) {
				child = children[i];
				
				if (s2 === '*')
					tag = child.tag;
				else {
					if (s2 !== child.tag)
						continue;
					tag = s2;
				}

				xd = new XOfflineData({tag:tag});

				if (child._children !== undefined)
					xd.setChildren(child._children);
				else
				if (child._content !== undefined)
					xd.setContent(child._content);
				else
					throw new Error('not data found...');

				res.push(xd);
			}
		
			return res;
		};
	/*
		// searchTextValue --------------------------------------------
		XOfflineData.prototype.searchTextValue = function(exp, src, tmpl) {
			var id, text;

			text = this.searchValue(exp);
			id = src.getString("id");
			
			return JsUtil.transformText(tmpl.fonts, id, text);
		};
	*/
		// searchValue ------------------------------------------------
		XOfflineData.prototype.searchValue = function(exp) {
			var el;

			if (exp === '.' || exp === 'text()')
				return this.data._content;

			if ((el = this.searchElement(exp)) === null)
				return '';

			return el.getTextContent();
		};
	/*
		// setChildren ------------------------------------------------
		XOfflineData.prototype.setChildren = function(children) {

			this.data['_children'] = children;
		};

		// setContent -------------------------------------------------
		XOfflineData.prototype.setContent = function(value) {

			this.data['_content'] = value;
		};
		
		// setIntAttr -------------------------------------------------
		XOfflineData.prototype.setIntAttr = function(attr, value) {

			this.data[attr] = value;
		};
	*/	
		// setStringAttr ----------------------------------------------
		XOfflineData.prototype.setStringAttr = function(attr, value) {

			this.data[attr] = value;
		};

		// write ------------------------------------------------------
		XOfflineData.prototype.write = function() {

			console.log(this.data);
		};

		// Constructor ------------------------------------------------
		{
			this.data = data;
		}
	};

	/* ================================================================	*
	 * XWData.js 														*
	 * ================================================================	*/

	var XWData = function(data) {

		// getAlign ---------------------------------------------------
		XWData.prototype.getAlign = function(attr) {
			var v;

			if ((v = this.data[attr]) === undefined)
				return 'center';

			return v;
		};

		// getAllChildElements ----------------------------------------
		XWData.prototype.getAllChildElements = function(attr) {
			var i, a, list;

			list = [];

			if ((a = this.data['_children']) === undefined)
				return list;

			for(i = 0; i < a.length; ++i)
				if (a[i].tag === attr)
					list.push(new XWData(a[i]));

			return list;
		};

		// getAllChildrenWidgets --------------------------------------
		XWData.prototype.getAllChildrenWidgets = function() {
			var i, a, list;

			list = [];

			if ((a = this.data['_children']) === undefined)
				return list;
				
			for(i = 0; i < a.length; ++i)
				list.push(new XWData(a[i]));

			return list;
		};

		// getAnchor --------------------------------------------------
		XWData.prototype.getAnchor = function(attr) {

			return this.data[attr];
		};

		// getAttrBool ------------------------------------------------
		XWData.prototype.getAttrBool = function(attr, def) {
			var s;

			if ((s = this.data[attr]) === undefined)
				return def;

			return s;
		};

		// getAttrInt -------------------------------------------------
		XWData.prototype.getAttrInt = function(attr, def) {
			var s;

			if ((s = this.data[attr]) === undefined)
				return def;

			return s;
		};

		// getAttrSize ------------------------------------------------
		XWData.prototype.getAttrSize = function(attr, def) {
			var v;

			if ((v = this.data[attr]) === undefined)
				return def;
			
			return new XSize(v[0], v[1]);
		};

		// getAttrString ----------------------------------------------
		XWData.prototype.getAttrString = function(attr, def) {
			var s;

			if ((s = this.data[attr]) === undefined)
				return def;

			return s;
		};

		// getAttributes ----------------------------------------------
		XWData.prototype.getAttributes = function() {
			var v, kvs;
		
			kvs = [];

			for(var k in this.data) {
				if ((v = this.data[k]) === undefined)
					continue;

				if (v !== null)
					kvs.push(new XKeyValue(k, v));
			}

			return kvs;
		};

		// getBool ----------------------------------------------------
		XWData.prototype.getBool = function(attr) {
			var v;

			if ((v = this.data[attr]) === undefined)
				return false;

			return v;
		};

		// getBorder --------------------------------------------------
		XWData.prototype.getBorder = function(attr) {
			var v;

			if ((v = this.data[attr]) === undefined)
				return new XBorder('black', new XEdge(0, 0, 0, 0));

			if (v.length === 2)
				return new XBorder(v[0], new XEdge(v[1], v[1], v[1], v[1]));

			return new XBorder(v[0], new XEdge(v[1], v[2], v[3], v[4]));
		};

		// getColor ---------------------------------------------------
		XWData.prototype.getColor = function(attr) {
			var v;

			switch(attr) {
			case 'caption':
			case 'even':
			case 'odd':
				if ((v = this.data[attr]) === undefined)
					return UILib.NO_COLOR;

				return v;
			}
			
			if ((v = this.data[attr]) === undefined)
				return 'black';

			return v;
		};

		// getEdge ----------------------------------------------------
		XWData.prototype.getEdge = function(attr) {
			var v;

			if ((v = this.data[attr]) === undefined)
				return new XEdge(0, 0, 0, 0);

			if (v.length === 1)
				return new XEdge(v[0], v[0], v[0], v[0]);

			return new XEdge(v[0], v[1], v[2], v[3]);
		};

		// getFirstChild ----------------------------------------------
		XWData.prototype.getFirstChild = function() {
			var a;

			if ((a = this.data['_children']) === undefined)
				return null;

			return new XWData(a[0]);
		};

		// getFirstChildWithTag ---------------------------------------
		XWData.prototype.getFirstChildWithTag = function(attr) {
			var a, i;

			if ((a = this.data['_children']) === undefined)
				return null;

			for(i = 0; i < a.length; ++i)
				if (a[i].tag === attr)
					return new XWData(a[i]);

			return null;
		};

		// getFont ------------------------------------------------
		XWData.prototype.getFont = function() {

			return this.data.fontid;
		};

		// getInt -----------------------------------------------------
		XWData.prototype.getInt = function(attr) {
			var v;

			if ((v = this.data[attr]) === undefined)
				return 0;

			return v;
		};

		// getIntArray ------------------------------------------------
		XWData.prototype.getIntArray = function(attr, length) {
			var v;

			if ((v = this.data[attr]) === undefined)
				return UILib.newArray(length, 0);

			return v;
		};

		// getIntArrayList --------------------------------------------
		XWData.prototype.getIntArrayList = function(attr, sep, max) {
			var v;

			if ((v = this.data[attr]) === undefined)
				return null;

			return v;
		};

		// getLabel ---------------------------------------------------
		XWData.prototype.getLabel = function() {

			return new MLabel(this.data);
		};

		// getPercent -------------------------------------------------
		XWData.prototype.getPercent = function(attr) {
			var v;

			if ((v = this.data[attr]) === undefined)
				return 0;

			return v;
		};

		// getOrientation ---------------------------------------------
		XWData.prototype.getOrientation = function(attr) {
			var v;

			if ((v = this.data[attr]) === undefined)
				return 0;

			return v;
		};

		// getPercentArray --------------------------------------------
		XWData.prototype.getPercentArray = function(attr, value, length) {
			var a, i, n, s, error, idx, total, values;

			if (value === null)
				return UILib.newArray(length, 1.0 / length);

			a = UILib.newArray(length, 0);

			error = false;
			total = 0;
			values = value.split(',');
			
			for(i = 0; i < values.length; ++i) {
				s = values[i];

				if ((n = s.indexOf('=')) === -1) {
					idx = parseInt(s);
					a[idx] = 1.0;
				} else {
					idx = parseInt(s.substring(0, n));
					a[idx] = parseFloat(s.substring(n + 1));
				}

				if (idx < 0 || idx >= length || a[idx] < 0) {
					error = true;
					break;
				}

				total += a[idx];
			}

			if  (error || total === 0)
				return UILib.newArray(length, 1.0 / length);

			for(i = 0; i < length; ++i)
				a[i] /= total;

			return a;
		};

		// getRefWidget -----------------------------------------------
		XWData.prototype.getRefWidget = function() {
			var ref;

			ref = this.data.ref;
			
			if (UILib.stringsMatch(ref, "template:"))
				return MW.getTemplateWidget(this.data.id, ref.substring(9));

			throw new Error('ref not supported: ' + ref);
		};

		// getSize ----------------------------------------------------
		XWData.prototype.getSize = function(attr) {
			var v;

			if ((v = this.data[attr]) === undefined)
				return new XSize(0, 0);
			
			return new XSize(v[0], v[1]);
		};

		// getString --------------------------------------------------
		XWData.prototype.getString = function(attr) {
			var v;

			if ((v = this.data[attr]) !== undefined)
				return v;

			switch(attr) {
			case 'weightm':
				return 'width';
			}

			throw new Error('not yet implemented default value for: ' + attr);
		};

		// getStringArray ---------------------------------------------
		XWData.prototype.getStringArray = function(attr) {
			var v;

			if ((v = this.data[attr]) === undefined)
				return null;

			return v;
		};

		// getTag -----------------------------------------------------
		XWData.prototype.getTag = function() {

			return(this.data['tag']);
		};

		// getText ----------------------------------------------------
		XWData.prototype.getText = function() {

			return new MText(this.data);
		};

		// getXData ---------------------------------------------------
		XWData.prototype.getXData = function(value) {

			return MW.getWidgetData(value);
		};

		// isRawWidget ------------------------------------------------
		XWData.prototype.isRawWidget = function() {

			return false;
		};
		
		// toUIXWData -------------------------------------------------
		XWData.prototype.toUIXWData = function(data, force) {

			return DataToJs.transformToXWData(data, force);
		};

		// Constructor ------------------------------------------------
		{
			this.data = data;
		}
	};

	/* ================================================================	*
	 * DataToJs.js 														*
	 * ================================================================	*/

	var DataToJs = function(force) {

		// copyAttributes -------------------------------------------------
		DataToJs.prototype.copyAttributes = function(data, res) {

			for(var k in data)
				res[k] = data[k];
		};

		// getAnchor ------------------------------------------------------
		DataToJs.prototype.getAnchor = function(v) {

			if (v === undefined)
				return 'center';

			return v.toLowerCase();
		};

		// getBool --------------------------------------------------------
		DataToJs.prototype.getBool = function(v) {

			if (v === undefined)
				return undefined;

			return v === 'true';
		};

		// getBorder ------------------------------------------------------
		DataToJs.prototype.getBorder = function(v) {
			var n, s, t;

			if (v === undefined)
				return ['black',0];

			if ((n = v.indexOf(' ')) === -1) {
				s = v.toLowerCase();
				return [s,1];
			}

			s = v.substring(0, n).toLowerCase();
			t = this.getEdge(v.substring(n + 1));

			return [s, t[0], t[1], t[2], t[3]];
		};

		// getColor -------------------------------------------------------
		DataToJs.prototype.getColor = function(attr, v) {

			if (v === undefined) {
				switch(attr) {
				case 'bg':
					return 'white';
				case 'fg':
				case 'selection':
					return 'black';
				case 'caption':
				case 'even':
				case 'odd':
					return undefined;
				}

				throw new Error('invalid attribute: ' + attr);
			}

			return v.toLowerCase();
		};

		// getData --------------------------------------------------------
		DataToJs.prototype.getData = function(v) {

			if (v === undefined)
				return null;

			return v;
		};

		// getEdge --------------------------------------------------------
		DataToJs.prototype.getEdge = function(v) {
			var n, fields, top, left, bottom, right;

			if (v === undefined)
				return [0];

			fields = v.split(':');

			switch(fields.length) {
			case 1:
				n = parseInt(fields[0]);
				return [n,n,n,n];
			case 4:
				top = parseInt(fields[0]);
				right = parseInt(fields[1]);
				bottom = parseInt(fields[2]);
				left = parseInt(fields[3]);
				return [top,right,bottom,left];
			}

			throw new Error('not an edge: ' + v);
		};

		// getInt ---------------------------------------------------------
		DataToJs.prototype.getInt = function(v) {

			if (v === undefined)
				return undefined;

			return parseInt(v);
		};

		// getIntArray ----------------------------------------------------
		DataToJs.prototype.getIntArray = function(v, length) {
			var a, i, n, s, idx, fields;

			if (v === undefined)
				return undefined;

			a = [];
			for(i = 0; i < length; ++i)
				a[i] = 0;

			fields = v.split(',');
			for(i = 0; i < fields.length; ++i) {
				s = fields[i];

				if ((n = s.indexOf('=')) === -1)
					continue;

				idx = parseInt(s.substring(0, n));
				a[idx] = parseInt(s.substring(n + 1));
			}
			
			return a;
		};

		// getIntArrayList ------------------------------------------------
		DataToJs.prototype.getIntArrayList = function(v, sep, max) {
			var i, j, l, array, items, list;

			if (v === undefined)
				return undefined;

			if (v === 'all') {
				list = [];
				l = list[0] = [];
				for(i = 0; i < max; ++i)
					l[i] = i;
				return list;
			}

			list = [];
			items = v.split(',');
			for(i = 0; i < items.length; ++i) {
				array = items[i].split(sep);
				list[i] = [];
				for(j = 0; j < array.length; ++j)
					list[i][j] = parseInt(array[j]);
			}

			return list;
		};

		// getOrientation -------------------------------------------------
		DataToJs.prototype.getOrientation = function(v) {

			switch(v) {
			case 'horizontal':
				return XLib.Horizontal;
			case 'vertical':
				return XLib.Vertical;
			default:
				throw new Error('invalid orientation: ' + v);
			}
		};

		// getPercent -----------------------------------------------------
		DataToJs.prototype.getPercent = function(v) {
			var p;

			if (v === undefined)
				return undefined;

			if (v === 'golden')
				return 100 / MWLib.GoldenRatio;

			p = parseInt(v);

			return p;
		};

		// getSize --------------------------------------------------------
		DataToJs.prototype.getSize = function(v, def) {
			var n, w, h;

			if (v === undefined)
				return def;

			if ((n = v.indexOf('x')) === -1)
				throw new Error('not a size: ' + v);

			w = parseInt(v.substring(0, n));
			h = parseInt(v.substring(n + 1));
			return [w,h];
		};

		// getString ------------------------------------------------------
		DataToJs.prototype.getString = function(v) {

			if (v === undefined)
				return undefined;

			return v;
		};

		// getStringArray -------------------------------------------------
		DataToJs.prototype.getStringArray = function(v) {
			var a, i, r;

			if (v === undefined)
				return undefined;

			r = [];
			a = v.split(':');
			for(i = 0; i < a.length; ++i)
				r.push(a[i]);

			return r;
		};

		// transform ------------------------------------------------------
		DataToJs.prototype.transform = function(data) {
			var res;

			res = {};

			switch(data.tag) {
			case 'checkbox':
			case 'date':
			case 'iconlink':
			case 'radiobutton':
			case 'table':
			case 'tree':
				if (this.force)
					break;

				this.copyAttributes(data, res);
				return res;
			}

			this.transformWidget(data, res);

			switch(res.tag) {
			case 'area':
				break;
			case 'button':
				res.text = this.getString(data.text);
				res.action = this.getString(data.action);
				res.data = this.getData(data.data);
				break;
			case 'checkbox-item':
				res.key = this.getString(data.key);
				res.group = this.getString(data.group);
				break;
			case 'combobox':
				this.transformComboBox(data, res);
				break;
			case 'dropdown':
				this.transformDropDown(data, res);
				break;
			case 'flow':
			case 'date':
				this.transformFlow(data, res);
				break;
			case 'grid':
				this.transformGrid(data, res);
				break;
			case 'hflow':
			case 'checkbox':
			case 'iconlink':
			case 'radiobutton':
				this.transformHorFlow(data, res);
				break;
			case 'label':
				this.transformLabel(data, res);
				break;
			case 'labelbutton':
				this.transformLabel(data, res);
				res.action = this.getString(data.action);
				res.active = this.getString(data.active);
				res.over = this.getString(data.over);
				res.data = this.getData(data.data);
				break;
			case 'link':
				this.transformLabel(data, res);
				res.action = this.getString(data.action);
				res.data = this.getData(data.data);
				break;
			case 'menubar':
				this.transformMenuBar(data, res);
				break;
			case 'password':
				res.ncols = this.getInt(data.ncols);
				break;
			case 'pictbutton':
				res.ref = this.getString(data.ref);
				res.action = this.getString(data.action);
				res.active = this.getString(data.active);
				res.over = this.getString(data.over);
				res.data = this.getData(data.data);
				res.ratio = this.getString(data.ratio);
				res.naturalsz = this.getSize(data.naturalsz);
				break;
			case 'picture':
				res.ref = this.getString(data.ref);
				res.ratio = this.getString(data.ratio);
				break;
			case 'radiobutton-item':
				res.key = this.getString(data.key);
				res.group = this.getString(data.group);
				break;
			case 'scroll':
			case 'table':
				this.transformScrollPane(data, res);
				break;
			case 'simplegrid':
				this.transformSimpleGrid(data, res);
				break;
			case 'splitpane':
				this.transformSplitPane(data, res);
				break;
			case 'stack':
				this.transformStack(data, res);
				break;
			case 'text':
				res.text = data.text;
				res.widthmin = this.getInt(data.widthmin);
				res.widthmax = this.getInt(data.widthmax);
				break;
			case 'textarea':
				res.ncols = this.getInt(data.ncols);
				res.nrows = this.getInt(data.nrows);
				break;
			case 'textfield':
				res.ncols = this.getInt(data.ncols);
				break;
			case 'tree':
				this.transformTree(data, res);
				break;
			case 'treerow':
				this.transformHorFlow(data, res);
				res.depth = this.getInt(data.depth);
				res.selection = this.getString(data.selection);
				res.value = this.getString(data.value);
				break;
			case 'treearea':
				break;
			case 'treelabel':
				this.transformLabel(data, res);
				res.selection = this.getString(data.selection);
				break;
			case 'treetoggle':
				res.mode = this.getString(data.mode);
				res.selection = this.getString(data.selection);
				break;
			case 'vflow':
				this.transformVerFlow(data, res);
				break;
			default:
				throw new Error('not yet implemented: ' + res.tag);
			}

			return res;
		};

		// transformWidget ------------------------------------------------
		DataToJs.prototype.transformWidget = function(data, res) {

			if ((res.id = data.id) === undefined)
				res.id = MW.getTemplateWidgetId();

			res.tag = this.getString(data.tag);
			res.props = this.getString(data.props);

			res.margin = this.getEdge(data.margin);
			res.border = this.getBorder(data.border);
			res.padding = this.getEdge(data.padding);
			res.size = this.getSize(data.size, [0,0]);

			if (data.expand !== undefined)
				res.expand = this.getPercent(data.expand);
			else {
				res.expandh = this.getPercent(data.expandh);
				res.expandv = this.getPercent(data.expandv);
			}

			res.anchor = this.getAnchor(data.anchor);
			res.fg = this.getColor('fg', data.fg);
			res.bg = this.getColor('bg', data.bg);
			res.responsive = this.getStringArray(data.responsive);
		};

		// transformComboBox ----------------------------------------------
		DataToJs.prototype.transformComboBox = function(data, res) {

			res['_children'] = data['_children'];
		};

		// transformDropDown ----------------------------------------------
		DataToJs.prototype.transformDropDown = function(data, res) {
			var dst;

			this.transformLabel(data, res);
			res.action = this.getString(data.action);
			res.active = this.getString(data.active);
			res.over = this.getString(data.over);
			res.data = this.getData(data.data);

			dst = [];
			dst.push(this.transform(data['_children'][0]));
			
			res['_children'] = dst;
		};
		
		// transformFlow --------------------------------------------------
		DataToJs.prototype.transformFlow = function(data, res) {
			var i, c1, c2, src, dst;

			res.gap = this.getInt(data.gap);
			res.interline = this.getInt(data.interline);
			res.align = this.getString(data.align);

			/*
			 * children
			 */
			dst = [];
			src = data['_children'];
			for(i = 0; i < src.length; ++i) {
				c1 = src[i];
				c2 = this.transform(c1);
				dst.push(c2);
			}

			res['_children'] = dst;
		};

		// transformGrid --------------------------------------------------
		DataToJs.prototype.transformGrid = function(data, res) {
			var i, c1, c2, src, dst;

			res.ncols = this.getInt(data.ncols);
			res.nrows = this.getInt(data.nrows);

			if (data.gap !== undefined)
				res.gap = this.getInt(data.gap);
			else {
				res.gapw = this.getInt(data.gapw);
				res.gaph = this.getInt(data.gaph);
			}

			res.caption = this.getColor('caption', data.caption);
			res.even = this.getColor('even', data.even);
			res.odd = this.getColor('odd', data.odd);

			res.colsmin = this.getIntArray(data.colsmin, res.ncols);
			res.rowsmin = this.getIntArray(data.rowsmin, res.nrows);
			res.colseq = this.getIntArrayList(data.colseq, '=', res.ncols);
			res.rowseq = this.getIntArrayList(data.rowseq, '=', res.nrows);
			res.weightm = this.getString(data.weightm);
			res.weighth = this.getString(data.weighth);
			res.weightv = this.getString(data.weightv);

			/*
			 * cell/shrink...
			 */
			dst = [];
			src = data['_children'];
			for(i = 0; i < src.length; ++i) {
				c1 = src[i];
				c2 = {tag:c1.tag};

				switch(c1.tag) {
				case 'cell':
					c2.size = this.getSize(c1.size, [1,1]);
					if (c1._children !== undefined)
						c2._children = [this.transform(c1._children[0])];
					break;
				case 'shrink':
					this.transformGridShrink(c1, c2);
					break;
				default:
					throw new Error('not yet implemented: ' + c1.tag);
				}

				dst.push(c2);
			}
			res['_children'] = dst;
		};

		// transformGridShrink --------------------------------------------
		DataToJs.prototype.transformGridShrink = function(src, dst) {
			var a, i, c1, c2, cells;

			dst.key = this.getString(src.key);
			dst.nlines = this.getInt(src.nlines);
			dst.ncols = this.getInt(src.ncols);
			dst.nrows = this.getInt(src.nrows);

			if (src.gap !== undefined)
				dst.gap = this.getInt(src.gap);
			else {
				dst.gapw = this.getInt(src.gapw);
				dst.gaph = this.getInt(src.gaph);
			}

			dst.colsmin = this.getIntArray(src.colsmin, dst.ncols);
			dst.rowsmin = this.getIntArray(src.rowsmin, dst.nrows);
			dst.colseq = this.getIntArrayList(src.colseq, '=', dst.ncols);
			dst.rowseq = this.getIntArrayList(src.rowseq, '=', dst.nrows);
			dst.weighth = this.getString(src.weighth);
			dst.weightv = this.getString(src.weightv);

			/*
			 * cells...
			 */
			a = [];
			cells = src['_children'];
			for(i = 0; i < cells.length; ++i) {
				c1 = cells[i];
				c2 = {tag:'cell'};
				c2.offset = this.getInt(c1.offset);
				c2.size = this.getSize(c1.size, [1,1]);
				c2.anchor = this.getAnchor(c1.anchor);
				a.push(c2);
			}
			dst['_children'] = a;
		};

		// transformHorFlow -----------------------------------------------
		DataToJs.prototype.transformHorFlow = function(data, res) {
			var i, c1, c2, src, dst;

			res.weight = this.getInt(data.weight);
			res.gap = this.getInt(data.gap);
			res.align = this.getString(data.align);

			/*
			 * children
			 */
			dst = [];
			src = data['_children'];
			for(i = 0; i < src.length; ++i) {
				c1 = src[i];
				c2 = this.transform(c1);
				dst.push(c2);
			}

			res['_children'] = dst;
		};

		// transformLabel -------------------------------------------------
		DataToJs.prototype.transformLabel = function(data, res) {
			var html, sb, fontid;

			res.nowrap = (data.nowrap === 'true');
			
			// if label is in XCheckBox, XRadioButton or XIconLink then use data.font
			fontid = (data.fontid !== undefined) ? data.fontid : data.font;

			sb = [];
			sb.push('<div id="');
			sb.push(res.id);
			sb.push('"><p id="');
			sb.push(res.id);
			sb.push('_label" class="mw-label mwf-');
			sb.push(fontid);

			if (res.nowrap)
				sb.push(' mw-nowrap');

			sb.push(' mw-');
			
			if (data.align !== undefined)
				sb.push(data.align);
			else
				sb.push('center');
			
			if (data.tag === 'link')
				sb.push(' mw-link');

			sb.push('">');
	//		html = data.text.replace(/\\n/g, '<br>').replace(/\\\\/g, '\\');
			html = data.text;
			sb.push(html);
			sb.push('</p></div>');

			res.text = sb.join('');
		};

		// transformMenuBar -----------------------------------------------
		DataToJs.prototype.transformMenuBar = function(data, res) {
			var i, j, c, src, dst, children;

			res.gap = this.getInt(data.gap);
			res.interline = this.getInt(data.interline);

			/*
			 * left/right...
			 */
			dst = [];
			src = data['_children'];
			for(i = 0; i < src.length; ++i) {
				c = {tag:src[i].tag, _children:[]};

				children = src[i]._children;
				for(j = 0; j < children.length; ++j)
					c._children[j] = this.transform(children[j]);

				dst.push(c);
			}

			res['_children'] = dst;
		};

		// transformScrollPane --------------------------------------------
		DataToJs.prototype.transformScrollPane = function(data, res) {
			var wt, dst;

			// atribute bg & padding not allowed...
			//res.bg = undefined;
			res.padding = undefined;

			dst = [];
			wt = this.transform(data['_children'][0]);

			// attribute margin & border not allowed...
			//wt.margin = undefined;
			wt.border = undefined;
			
			dst.push(wt);
			
			res['_children'] = dst;
		};

		// transformSimpleGrid --------------------------------------------
		DataToJs.prototype.transformSimpleGrid = function(data, res) {
			var i, c1, c2, src, dst;

			if (data.gap !== undefined)
				res.gap = this.getInt(data.gap);
			else {
				res.gapw = this.getInt(data.gapw);
				res.gaph = this.getInt(data.gaph);
			}

			res.maxcols = this.getInt(data.maxcols);
			res.mincols = this.getInt(data.mincols);
			res.compact = this.getBool(data.compact);
			res.align = this.getString(data.align);

			/*
			 * children...
			 */
			dst = [];
			src = data['_children'];
			for(i = 0; i < src.length; ++i) {
				c1 = src[i];
				c2 = this.transform(c1);
				dst.push(c2);
			}

			res['_children'] = dst;
		};

		// transformSplitPane ---------------------------------------------
		DataToJs.prototype.transformSplitPane = function(data, res) {
			var i, c1, c2, dst, src;

			res.gap = this.getInt(data.gap);
			res.orientation = this.getOrientation(data.orientation);
			res.buttonc = this.getColor("buttonc", data.buttonc);
			res.weight = this.getString(data.weight);

			/*
			 * children...
			 */
			dst = [];
			src = data['_children'];
			for(i = 0; i < src.length; ++i) {
				c1 = src[i];
				c2 = this.transform(c1);
				dst.push(c2);
			}

			res['_children'] = dst;
		};

		// transformStack -------------------------------------------------
		DataToJs.prototype.transformStack = function(data, res) {
			var i, c1, c2, dst, src;

			/*
			 * children...
			 */
			dst = [];
			src = data['_children'];
			for(i = 0; i < src.length; ++i) {
				c1 = src[i];
				c2 = this.transform(c1);
				dst.push(c2);
			}

			res['_children'] = dst;
		};

		// transformTree --------------------------------------------------
		DataToJs.prototype.transformTree = function(data, res) {
			var a, i, r, bg, fg, sel, fontid, drow, row, src, dview, view, tview;

			dview = data['_children'][0];

			if (dview['_children'] !== undefined) {
				drow = dview['_children'][0];
				fg = this.getColor('fg', drow.fg);
				sel = this.getColor('selection', drow.selection);
			} else {
				fg = this.getColor('fg', 'white');
				sel = this.getColor('selection', 'red');
			}

			bg = this.getColor('bg', dview.bg);
			fontid = this.getString(data.fontid);

	//		res.bg = undefined;
			res.padding = undefined;

			// treeview...
			view = {};
			view.anchor = 'center';
			view.bg = bg;
			view.expandh = 100;
			view.gap = this.getInt(data.vgap);
			view.id = res.id + '_view';
			view.tag = 'treeview';
			view.weight = -1;
			view.action = this.getString(dview.action);

			r = [];
			r.push(view);
			res['_children'] = r;

			// treeview & treerows...
			r = [];

			tview = data._children[0];

			if ((a = tview._children) === undefined) {
				view['_children'] = r;
				return;
			}

			for(i = 0; i < a.length; ++i) {
				src = a[i];
				row = {};
				row.align = 'center';
				row.anchor = 'center';
				row.bg = bg;
				row.fg = fg;
				row.depth = this.getInt(src.depth);
				row.expandh = '100%';
				row.fontid = fontid;
				row.gap = this.getInt(data.hgap);
				row.labelminw = this.getString(src.labelminw);
				row.id = res.id + '_row' + i;
				row.mode = 'none';
				row.padding = this.getString(src.padding);
				row.selection = sel;
				row.tag = 'treerow';
				row.text = src.text;
				row.value = this.getString(a[i].value);

				r.push(row); 
			}

			view['_children'] = r;
		};
		
		// transformVerFlow -----------------------------------------------
		DataToJs.prototype.transformVerFlow = function(data, res) {
			var i, c1, c2, src, dst;

			res.gap = this.getInt(data.gap);
			res.weight = this.getInt(data.weight);

			/*
			 * children
			 */
			dst = [];
			src = data['_children'];
			for(i = 0; i < src.length; ++i) {
				c1 = src[i];
				c2 = this.transform(c1);
				dst.push(c2);
			}

			res['_children'] = dst;
		};

		// Constructor ----------------------------------------------------
		{
			this.force = force;
		}
	};

	// search -------------------------------------------------------------
	DataToJs.search = function(data, id) {
		var i, r, children;

		if (data.id === id)
			return data;

		if (data._children === undefined || data._children === null)
			return null;

		children = data._children;

		for(i = 0; i < children.length; ++i)
			if ((r = DataToJs.search(children[i], id)) !== null)
				return r;

		return(null);
	};

	// transformFromXElement ----------------------------------------------
	DataToJs.transformFromXElement = function(data, force) {
		var t;

		t = new DataToJs(force);
		return new XWData(t.transform(data));
	};

	// transformToXWData --------------------------------------------------
	DataToJs.transformToXWData = function(xdata, force) {
		var t;

		t = new DataToJs(force);
		return new XWData(t.transform(xdata.data));
	};

	/* ================================================================	*
	 * HtmlWriter.js 													*
	 * ================================================================	*/

	var HtmlWriter = function() {
	};

	// createArea ---------------------------------------------------------
	HtmlWriter.createArea = function(sb, wt) {

		sb.push('<div id="');
		sb.push(wt.id);
		sb.push('"></div>');
	};

	// createButton -------------------------------------------------------
	HtmlWriter.createButton = function(sb, wt) {

		sb.push('<div id="');
		sb.push(wt.id);
		sb.push('"><button id="');
		sb.push(wt.id + '_input');
		sb.push('" class="mw">');
		sb.push(wt.text);
		sb.push('</button></div>');
	};

	// createCombobox -----------------------------------------------------
	HtmlWriter.createComboBox = function(sb, wt) {
		var i, opt, options;

		sb.push('<div id="');
		sb.push(wt.id);
		sb.push('"><select id="');
		sb.push(wt.id + '_input');
		sb.push('" class="mw">');

		options = wt.options;
		for(i = 0; i < options.length; ++i) {
			opt = options[i];

			sb.push('\n');
			sb.push('<option value="' + opt.key + '">');
			sb.push(opt.value);
			sb.push('</option>');
		}

		sb.push('</select></div>');
	};

	// createCheckBoxItem -------------------------------------------------
	HtmlWriter.createCheckBoxItem = function(sb, wt) {

		sb.push('<div id="');
		sb.push(wt.id);
		sb.push('"><input');
		sb.push(' id="' + wt.id + '_input"');
		sb.push(' type="checkbox"');
		sb.push(' name="' + wt.group + '"');
		sb.push(' value="' + wt.key + '"');
		sb.push('></div>');
	};

	// createChildren -----------------------------------------------------
	HtmlWriter.createChildren = function(sb, wt) {
		var i, children;

		sb.push('<div id="');
		sb.push(wt.id);
		sb.push('">');

		children = wt.children;
		for(i = 0; i < children.length; ++i) {
			sb.push('\n');
			HtmlWriter.createWidget(sb, children[i]);
		}
		
		sb.push('\n</div>');
	};

	// createDropDown -----------------------------------------------------
	HtmlWriter.createDropDown = function(sb, wt) {

		sb.push(wt.uilabel.text);
		wt.uilabel.text = null;

		sb.push('\n');
		
		HtmlWriter.createWidget(sb, wt.popup);
	};

	// createLabel --------------------------------------------------------
	HtmlWriter.createLabel = function(sb, wt) {

		sb.push(wt.uilabel.text);
		wt.uilabel.text = null;
	};

	// createLink ---------------------------------------------------------
	HtmlWriter.createLink = function(sb, wt) {

		sb.push(wt.uilabel.text);
		wt.uilabel.text = null;
	};

	// createPassword -----------------------------------------------------
	HtmlWriter.createPassword = function(sb, wt) {

		sb.push('<div id="');
		sb.push(wt.id);
		sb.push('"><input id="');
		sb.push(wt.id + '_input');
		sb.push('" type="password" size="');
		sb.push(wt.ncols);
		sb.push('"></div>');
	};

	// createPictButton ---------------------------------------------------
	HtmlWriter.createPictButton = function(sb, wt) {
		var img, ref;

		if (wt.ref.startsWith('cache:')) {
			ref = wt.ref.substring(6);
			if ((img = MW.HttpFiles.getImage(ref)) === null)
				throw new Error('images not found: ' + ref);
			ref = img.src;
		} else
		if (wt.ref.startsWith('offline:'))
			ref = wt.ref.substring(8);
		else
			ref = 'images/' + wt.ref;

		sb.push('<div id="');
		sb.push(wt.id);
		sb.push('"><img id="');
		sb.push(wt.id + '_pict');
		sb.push('" src="');
		sb.push(ref);
		sb.push('"></div>');
	};

	// createPicture ------------------------------------------------------
	HtmlWriter.createPicture = function(sb, wt) {
		var imgId;

		imgId = wt.id;

		if (wt.ratio !== null) {
			sb.push('<div id="');
			sb.push(wt.id);
			sb.push('">');
			imgId += '_img';
		}
			
		sb.push('<img id="');
		sb.push(imgId);

		if (wt.ref !== null) {
			if (!wt.ref.startsWith('var:')) { 
				sb.push('" src="images/');
				sb.push(wt.ref);
			}
		}
		
		sb.push('" alt="">');

		if (wt.ratio !== null)
			sb.push('</div>');
	};

	// createPlugin -------------------------------------------------------
	HtmlWriter.createPlugin = function(sb, wt) {
		var html;

		if ((html = XPluginLib.html[wt.id]) === undefined)
			throw new Error('html for plugin not found: ' + wt.id);

		sb.push(html);
	/*
		sb.push('<div id="');
		sb.push(wt.id);
		sb.push('">Hello World</canvas>');
	*/
	};

	// createPluginDisplay ------------------------------------------------
	HtmlWriter.createPluginDisplay = function(sb, wt) {

		sb.push('<canvas id="');
		sb.push(wt.id);
		sb.push('"></canvas>');
	};

	// createPluginSlide --------------------------------------------------
	HtmlWriter.createPluginSlide = function(sb, wt) {

		sb.push('<canvas id="');
		sb.push(wt.id);
		sb.push('"></canvas>');
	};

	// createPluginTocEdit ------------------------------------------------
	HtmlWriter.createPluginTocEdit = function(sb, wt) {

		sb.push('<canvas id="');
		sb.push(wt.id);
		sb.push('" width="1000" height="1000"></canvas>');
	};

	// createRadioButtonItem ----------------------------------------------
	HtmlWriter.createRadioButtonItem = function(sb, wt) {

		sb.push('<div id="');
		sb.push(wt.id);
		sb.push('"><input');
		sb.push(' id="' + wt.id + '_input"');
		sb.push(' type="radio"');
		sb.push(' name="' + wt.group + '"');
		sb.push(' value="' + wt.key + '"');
		sb.push('></div>');
	};

	// createScrollPane ---------------------------------------------------
	HtmlWriter.createScrollPane = function(sb, wt) {

		sb.push('<div id="');
		sb.push(wt.id);
		sb.push('" style="overflow-y:scroll">');

		sb.push('\n');
		HtmlWriter.createWidget(sb, wt.view);

		sb.push('\n</div>');
	};

	// createText ---------------------------------------------------------
	HtmlWriter.createText = function(sb, wt) {

		sb.push(wt.uitext.text);
	};

	// createTextArea -----------------------------------------------------
	HtmlWriter.createTextArea = function(sb, wt) {
		
		sb.push('<div id="');
		sb.push(wt.id);
		sb.push('"><textarea id="');
		sb.push(wt.id + '_input');
		sb.push('" class="mw" cols="');
		sb.push(wt.ncols);
		sb.push('" rows="');
		sb.push(wt.nrows);

		if (wt.nospellcheck)
			sb.push('" spellcheck="false');

		sb.push('"></textarea></div>');
	};

	// createTextField ----------------------------------------------------
	HtmlWriter.createTextField = function(sb, wt) {

		sb.push('<div id="');
		sb.push(wt.id);
		sb.push('"><input id="');
		sb.push(wt.id + '_input');
		sb.push('" type="text" size="');
		sb.push(wt.ncols);
		sb.push('"></div>');
	};

	// createTree ---------------------------------------------------------
	HtmlWriter.createTree = function(sb, wt) {

		sb.push('<div id="');
		sb.push(wt.id);
		sb.push('" style="overflow-y:scroll">');

		sb.push('\n');
		HtmlWriter.createTreeRows(sb, wt.view);

		sb.push('\n</div>');
	};

	// createTreeRows -----------------------------------------------------
	HtmlWriter.createTreeRows = function(sb, wt) {
		var i, all;

		sb.push('<div id="');
		sb.push(wt.id);
		sb.push('">');

		all = wt.all;
		for(i = 0; i < all.length; ++i) {
			sb.push('\n');
			HtmlWriter.createWidget(sb, all[i]);
		}
		
		sb.push('\n</div>');
	};

	// createTreeToggle ---------------------------------------------------
	HtmlWriter.createTreeToggle = function(sb, wt) {

		sb.push('<canvas id="');
		sb.push(wt.id);
		sb.push('" width="100" height="100"></canvas>');
	};

	// createWidget -------------------------------------------------------
	HtmlWriter.createWidget = function(sb, wt) {

		switch(wt.type) {
		case 'area':
		case 'treearea':
			HtmlWriter.createArea(sb, wt);
			break;
		case 'button':
			HtmlWriter.createButton(sb, wt);
			break;
		case 'checkbox-item':
			HtmlWriter.createCheckBoxItem(sb, wt);
			break;
		case 'combobox':
			HtmlWriter.createComboBox(sb, wt);
			break;
		case 'dropdown':
			HtmlWriter.createDropDown(sb, wt);
			break;
		case 'flow':
		case 'date':
		case 'grid':
		case 'hflow':
		case 'checkbox':
		case 'iconlink':
		case 'radiobutton':
		case 'menubar':
		case 'simplegrid':
		case 'splitpane':
		case 'stack':
		case 'treerow':
		case 'treeview':
		case 'vflow':
			HtmlWriter.createChildren(sb, wt);
			break;
		case 'label':
		case 'labelbutton':
		case 'treelabel':
			HtmlWriter.createLabel(sb, wt);
			break;
		case 'link':
			HtmlWriter.createLink(sb, wt);
			break;
		case 'password':
			HtmlWriter.createPassword(sb, wt);
			break;
		case 'picture':
			HtmlWriter.createPicture(sb, wt);
			break;
		case 'pictbutton':
			HtmlWriter.createPictButton(sb, wt);
			break;
		case 'plugin':
			HtmlWriter.createPlugin(sb, wt);
			break;
		case 'plugin-display':
			HtmlWriter.createPluginDisplay(sb, wt);
			break;
		case 'plugin-slide':
			HtmlWriter.createPluginSlide(sb, wt);
			break;
		case 'plugin-toc-edit':
			HtmlWriter.createPluginTocEdit(sb, wt);
			break;
		case 'radiobutton-item':
			HtmlWriter.createRadioButtonItem(sb, wt);
			break;
		case 'scroll':
		case 'table':
			HtmlWriter.createScrollPane(sb, wt);
			break;
		case 'text':
			HtmlWriter.createText(sb, wt);
			break;
		case 'textarea':
			HtmlWriter.createTextArea(sb, wt);
			break;
		case 'textfield':
			HtmlWriter.createTextField(sb, wt);
			break;
		case 'tree':
			HtmlWriter.createTree(sb, wt);
			break;
		case 'treetoggle':
			HtmlWriter.createTreeToggle(sb, wt);
			break;
		default:
			throw new Error('not yet implemented: ' + wt.type);
		}
	};

	/* ================================================================	*
	 * WebcamCfg.js 													*
	 * ================================================================	*/

	var WebcamCfg = function(webcam) {

		// changeDevices -----------------------------------------------------
		WebcamCfg.prototype.changeDevices = function(evt) {
			var el, id, idx;

			el = DocLib.getElement('webcam-amenu');
			if ((idx = el.selectedIndex) === -1)
				id = null;
			else
				id = el.options[el.selectedIndex].value;
			this.webcam.constraints.audio.deviceId = id;

			el = DocLib.getElement('webcam-vmenu');
			if ((idx = el.selectedIndex) === -1)
				id = null;
			else
				id = el.options[el.selectedIndex].value;
			this.webcam.constraints.video.deviceId = id;

			this.restartWebcam();
		};

		// changeFps ---------------------------------------------------------
		WebcamCfg.prototype.changeFps = function(evt) {
			var v;

			v = this.webcam.constraints.video;

			switch(evt.srcElement.id) {
			case 'webcam-1fps':
				v.frameRate = 1;
				break;
			case 'webcam-3fps':
				v.frameRate = 3;
				break;
			case 'webcam-5fps':
				v.frameRate = 5;
				break;
			case 'webcam-12fps':
				v.frameRate = 12;
				break;
			default:
				throw new Error('not yet implemented: ' + evt.srcElement.id);
			}

			this.restartWebcam();
		};

		// changeSize --------------------------------------------------------
		WebcamCfg.prototype.changeSize = function(evt) {
			var v;

			v = this.webcam.constraints.video;

			switch(evt.srcElement.id) {
			case 'webcam-qqvga':
				v.width = 160;
				v.height = 120;
				break;
			case 'webcam-qvga':
				v.width = 320;
				v.height = 240;
				break;
			case 'webcam-vga':
				v.width = 640;
				v.height = 480;
				break;
			}

			this.restartWebcam();
		};

		// init --------------------------------------------------------------
		WebcamCfg.prototype.init = function() {
			var el;

			this.bandwidth = DocLib.getElement('webcam-bandwidth');

			let cfg = this;

			// microphone available...
			el = DocLib.getElement('webcam-amenu');
			el.addEventListener('click', function(evt) {

				cfg.changeDevices(evt);
			});

			// webcam available...
			el = DocLib.getElement('webcam-vmenu');
			el.addEventListener('click', function(evt) {

				cfg.changeDevices(evt);
			});

			// webcam resolutions...
			el = DocLib.getElement('webcam-qqvga');
			el.checked = true;
			
			for(var sz of ['qqvga', 'qvga', 'vga']) {
				el = DocLib.getElement('webcam-' + sz);
				el.addEventListener('click', function(evt) {

					cfg.changeSize(evt);
				});
			}

			// frame per seconds...
			el = DocLib.getElement('webcam-5fps');
			el.checked = true;

			for(var fps of ['1fps', '3fps', '5fps', '12fps']) {
				el = DocLib.getElement('webcam-' + fps);
				el.addEventListener('click', function(evt) {

					cfg.changeFps(evt);
				});
			}
		};

		// initDevices -------------------------------------------------------
		WebcamCfg.prototype.initDevices = function() {

			let cfg = this;

			navigator.mediaDevices.enumerateDevices()
				.then(function(devicesInfo) {				// success...

					cfg.setDevices(devicesInfo);
				}, function(raison) {						// failure...

					XApplication.logWarning("=====> enumerateDevices failure");
					XApplication.logWarning(raison);
				}).catch(function(raison) {					// exception...

					XApplication.logWarning("=====> enumerateDevices exception");
					XApplication.logWarning(raison);
			});
		};

		// restartWebcam -----------------------------------------------------
		WebcamCfg.prototype.restartWebcam = function() {

			this.running = false;

			this.bandwidth.innerHTML = "??? KBs";

			this.setInputDisabled(true);
			this.webcam.stop();
			this.webcam.start('config');
		};

		// recorderStart -----------------------------------------------------
		WebcamCfg.prototype.recorderStart = function(stream) {
			var isz, options, recorder, sizes, start;

			options = {
	//			mimeType: 'video/webm'
				mimeType: 'video/webm;codecs=vp9'
			};

	/*
	var types = ["video/webm", 
	             "audio/webm", 
	             "video/webm\;codecs=vp8", 
	             "video/webm\;codecs=vp9", 
	             "video/webm\;codecs=daala", 
	             "video/webm\;codecs=h264", 
	             "audio/webm\;codecs=opus", 
	             "video/mpeg"];

	for (var i in types) { 
	  console.log( "Is " + types[i] + " supported? " + (MediaRecorder.isTypeSupported(types[i]) ? "Maybe!" : "Nope :(")); 
	}
	*/
			recorder = new MediaRecorder(stream, options);
			this.webcam.recorder = recorder;
			
			isz = 0;
			sizes = Array(10);
			start = new Date().getTime();

			this.running = true;

			let cfg = this;

			recorder.ondataavailable = function(evt) {
				var i, n, v, len;

				if (!cfg.running)
					return;
				
	//			length += evt.data.size;
	//			t = new Date().getTime() - start;
	//			v = length / t * 1000 / 1024;

				//console.log(evt.data.size);
				sizes[isz] = evt.data.size;
				if (++isz === sizes.length)
					isz = 0;

				n = len = 0;
				for (i = 0; i < sizes.length; ++i) {
					if (sizes[i] === undefined)
						continue;

					len += sizes[i];
					++n;
				}

				if (n === 0) {
					cfg.bandwidth.innerHTML = '??? KBs';
					return;
				}
				
				v = len / n / 1024;
				cfg.bandwidth.innerHTML = v.toFixed(2) + ' KBs';
			};

			recorder.onerror = function(evt) {

				XApplication.logWarning("recorder error: " + evt.error.name);
			};
			
			recorder.start(1000);
		};

		// setDevices --------------------------------------------------------
		WebcamCfg.prototype.setDevices = function(devicesInfo) {
			var el, amenu, vmenu, alabel, vlabel;

			alabel = vlabel = null;

			if (this.webcam.audioTrack !== null)
				alabel = this.webcam.audioTrack.label;

			if (this.webcam.videoTrack !== null)
				vlabel = this.webcam.videoTrack.label;

			amenu = DocLib.getElement('webcam-amenu');
			XUtil.removeAllChildren(amenu);
			vmenu = DocLib.getElement('webcam-vmenu');
			XUtil.removeAllChildren(vmenu);

			for(var info of devicesInfo) {
				switch(info.kind) {
				case 'audioinput':
					el = document.createElement('option');
					el.setAttribute('style', 'padding: 5px');
					el.text = info.label;
					el.value = info.deviceId;
					amenu.appendChild(el);

					if (alabel !== null && alabel === el.text)
						amenu.value = el.value;
					break;
				case 'videoinput':
					el = document.createElement('option');
					el.setAttribute('style', 'padding: 5px');
					el.text = info.label;
					el.value = info.deviceId;
					vmenu.appendChild(el);

					if (vlabel !== null && vlabel === el.text)
						vmenu.value = el.value;
					break;
				}
			}

			this.reset = false;
		};

		// setInputDisabled --------------------------------------------------
		WebcamCfg.prototype.setInputDisabled = function(disabled) {
			var el;

			for(var id of this.inputs) {
				el = DocLib.getElement('webcam-' + id);
				el.disabled = disabled;
			};
		};

		// start -------------------------------------------------------------
		WebcamCfg.prototype.start = function() {

			if (this.bandwidth === null)
				this.init();

			this.reset = true;

			this.setInputDisabled(true);
			this.bandwidth.innerHTML = "??? KBs";
		};

		// videoStarted ------------------------------------------------------
		WebcamCfg.prototype.videoStarted = function(stream) {

			this.setInputDisabled(false);
			
			this.webcam.visualizer.start('webcam-audio', stream);

			this.recorderStart(stream);

			if (!this.reset)
				return;

			this.initDevices();
		};

		// Constructor -------------------------------------------------------
		{
			this.webcam = webcam;
			this.reset = false;
			this.running = false;

			this.bandwidth = null;

			this.inputs = [
				'amenu', 'vmenu', 'qqvga', 'qvga', '1fps', '3fps', '5fps', '12fps'
			];
		}
	};

	/* ================================================================	*
	 * AudioVisualizer.js 												*
	 * ================================================================	*/

	var AudioVisualizer = function() {

		// start -------------------------------------------------------------
		AudioVisualizer.prototype.start = function(id, stream) {
			var actxt, analyser, data, source;
			var ctxt, st, width, height;

			this.canvas = DocLib.getElement(id);
			this.running = true;

			actxt = new window.AudioContext();
			analyser = actxt.createAnalyser();
			data = new Uint8Array(analyser.frequencyBinCount);

			source = actxt.createMediaStreamSource(stream);
			source.connect(analyser);

			st = window.getComputedStyle(this.canvas);
			width = Math.ceil(parseFloat(st.getPropertyValue('width')));
			height = Math.ceil(parseFloat(st.getPropertyValue('height')));

			ctxt = this.canvas.getContext("2d");
			ctxt.canvas.width = width;
			ctxt.canvas.height = height;

			requestAnimationFrame(draw);

			let av = this;

			function draw() {
				var d, v, x, y, w, h, off;

				if (!av.running)
					return;

				w = ctxt.canvas.width;
				h = ctxt.canvas.height;
				ctxt.fillStyle = '#282828';
				ctxt.fillRect(0, 0, w, h);

				analyser.getByteTimeDomainData(data);

				ctxt.lineWidth = 1;
				ctxt.strokeStyle = 'white';

				ctxt.beginPath();

				off = 0;
				d = 1000 / w;

				for(x = 0; x < w; ++x) {
					off += d;
					v = data[Math.round(off)] / 128.0;
					y = v * h / 2;

					if (x === 0)
						ctxt.moveTo(x, y);
					else
						ctxt.lineTo(x, y);
				}

				ctxt.lineTo(w, h / 2);
				ctxt.stroke();

				requestAnimationFrame(draw);
			}
		};

		// stop --------------------------------------------------------------
		AudioVisualizer.prototype.stop = function() {
			var w, h, ctxt;

			this.running = false;

			if (this.canvas === null)
				return;
					
			ctxt = this.canvas.getContext("2d");

			w = ctxt.canvas.width;
			h = ctxt.canvas.height;
			ctxt.fillStyle = '#282828';
			ctxt.fillRect(0, 0, w, h);
		};

		// Constructor -------------------------------------------------------
		{
			this.running = false;
			this.canvas = null;
		}
	};

	/* ================================================================	*
	 * TransmissionHandler.js 											*
	 * ================================================================	*/

	var TransmissionHandler = function(xappl) {

		// start -------------------------------------------------------------
		TransmissionHandler.prototype.start = function(stream) {
			var options, url;

			options = {
	//			mimeType: 'video/webm'
				mimeType: 'video/webm;codecs=vp9'
			};

			this.recorder = new MediaRecorder(stream, options);

			if ((this.hash = this.xappl.getVar('#lecture-info/session-hash')) === '')
				throw new Error("can't get session-hash");
		
			url = xappl.getEndpoint() + '-transmission';
			XWebSocket.open('transmission', url, this);
		};

		// stop --------------------------------------------------------------
		TransmissionHandler.prototype.stop = function() {

			if (this.recorder !== null) {
				this.recorder.stop();
				this.recorder = null;
			}

			if (this.ws !== null) {
				this.ws.close();
				this.ws = null;
			}
		};

		// startRecorder -----------------------------------------------------
		TransmissionHandler.prototype.startRecorder = function() {

			let vh = this;

			this.recorder.ondataavailable = function(evt) {

				if (vh.ws === null)
					return;

				//console.log(evt.data.size);
				vh.ws.sendData(evt.data);
			};

			this.recorder.onerror = function(evt) {
			};

			this.recorder.start(1000);
		};

		// wsClose -----------------------------------------------------------
		TransmissionHandler.prototype.wsClose = function(ws) {

			XApplication.logMsg('TransmissionHandler: wsClose');
		};

		// wsErrorOnClose ----------------------------------------------------
		TransmissionHandler.prototype.wsErrorOnClose = function(ws, code, reason) {

			XApplication.logMsg('TransmissionHandler: wsErrorOnClose [' + reason + ']');

			this.xappl.putVar('messages-abort-msg',  '[#' + code + '] ' + reason);
			this.xappl.taskExecTask('messages-abort');
		};

		// wsOpen ------------------------------------------------------------
		TransmissionHandler.prototype.wsOpen = function(ws) {
			var args, data, encoder;

			this.ws = ws;

			args = 'session-hash=' + this.hash;

			encoder = new TextEncoder();
			data = encoder.encode(args);

			ws.sendData(data);

			this.startRecorder();
		};

		// Constructor -------------------------------------------------------
		{
			this.xappl = xappl;

			this.ws = null;
			this.recorder = null;
		}
	};

	/* ================================================================	*
	 * RecordingHandler.js 												*
	 * ================================================================	*/

	var RecordingHandler = function(xappl, webcam, messages) {

		// getCurrentTime ----------------------------------------------------
		RecordingHandler.prototype.getCurrentTime = function() {

			if (this.startingTime === -1)
				return -1;

			return (Date.now() - this.startingTime) / 1000;
		};

		// getStreamId -------------------------------------------------------
		RecordingHandler.prototype.getStreamId = function() {

			return this.streamId;
		};

		// start -------------------------------------------------------------
		RecordingHandler.prototype.start = function(stream) {
			var options, url;

			options = {
	//			mimeType: 'video/webm'
				mimeType: 'video/webm;codecs=vp9'
			};

			this.recorder = new MediaRecorder(stream, options);

			if ((this.hash = this.xappl.getVar('#lecture-info/session-hash')) === '')
				throw new Error("can't get session-hash");
		
			url = xappl.getEndpoint() + '-recording';
			XWebSocket.open('recording', url, this);
		};

		// stop --------------------------------------------------------------
		RecordingHandler.prototype.stop = function() {

			if (this.recorder !== null) {
				if (this.recorder.state !== 'inactive')
					this.recorder.stop();

				this.recorder = null;
			}

			if (this.ws !== null) {
				this.ws.close();
				this.ws = null;
			}

			this.display.stopRecording();
		};

		// startRecorder -----------------------------------------------------
		RecordingHandler.prototype.startRecorder = function() {

			let vh = this;

			this.recorder.ondataavailable = function(evt) {
				var i, amounts, nbytes, status;

				if (vh.ws === null)
					return;

				nbytes = vh.ws.getBufferedAmount();

				vh.ws.sendData(evt.data);

				amounts = vh.amounts;
				
				status = 'green';

				if (nbytes > 0) {
					for(i = 0; i < amounts.length; ++i) {
						if ((nbytes -= amounts[i]) < 0)
							break;

						if (i === 5)
							status = 'orange';
					}

					if (i === amounts.length)
						status = 'red';
				}

				if (status !== vh.networkStatus) {
					vh.networkStatus = status;
					vh.xappl.taskExecTask('recording-network-' + status);
					messages.send('network-status-' + status);
				}

				amounts.splice(0, 0, evt.data.size);

				if (amounts.length === vh.amountsMax)
					amounts.pop();
			};

			this.recorder.onerror = function(evt) {
			};

			this.startingTime = Date.now();
			this.recorder.start(1000);
			
			this.networkStatus = 'green';
			this.xappl.taskExecTask('recording-network-green');

			if (this.display === null)
				if ((this.display = MW.WLib.getWidget('transmitter', 'transmitter-display')) === null)
					throw new Error("widget not found: transmitter-display");

			this.display.startRecording();
		};

		// wsClose -----------------------------------------------------------
		RecordingHandler.prototype.wsClose = function(ws) {

			if (this.ws !== ws)
				return;

			XApplication.logMsg('RecordingHandler: wsClose');
		};

		// wsErrorOnClose ----------------------------------------------------
		RecordingHandler.prototype.wsErrorOnClose = function(ws, code, reason) {

			if (this.ws !== ws)
				return;

			XApplication.logMsg('RecordingHandler: wsErrorOnClose [' + reason + ']');

			this.xappl.putVar('messages-abort-msg',  '[#' + code + '] ' + reason);
			this.xappl.taskExecTask('messages-abort');
		};

		// wsMessage ---------------------------------------------------------
		RecordingHandler.prototype.wsMessage = function(ws, doc) {

			if (this.ws !== ws)
				return;

			this.streamId = doc.documentElement.getAttribute('streamId');
			this.webcam.websocketConnected();

			this.startRecorder();
			this.xappl.taskExecTask('recording-connected');
		};

		// wsOpen ------------------------------------------------------------
		RecordingHandler.prototype.wsOpen = function(ws) {
			var args, data, encoder, slide;

			this.ws = ws;

			slide = this.xappl.uiappl.tplugin.slidename;
			args = 'session-hash=' + this.hash + ',slide=' + slide;
			args += ',resolution=' + this.webcam.getResolution();
			args += ',frame-rate=' + this.webcam.getFrameRate();

			encoder = new TextEncoder();
			data = encoder.encode(args);

			ws.sendData(data);

			this.xappl.taskExecTask('recording-websocket-open');
	//		this.startRecorder();

			this.amounts = new Array();
		};

		// Constructor -------------------------------------------------------
		{
			this.xappl = xappl;
			this.webcam = webcam;
			this.messages = messages;

			this.ws = null;
			this.recorder = null;
			this.display = null;

			this.startingTime = -1;
			this.streamId = null;
			this.amounts = null;
			this.amountsMax = 20;
			this.networkStatus = null;
		}
	};

	/* ================================================================	*
	 * WebcamCheckHandler.js 											*
	 * ================================================================	*/

	var WebcamCheckHandler = function(xappl, visualizer) {

		// play --------------------------------------------------------------
		WebcamCheckHandler.prototype.play = function() {

			if (this.video === null) {
				this.video = DocLib.getElement('webcam-check-video-player');

				let wh = this;

				this.video.addEventListener('ended', function() {

					if (wh.videoStop)
						return;

					wh.xappl.taskExecTask('webcam-check-stop');
				});
			}

			this.video.src = this.videoURL;
			this.video.muted = false;
			
			this.videoStop = false;
		};

		// start -------------------------------------------------------------
		WebcamCheckHandler.prototype.start = function(stream) {
			var options;

			options = {
				mimeType: 'video/webm;codecs=vp9'
			};

			this.recorder = new MediaRecorder(stream, options);
			this.chunks = [];

			let wh = this;

			this.recorder.ondataavailable = function(evt) {

				wh.chunks.push(evt.data);
			};

			this.recorder.onstop = function(evt) {
				var blob;

				blob = new Blob(wh.chunks, { 'type' : 'video/webm;codecs=vp9'});
				wh.videoURL = URL.createObjectURL(blob);
			};
			
			this.recorder.start(1000);
			this.visualizer.start('webcam-check-audio-canvas', stream);

			this.timerId = setTimeout(evt => {

				if (wh.timerId === -1)
					return;

				wh.timerId = -1;
				wh.recorder.stop();
				wh.xappl.taskExecTask('webcam-check-stop');
			}, 10000);
		};

		// stop --------------------------------------------------------------
		WebcamCheckHandler.prototype.stop = function() {

			if (this.video !== null && this.video.src !== null) {
				this.video.src = '';
				this.video.muted = true;
				this.videoStop = true;
			}

			if (this.timerId !== -1) {
				clearTimeout(this.timerId);
				this.timerId = -1;
			}
		};

		// Constructor -------------------------------------------------------
		{
			this.xappl = xappl;
			this.visualizer = visualizer;

			this.recorder = null;
			this.chunks = null;
			this.video = null;
			this.videoURL = null;
			this.videoStop = true;
			this.timerId = -1;
		}
	};

	/* ================================================================	*
	 * Webcam.js 														*
	 * ================================================================	*/

	var Webcam = function(xappl, messages) {

		// getCurrentTime ----------------------------------------------------	
		Webcam.prototype.getCurrentTime = function() {

	//		return this.media.currentTime;
			return this.rhandler.getCurrentTime();
		};
		
		// getFrameRate ------------------------------------------------------	
		Webcam.prototype.getFrameRate = function() {
			var settings;

			if (this.videoTrack === null)
				return '0';

			if ((settings = this.videoTrack.getSettings()) === null)
				return '0';

			return settings.frameRate;
		};

		// getResolution -----------------------------------------------------	
		Webcam.prototype.getResolution = function() {
			var settings;

			if (this.videoTrack === null)
				return '0x0';

			if ((settings = this.videoTrack.getSettings()) === null)
				return '0';
		
			return settings.width + 'x' + settings.height;
		};

		// getStreamId -------------------------------------------------------	
		Webcam.prototype.getStreamId = function() {

			return this.rhandler.getStreamId();
		};

		// goError -----------------------------------------------------------	
		Webcam.prototype.goError = function(msg) {

			this.xappl.putVar('webcam-status', 'error');
			this.stop();

			this.xappl.putVar('webcam-error', msg);
			this.xappl.taskExecTask('webcam-error');
		};

		// start -------------------------------------------------------------
		Webcam.prototype.start = function(mode) {
			var v, w, h, constraints, status;

			status = this.xappl.getVar('webcam-status');

			if (mode === 'recording') {
				if (status !== 'running')
					throw new Error('invalid status: ' + status);

				if (this.videoTrack !== null) {
					this.videoTrack.stop();
					this.videoTrack = null;
				}

				status = 'idle';
				this.xappl.putVar('webcam-status', 'idle');
				this.media.srcObject = null;
				this.media = null;
			}

			if (status !== 'idle')
				throw new Error('invalid status: ' + status);

			this.mode = mode;

			switch(mode) {
			case 'check':
			case 'config':
				if (mode === 'config') {
					this.cfg.start();
					this.xappl.runTask('webcam-disabled-button');

					this.media = DocLib.getElement('webcam-video');
				} else {
					this.xappl.runTask('webcam-check-disabled-button');

					this.media = DocLib.getElement('webcam-check-video-player');
				}

				this.media.muted = true;
				this.media.style.position = 'relative';

				constraints = this.constraints;

				v = constraints.video;

				if ((w = v.width) > 400)
					w = 400;
				this.media.style.width = w + 'px';

				if ((h = v.height) > 300)
					h = 300;
				this.media.style.height = h + 'px';

				this.media.style.left = ((400 - w) / 2) + 'px';
				this.media.style.top = ((300 - h) / 2) + 'px';
				break;
			case 'recording':
				this.media = DocLib.getElement('transmitter-video');
				this.media.muted = true;

				constraints = this.constraints;
				break;
			case 'transmission':
				this.media = DocLib.getElement('transmitter-video');

				constraints = {
					video: this.constraints.video
				};
				break;
			default:
				throw new Error('invalid mode: ' + this.mode);
			}

			this.xappl.putVar('webcam-status', 'starting');

			let wm = this;
			let error = false;

			//console.log(constraints.video.width + 'x' + constraints.video.height);
			navigator.mediaDevices.getUserMedia(constraints)
				.then(function(stream) {							// success...
					var settings, status, tracks;

					try {
						status = wm.xappl.getVar('webcam-status');

						tracks = stream.getAudioTracks();
						if (tracks.length > 0)
							wm.audioTrack = tracks[0];

						tracks = stream.getVideoTracks();
						if (tracks.length > 0)
							wm.videoTrack = tracks[0];

						if (wm.mode !== 'transmission' && wm.audioTrack === null) {
							wm.videoTrack.stop();
							wm.videoTrack = null;

							wm.stop();
							wm.xappl.putVar('webcam-error', 'no-audio');
							wm.xappl.taskExecTask('webcam-error');
							return;
						}

						if (wm.videoTrack === null) {
							wm.audioTrack.stop();
							wm.audioTrack = null;

							wm.stop();
							wm.xappl.putVar('webcam-error', 'no-video');
							wm.xappl.taskExecTask('webcam-error');
							return;
						}
					
						if (status === 'abort') {
							wm.stop();
							return;
						}

						wm.media.srcObject = stream;
						wm.xappl.putVar('webcam-status', 'running');

						settings = wm.videoTrack.getSettings();
						XApplication.logWarning('Video settings: ' + settings.width + 'x' + settings.height);
						XApplication.logWarning('                frameRate: ' + settings.frameRate);

						switch(wm.mode) {
						case 'check':
							wm.chandler.start(stream);
							wm.xappl.taskExecTask('webcam-check-start-ok');
							return;
						case 'config':
							wm.cfg.videoStarted(stream);
							wm.xappl.taskExecTask('webcam-enabled-button');
							return;
						case 'recording':
	//						wm.visualizer.start('transmitter-audio', stream);
							wm.stream = stream;
							wm.rhandler.start(stream);
	//						wm.xappl.taskExecTask('recording-start-ok');
							return;
						case 'transmission':
							wm.thandler.start(stream);
							wm.xappl.taskExecTask('transmission-start-ok');
							return;
						}
					}
					catch(err) {
						HttpError.send(err);
					}
		
					throw new Error('internal error: ' + wm.mode);
				}, function(raison) {								// failure...

					XApplication.logWarning('=====> getUserMedia failure');
					XApplication.logWarning(raison.name);

					if (!error) {
						error = true;
						wm.goError(raison.name);
					}
				}).catch(function(raison) {							// error...

					XApplication.logWarning('=====> getUserMedia exception');
					XApplication.logWarning(raison);

					if (!error) {
						error = true;
						wm.goError(raison.message);
					}
				});
		};

		// stop --------------------------------------------------------------
		Webcam.prototype.stop = function() {
			var status;

			this.stream = null;

			status = this.xappl.getVar('webcam-status');

			switch(status) {
			case 'idle':
				return;
			case 'starting':
				this.xappl.putVar('webcam-status', 'abort');
				return;
			case 'running':
			case 'abort':
			case 'error':
				break;
			default:
				throw new Error('invalid status: ' + status);
			}

			switch(this.mode) {
			case 'check':
			case 'config':
			case 'recording':
				if (this.mode === 'recording')
					this.rhandler.stop();

				this.visualizer.stop();

				if (this.audioTrack !== null) {
					this.audioTrack.stop();
					this.audioTrack = null;
				}

				if (this.videoTrack !== null) {
					this.videoTrack.stop();
					this.videoTrack = null;
				}
				break;
			case 'transmission':
				this.thandler.stop();

				if (this.videoTrack !== null) {
					this.videoTrack.stop();
					this.videoTrack = null;
				}
				break;
			default:
				throw new Error('not yet implemented: ' + this.mode);
			}

			this.mode = 'idle';
			this.xappl.putVar('webcam-status', 'idle');
			this.media.srcObject = null;
			this.media = null;
		};

		// websocketConnected ------------------------------------------------
		Webcam.prototype.websocketConnected = function() {

			this.visualizer.start('transmitter-audio', this.stream);
		};

		// Constructor -------------------------------------------------------
		{
			this.xappl = xappl;

			this.mode = 'idle';

			this.constraints = {
				audio: {
					deviceId: null,
					sampleSize: 16,
					channelCount: 1
				},
				video: {
					deviceId: null,
					width: 160,
					height: 120,
	//				width: 640,
	//				height: 480,
					frameRate: 5
				}
			};

	//		console.log('----------------------------------- Warning ----------------------------------');
			
			this.media = null;
			this.visualizer = null;
			this.stream = null;
			this.audioTrack = null;
			this.videoTrack = null;

			this.cfg = new WebcamCfg(this);
			this.visualizer = new AudioVisualizer();
			this.chandler = new WebcamCheckHandler(xappl, this.visualizer);
			this.thandler = new TransmissionHandler(xappl);
			this.rhandler = new RecordingHandler(xappl, this, messages);
		}
	};

	/* ================================================================	*
	 * MessagesHandler.js 												*
	 * ================================================================	*/

	var MessagesHandler = function(appl, xappl, prefix, usession) {

		// add ---------------------------------------------------------------
		MessagesHandler.prototype.add = function(doc) {
			var el, entry, root;

			el = new XmlData(doc, doc.documentElement);
			entry = el.getFirstChildWithTag('entry');

			root = new XmlData(this.msgs, this.msgs.documentElement);
			el = root.getFirstChild();
			el.importAsFirstChild(entry);

			this.update();
		};

		// hidden ------------------------------------------------------------
		MessagesHandler.prototype.hidden = function() {
			var el, entries, msgid, off, xd;

			msgid = parseInt(this.xappl.getVar('action-data'));
			
			xd = new XmlData(this.msgs, this.msgs.documentElement);
			entries = xd.searchElements('/all-messages/messages/entry');

			for(var entry of entries) {
				if (entry.getInt('id') !== msgid)
					continue;

				entry.setStringAttr('hidden', 'true');

				el = this.scroll.uiwt.element;
				off = el.scrollTop;
				this.update();
				el.scrollTop = off;
				return;
			}
		};

		// init --------------------------------------------------------------
		MessagesHandler.prototype.init = function(msgs) {

			this.msgs = msgs;
		
			this.update();
		};

		// refresh -----------------------------------------------------------
		MessagesHandler.prototype.refresh = function() {
			var xd, entries;

			xd = new XmlData(this.msgs, this.msgs.documentElement);
			entries = xd.searchElements('/all-messages/messages/entry');

			for(var el of entries)
				el.removeAttr('hidden');

			this.update();
		};

		// send --------------------------------------------------------------
		MessagesHandler.prototype.send = function(msg) {

			this.ws.sendMsg(msg);
		};

		// start -------------------------------------------------------------
		MessagesHandler.prototype.start = function() {
			var url;

			if ((this.hash = this.xappl.getVar('#lecture-info/session-hash')) === '')
				throw new Error("can't get session-hash");
		
			url = xappl.getEndpoint() + '-events';
			XWebSocket.open('events', url, this);
		};

		// stop --------------------------------------------------------------
		MessagesHandler.prototype.stop = function() {

			if (this.ws === null)
				return;

			this.ws.close();
			this.ws = null;
		};

		// update ------------------------------------------------------------
		MessagesHandler.prototype.update = function() {
			var el, label, root, view, xd;

			root = this.msgs.documentElement;
			el = root.firstElementChild;

			label = xappl.getVar('xmessages-list-label');
			xd = XMessagesList.getNewView(this.prefix, this.scroll, new XmlData(this.msgs, el), label);
			view = XLib.createWidget(this.appl.getXWData(xd));
			MW.WLib.updateScrollPaneView(this.scroll, view, true);
		};

		// wsClose -----------------------------------------------------------
		MessagesHandler.prototype.wsClose = function(ws) {

			XApplication.logMsg('MessagesHandler: close');
		};

		// wsErrorOnClose ----------------------------------------------------
		MessagesHandler.prototype.wsErrorOnClose = function(ws, code, reason) {

			XApplication.logMsg('MessagesHandler: wsErrorOnClose');

			if (reason === 'new-speaker') {
				this.xappl.taskExecTask('messages-new-speaker');
				return;
			}

			this.xappl.putVar('messages-abort-msg',  '[#' + code + '] ' + reason);
			this.xappl.taskExecTask('messages-abort');
		};

		// wsMessage ---------------------------------------------------------
		MessagesHandler.prototype.wsMessage = function(ws, doc) {
			var root;

			root = doc.documentElement;

			switch(root.tagName) {
			case 'all-messages':
				this.init(doc);
				break;
			case 'message-empty':
				this.xappl.taskExecTask(this.prefix + '-new-message-empty');
				break;
			case 'new-message':
				this.add(doc);
				break;
			case 'lecture-broadcast':
				this.xappl.taskExecTask('lecture-broadcast');
				break;
			case 'lecture-init':
				this.usession.eventInit('receiver', new XmlData(doc, doc.documentElement));
				break;
			case 'lecture-seek':
				this.usession.eventSeek(doc);
				break;
			case 'lecture-status':
				this.usession.eventStatus(doc);
				break;
			case 'lecture-stream':
				this.usession.eventStream(doc);
				break;
			case 'reload-slides':
				this.xappl.taskExecTask('reload-slides');
				break;
			default:
				throw new Error('not yet implemented: ' + root.tagName);
			}
		};

		// wsOpen ------------------------------------------------------------
		MessagesHandler.prototype.wsOpen = function(ws) {

			this.ws = ws;
			this.scroll = MW.WLib.getWidget(this.prefix, this.prefix + '-msgs');

			ws.sendMsg('session-hash=' + this.hash);
		};

		// Constructor -------------------------------------------------------
		{
			this.appl = appl;
			this.xappl = xappl;
			this.prefix = prefix;
			this.usession = usession;

			this.ws = null;
		}
	};

	/* ================================================================	*
	 * BrowserPlugin.js 												*
	 * ================================================================	*/

	var BrowserPlugin = function(xappl) {

		// exec --------------------------------------------------------------
		BrowserPlugin.prototype.exec = function(ref) {

			switch(ref) {
			case 'reload-english':
				this.reload('en');
				break;
			case 'reload-french':
				this.reload('fr');
				break;
			case 'reload-portuguese':
				this.reload('pt');
				break;
			case 'reload-russian':
				this.reload('ru');
				break;
			case 'website-admin':
				this.websiteAdmin();
				break;
			case 'website-java':
				this.websiteJava();
				break;
			case 'website-users':
				this.websiteUsers();
				break;
			default:
				throw new Error('not yet implemented[BrowserPlugin]: ' + ref);
			}
		};

		// reload ------------------------------------------------------------
		BrowserPlugin.prototype.reload = function(lang) {
			var n, s, path, url;

			s = window.location.pathname;
			if ((n = s.lastIndexOf('/')) === -1)
				throw new Error('invalid pathname: ' + s);
			path = s.substring(0, n + 1);

			url = window.location.origin + path;
			
			if (lang !== 'fr')
				url += 'index-' + lang + '.html';

			window.location.href = url;
		};

		// websiteAdmin ------------------------------------------------------
		BrowserPlugin.prototype.websiteAdmin = function() {
			var n1, n2, s, filename, lang, path, url;

			s = window.location.pathname;
			if ((n1 = s.lastIndexOf('/')) === -1)
				throw new Error('invalid pathname: ' + s);
			path = s.substring(0, n1 + 1);

			url = window.location.origin + path;

			filename = s.substring(n1);
			if (filename === '/')
				lang = 'fr';
			else {
				if ((n1 = s.lastIndexOf('-')) === -1)
					throw new Error('invalid pathname: ' + s);

				if ((n2 = s.lastIndexOf('.')) === -1)
					throw new Error('invalid pathname: ' + s);

				lang = s.substring(n1 + 1, n2);
			}

			url += 'admin-' + lang + '.html';

			window.location.href = url;
		};

		// websiteJava -------------------------------------------------------
		BrowserPlugin.prototype.websiteJava = function() {

			window.location.href = 'http://raft1.unige.ch/dudal/';
		};

		// websiteUsers ------------------------------------------------------
		BrowserPlugin.prototype.websiteUsers = function() {
			var s, n1, n2, filename, lang, no, path, url;

			no = this.xappl.getVar('action-data');

			s = window.location.pathname;
			if ((n1 = s.lastIndexOf('/')) === -1)
				throw new Error('invalid pathname: ' + s);
			path = s.substring(0, n1 + 1);

			url = window.location.origin + path;

			filename = s.substring(n1);
			if (filename === '/')
				lang = 'fr';
			else {
				if ((n1 = s.lastIndexOf('-')) === -1)
					throw new Error('invalid pathname: ' + s);

				if ((n2 = s.lastIndexOf('.')) === -1)
					throw new Error('invalid pathname: ' + s);

				lang = s.substring(n1 + 1, n2);
			}

			url += 'course?' + no;
			
			if (lang !== 'fr')
				url += '&lang=' + lang;

			window.location.href = url;
		};

		// Constructor -------------------------------------------------------
		{
			this.xappl = xappl;
		}
	};

	/* ================================================================	*
	 * MigrationViewerPlugin.js 										*
	 * ================================================================	*/

	var MigrationViewerPlugin = function(appl, xappl) {

		// exec --------------------------------------------------------------
		MigrationViewerPlugin.prototype.exec = function(ref) {

			switch(ref) {
			case 'messages-hidden':
				this.hidden();
				break;
			case 'messages-refresh':
				this.refresh();
				break;
			case 'start':
				this.start();
				break;
			case 'wordpress-link':
				this.showWordPressLink();
				break;
			default:
				throw new Error('not yet implemented[MigrationViewerPlugin]: ' + ref);
			}
		};

		// hidden ------------------------------------------------------------
		MigrationViewerPlugin.prototype.hidden = function() {
			var el, entries, msgid, off;

			msgid = parseInt(this.xappl.getVar('action-data'));

			entries = this.msgs.searchElements('entry');

			for(var entry of entries) {
				if (entry.getInt('id') !== msgid)
					continue;

				entry.setStringAttr('hidden', 'true');

				// build messages...
				el = this.scroll.uiwt.element;
				off = el.scrollTop;
				this.update();
				el.scrollTop = off;
				return;
			}
		};

		// refresh -----------------------------------------------------------
		MigrationViewerPlugin.prototype.refresh = function() {
			var entries;

			entries = this.msgs.searchElements('entry');

			for(var el of entries)
				el.removeAttr('hidden');

			this.update();
		};

		// showWordPressLink -------------------------------------------------
		MigrationViewerPlugin.prototype.showWordPressLink = function() {
			var url;

			url = this.xappl.parseText('${#migration-view/url}');
			window.open(url, 'dudal-wordpress');
			
		};

		// start -------------------------------------------------------------
		MigrationViewerPlugin.prototype.start = function() {
			var data, xd;

			this.label = xappl.getVar('xmessages-list-label');
			this.scroll = MW.WLib.getWidget('migration-view', 'receiver-msgs');
			
			// get messages...
			data = this.xappl.getData('migration-view');
			this.msgs = data.getFirstChildWithTag('messages');

			// update messages...
			this.update();

			/*
			 * init video events..
			 */
			xd = data.getFirstChildWithTag('lecture-init');
			this.appl.rplugin.usession.eventInit('migration', xd);
		};
		
		// update ------------------------------------------------------------
		MigrationViewerPlugin.prototype.update = function() {
			var view, wt;

			view = XMessagesList.getNewView('migration-view', this.scroll, this.msgs, this.label);
			wt = XLib.createWidget(this.appl.getXWData(view));
			MW.WLib.updateScrollPaneView(this.scroll, wt, true);
		};

		// Constructor -------------------------------------------------------
		{
			this.appl = appl;
			this.xappl = xappl;
			this.label = null;
			this.scroll = null;
			this.msgs = null;
		}
	};

	/* ================================================================	*
	 * OfflineViewerPlugin.js 											*
	 * ================================================================	*/

	var OfflineViewerPlugin = function(appl, xappl) {

		// download-course ---------------------------------------------------
		OfflineViewerPlugin.prototype.downloadCourse = function() {
			var xhttp;

			xhttp = new XMLHttpRequest();

			let me = this;

			xhttp.onreadystatechange = function() {
				var a, id, name, url;

				if (xhttp.readyState !== 4)				// 4 DONE ...
					return;

				if (xhttp.responseText !== 'check-ok') {
					me.xappl.taskExecTask('#http-timeout');
					return;
				}

				id = me.xappl.getVar('#lecture-info/lectureId');

				name = 'dudal-' + XInteger.toPadding(id, 5) + '.zip';
				url = 'zip/' + me.hashSession + '/' + name;

				a = document.createElement('a');
				document.body.appendChild(a);
				a.style = "display: none";
				a.href = url;
				a.download = name;
				a.click();
				document.body.removeChild(a);
	//			window.location.assign(url);
			};

			xhttp.open('GET', 'zip/' + this.hashSession);
			xhttp.send();

		};

		// exec --------------------------------------------------------------
		OfflineViewerPlugin.prototype.exec = function(ref, nextTask) {

			switch(ref) {
			case 'cache-offline':
				MW.HttpFiles.cacheOffline(ref, nextTask);
				break;
			case 'download-course':
				this.downloadCourse();
				break;
			case 'messages-hidden':
				this.hidden();
				break;
			case 'messages-refresh':
				this.refresh();
				break;
			case 'start':
				this.start();
				break;
			case 'video-pause':
				this.appl.rplugin.usession.eventAction('receiver-pause');
				break;
			case 'video-play':
				this.appl.rplugin.usession.eventAction('receiver-play');
				break;
			default:
				throw new Error('not yet implemented[OfflineViewerPlugin]: ' + ref);
			}
		};

		// hidden ------------------------------------------------------------
		OfflineViewerPlugin.prototype.hidden = function() {
			var el, entries, msgid, off;

			msgid = parseInt(this.xappl.getVar('action-data'));

			entries = this.msgs.getChildren('entry');

			for(var entry of entries) {
				if (entry.getInt('id') !== msgid)
					continue;

				entry.setStringAttr('hidden', 'true');

				// build messages...
				el = this.scroll.uiwt.element;
				off = el.scrollTop;
				this.update();
				el.scrollTop = off;
				return;
			}
		};

		// refresh -----------------------------------------------------------
		OfflineViewerPlugin.prototype.refresh = function() {
			var entries;

			entries = this.msgs.getChildren('entry');

			for(var entry of entries)
				entry.removeAttr('hidden');

			this.update();
		};

		// setHash -----------------------------------------------------------
		OfflineViewerPlugin.prototype.setHash = function(hash) {

			this.hashSession = hash;
		};

		// start -------------------------------------------------------------
		OfflineViewerPlugin.prototype.start = function() {
			var data, xd;

			this.label = xappl.getVar('xmessages-list-label');
			this.scroll = MW.WLib.getWidget('offline', 'offline-msgs');
			
			// get messages...
			data = this.xappl.getData('lecture-info');
			this.msgs = data.getFirstChildWithTag('messages');

			// update messages...
			this.update();

			/*
			 * init video events..
			 */
			xd = data.getFirstChildWithTag('lecture-init');

			this.appl.rplugin.usession.eventInit('offline', xd);
		};

		// update ------------------------------------------------------------
		OfflineViewerPlugin.prototype.update = function() {
			var view, wt;

			view = XMessagesList.getNewView('offline', this.scroll, this.msgs, this.label);
			wt = XLib.createWidget(this.appl.getXWData(view));
			MW.WLib.updateScrollPaneView(this.scroll, wt, true);
		};

		// Constructor -------------------------------------------------------
		{
			this.appl = appl;
			this.xappl = xappl;
			this.hashSession = null;
			this.label = null;
			this.scroll = null;
			this.msgs = null;
		}
	};

	/* ================================================================	*
	 * ReceiverPlugin.js 												*
	 * ================================================================	*/

	var ReceiverPlugin = function(appl, xappl) {

		// exec --------------------------------------------------------------
		ReceiverPlugin.prototype.exec = function(ref, value) {
			var hash;

			switch(ref) {
			case 'message-send':
				this.messageSend(value);
				break;
			case 'messages-hidden':
				this.messages.hidden();
				break;
			case 'messages-refresh':
				this.messages.refresh();
				break;
			case 'messages-start':
				this.messages.start();
				break;
			case 'messages-stop':
				this.messages.stop();
				break;
			case 'ping-start':
				this.pingStart();
				break;
			case 'ping-stop':
				this.pingStop();
				break;
			case 'receiver-abort':
				this.messages.stop();
				break;
			case 'receiver-pause':
				this.usession.eventLectureMode('receiver-pause', -1);
				break;
			case 'receiver-play':
				this.usession.eventLectureMode('receiver-play', -1);
				break;
			case 'receiver-start-live':
				this.usession.eventLectureMode('receiver-start-live', -1);
				break;
			case 'reload-english':
				this.reload('en');
				break;
			case 'reload-french':
				this.reload('fr');
				break;
			case 'reload-portuguese':
				this.reload('pt');
				break;
			case 'reload-russian':
				this.reload('ru');
				break;
			case 'reload-browser':
				this.reloadBrowser();
				break;
			case 'reload-page':
				location.reload(true);
				break;
			case 'reload-slides':
				this.exec('messages-stop', null);
				this.exec('video-stop', null);
				this.exec('ping-stop', null);
				this.appl.setXApplication(this.xappl);
				break;
			case 'set-hash-session':
				hash = this.xappl.getVar('#lecture-info/session-hash');
				MW.HttpFiles.setHash(hash);
				this.usession.setHash(hash);
				this.appl.oplugin.setHash(hash);
				break;
			case 'video-stop':
				this.usession.eventAction('video-stop');
				break;
			default:
				throw new Error('not yet implemented[ReceiverPlugin]: ' + ref);
			}
		};
	/*
		// display -----------------------------------------------------------
		ReceiverPlugin.prototype.displayCourse = function() {
			var s, url;

			s = this.xappl.getVar('action-data');
			if (s.startsWith('entry-') === false)
				throw new Error('invalid value of action-data: ' + s);

			url = window.location.origin + window.location.pathname + 'course?';
			url += s.substring(6);
			
			window.location.href = url;
		};

	*/
		// messageSend -------------------------------------------------------
		ReceiverPlugin.prototype.messageSend = function(varname) {

			this.messages.send('message:' + this.xappl.parseText(varname));
		};

		// pingStart ---------------------------------------------------------
		ReceiverPlugin.prototype.pingStart = function() {

			let me = this;

			this.timer = setInterval(function() {
				var xhttp;

				if ((++me.timerCounter % 2) === 0) {
					console.log('websocket-ping...');

					try {
						me.messages.send('ping');
					}
					catch(err) {
						clearInterval(me.timer);
						return;
					}

					return;
				}

				xhttp = new XMLHttpRequest();
				
				xhttp.onreadystatechange = function() {

					if (this.readyState === 4 && this.status === 200)
						console.log('servlet-ping ok...');
				};

				xhttp.open('GET', me.xappl.getUrl() + '?ping', true);
				xhttp.send();
			}, 60000);		//60 secs...
		};

		// pingStop ----------------------------------------------------------
		ReceiverPlugin.prototype.pingStop = function() {

			if (this.timer !== null)
				clearInterval(this.timer);
		};

		// reload ------------------------------------------------------------
		ReceiverPlugin.prototype.reload = function(lang) {
			var n, qs, url, lectureId;

			qs = window.location.search;

			if (qs.charAt(0) !== '?')
				return;

			if ((n = qs.indexOf('&')) === -1)
				lectureId = parseInt(qs.substring(1));
			else
				lectureId = parseInt(qs.substring(1, n));

			url = window.location.origin + window.location.pathname + '?';
			url += lectureId;

			if (lang !== 'fr')
				url += '&lang=' + lang;

			window.location.href = url;
		};

		// reloadBrowser -----------------------------------------------------
		ReceiverPlugin.prototype.reloadBrowser = function() {
			var n, pathname, search, url;

			pathname = window.location.pathname;
			if ((n = pathname.lastIndexOf('/')) === -1)
				return;
			url = window.location.origin + pathname.substring(0, n + 1);

			search = window.location.search;			// ?626&lang=en
			if ((n = search.indexOf('&lang=')) !== -1)
				if (n + 8 === search.length)
					url += 'index-' + search.substring(n + 6) + '.html';

			window.location.href = url;
		};

		// Constructor -------------------------------------------------------
		{
			var v;

			this.appl = appl;
			this.xappl = xappl;

			this.usession = new UserSessionHandler(xappl);
			this.messages = new MessagesHandler(appl, xappl, 'receiver', this.usession);
			this.usession.setMessages(this.messages);

			this.timer = null;
			this.timerCounter = 0;

			// cookie...
			v = decodeURIComponent(DocLib.getCookie('DUDAL-EMAIL'));
			xappl.putVar('DUDAL-EMAIL', v);
		}
	};

	/* ================================================================	*
	 * StatsPlugin.js 													*
	 * ================================================================	*/

	var StatsPlugin = function() {

		// display -----------------------------------------------------------
		StatsPlugin.prototype.display = function(url) {
			var log, xhttp;

			log = DocLib.getElement('stats-plugin-area');

			xhttp = new XMLHttpRequest();
			
			xhttp.onreadystatechange = function() {

				switch(xhttp.readyState) {
				case 3:										// loading...
				case 4:										// done...
					log.innerHTML = xhttp.responseText;
				}
			};

			xhttp.open('GET', url + '?' + "stats-log", true);
			xhttp.send();
		};

		// Constructor -------------------------------------------------------
		{
		}
	};

	/* ================================================================	*
	 * TransmitterPlugin.js 											*
	 * ================================================================	*/

	var TransmitterPlugin = function(appl, xappl) {
		
		// exec --------------------------------------------------------------
		TransmitterPlugin.prototype.exec = function(ref, value) {

			switch(ref) {
			case 'check-start':
				this.webcam.start('check');
				break;
			case 'check-stop':
				this.mode = 'idle';
				this.webcam.stop();
				this.webcam.chandler.stop();
				break;
			case 'check-play':
				this.webcam.chandler.play();
				break;
			case 'config-start':
				this.webcam.start('config');
				break;
			case 'config-stop':
			case 'recording-stop':
			case 'transmission-stop':
				this.mode = 'idle';
				this.webcam.stop();
	//			this.display.transmissionStop();
				break;
			case 'display-init':
				if ((this.display = MW.WLib.getWidget('transmitter', 'transmitter-display')) === null)
					throw new Error("widget not found: transmitter-display");
				break;
			case 'lecture-videos-start':
				this.lectureVideosStart(value);
				break;
			case 'lecture-videos-stop':
				this.lectureVideosStop();
				break;
			case 'lecture-view':
				this.lectureView(value);
				break;
			case 'message-send':
				this.messageSend(value);
				break;
			case 'messages-hidden':
				this.messages.hidden();
				break;
			case 'messages-refresh':
				this.messages.refresh();
				break;
			case 'messages-start':
				this.messages.start();
				break;
			case 'messages-stop':
				this.messages.stop();
				break;
			case 'ping-start':
				this.pingStart();
				break;
			case 'ping-stop':
				this.pingStop();
				break;
			case 'pointer':
				this.pointerSendPosition(value);
				break;
			case 'recording-start':
				this.mode = 'recording';
				this.slidename = this.display.getCurrentSlide();
				this.webcam.start('recording');
				break;
			case 'reload-english':
				this.reload('en');
				break;
			case 'reload-french':
				this.reload('fr');
				break;
			case 'reload-portuguese':
				this.reload('pt');
				break;
			case 'reload-russian':
				this.reload('ru');
				break;
			case 'reload-browser':
				this.reloadBrowser();
				break;
			case 'session-offline':
				this.messages.send('session-offline');
				break;
			case 'slide-down':
				this.slideKey(true);
				break;
			case 'slide-up':
				this.slideKey(false);
				break;
			case 'slide-update':
				this.slideUpdate();
				break;
			case 'transmission-abort':
				this.messages.stop();
				this.webcam.stop();
				break;
			case 'transmission-start':
				this.mode = 'transmission';
				this.webcam.start('transmission');
				break;
			default:
				throw new Error('not yet implemented[TransmitterPlugin]: ' + ref);
			}
		};

		// lectureVideosStart ------------------------------------------------
		TransmitterPlugin.prototype.lectureVideosStart = function(value) {
			var el, codec, data, duration, item, partId, url, video;

			partId = this.xappl.parseText(value);
			data = this.xappl.getData('lecture-videos');
			item = data.searchElement('videos/stream[@partId=\'' + partId + '\']');
			duration = item.getAttrString('duration', null);
			codec = item.getAttrString('codec', null);

			el = DocLib.getElement('lecture-videos-show-partid_label');
			el.textContent = codec + ': ' + partId + ' - ' + duration;

			url = 'course/video-remove/';
			url += this.xappl.getVar('lectureId');
			url += '/';
			url += partId.substring(1);
			url += '.' + codec;

			video = DocLib.getElement('webcam-video-remove');
			video.src = url;
			video.controls = true;
			video.play();
		};

		// lectureVideosStop -------------------------------------------------
		TransmitterPlugin.prototype.lectureVideosStop = function() {
			var video;

			video = DocLib.getElement('webcam-video-remove');
			video.pause();
			video.src = '';
		};

		// lectureView -------------------------------------------------------
		TransmitterPlugin.prototype.lectureView = function(value) {
			var url;

			url = this.xappl.parseText(value);
			Log.info('url: ' + url);
			window.open(url);
		};

		// messageSend -------------------------------------------------------
		TransmitterPlugin.prototype.messageSend = function(varname) {

			this.messages.send('message:' + this.xappl.parseText(varname));
		};

		// pingStart ---------------------------------------------------------
		TransmitterPlugin.prototype.pingStart = function() {

			let me = this;

			this.timer = setInterval(function() {
				var xhttp;

				if ((++me.timerCounter % 2) === 0) {
					console.log('websocket-ping...');

					try {
						me.messages.send('ping');
					}
					catch(err) {
						clearInterval(me.timer);
						return;
					}

					return;
				}

				xhttp = new XMLHttpRequest();
				
				xhttp.onreadystatechange = function() {

					if (this.readyState === 4 && this.status === 200)
						console.log('ping ok...');
				};

				xhttp.open('GET', me.xappl.getUrl() + '?ping', true);
				xhttp.send();
			}, 60000);			// 60 secs
		};

		// pingStop ----------------------------------------------------------
		TransmitterPlugin.prototype.pingStop = function() {

			if (this.timer !== null)
				clearInterval(this.timer);
		};

		// pointerSendPosition -----------------------------------------------
		TransmitterPlugin.prototype.pointerSendPosition = function(value) {
			var t, time;

			t = this.webcam.getCurrentTime();

			if (value !== 'END')
				--t;
			
			time = Math.round(t * 10) / 10;
			//console.log('pointer:' + time + ':' + value + ':' + this.webcam.getStreamId());
			this.messages.send('pointer:' + time + ':' + value + ':' + this.webcam.getStreamId());
		};

		// reload ------------------------------------------------------------
		TransmitterPlugin.prototype.reload = function(lang) {
			var n, s, url;
			
			url = window.location.origin;

			s = window.location.pathname;
			n = s.lastIndexOf('/');
			url += s.substring(0, n + 1);

			url += "admin-" + lang + ".html";

			window.location.href = url;
		};

		// reloadBrowser -----------------------------------------------------
		TransmitterPlugin.prototype.reloadBrowser = function() {
			var n, pathname, lang, url;

			pathname = window.location.pathname;
			if ((n = pathname.lastIndexOf('/')) === -1)
				return;

			url = window.location.origin + pathname.substring(0, n + 1);

			if ((lang = XUtil.getLang()) !== 'fr')
				url += 'index-' + lang + '.html';

			window.location.href = url;
		};

		// slideKey ----------------------------------------------------------
		TransmitterPlugin.prototype.slideKey = function(down) {
			var i, n, el, current, data, imgs, link, max, path, time, top;

			current = this.display.getCurrentSlide();

			data = this.xappl.getData('lecture-info');
			imgs = data.searchElements('toc/img');

			for(i = 0; i < imgs.length; ++i) {
				path = imgs[i].getTextContent();
				if ((n = path.indexOf('/')) === -1)
					continue;

				link = path.substring(n + 1);
				if (link === current)
					break;
			}

			if (i === imgs.length)
				return;

			if (down) {
				if (++i === imgs.length)
					return;
			} else {
				if (--i < 0)
					return;
			}

			path = imgs[i].getTextContent();

			this.display.doAction('set-slide', 'cache:' + path);

			el = DocLib.getElement('transmitter-toc');
			max = el.scrollHeight - el.clientHeight;
			top = max * i / (imgs.length - 1);
			el.scrollTo(0, top);

			if (this.mode !== 'recording')
				return;

			if ((n = path.indexOf('/')) === -1)
				throw new Error("invalid path: " + path);
			link = path.substring(n + 1);

			this.slidename = link;
			time = Math.round(this.webcam.getCurrentTime() * 10) / 10;
			this.messages.send('slide:' + time + ':' + this.slidename + ':' + this.webcam.getStreamId());
		};

		// slideUpdate -------------------------------------------------------
		TransmitterPlugin.prototype.slideUpdate = function() {
			var name, time;

			if (this.mode !== 'recording')
				return;

			if ((name = this.display.getCurrentSlide()) === this.slidename)
				return;

			this.slidename = name;

			time = Math.round(this.webcam.getCurrentTime() * 10) / 10;
			//console.log('slide:' + time + ':' + this.slidename + ':' + this.webcam.getStreamId());
			this.messages.send('slide:' + time + ':' + this.slidename + ':' + this.webcam.getStreamId());
		};

		// Constructor -------------------------------------------------------
		{
			this.appl = appl;
			this.xappl = xappl;

			this.messages = new MessagesHandler(appl, xappl, 'transmitter', null);
			this.webcam = new Webcam(xappl, this.messages);
			this.display = null;

			this.mode = 'idle';
			this.slidename = null;
			this.timer = null;
			this.timerCounter = 0;
		}
	};

	/* ================================================================	*
	 * HealthPlugin.js 													*
	 * ================================================================	*/

	var HealthPlugin = function(xappl) {

		// display -----------------------------------------------------------
		HealthPlugin.prototype.display = function(url, ref) {
			var log, me, xhttp;

			log = DocLib.getElement('health-plugin-area');
			log.innerHTML = '';

			xhttp = new XMLHttpRequest();

			me = this.xappl;
			xhttp.onreadystatechange = function() {

				switch(xhttp.readyState) {
				case 3:										// loading...
					if (xhttp.status !== 200)
						return;
				case 4:										// done...
					if (xhttp.status !== 200) {
						me.taskExecTask('#http-timeout');
						return;
					}

					console.log(xhttp.readyState + ' ' + xhttp.status);
					log.innerHTML = xhttp.responseText;
				}
			};

			xhttp.open('GET', url + '?health-' + ref, true);
			xhttp.send();
		};

		// Constructor -------------------------------------------------------
		{
			this.xappl = xappl;
		}
	};

	/* ================================================================	*
	 * WordPressPlugin.js 												*
	 * ================================================================	*/

	var WordPressPlugin = function(xappl) {

		// download ----------------------------------------------------------
		WordPressPlugin.prototype.download = function(url) {
			var log, me, xhttp;

			log = DocLib.getElement('wordpress-plugin-area');
			log.innerHTML = '';

			xhttp = new XMLHttpRequest();

			me = this.xappl;
			xhttp.onreadystatechange = function() {

				switch(xhttp.readyState) {
				case 3:										// loading...
					if (xhttp.status !== 200)
						return;

				case 4:										// done...
					if (xhttp.status !== 200) {
						me.taskExecTask('#http-timeout');
						return;
					}

					log.innerHTML = xhttp.responseText;
				}
			};

			xhttp.open('GET', url + '?wordpress-download', true);
			xhttp.send();
		};

		// Constructor -------------------------------------------------------
		{
			this.xappl = xappl;
		}
	};

	/* ================================================================	*
	 * UserSessionHandler.js 											*
	 * ================================================================	*/

	var UserSessionHandler = function(xappl) {

		// eventAction -------------------------------------------------------
		UserSessionHandler.prototype.eventAction = function(name) {
			var l, d, t, plugin;

			if (this.xslide === null)
				return;

			l = this.xslide.layout;
			plugin = this.xslide.plugin;

			switch(name) {
			case 'receiver-pause':
				this.video.oncanplay = null;
				this.video.ontimeupdate = null;
				this.video.onended = null;
				
				switch(this.lmode) {
				case 'live':
					l.setVideoMode('live');
					this.liveStartStream();
					return;
				}

				l.setVideoMode('video-pause');
				this.video.pause();

				this.nextEvent = 0;
				this.lastLink = null;

				plugin.repaint();
				if (this.ctxt === 'offline')
					this.xappl.taskWidgetSet('offline', 'receiver-mode', 'receiver-play');
				return;
			case 'receiver-play':
				if (l.isVideoInPause()) {
					l.setVideoMode('video-running');
					this.videoSetCallback();
					this.video.play();
					plugin.repaint();

					if (this.ctxt === 'offline')
						this.xappl.taskWidgetSet('offline', 'receiver-mode', 'receiver-pause');
					return;
				}

				if (l.isLivePlaying())
					this.liveStopStream();

				l.setVideoMode('video-starting');
				this.videoAskOffset(this.xslide.getCurrentTimeInMillis());
				this.showWidgetMedia('receiver-video');

				if (this.ctxt === 'offline')
					this.xappl.taskWidgetSet('offline', 'receiver-mode', 'receiver-pause');
				return;
			case 'video-seek':
				if (l.isVideoPlaying())
					this.videoStopStream();
				
				d = this.xslide.getDurationInMillis();
				t = l.getCursorDraggingTime();

				if (t + 5000 < d || this.lmode === 'recording') {
					if (l.isLivePlaying())
						this.liveStopStream();
				
					l.setVideoMode('video-starting');
					this.videoAskOffset(t);
					this.showWidgetMedia('receiver-video');

					if (this.ctxt === 'offline')
						this.xappl.taskWidgetSet('offline', 'receiver-mode', 'receiver-pause');
					return;
				}
				
				this.xslide.setCurrentTime(d);

				switch(this.lmode) {
				case 'live':
					if (l.isLivePlaying()) {
						l.setVideoMode('live-end');
						return;
					}

					l.setVideoMode('live-end');
					this.liveStartStream();
					plugin.repaint();
					return;
				}

				l.setVideoMode('video-end');
				this.showWidgetMedia('receiver-info');
				return;
			case 'video-end':
				this.videoStopStream();

				switch(this.lmode) {
				case 'live':
					l.setVideoMode('live-end');
					this.liveStartStream();
					plugin.repaint();
					return;
				}

				l.setVideoMode('video-end');
				this.showWidgetMedia('receiver-info');
				return;
			case 'video-stop':
				this.videoStopStream();
				return;
			}

			XApplication.logMsg('UserSessionHandler[eventAction]: ' + name);
		};

		// eventInit ---------------------------------------------------------
		UserSessionHandler.prototype.eventInit = function(ctxt, data) {
			var i, l, children, duration, id, part;

			switch(ctxt) {
			case 'migration':
				this.migrationId = this.xappl.getVar('migrationId');
				this.xslide = MW.WLib.getWidget('migration-view', 'migration-view-slide');
				id = this.migrationId;
				break;
			case 'offline':
				this.xslide = MW.WLib.getWidget('offline', 'offline-slide');
				id = 'slides';
				break;
			case 'receiver':
				this.lectureId = this.xappl.getVar('#lecture-info/lectureId');
				this.xslide = MW.WLib.getWidget('receiver', 'receiver-slide');
				id = this.lectureId;
				break;
			default:
				throw new Error('invalid ctxt: ' + ctxt);
			}

			this.ctxt = ctxt;
			this.video = DocLib.getElement('receiver-video');

			// video parts...
			children = data.getAllChildren();
			this.parts = [];

			l = this.xslide.layout;

			if (children.length === 0)
				l.setVideoRecording(false);
			else {
				l.setVideoRecording(true);

				for(i = 0; i < children.length; ++i) {
					part = new XVideoPart(children[i], id);
					this.parts.push(part);
				}
			}

			duration = data.getInt('duration');

			this.xslide.setDuration(duration);
			this.xslide.initVideoParts(this.parts);
		};

		// eventSeek ---------------------------------------------------------
		UserSessionHandler.prototype.eventSeek = function(doc) {
			var i, l, data, ipart, timecode, starting, value;

			data = new XmlData(doc, doc.documentElement);

			l = data.getAllChildren();
			for(i = 0; i < l.length; ++i) {
				value = parseInt(l[i].getTextContent());

				switch(l[i].getTag()) {
				case 'part':
					ipart = value;
					break;
				case 'timecode':
					timecode = value;
					break;
				case 'starting':
					starting = value / 1000;
					break;
				}
			}

	//		console.log('eventSeek: ipart=' + ipart + ',timecode=' + timecode + ',starting=' + starting);
			this.xslide.layout.setVideoMode('video-running');
			this.videoStartStream(ipart, timecode, starting);
		};

		// eventStatus -------------------------------------------------------
		UserSessionHandler.prototype.eventStatus = function(doc) {
			var l, t, data, mode, old, plugin;

			l = this.xslide.layout;
			plugin = this.xslide.plugin;

			old = this.lmode;
			data = new XmlData(doc, doc.documentElement);
			mode = data.getString('mode');

			XApplication.logMsg('UserSessionHandler[eventStatus]: ' + this.lmode + ' --> ' + mode);

			this.xslide.setDuration(data.getInt('duration'));

			if (mode !== 'recording' && this.timer !== null) {
				clearInterval(this.timer);
				this.timer = null;
			}

			switch(mode) {
			case 'offline':
				this.lmode = 'offline';

				if (data.getAttrBool('broadcast', false) === true)
					this.showWidgetMode('mode-broadcast');
				else
					this.showWidgetMode('mode-offline');

				switch(old) {
				case 'starting':
					if (!l.hasVideoRecording()) {
						l.setVideoMode('no-video');
						this.showWidgetMedia('receiver-info');
						return;
					}

					l.setVideoMode('video-running');
					this.videoStartStream(0, 0, 0);
					this.showWidgetMedia('receiver-video');
					return;
				case 'offline':
					return;
				case 'live':
					if (!l.isLivePlaying())
						return;

					if (l.isVideoAtEnd())
						l.setVideoMode('video-end');
					else
						l.setVideoMode('no-video');

					this.liveStopStream();
					this.showWidgetMedia('receiver-info');
					return;
				case 'online':
					return;
				}
				break;
			case 'live':
				this.lmode = 'live';
				this.showWidgetMode('mode-live');

				if (l.isVideoPlaying())
					return;

				if (l.isVideoAtEnd())
					l.setVideoMode('live-end');
				else
					l.setVideoMode('live');
				
				this.liveStartStream();
				this.showWidgetMedia('receiver-video');
				return;
			case 'recording':
				l.setVideoRecording(true);
				this.recordingStatus(data);
				this.xappl.taskWidgetShow('receiver', 'receiver-download', false);

				t = this.xslide.getDurationInMillis();
				this.xslide.layout.setRecordingStartingAt(t, new Date().getTime());

				switch(old) {
				case 'live':
					if (l.isLivePlaying())
						this.liveStopStream();
					break;
				case 'recording':
					return;
				}

				this.lmode = 'recording';
				this.showWidgetMode('mode-recording');
				
				this.startRecordingTimer();

				if (l.isVideoPlaying())
					return;

				switch(old) {
				case 'starting':
					l.setVideoMode('video-starting');
					this.videoAskOffset(t);
					this.showWidgetMedia('receiver-video');
					return;
				}

				this.xslide.setCurrentTime(t);

				l.setVideoMode('video-running');
				this.videoStartStream(this.parts.length - 1, 0, 0);

				this.showWidgetMedia('receiver-video');
				plugin.repaint();
				return;
			case 'online':
				this.lmode = 'online';
				this.showWidgetMode('mode-live');
				plugin.repaint();
				return;
			}

			XApplication.logMsg('UserSessionHandler[eventStatus]: ' + this.lmode + ' --> ' + mode);
		};

		// eventStream -------------------------------------------------------
		UserSessionHandler.prototype.eventStream = function(doc) {
			var i, data, id, ipart, live, part, parts;

			parts = this.parts;

			data = new XmlData(doc, doc.documentElement);
			ipart = data.getInt('ipart');
			live = data.getBool('live');

			id = XInteger.toPadding(ipart, 3);
			for(i = 0; i < parts.length; ++i) {
				part = parts[i];

				if (id === part.id) {
					parts[i].live = live;
					return;
				}
			}

			XApplication.logMsg('UserSessionHandler[eventStram]: ipart not found: ' + id);
		};

		// liveStartStream ----------------------------------------------------
		UserSessionHandler.prototype.liveStartStream = function() {
			var t;

			t = new Date().getTime();
			
			this.video.src = 'course/live-' + this.hashSession + '-' + t + '.webm';
			this.video.play();

			this.xslide.layout.setVideoLoading(true);
			this.xslide.plugin.repaint();

			let me = this;

			this.video.oncanplay = function() {

				me.xslide.layout.setVideoLoading(false);
				me.xslide.plugin.repaint();
			};
		};

		// liveStopStream -----------------------------------------------------
		UserSessionHandler.prototype.liveStopStream = function() {

			this.video.src = 'course/video-empty.webm';
			this.video.oncanplay = null;
		};

		// onended ------------------------------------------------------------
		UserSessionHandler.prototype.onended = function() {

			this.xslide.plugin.eventVideoEnded();
			
			if (this.ipart +  1 === this.parts.length) {
				this.eventAction('video-end');
				return;
			}

			this.videoStartStream(this.ipart + 1, 0, 0);
		};

		// ontimeupdate -------------------------------------------------------
		UserSessionHandler.prototype.ontimeupdate = function() {
			var t0, t1;

			// video terminated...
			if ((t0 = this.video.currentTime) === 0)
				return;

			t0 += this.startingOffset;
			this.processEvent(t0);

			if ((t1 = Math.round(t0)) === this.lastTimeDisplayed)
				return;

			this.lastTimeDisplayed = t1;
			this.xslide.setCurrentTime(t1 * 1000);

			if (this.xslide.layout.isTimebarVisible())
				this.xslide.plugin.repaint();
		};

		// processEvent --------------------------------------------------------
		UserSessionHandler.prototype.processEvent = function(t) {
			var cpart, evt, events, link, millis, route;

			millis = t * 1000;

			cpart = this.parts[this.ipart];
			events = cpart.events;

			link = null;
			route = null;

			while(this.nextEvent < events.length) {
				evt = events[this.nextEvent];

				if (millis < evt.timecode)
					break;

				switch(evt.type) {
				case "slide":
					link = evt.link;
					break;
				case "pointer":
					route = evt.route;
					break;
				}
				
				++this.nextEvent;
			}

			if (link !== null && link !== this.lastLink) {
				this.xslide.doAction('load', 'cache:' + link);
				this.lastLink = link;
			}

			if (route !== null)
				this.xslide.doAction('route', route);
		};

		// recordingStatus ---------------------------------------------------
		UserSessionHandler.prototype.recordingStatus = function(data) {
			var len, last, origin, part;

			part = new XVideoPart(data.getFirstChild(), this.lectureId);
			this.xslide.initEvents(part);

			if ((len = this.parts.length) === 0)
				last = null;
			else
				last = this.parts[len - 1];

			if ((origin = data.getAttrString('origin', null)) === null) {
				if (last !== null && last.id === part.id) {	// replace part...
					last = part;
					return;
				}

				this.parts.push(part);						// insert new part...
				return;
			}

			if (last === null) {
				Log.info('UserSessionHandler.recordingStatus: empty parts...');
				return;
			}

			switch(origin) {
			case 'pointer':
				last.addPointerEvent(part.events[0]);
				break;
			default:
				last.mergeSlides(part.events);
			}
		};

		// setHash ----------------------------------------------------
		UserSessionHandler.prototype.setHash = function(hash) {

			this.hashSession = hash;
		};

		// setMessages -------------------------------------------------------
		UserSessionHandler.prototype.setMessages = function(messages) {

			this.messages = messages;
		};

		// showWidgetMedia ---------------------------------------------------
		UserSessionHandler.prototype.showWidgetMedia = function(id) {

			switch(this.ctxt) {
			case 'migration':
				this.xappl.uiappl.widgetExec({
					action: 'show-widget',
					ref: 'migration-view#receiver-media',
					value: id});
				break;
			case 'offline':
				this.xappl.uiappl.widgetExec({
					action: 'show-widget',
					ref: 'offline#receiver-media',
					value: id});
				break;
			case 'receiver':
				this.xappl.uiappl.widgetExec({
					action: 'show-widget',
					ref: 'receiver#receiver-media',
					value: id});
				break;
			default:
				throw new Error('invalid ctxt: ' + this.ctxt);
			}
		};

		// showWidgetMode ----------------------------------------------------
		UserSessionHandler.prototype.showWidgetMode = function(id) {

			this.xappl.uiappl.widgetExec({
				action: 'show-widget',
				ref: 'receiver#receiver-mode',
				value: id});
		};

		// startRecordingTimer -------------------------------------------------
		UserSessionHandler.prototype.startRecordingTimer = function() {

			let me = this;

			this.timer = setInterval(function() {

				me.updateDuration();
			}, 1000);
		};

		// updateDuration ----------------------------------------------------
		UserSessionHandler.prototype.updateDuration = function() {
			var l, off, t0, t1;

			l = this.xslide.layout;

			off = l.getRecordingOffset();
			t0 = l.getRecordingTime();
			t1 = new Date().getTime();

			this.xslide.setDuration(off + (t1 - t0));

			if (!l.isTimebarVisible())
				return;
			
			if (l.isVideoPlaying())
				if (this.video.readyState >= 3)
					return;

	//		console.log('updateDuration - repaint ---------------------------');
			this.xslide.plugin.repaint();
		};

		// videoAskOffset ----------------------------------------------------
		UserSessionHandler.prototype.videoAskOffset = function(t) {
			var ipart, off, part, parts;

			this.xslide.setCurrentTime(t);
			
			switch(this.ctxt) {
			case 'migration':
				this.xslide.layout.setVideoMode('video-running');
				this.videoStartStream(0, 0, t);
				break;
			case 'offline':
				this.xslide.layout.setVideoMode('video-running');

				off = 0;
				parts = this.parts;
				for(ipart = 0; ipart < parts.length; ++ipart) {
					part = parts[ipart];
					if (part.starting + part.duration > t) {
						off = t - part.starting;
						break;
					}
				}

				if (ipart === parts.length) {
					ipart = parts.length;
					off = parts[ipart].duration;
				}

				this.videoStartStream(ipart, 0, off);
				break;
			case 'receiver':
				this.messages.send('lecture-seek:' + t);
				break;
			default:
				throw new Error('invalid ctxt: ' + this.ctxt);
			}
		};

		// videoSetCallback ---------------------------------------------------
		UserSessionHandler.prototype.videoSetCallback = function() {
			
			let me = this;

			if (!this.video.paused) {
				this.xslide.layout.setVideoLoading(true);

				this.video.oncanplay = function() {

					me.xslide.layout.setVideoLoading(false);
				};
			}

			this.video.ontimeupdate = function() {

				me.ontimeupdate();
			};

			this.video.onended = function() {

				me.onended();
			};
		};

		// videoStartStream ---------------------------------------------------
		UserSessionHandler.prototype.videoStartStream = function(ipart, timecode, startingTime) {
			var t, cpart, url;
			
			this.ipart = ipart;
			cpart = this.parts[ipart];

			// time bar...
			this.startingOffset = (cpart.starting + timecode) / 1000.0;
			this.lastTimeDisplayed = -1;

			// swap slide....
			this.lastLink = null;
			this.nextEvent = 0;

			// start video....
			switch(this.ctxt) {
			case 'migration':
				url = 'migration-course/video-' + this.migrationId + '.mp4';

				if (startingTime !== 0)
					url += '#t=' + startingTime / 1000;

				break;
			case 'offline':
				url = 'videos/' + XInteger.toPadding(cpart.id, 3) + '-final.mp4';

				if (startingTime !== 0)
					url += '#t=' + startingTime / 1000;
				break;
			case 'receiver':
				if (cpart.live) {
					t = new Date().getTime();
					url = 'course/video-' + this.hashSession + '-' + cpart.id + '-' + timecode + '-' + t + '.webm';
				} else
					url = 'course/video-' + this.hashSession + '-' + cpart.id + '-final.mp4';

				if (startingTime !== 0)
					url += '#t=' + startingTime;
				break;
			default:
				throw new Error('invalid ctxt: ' + this.ctxt);
			}

			XApplication.logMsg('UserSessionHandler[play]: ' + url);
			this.video.src = url;
			this.video.play();

			this.videoSetCallback();
		};

		// videoStopStream ----------------------------------------------------
		UserSessionHandler.prototype.videoStopStream = function() {

			if (this.ctxt !== 'offline')
				this.video.src = 'course/video-empty.webm';
			else {
				this.video.src = '';
				this.xappl.taskWidgetSet('offline', 'receiver-mode', 'receiver-noplay');
			}

			this.video.oncanplay = null;
			this.video.ontimeupdate = null;
			this.video.onended = null;
		};

		// Constructor -------------------------------------------------------
		{
			var v;

			this.xappl = xappl;
			this.hashSession = null;
			this.messages = null;

			this.xslide = null;
			this.video = null;
			this.importId = -1;
			this.lectureId = -1;

			this.ctxt = null;
			this.lmode = 'starting';
			this.parts = null;
			this.ipart = -1;

			// time bar...
			this.startingOffset = 0;
			this.lastTimeDisplayed = 0;

			// swap slides...
			this.lastLink = null;
			this.nextEvent = 0;

			// timer...
			this.timer = null;

			// cookie...
			v = decodeURIComponent(DocLib.getCookie('DUDAL-EMAIL'));
			xappl.putVar('DUDAL-EMAIL', v);

			v = decodeURIComponent(DocLib.getCookie('DUDAL-USERNAME'));
			xappl.putVar('DUDAL-USERNAME', v);
		}
	};

	/* ================================================================	*
	 * LecturesMove.js 													*
	 * ================================================================	*/

	var LecturesMove = function(xappl) {

		// exec --------------------------------------------------------------
		LecturesMove.prototype.exec = function(ref) {
			var el, no, sb, lectures, list;

			switch(ref) {
			case 'list':
				lectures = this.xappl.getData('folder-list');

				sb = new XStringBuilder();

				list = this.xappl.getVar('lectures-move-list').split(',');
				for(no of list) {
					el = lectures.searchElement('/magicw/lecture[@id=\'' + no + '\']');

					sb.append(' • ')
					sb.append(el.getString('title'));
					sb.append(' [');
					sb.append(el.getString('date'));
					sb.append(']\n');
				}

				DocLib.getElement('lectures-move-dst_lectures_input').value = sb.getText();
				return;
			default:
				throw new Error('not yet implemented[LecturesMove]: ' + ref);
			}
		};

		// Constructor -------------------------------------------------------
		{
			this.xappl = xappl;
		}
	};

	/* ================================================================	*
	 * UIApplication.js 												*
	 * ================================================================	*/

	var UIApplication = function(roots, widgets) {

		// cacheClear -------------------------------------------------
		UIApplication.prototype.cacheClear = function() {
		};
		
		// cacheDownload ----------------------------------------------
		UIApplication.prototype.cacheDownload = function(dataId, files, nextTask) {

			MW.HttpFiles.cacheDownload(dataId, files, nextTask);
		};
		
		// cardLoad ---------------------------------------------------
		UIApplication.prototype.cardLoad = function(id) {
		};
		
		// cardShow ---------------------------------------------------
		UIApplication.prototype.cardShow = function(id) {

			if (MW.isLiveWidget(id))
				MW.createLiveWidget(id);
			
			if (this.cardId !== null)
				DocLib.hide(this.cardId);

			MW.resetLayout(id);

			this.cardId = id;
			DocLib.show(this.cardId);

			MW.setRoot(id);
		};

		// cardsHide --------------------------------------------------
		UIApplication.prototype.cardsHide = function() {

			for(var id in this.roots)
				DocLib.hide(id);
		};

		// cardUpdate -------------------------------------------------
		UIApplication.prototype.cardUpdate = function(action, xvars) {
			var el, path, visible, wt;

			visible = MW.WLib.isCurrentCardId(action.cardId);
			
			if ((wt = MW.WLib.getWidget(action.cardId, action.widgetId)) === null)
				throw new Error('widget not found: ' + action.widgetId);

			switch(wt.type) {
			case 'label':
			case 'labelbutton':
				el = DocLib.getElement(wt.id + '_label');
				el.innerHTML = this.xappl.parseText(action.value);
				this.roots[action.cardId].setUpdateUI(true);
				break;
			case 'picture':
				MW.WLib.widgetUpdatePicture(wt, action.value);
				break;
			case 'table':
				MW.WLib.widgetUpdateTable(wt, action.xref, visible);
				this.roots[action.cardId].setUpdateUI(true);
				break;
			case 'tree':
				path = null;
				if (action.value !== null)
					path = xvars.parse(action.value);

				MW.WLib.widgetUpdateTree(wt, action.xref, path, visible);
				this.roots[action.cardId].setUpdateUI(true);
				break;
			default:
				throw new Error('not yet implemented: ' + wt.type);
			}
		};

		// cleanFormField ---------------------------------------------
		UIApplication.prototype.cleanFormField = function(type, id, fieldId) {
			var i, k, l, el, exp, root, wt;

			switch(type) {
			case 'checkbox':
				k = id + '_' + fieldId;
				exp = 'input[type=checkbox][name="' + k + '"]:checked';

				l = document.querySelectorAll(exp);
				for(i = 0; i < l.length; ++i)
					l[i].checked = false;
				return;
			case 'date':
				root = MW.getCard(id);

				k = id + '_' + fieldId;
				if ((wt = XLib.search(root, k)) === null)
					throw new Error('widget not found: ' + k);

				wt.uiwt.uidate.reset();
				return;
			case 'hidden':
				return;
			case 'radiobutton':
				k = id + '_' + fieldId;
				exp = 'input[type=radio][name="' + k + '"]:checked';

				if ((el = document.querySelector(exp)) === null)
					return;

				el.checked = false;
				return;
			case 'tree':
				return;
			default:
				k = id + '_' + fieldId + '_input';
				DocLib.formSetValue(k, type, '');
			}
		};

		// getFormField -----------------------------------------------
		UIApplication.prototype.getFormField = function(type, id, fieldId, itemId) {

			return DocLib.formGetValue(type, id, fieldId, itemId);
		};

		// getFormFields ----------------------------------------------
		UIApplication.prototype.getFormFields = function(type, id, fieldId) {

			return DocLib.formGetValues(type, id, fieldId);
		};

		// getTemplate ------------------------------------------------
		UIApplication.prototype.getTemplate = function(ref) {
			var tmpl;

			if ((tmpl = this.templates[ref]) === undefined)
				throw new Error('template not found: ' + ref);

			return tmpl;
		};

		// html5Action ------------------------------------------------
		UIApplication.prototype.html5Action = function(ref) {

			switch(ref) {
			case 'set-warning-before-leaving':
				window.onbeforeunload = function() {
					return 'Are you sure you want to leave?';
				};
				break;
			case 'unset-warning-before-leaving':
				window.onbeforeunload = undefined;
				break;
			default:
				throw new Error('not yet implemented: ' + ref);
			}
		};
		
		// httpPost ---------------------------------------------------
		UIApplication.prototype.httpPost = function(xdoc, callback) {
			var cl;

			cl = new HttpClient(this.url + '?magicw-xml', this.xappl);
			cl.post(xdoc, callback);
		};

		// httpRequest ------------------------------------------------
		UIApplication.prototype.httpRequest = function(req, xdoc, reqs) {
			var cl, url;

			url = this.url + '?magicw-xml';

			cl = new HttpClient(url, this.xappl);
			cl.xmlPost(req, xdoc, reqs);
		};

		// getKey -----------------------------------------------------
		UIApplication.prototype.getKey = function(id) {

			return this.xappl.getVar(id);
		};

		// getXForm ---------------------------------------------------
		UIApplication.prototype.getXForm = function(ref) {

			return this.forms.get(ref);
		};

		// getXWData --------------------------------------------------
		UIApplication.prototype.getXWData = function(data) {

			return DataToJs.transformToXWData(data, false);
		};

		// pluginAction -----------------------------------------------
		UIApplication.prototype.pluginAction = function(widgetId, ref, value) {
			var wt, xd, cardId, view;

			switch(widgetId) {
			case 'lecture-edit-toc':
				wt = MW.WLib.getWidget('lecture-edit', widgetId);
				wt.doAction(ref);
				return;
			case 'migration-view-toc':
			case 'receiver-toc':
			case 'transmitter-toc':
			case 'offline-toc':
				switch(widgetId) {
				case 'migration-view-toc':
					cardId = 'migration-view';
					break;
				case 'receiver-toc':
					cardId = 'receiver';
					break;
				case 'transmitter-toc':
					cardId = 'transmitter';
					break;
				case 'offline-toc':
					cardId = 'offline';
					break;
				}

				wt = MW.WLib.getWidget(cardId, widgetId);
				xd = XThumbnails.getNewView(cardId, wt, this.xappl);
				view = XLib.createWidget(this.getXWData(xd));
				MW.WLib.updateScrollPaneView(wt, view, false);
				return;
			case 'migration-view-slide':
				wt = MW.WLib.getWidget('migration-view', widgetId);
				wt.doAction(ref, value);
				return;
			case 'receiver-slide':
				wt = MW.WLib.getWidget('receiver', widgetId);
				wt.doAction(ref, value);
				return;
			case 'transmitter-display':
				wt = MW.WLib.getWidget('transmitter', widgetId);
				wt.doAction(ref, value);
				return;
			case 'offline-slide':
				wt = MW.WLib.getWidget('offline', widgetId);
				wt.doAction(ref, value);
				return;
			default:
				throw new Error('not yet implemented: ' + widgetId);
			}
		};

		// pluginObject -----------------------------------------------
		UIApplication.prototype.pluginObject = function(id, ref, value) {

			switch(id) {
			case 'browser':
				this.bplugin.exec(ref);
				return;
			case 'health-log':
				this.hplugin.display(this.url, ref);
				return;
			case 'lectures-move':
				this.lplugin.exec(ref);
				break;
			case 'migration-viewer':
				this.mplugin.exec(ref);
				return;
			case 'offline-viewer':
				this.oplugin.exec(ref, value);
				return;
			case 'receiver':
				this.rplugin.exec(ref, value);
				return;
			case 'stats-log':
				this.splugin.display(this.url);
				return;
			case 'transmitter':
				this.tplugin.exec(ref, value);
				return;
			case 'wordpress-log':
				this.wplugin.download(this.url);
				return;
			default:
				throw new Error('not yet implemented: ' + id + "/" + ref);
			}
		};

		// popupShow --------------------------------------------------
		UIApplication.prototype.popupShow = function(popupId, wt) {

			wt.id = popupId;
			MW.createPopup(wt);

			this.cardShow(wt.id);
		};

		// setFormField -----------------------------------------------
		UIApplication.prototype.setFormField = function(cardId, type, id, fieldId, value) {
			var k;

			switch(type) {
			case 'checkbox':
				k = id + '_' + fieldId + '_' + value + '_checkbox_input';
				break;
			case 'date':
				k = id + '_' + fieldId;
				break;
			case 'hidden':
				k = id + '#' + fieldId;
				this.xappl.putVar(k, value);
				return;
			case 'label':
				k = id + '_' + fieldId + '_label';

				if (cardId !== null)
					this.roots[cardId].setUpdateUI(true);
				else
					this.roots[id].setUpdateUI(true);
				break;
			case 'radiobutton':
				k = id + '_' + fieldId + '_' + value + '_radiobutton_input';
				break;
			default:
				k = id + '_' + fieldId + '_input';
			}

			if (type !== 'password')
				console.log(k + "[" + type + "]: " + value);
			else
				console.log(k + "[" + type + "]: ********");

			DocLib.formSetValue(k, type, value.replace(/\\n/g, '\n'));
		};

		// setFormTree -------------------------------------------------------
		UIApplication.prototype.setFormTree = function(fst) {
			var ref, wt, xd;

			ref = fst.formId + '_' + fst.inputId;

			if ((xd = this.wforms.get(ref)) === null)
				throw new Error('widget not found: ' + ref);

			xd.props = 'xref=' + fst.xref;
			wt = XLib.createWidget(new XWData(xd));

			this.widgets.treeUpdate(fst.cardId, wt);
		};

		// setTemplates ------------------------------------------------------
		UIApplication.prototype.setTemplates = function(templates) {

			this.templates = templates;
		};

		// setUrl ------------------------------------------------------------
		UIApplication.prototype.setUrl = function(url) {

			this.url = url;
		};

		// setWidgetForms ----------------------------------------------------
		UIApplication.prototype.setWidgetForms = function(data) {
			var i;

			for(i = 0; i < data.length; ++i)
				this.wforms.put(data[i].id, data[i]);
		};

		// setXApplication --------------------------------------------
		UIApplication.prototype.setXApplication = function(xappl) {

			this.xappl = xappl;

			this.bplugin = new BrowserPlugin(xappl);
			this.rplugin = new ReceiverPlugin(this, xappl);
			this.mplugin = new MigrationViewerPlugin(this, xappl);
			this.oplugin = new OfflineViewerPlugin(this, xappl);
			this.tplugin = new TransmitterPlugin(this, xappl); 
			this.hplugin = new HealthPlugin(xappl);
			this.wplugin = new WordPressPlugin(xappl);
			this.splugin = new StatsPlugin();
			this.lplugin = new LecturesMove(xappl);
		};

		// setXForms --------------------------------------------------
		UIApplication.prototype.setXForms = function(data) {
			var i, j, child, children, xdata, xform;

			xdata = new XData(data);
			children = xdata.getAllChildren();

			for(i = 0; i < children.length; ++i) {
				child = children[i];

				xform = new XForm(child);
				this.forms.put(xform.id, xform);
			}
		};

		// templateUpdate ---------------------------------------------
		UIApplication.prototype.templateUpdate = function(id, ref, templateId, visible) {

			MW.templateUpdate(id, ref, templateId, visible);
		};

		// uploadFiles ------------------------------------------------
		UIApplication.prototype.uploadFiles = function(attrs) {

			this.uploadFilesTask.openChooser(attrs);
		};

		// waitingStart -----------------------------------------------
		UIApplication.prototype.waitingStart = function() {

			if (document.body.style.cursor === 'wait')
				return;

			document.body.style.cursor = 'wait';
			DocLib.getElement('mw-waitingOverlay').style.display = 'block';
		};

		// waitingStop ------------------------------------------------
		UIApplication.prototype.waitingStop = function() {

			document.body.style.cursor = 'default';
			DocLib.getElement('mw-waitingOverlay').style.display = 'none';
		};

		// widgetConfig -----------------------------------------------
		UIApplication.prototype.widgetConfig = function(id, ref, value) {

			MW.widgetConfigModify(id, ref, value);
		};

		// widgetClean ------------------------------------------------
		UIApplication.prototype.widgetClean = function(cardId, widgetId) {

			this.widgetSet(cardId, widgetId, '');
		};

		// widgetExec -------------------------------------------------
		UIApplication.prototype.widgetExec = function(xwa) {
			var n, child, style, wt, cardId, widgetId;

			if ((n = xwa.ref.indexOf('#')) === -1)
				throw new Error('invalid ref: ' + xwa.ref);

			cardId = xwa.ref.substring(0, n);
			widgetId = xwa.ref.substring(n + 1);

			if ((wt = MW.WLib.getWidget(cardId, widgetId)) === null)
				throw new Error('widget not found: ' + cardId + '/' + widgetId);

			style = wt.uiwt.element.style;

			switch(xwa.action) {
			case 'background':
				style.background = xwa.value;
				return;
			case 'disabled':
				wt.disabled = true;
				style.cursor = 'default';
				style.background = 'gainsboro';
				child = DocLib.getElement(wt.id + '_label');
				child.style.cursor = 'default';
				break;
			case 'enabled':
				wt.disabled = false;
				style.cursor = 'pointer';
				style.background = wt.bg;
				child = DocLib.getElement(wt.id + '_label');
				child.style.cursor = 'pointer';
				break;
			case 'hide':
				wt.hidden = true;
				style.display = 'none';
				MW.layoutReset();
				break;
			case 'show':
				wt.hidden = false;
				style.display = 'block';
				MW.layoutReset();
				break;
			case 'visibility-off':
				wt.visibility = false;
				style.visibility = 'hidden';
				return;
			case 'visibility-on':
				wt.visibility = true;
				style.visibility = 'visible';
				return;
			case 'show-widget':
				wt.showWidget(xwa.value);
				return;
			default:
				throw new Error('action not yet implemented: ' + xwa.action);
			}
		};

		// widgetGet --------------------------------------------------
		UIApplication.prototype.widgetGet = function(cardId, widgetId) {
			var a, el, group, id, list, wt;

			if ((wt = MW.WLib.getWidget(cardId, widgetId)) === null)
				throw new Error('widget not found: ' + cardId + '/' + widgetId);

			switch(wt.type) {
			case 'table':
				if ((group = wt.getGroup()) === null)
					throw new Error("group not defined in widget: " + widgetId);

				a = [];
				list = document.querySelectorAll('input[name="' + group + '"]:checked');
				for(var item of list)
					a.push(item.value);

				return a.join(',');
			case 'textarea':
				id =  wt.id + '_input';
				if ((el = DocLib.getElement(id)) === null)
					throw new Error('widget not found: ' + cardId + '/' + id);
				
				return el.value;
			case 'tree':
				return wt.view.getSelectedValue();
			default:
				throw new Error('not yet implemented: ' + wt.type);
			}
		};

		// widgetSet --------------------------------------------------
		UIApplication.prototype.widgetSet = function(cardId, widgetId, value) {
			var el, wt;

			if ((wt = MW.WLib.getWidget(cardId, widgetId)) === null)
				throw new Error('widget not found: ' + widgetId);

			switch(wt.type) {
			case 'textarea':
				el = DocLib.getElement(wt.id + '_input');
				el.value = value;
				break;
			case 'stack':
				wt.showWidget(value);
				break;
			default:
				throw new Error('not yet implemented: ' + wt.type);
			}
		};

		// widgetShow -------------------------------------------------
		UIApplication.prototype.widgetShow = function(cardId, widgetId, visible) {
			var el, wt;

			if ((wt = MW.WLib.getWidget(cardId, widgetId)) === null)
				throw new Error('widget not found: ' + widgetId);

			el = DocLib.getElement(wt.id);
			el.style.display = visible ? 'block' : 'none';
		};

		// wizardTask -------------------------------------------------
		UIApplication.prototype.wizardTask = function(wizardId, taskId) {
			var root, wt;

			root = MW.getRootWidget();

			if ((wt = XLib.search(root, wizardId)) === null)
				throw new Error('widget not fountd: ' + wizardId);

			if (wt.type !== 'stack')
				throw new Error('widget not a stack: ' + wt.type);

			wt.task(taskId);
		};

		// Constructor -------------------------------------------------------
		{
			this.roots = roots;
			this.widgets = widgets;

			this.forms = new XMap();
			this.wforms = new XMap();
			this.templates = null;
			this.url = null;
			this.cardId = null;

			this.uploadFilesTask = new UploadFilesTask();
		};
	};

	/* ================================================================	*
	 * WidgetsLib.js 													*
	 * ================================================================	*/

	var WidgetsLib = function(roots) {

		// getWidget --------------------------------------------------
		WidgetsLib.prototype.getWidget = function(cardId, widgetId) {
			var card, wt;

			if ((card = this.roots[cardId]) === undefined)
				throw new Error('cardId not found: ' + cardId);

			if ((wt = XLib.search(card.widget, widgetId)) !== null)
				return wt;

			throw new Error('widgetId not found: ' + cardId + ' (' + cardId + ',' + widgetId + ')');
		};

		// treeUpdate -----------------------------------------------------
		WidgetsLib.prototype.treeUpdate = function(cardId, wt) {
			var cur, html, sb;

			sb = [];
			HtmlWriter.createWidget(sb, wt.view);
			html = sb.join('');

			cur = this.getWidget(cardId, wt.id);
			cur.uiwt.element.innerHTML = html;

			MW.initWidget(wt.view, cur);

			cur.replace(wt);
			cur.layoutUpdateUI(false);
		};

		// Constructor ----------------------------------------------------
		{
			this.roots = roots;
		}
	};

	/* ================================================================	*
	 * MagicW.js 														*
	 * ================================================================	*/

	// actionEventResized ------------------------------------------------
	MagicW.prototype.actionEventResized = function() {
		var w, h;

		w = window.innerWidth;
		h = window.innerHeight;

		this.xroot.doLayout(0, 0, w, h);

		// vertical scrollbar is present?
		if (w === document.documentElement.clientWidth)
			return;

		w = document.documentElement.clientWidth;
		h = document.documentElement.clientHeight;
		this.xroot.doLayout(0, 0, w, h);
	};
	/*
	// actionPictButtonLoaded --------------------------------------------
	MagicW.prototype.actionPictButtonLoaded = function(id) {
		var el;

		el = DocLib.getElement(id);
		el.style.visibility = 'visible';
	};
	*/
	// actionPopupEvent --------------------------------------------------
	MagicW.prototype.actionPopupEvent = function(evt, wt) {
		var  x, y, l, el, popup, offx, offy, winw, winh, xright, xbottom;

		if (!wt.jsVisible) {
			/*
			 * widget parent...
			 */
			el = DocLib.getElement(wt.id);
			el.style.background = wt.over;

			x = el.offsetLeft;
			y = el.offsetTop + wt.cheight;

			l = el.getBoundingClientRect();
			offx = l.x;
			offy = l.y + wt.cheight;
			
			/*
			 * focus...
			 */
			if ((el = document.activeElement) !== document.body)
				el.blur();

			/*
			 * popup...
			 */
			popup = wt.popup;
			el = DocLib.getElement(popup.id);
			el.style.display = 'block';
			el.style.zIndex = 1;

			XLib.doLayout(popup, wt.cwidth);

			/*
			 * popup location...
			 */
			winw = document.documentElement.clientWidth; 
			winh = document.documentElement.clientHeight;

			if ((xright = offx + popup.cwidth) > winw) {
				if ((x -= xright - winw) < 0)
					x = 0;
			}

			if ((xbottom = offy + popup.cheight) > winh) {
				if ((y -= xbottom - winh) < 0)
					y = 0;
			}

			popup.layoutSetLocation(x, y, 'center');


			/*
			 * popup area
			 */
			l = el.getBoundingClientRect();

			wt.jsVisible = true;
			wt.jsPopupLoc = new XRectangle(l.x, l.y, l.width, l.height);
			return;
		}

		if (evt.type === 'mouseenter')
			return;

		if (wt.jsPopupLoc.isInside(evt.clientX, evt.clientY))
			return;
		
		el = DocLib.getElement(wt.id);
		el.style.background = wt.bg;
		
		el = DocLib.getElement(wt.popup.id);
		el.style.display = 'none';

		wt.jsVisible = false;
	};

	// actionTaskEvent ---------------------------------------------------
	MagicW.prototype.actionTaskEvent = function(wt, data) {

		if (wt.action === null) {
			Log.info("no task defined: " + wt.id);
			return;
		}

		try {
			this.xappl.taskNotifyEvent(wt.action, data);
		}
		catch(err) {
			HttpError.send(err);
			this.xappl.runTask('#http-jserror');
		}
	};

	// actionTaskId ------------------------------------------------------
	MagicW.prototype.actionTaskId = function(taskid, data) {

		if (taskid === null) {
			Log.info("no task defined: " + taskid);
			return;
		}

		this.xappl.taskNotifyEvent(taskid, data);
	};

	// createLiveWidget --------------------------------------------------
	MagicW.prototype.createLiveWidget = function(id) {
		var d, wt;

		if (this.widgets[id] === undefined)
			throw new Error('invalid widget name: ' + id);

		if ((d = this.widgets[id]) === null)
			throw new Error('not a live widget:' + id);
		
		wt = XLib.createWidget(new XWData(d.data));

		if (XApplication.isModeDebug())
			MWLib.checkWidgetId(wt);
		
		this.createWidget(wt);
	};

	// createPopup ------------------------------------------------------
	MagicW.prototype.createPopup = function(wt) {

		this.createWidget(wt);

		this.widgets[wt.id] = null;
	};

	// createWidget -----------------------------------------------------
	MagicW.prototype.createWidget = function(wt) {
		var el, body, fragment, html, sb, range;

		body = document.getElementsByTagName('BODY')[0];

		if ((el = DocLib.getElement(wt.id)) !== null)
			body.removeChild(el);

		sb = [];
		HtmlWriter.createWidget(sb, wt);
		html = sb.join('');

		range = document.createRange();
		range.selectNode(body);
		fragment = range.createContextualFragment(html);
		document.body.appendChild(fragment);

		this.initWidget(wt, null);
		this.roots[wt.id] = new XRoot(wt); 
	};

	// doUpdate -----------------------------------------------------------
	MagicW.prototype.doUpdateUI = function(wt) {
		var i, el;

		switch(wt.type) {
		case 'label':
			el = DocLib.getElement(wt.id);
			wt.uilabel.init(el);
			break;
		}

		if (wt.children !== null)
			for(i = 0; i < wt.children.length; ++i)
				this.doUpdateUI(wt.children[i], wt);
	};

	// getCard -----------------------------------------------------------
	MagicW.prototype.getCard = function(id) {
		var xroot;
		
		if ((xroot = this.roots[id]) === undefined)
			throw new Error('card should be a live widget...');

		return xroot.widget;
	};

	// getEdge -----------------------------------------------------------
	MagicW.prototype.getEdge = function(v) {
		
		return v.top + 'px ' + v.right + 'px ' + v.bottom + 'px ' + v.left + 'px';
	};

	// getRootWidget -----------------------------------------------------
	MagicW.prototype.getRootWidget = function() {

		return this.xroot.widget;
	};

	// getTemplateWidget -------------------------------------------------
	MagicW.prototype.getTemplateWidget = function(id, ref) {
		var tmpl, wdata, xtmpl;

		if (this.templates === null)
			throw new Error('a template (' + ref + ') is defined inside a widget: use props="tref=..."');
		
		if ((tmpl = this.templates[ref]) === undefined)
			throw new Error('template not found: ' + ref);

		xtmpl = XTemplate.getWidget(this.xappl, tmpl);
		if (xtmpl.data.id === undefined)
			xtmpl.data.id = id;

		wdata = DataToJs.transformToXWData(xtmpl, false);

		return XLib.createWidget(wdata);
	};

	// getTemplateWidgetId -----------------------------------------------
	MagicW.prototype.getTemplateWidgetId = function() {
		var id;

		id = '_md' + this.templateWidgetId.toString(16).padStart(6, '0');
		
		this.templateWidgetId++;

		return id;
	};

	// getWidgetData -----------------------------------------------------
	MagicW.prototype.getWidgetData = function(id) {

		if (id === 'http-request')
			return null;

		return this.xappl.getData(id);
	};

	// initWidget --------------------------------------------------------
	MagicW.prototype.initWidget = function(wt, parent) {
		var i, w, h, el, child, label, pict, popup, style;

		el = DocLib.getElement(wt.id);

		if (el === null)
			throw new Error('widget not found: ' + wt.id);
		
		wt.uiwt = new UIWidget(el, wt, parent);
		
		style = el.style;
		style.position = 'absolute';

		style.background = wt.bg;
		style.margin = this.getEdge(wt.margin);
		style.padding = this.getEdge(wt.padding);
		style.borderWidth = this.getEdge(wt.border.thickness);
		style.borderStyle = 'solid';
		style.borderColor = wt.border.color;

		switch(wt.id) {
		case 'transmitter-toc_vflow':
			el.setAttribute('style', 'outline-color: black');

			el.tabIndex = 0;

			el.addEventListener('keyup', function(evt) {

				switch(evt.key) {
				case 'ArrowUp':
					MW.xappl.runTask('transmitter-slide-up');
					break;
				case 'ArrowDown':
					MW.xappl.runTask('transmitter-slide-down');
					break;
				}
			});
		}

		switch(wt.type) {
		case 'labelbutton':
			style.userSelect = 'none';
			break;
		}

		switch(wt.type) {
		case 'area':
		case 'flow':
		case 'date':
		case 'grid':
		case 'hflow':
		case 'checkbox':
		case 'iconlink':
		case 'radiobutton':
		case 'pictbutton':
		case 'picture':
		case 'plugin-display':
		case 'plugin-slide':
		case 'plugin-toc-edit':
		case 'menubar':
		case 'scroll':
		case 'simplegrid':
		case 'splitpane':
		case 'stack':
		case 'table':
		case 'tree':
		case 'treearea':
		case 'treetoggle':
		case 'treerow':
		case 'treeview':
		case 'vflow':
			switch(wt.type) {
			case 'date':
				wt.uiwt.uidate = new MDate(wt.id);
				break;
			case 'plugin':
				break;
			case 'plugin-display':
				wt.appl = this.xappl;
				wt.plugin = new XDisplayPlugin(wt);
				break;
			case 'plugin-slide':
				wt.appl = this.xappl;
				wt.plugin = new XSlidePlugin(wt);
				break;
			case 'plugin-toc-edit':
				wt.appl = this.xappl;
				wt.plugin = new JsTocEdit(wt);
				break;
			case 'scroll':
			case 'table':
			case 'tree':
				wt.uiwt.scrollw = MWLib.getScrollbarWidth();
				break;
			case 'treetoggle':
				wt.uiwt.uitoggle = new MTreeToggle(wt);
				wt.uiwt.uitoggle.repaint();
				break;
			}

			style.width = wt.minWidth + 'px';
			style.height = wt.minHeight + 'px';
			break;
		case 'button':
		case 'combobox':
		case 'checkbox-item':
		case 'password':
		case 'radiobutton-item':
		case 'textarea':
		case 'textfield':
			switch(wt.type) {
			case 'textarea':
				if (wt.inputc !== 'white')
					wt.uiwt.child.style.background = wt.inputc;
				break;
			}

			style = window.getComputedStyle(el);

			if ((w = Math.ceil(parseFloat(style.getPropertyValue('width')))) > wt.minWidth)
				wt.minWidth = w;
			
			if ((h = Math.ceil(parseFloat(style.getPropertyValue('height')))) > wt.minHeight)
				wt.minHeight = h;

			if (wt.fg !== 'black')
				wt.uiwt.child.style.color = wt.fg;
			break;
		case 'dropdown':
		case 'label':
		case 'treelabel':
			wt.uilabel.init(el);
			break;
		case 'labelbutton':
		case 'link':
			label = DocLib.getElement(wt.id + '_label');
			label.style.cursor = 'pointer';
			wt.uilabel.init(el);
			break;
		case 'plugin':
			style = window.getComputedStyle(el);

			if ((w = Math.ceil(parseFloat(style.getPropertyValue('width')))) > wt.minWidth)
				wt.minWidth = w;
			
			if ((h = Math.ceil(parseFloat(style.getPropertyValue('height')))) > wt.minHeight)
				wt.minHeight = h;
			break;
		case 'text':
			wt.uitext.init(el);
			break;
		default:
			throw new Error('not yet implemented: ' + wt.type);
		}

		switch(wt.type) {
		case 'button':
			el.onclick = function(event) {

				MW.actionTaskEvent(wt, wt.data);
			};
			break;
		case 'dropdown':
			this.initWidget(wt.popup, null);

			wt.jsVisible = false;
			wt.jsPopupLoc = null;
			
			el.onmouseenter = function(evt) {
				
				try {
					MW.actionPopupEvent(evt, wt);
				}
				catch(err) {
					HttpError.send(err);
					MW.xappl.runTask('#http-jserror');
				}
			};

			el.onmouseleave = el.onmouseenter;

			popup = DocLib.getElement(wt.popup.id);
			popup.onmouseleave = el.onmouseleave;
			break;
		case 'labelbutton':
			el.onclick = function(event) {

				if (!wt.disabled)
					MW.actionTaskEvent(wt, wt.data);
			};

			el.onmouseover = function() {

				try {
					if (!wt.disabled)
						el.style.background = wt.over;
				}
				catch(err) {
					HttpError.send(err);
					MW.xappl.runTask('#http-jserror');
				}
			};

			el.onmousedown = function() {

				try {
					if (!wt.disabled)
						el.style.background = wt.active;
				}
				catch(err) {
					HttpError.send(err);
					MW.xappl.runTask('#http-jserror');
				}
			};

			el.onmouseout = el.onmouseup = function() {

				try {
					if (!wt.disabled)
						el.style.background = wt.bg;
				}
				catch(err) {
					HttpError.send(err);
					MW.xappl.runTask('#http-jserror');
				}
			};
			break;
		case 'link':
			el.onclick = function(event) {

				MW.actionTaskEvent(wt, wt.data);
			};
			break;
		case 'pictbutton':
			el.onclick = function() {

				MW.actionTaskEvent(wt, wt.data);
			};

			pict = el.firstElementChild;
			pict.style.display = 'block';
			pict.style.margin = 'auto';

			el.onmouseover = function() {

				try {
					el.style.background = wt.over;
	//				pict.src = 'images/brighter/' + wt.ref;
	//				pict.style.opacity = 2.0;
					pict.style.filter = 'brightness(120%)';
				}
				catch(err) {
					HttpError.send(err);
					MW.xappl.runTask('#http-jserror');
				}
			};

			el.onmousedown = function() {

				try {
					el.style.background = wt.active;
	//				pict.src = 'images/' + wt.ref;
					pict.style.opacity = 0.4;
				}
				catch(err) {
					HttpError.send(err);
					MW.xappl.runTask('#http-jserror');
				}
			};

			el.onmouseout = el.onmouseup = function() {

				try {
					el.style.background = wt.bg;
	//				pict.src = 'images/' + wt.ref;
					pict.style.opacity = 1.0;
					pict.style.filter = 'none';
				}
				catch(err) {
					HttpError.send(err);
					MW.xappl.runTask('#http-jserror');
				}
			};

			break;
		case 'splitpane':
			wt.jsInside = false;
			wt.jsDragg = false;

			el.onmousedown = function(evt) {
				var x, y, r;

				try {
					r = el.getBoundingClientRect();
					x = evt.clientX - r.left + wt.margin.left;
					y = evt.clientY - r.top + wt.margin.top;

					wt.jsDragg = wt.mousePressed(x, y);
				}
				catch(err) {
					HttpError.send(err);
					MW.xappl.runTask('#http-jserror');
				}
			};

			el.onmouseup = function() {

				try {
					wt.mouseReset();
					wt.jsDragg = false;
				}
				catch(err) {
					HttpError.send(err);
					MW.xappl.runTask('#http-jserror');
				}
			};

			el.onmousemove = function(evt) {
				var x, y, r;

				try {
					if (wt.jsDragg)
						evt.preventDefault();
				
					r = el.getBoundingClientRect();
					x = evt.clientX - r.left + wt.margin.left;
					y = evt.clientY - r.top + wt.margin.top;

					if (evt.buttons === 0) {
						if (wt.mouseIsInsideDivider(x, y) === wt.jsInside)
							return;

						wt.jsInside = !wt.jsInside;

						if (wt.jsInside) {
							if (wt.orientation === XLib.Horizontal)
								el.style.cursor = 'w-resize';
							else
								el.style.cursor = 'n-resize';
						} else	
							el.style.cursor = 'default';
						
						return;
					}

					if (wt.mouseDragged(x, y))
						MW.actionEventResized();
				}
				catch(err) {
					HttpError.send(err);
					MW.xappl.runTask('#http-jserror');
				}
			};
			break;
		case 'treerow':
			el.style.cursor = 'default';
			
			el.onclick = function(evt) {

				try {
					wt.view.taskRowSelected(wt.idx);
				}
				catch(err) {
					HttpError.send(err);
					MW.xappl.runTask('#http-jserror');
				}
			};

			break;
		}

		switch(wt.type) {
		case 'treeview':
			for(i = 0; i < wt.all.length; ++i)
				this.initWidget(wt.all[i], wt);
			break;
		case 'stack':
			for(i = 0; i < wt.children.length; ++i) {
				child = wt.children[i];
				this.initWidget(child, wt);

				if (wt.view === child)
					continue;

				el = child.uiwt.element;
				el.style.visibility = 'hidden';
			}
			break;
		default:
			if (wt.children !== null)
				for(i = 0; i < wt.children.length; ++i)
					this.initWidget(wt.children[i], wt);
		}

		if (parent === null)
			if (this.firstWidget !== wt.id)
				style.display = 'none';
	};

	// isLiveWidget -------------------------------------------------------
	MagicW.prototype.isLiveWidget = function(id) {

		if (this.widgets[id] !== null)
			return true;

		return false;
	};

	// layoutReset -------------------------------------------------------
	MagicW.prototype.layoutReset = function() {

		this.actionEventResized();
	};

	// onload -------------------------------------------------------------
	MagicW.prototype.onload = function() {
		var bg, body, root, sb, wdata, wt;
		
		Log.setLevels('info');

		body = document.body;

		if (this.whtml !== '')
			body.innerHTML += this.whtml;

		/*
		 * add mw-waitingOverlay:
		 */
		body.innerHTML += '\n<div id="mw-waitingOverlay" class="mw-waitingOverlay"></div>';

		for(var id in this.widgets) {
			wt = this.widgets[id];
			
			if (wt.type === 'live')
				continue;

			this.initWidget(wt, null);
			root = new XRoot(wt); 

			this.roots[wt.id] = root;
			this.widgets[id] = null;		// ---> not a live widget
		}

		this.whtml = null;

		if (this.firstWidget !== null) {
			if ((root = this.roots[this.firstWidget]) !== undefined)
				this.xroot = root;
			else {
				if ((wdata = this.widgets[this.firstWidget]) === undefined)
					throw new Error('widget not found: ' + this.firstWidget);

				wt = XLib.createWidget(new XWData(wdata.data));
				this.firstWidget = wt.id;

				sb = [];
				HtmlWriter.createWidget(sb, wt);
				body.innerHTML += sb.join('');

				this.initWidget(wt, null);
				this.xroot = new XRoot(wt);
			}
		}

		if ((bg = this.xappl.getBackground()) !== null)
			document.body.style.backgroundColor = bg;
		
		window.addEventListener('resize', function() {
			MW.actionEventResized();
		});

		if (this.xroot !== null)
			this.actionEventResized();

		this.xappl.runTask('#start', null);
	};

	// resetLayout -------------------------------------------------------
	MagicW.prototype.resetLayout = function(id) {
		var xroot;
		
		if ((xroot = this.roots[id]) === undefined)
			throw new Error('card should be a live widget...');

		this.resetWidget(this.getCard(id));
	};

	// resetWidget -------------------------------------------------------
	MagicW.prototype.resetWidget = function(wt) {
		var i, children;

		wt.uiwt.layoutReset();

		if ((children = wt.children) === null)
			return;
		
		for(i = 0; i < children.length; ++i)
			this.resetWidget(children[i]);
	};

	// setFirstWidget ----------------II-----------------------------------
	MagicW.prototype.setFirstWidget = function(id) {

		this.firstWidget = id;
	};

	// setWidgetForms -----------------------------------------------------
	MagicW.prototype.setWidgetForms = function(data) {

		this.uiappl.setWidgetForms(data);
	};

	// setFonts -----------------------------------------------------------
	MagicW.prototype.setFonts = function(data) {

		XPluginLib.setFonts(data);
	};

	// setForms -----------------------------------------------------------
	MagicW.prototype.setForms = function(data) {

		this.uiappl.setXForms(data);
	};

	MagicW.prototype.setHtmlPlugins = function(data) {

		XPluginLib.setHtml(data);
	};

	// setOfflineVar -----------------------------------------------------
	MagicW.prototype.setOfflineVar = function(data) {

		this.xappl.putData(data.xref, new XOfflineData(data));
	};

	// setRoot ------------------------------------------------------------
	MagicW.prototype.setRoot = function(id) {

		this.xroot = this.roots[id];

		if (this.xroot.updateUI) {
			this.doUpdateUI(this.xroot.widget);
			this.xroot.doUpdateUI();
		}

		this.actionEventResized();
	};

	// setTasks -----------------------------------------------------------
	MagicW.prototype.setTasks = function(data) {

		this.xappl.init(new XData(data));

		this.uiappl.setUrl(this.xappl.getUrl());

		this.HttpFiles.initCanvasImages();
	};

	// setTemplates ------------------------------------------------------
	MagicW.prototype.setTemplates = function(data) {
		var i, child, children, ref, xdata;

		this.templates = {};

		xdata = new XData(data);
		children = xdata.getAllChildren();

		for(i = 0; i < children.length; ++i) {
			child = children[i];
			ref = child.getString('ref');
			this.templates[ref] = child;
		}

		this.uiappl.setTemplates(this.templates);
	};

	// setVar ------------------------------------------------------------
	MagicW.prototype.setVar = function(data) {
		var id;

		switch(data.tag) {
		case 'magicw':
			id = data.xref;
			break;
		case 'popup':
			id = data.tag;
			break;
		default:
			id = data._children[0].tag;
		}

		this.xappl.putData(id, new XData(data));
	};

	// setWidgets --------------------------------------------------------
	MagicW.prototype.setWidgets = function(list) {
		var i, sb, wt;

		sb = [];
		this.widgets = [];

		for(i = 0; i < list.length; ++i) {
			if (list[i].live === true) {
				this.widgets[list[i].id] = { type:'live',data:list[i] };
				continue;
			}

			wt = XLib.createWidget(new XWData(list[i]));
			this.widgets[wt.id] = wt;

			HtmlWriter.createWidget(sb, wt);
		}

		this.whtml = sb.join('');
	};

	// templateUpdate ----------------------------------------------------
	MagicW.prototype.templateUpdate = function(id, ref, templateId, visible) {
		var el, html, parent, root, sb, wt;

		wt = MW.getTemplateWidget(id, templateId);

		if ((root = this.roots[ref]) === undefined)
			throw new Error('card not found: ' + ref);

		switch(wt.type) {
		case 'tree':
			if ((parent = XLib.search(root.widget, id)) === null)
				throw new Error('widget not found: ' + id);

			sb = [];
			HtmlWriter.createWidget(sb, wt.view);
			html = sb.join('');

			el = parent.uiwt.element;
			el.innerHTML = html;
			this.initWidget(wt.view, parent);
			
			parent.replace(wt);
			parent.layoutUpdateUI(visible);
			return;
		}

		if ((parent = XLib.searchParent(root.widget, id)) === null)
			throw new Error('widget not found: ' + id);

		switch(parent.type) {
		case 'scroll':
			sb = [];
			HtmlWriter.createWidget(sb, wt);
			html = sb.join('');

			el = parent.uiwt.element;
			el.innerHTML = html;
			this.initWidget(wt, parent);
			
			parent.setView(wt);
			parent.layoutUpdateUI(visible);
			return;
		}

		throw new Error('not yet implemented: ' + parent.type);
	};

	// widgetConfigModify ------------------------------------------------
	MagicW.prototype.widgetConfigModify = function(id, ref, value) {
		var s, n1, n2, n3, item, wt;

		if ((item = this.widgets[id]) === undefined)
			throw new Error('widget not found: ' + id);

		if ((wt = DataToJs.search(item.data, ref)) === null)
			throw new Error('widget not found: ' + ref);

		switch(wt.tag) {
		case 'label':
			s = wt.text;
			n1 = s.indexOf('>');
			n2 = s.indexOf('>', n1 + 1);
			n3 = s.indexOf('<', ++n2);
			wt.text = s.substring(0, n2) + value + s.substring(n3);
			break;
		default:
			throw new Error('not supported: ' + wt.tag);
		}
	};

	// constructor
	{
		var n, id, qs;

		this.widgets = null;
		this.whtml = null;
		this.roots = [];
		this.xroot = null;
		this.templates = null;
		this.firstWidget = null;
		this.templateWidgetId = 0;

		this.WLib = new WLib(this.roots);

		this.wlib = new WidgetsLib(this.roots);
		this.uiappl = new UIApplication(this.roots, this.wlib);
		this.xappl = new XApplication(this.uiappl);
		this.uiappl.setXApplication(this.xappl);

		this.HttpFiles = new HttpFiles(this.xappl);

		// Browser detection:
		// Firefox 1.0+
		this.isFirefox = typeof InstallTrigger !== 'undefined';
		// Chrome 1 - 71
		this.isChrome = !!window.chrome && (!!window.chrome.webstore || !!window.chrome.runtime);

		console.log("firefox     : " + this.isFirefox);
		console.log("chrome      : " + this.isChrome);

		if (window.location.search !== '') {
			qs = window.location.search.substring(1);

			if ((n = qs.indexOf('&')) !== -1)
				id = qs.substring(0, n);
			else
				id = qs;

			this.xappl.putVar('DUDAL-LECTUREID', id);
			console.log("lectureId   : " + this.xappl.getVar('DUDAL-LECTUREID'));
		}
	}

	/* ================================================================	*
	 * JsTocEdit.js 													*
	 * ================================================================	*/

	var JsTocEdit = function(wt) {

		// cardUpdated -------------------------------------------------------
		JsTocEdit.prototype.cardUpdated = function() {

			MW.actionEventResized();
		};

		// dragover ----------------------------------------------------------
		JsTocEdit.prototype.dragover = function(evt) {
			var idx, info;

			this.idrag = -2;
			
			if ((info = wt.getPositionInfo(evt.offsetX, evt.offsetY)) === null)
				idx = -1;
			else
			if (!info.inside)
				idx = -1;
			else
				idx = info.idx;

			if (idx === this.iselected)
				return;

			this.iselected = idx;
			this.repaint();
		};

		// drawBorder --------------------------------------------------------
		JsTocEdit.prototype.drawBorder = function(ctx, x, y) {
			var i, w, h, border;

			w = this.wt.cellw;
			h = this.wt.cellh;
			border = this.wt.iborder;

			ctx.beginPath();

			for(i = 0; i < border; ++i) {
				ctx.rect(x++, y++, w, h);

				w -= 2;
				h -= 2;
			}

			ctx.stroke();
		};

		// drawImage ---------------------------------------------------------
		JsTocEdit.prototype.drawImage = function(ctx, img, x, y) {
			var r, v, w, h, border, imgw, imgh;

	 		border = this.wt.iborder;

			x += border;
			y += border;
			w = this.wt.cellw - 2 * border;
			h = this.wt.cellh - 2 * border;

			ctx.fillRect(x, y, w, h);

			if (img === null)
				return;

			imgw = img.width;
			imgh = img.height;

			r = imgw / imgh;

			if ((v = Math.round(w / r)) <= h) {
				y += (h - v) / 2;
				h = v;
			} else {
				v = Math.round(h * r);
				x += (w - v) / 2;
				w = v;
			}

			ctx.drawImage(img, x, y, w, h);
		};
		
		// drop --------------------------------------------------------------
		JsTocEdit.prototype.drop = function(evt) {
			var i, l, n, appl, files, list, name;

			files = evt.dataTransfer.files;

			list = [];

			for(i = 0; i < files.length; ++i) {
				name = files[i].name;
				if ((n = name.lastIndexOf('.')) === -1)
					continue;

				switch(name.substring(n).toLowerCase()) {
				case '.gif':
				case '.jpg':
				case '.jpeg':
				case '.png':
					list.push(files[i]);
					break;
				}
			}

			if (list.length === 0)
				return false;

			appl = this.wt.appl;

			if (this.iselected === -1)
				appl.putVar('lectureEditOffset', 'end');
			else
				appl.putVar('lectureEditOffset', this.iselected.toString());

			l = [];
			l.push(new XAttr('requestId', 'lecture-edit-add'));
			l.push(new XAttr('lectureId', this.wt.lectureId));
			l.push(new XAttr('nextTask', 'lecture-edit-add-download'));

			MW.uiappl.uploadFilesTask.init(new XAttrs(l));
			MW.uiappl.uploadFilesTask.exec(files);
			
			this.idrag = -1;
			this.iselected = -1;
		};

		// getSelected -------------------------------------------------------
		JsTocEdit.prototype.getSelected = function() {

			return this.selectedValue;
		};
		
		// init --------------------------------------------------------------
		JsTocEdit.prototype.init = function(files) {
			var i;

			this.files = files;

			this.images = [];
			for(i = 0; i < files.length; ++i)
				this.images[i] = MW.HttpFiles.getImage(files[i].path);
		};

		// mouseDown ---------------------------------------------------------
		JsTocEdit.prototype.mouseDown = function(evt) {
			var x, y, x0, y0, appl, pos;

			if (this.iselected === -1)
				return;

			appl = this.wt.appl;

			x = evt.offsetX;
			y = evt.offsetY;

			pos = wt.getImagePosition(this.iselected);

			x0 = pos.x + 10;
			y0 = pos.y + 10;
			if (x0 <= x && x < x0 + 22 && y0 <= y && y < y0 + 22) {
				this.selectedValue = this.iselected;
				appl.putVar('action-data', this.selectedValue.toString());
				appl.taskExecTask('lecture-edit-add');
				return;
			}

			x0 = pos.x + wt.cellw - 30;
			y0 = pos.y + 5;
			if (x0 <= x && x < x0 + 22 && y0 <= y && y < y0 + 32) {
				this.selectedValue = this.iselected;
				appl.putVar('lecture-edit-img', 'cache:' + this.files[this.iselected].path);
				appl.taskExecTask('lecture-edit-remove');
				return;
			}

			this.idrag = this.iselected;
			this.lastX = x;
			this.lastY = y;
		};

		// mouseDrag ---------------------------------------------------------
		JsTocEdit.prototype.mouseDrag = function(evt) {
			var x, y, info;

			if (this.idrag === -1)
				return;

			x = evt.offsetX;
			y = evt.offsetY;

			this.dx = x - this.lastX;
			this.dy = y - this.lastY;

			if ((info = wt.getPositionInfo(x, y)) !== null) {
				this.iselected = info.idx;
				this.inside = info.inside;
			} else
				this.inside = false;

			this.repaint();
		};

		// mouseMove ---------------------------------------------------------
		JsTocEdit.prototype.mouseMove = function(evt) {
			var idx, info;

			this.inside = false;

			if ((info = this.wt.getPositionInfo(evt.offsetX, evt.offsetY)) === null)
				idx = -1;
			else
			if (!info.inside)
				idx = -1;
			else
				idx= info.idx;

			if (this.inside !== -1)
				this.inside = true;

			if (this.iselected === idx)
				return;

			this.iselected = idx;

			this.repaint();
		};

		// mouseOut ----------------------------------------------------------
		JsTocEdit.prototype.mouseOut = function() {

			this.idrag = -1;
			this.iselected = -1;
			this.inside = false;
			this.repaint();
		};

		// mouseUp -----------------------------------------------------------
		JsTocEdit.prototype.mouseUp = function(evt) {

			if (this.idrag === -1)
				return;

			if (this.iselected !== this.idrag) {
				if (this.iselected === this.images.length)
					--this.iselected;

				this.moveImage(this.idrag, this.iselected);
				this.wt.moveImage(this.idrag, this.iselected);
			}

			this.idrag = -1;
			this.repaint();
		};

		// moveImage ---------------------------------------------------------
		JsTocEdit.prototype.moveImage = function(isrc, idst) {
			var i, images, tmp;

			images = this.images;
			tmp = images[isrc];

			if (idst < isrc) {
				for(i = isrc; i > idst; --i)
					images[i] = images[i - 1];
			} else {
				for(i = isrc; i < idst; ++i)
					images[i] = images[i + 1];
			}

			images[idst] = tmp;
		};
		
		// removeImage -------------------------------------------------------
		JsTocEdit.prototype.removeImage = function(files, idx) {
			var i, n, images, tmp;

			this.files = files;

			tmp = [];
			images = this.images;

			for(i = 0; i < idx; ++i)
				tmp[i] = images[i];

			n = images.length - 1;
			for(i = idx; i < n; ++i)
				tmp[i] = images[i + 1];

			this.images = tmp;
			
			this.iselected = -1;
			this.inside = false;
		};

		// repaint -----------------------------------------------------------
		JsTocEdit.prototype.repaint = function(w, h) {
			var i, p, x, y, ctx, images, pos, wt;

			if (arguments.length === 0) {
				ctx = this.canvas.getContext('2d');

				ctx.fillStyle = this.canvas.style.backgroundColor;
				ctx.fillRect(0, 0, ctx.canvas.width, ctx.canvas.height);
			} else {
				ctx = this.canvas.getContext('2d');
				ctx.canvas.width = w;
				ctx.canvas.height = h;
			}
			
			ctx.lineWidth = 1;
			ctx.fillStyle = '#000000';
			ctx.strokeStyle = '#ff0000';
			
			wt = this.wt;
			images = this.images;

			p = null;

			for(i = 0; i < images.length; i++) {
				pos = wt.getImagePosition(i);

				if (i === this.idrag) {
					p = pos;
					continue;
				}

				x = pos.x;
				y = pos.y;

				this.drawImage(ctx, images[i], x, y);

				if (i !== this.iselected)
					continue;

				if (this.idrag === -1 && !this.inside)
					continue;

				this.drawBorder(ctx, x, y);
			
				if (this.idrag === -1) {
					ctx.drawImage(this.add, x + 10, y + 10);
					ctx.drawImage(this.trash, x + this.wt.cellw - 35, y + 5);
				}
			}

			if (p === null)
				return;

			x = p.x + this.dx;
			y = p.y + this.dy;

			this.drawImage(ctx, images[this.idrag], x, y);
		};

		// Constructor -------------------------------------------------------
		{
			var me, parent;

			this.wt = wt;
			this.canvas = wt.uiwt.element;

			this.trash = MW.HttpFiles.getImage('MW.trash-32.png');

			this.iselected = -1;
			this.inside = false;
			this.idrag = -1;

			this.add = MW.HttpFiles.getImage('MW.add-22.png');

			// events...
			this.canvas.onmousemove = function(evt) {

				try {
					if (evt.buttons === 1)
						wt.plugin.mouseDrag(evt);
					else
						wt.plugin.mouseMove(evt);
				}
				catch(err) {
					HttpError.send(err);
					MW.xappl.runTask('#http-jserror');
				}
			};

			this.canvas.onmousedown = function(evt) {

				try {
					wt.plugin.mouseDown(evt);
				}
				catch(err) {
					HttpError.send(err);
					MW.xappl.runTask('#http-jserror');
				}
			};

			this.canvas.onmouseup = function(evt) {

				try {
					wt.plugin.mouseUp(evt);
				}
				catch(err) {
					HttpError.send(err);
					MW.xappl.runTask('#http-jserror');
				}
			};

			this.canvas.onmouseout = function(evt) {

				try {
					wt.plugin.mouseOut();
				}
				catch(err) {
					HttpError.send(err);
					MW.xappl.runTask('#http-jserror');
				}
			};

			me = this;
			parent = this.canvas.parentNode;

			this.canvas.addEventListener('dragover', function(evt) {

				evt.preventDefault();
				wt.plugin.dragover(evt);
			}, false);

			this.canvas.addEventListener('drop', function(evt) {

				evt.preventDefault();
				evt.stopPropagation();
				me.drop(evt);
			}, false);

			parent.addEventListener('dragover', function(evt) {

				evt.preventDefault();
			}, false);

			parent.addEventListener('drop', function(evt) {

				evt.preventDefault();
				me.drop(evt);
			}, false);
		}
	};

	/* ================================================================	*
	 * XDisplayPlugin.js 												*
	 * ================================================================	*/

	var XDisplayPlugin = function(wt) {

		// buttonSetLayout ---------------------------------------------------
		XDisplayPlugin.prototype.buttonSetLayout = function() {
			var ctx;

			ctx = this.canvas.getContext('2d');
			ctx.font = "bold 13px sans-serif";

			this.label1w = Math.ceil(ctx.measureText(this.wt.actionLabel1).width);
			this.label2w = Math.ceil(ctx.measureText(this.wt.actionLabel2).width);

			if (this.label1w > this.label2w)
				this.buttonw = this.label1w + 40;
			else
				this.buttonw = this.label2w + 40;

			this.buttony = 15;
			this.buttonh = 30;
		};

		// isOnButton --------------------------------------------------------
		XDisplayPlugin.prototype.isOnButton = function(x, y) {

			if (x < this.canvas.width - this.buttonw)
				return false;

			return y < this.buttonh;
		};

		// pointerPosition ---------------------------------------------------
		XDisplayPlugin.prototype.pointerPosition = function() {
			var x, y, v, px, py, vp, value;

			vp = this.viewport;

			if ((x = this.cursorX - vp.x) < 0)
				v = 'I';
			else
			if ((y = this.cursorY - vp.y) < 0)
				v = 'I';
			else
			if (x >= vp.w || y >= vp.h)
				v = 'I';
			else {
				px = Math.ceil(x * this.img.width / vp.w);
				py = Math.ceil(y * this.img.height / vp.h);

				if (px === this.lastX && py === this.lastY)
					v = 'I';
				else {
					v = px + ',' + py;
					this.lastX = px;
					this.lastY = py;
				}
			}

			if (this.entries.length === 0)
				this.entries.push(v);
			else
				this.entries.push(' ' + v);

			if (this.entries.length < 5)
				return;

			value = this.entries.join('');
			if (value !== 'I I I I I') {
				++this.counter;
				this.wt.sendPointerPositions(value);
			}

			this.entries = [];
		};

		// repaint -----------------------------------------------------------
		XDisplayPlugin.prototype.repaint = function(width, height) {
			var x, r, ctx, img, text, textw;

			ctx = this.canvas.getContext('2d');

			if (arguments.length === 2) {
				ctx.canvas.width = width;
				ctx.canvas.height = height;
			} else {
				width = ctx.canvas.width;
				height = ctx.canvas.height;
			}

			ctx.fillStyle = this.canvas.style.backgroundColor;
			ctx.fillRect(0, 0, width, height);

			if ((img = this.img) === null)
				return;

			r = XUILib.getImagePosition(width, height, img.width, img.height);
			ctx.drawImage(this.img, r.x, r.y, r.w, r.h);

			if (this.buttonStatus === 'hidden')
				return;

			this.viewport = r;

			switch(this.buttonStatus) {
			case 'annotate':
				text = this.wt.actionLabel1;
				textw = this.label1w;
				break;
			case 'dragging':
				text = this.wt.actionLabel2;
				textw = this.label2w;
				break;
			default:
				throw new Error("not yet implemented: " + this.buttonStatus);
			}

			x = width - this.buttonw;

			if (this.buttonInside) {
				if (this.buttonDown)
					ctx.fillStyle = '#484848';
				else
					ctx.fillStyle = 'dimgray';
			} else
				ctx.fillStyle = '#282828';

			ctx.fillRect(x, 0, this.buttonw, this.buttonh);

			x += (this.buttonw - textw) / 2;

			ctx.font = "bold 13px sans-serif";
			ctx.textBaseline = 'middle';
			ctx.fillStyle = 'white';

			ctx.fillText(text, x, this.buttony);
		};

		// setImage ----------------------------------------------------------
		XDisplayPlugin.prototype.setImage = function(path) {

			this.img = MW.HttpFiles.getImage(path);
		};

		// start -------------------------------------------------------------
		XDisplayPlugin.prototype.start = function() {

			this.buttonStatus = 'hidden';
			this.buttonInside = false;
			this.buttonDown = false;

			this.canvas.style.cursor = 'default';
		};

		// startPointerDragging ----------------------------------------------
		XDisplayPlugin.prototype.startPointerDragging = function() {

			this.isDragging = true;

			if (this.timer !== null)
				clearInterval(this.timer);

			this.entries = [];
			this.counter = 0;
			this.pointerPosition();

			let me = this;
			this.timer = setInterval(function() {

				me.pointerPosition();
			}, 200);
		};

		// stopPointerDragging -----------------------------------------------
		XDisplayPlugin.prototype.stopPointerDragging = function(repaint) {

			this.isDragging = false;

			if (this.timer !== null)
				clearInterval(this.timer);

			this.canvas.style.cursor = 'default';

			if (repaint) {
				this.buttonStatus = 'annotate';
				this.repaint();
			}

			if (this.counter > 0)
				this.wt.sendPointerPositions('END');

			this.timer = null;
			this.entries = null;
			this.counter = -1;
		};

		// startRecording ----------------------------------------------------
		XDisplayPlugin.prototype.startRecording = function() {

			this.buttonStatus = 'annotate';
			this.buttonInside = false;
			this.buttonDown = false;

			this.repaint();
		};

		// stopRecording -----------------------------------------------------
		XDisplayPlugin.prototype.stopRecording = function() {

			this.stopPointerDragging(false);

			this.buttonStatus = 'hidden';
			this.repaint();
		};

		// Constructor -------------------------------------------------------
		{
			this.wt = wt;
			this.canvas = wt.uiwt.element;

			this.img = null;

			this.buttonStatus = 'hidden';
			this.buttonInside = false;
			this.buttonDown = false;

			this.isDragging = false;
			this.timer = null;
			this.eventStart = -1;
			this.cursorX = this.cursorY = -1;
			this.viewport = null;
			this.lastX = this.lastY = -1;
			this.entries = null;
			this.counter = -1;

			this.buttonSetLayout(wt);

			/*
			 * events...
			 */
			let me = this;

			/*
			 * onmousedown:
			 */
			this.canvas.onmousedown = function(evt) {

				if (me.buttonStatus === 'hidden')
					return;

				if (!(me.buttonInside = me.isOnButton(evt.offsetX, evt.offsetY)))
					return;

				me.buttonDown = true;
				me.repaint();
			};

			/*
			 * onmousemove:
			 */
			this.canvas.onmousemove = function(evt) {
				var last;

				if (me.buttonStatus === 'hidden')
					return;

				if (me.buttonStatus === 'dragging') {
					me.cursorX = evt.offsetX;
					me.cursorY = evt.offsetY;
				}

				last = me.buttonInside;
				me.buttonInside = me.isOnButton(evt.offsetX, evt.offsetY);

				if (last === me.buttonInside)
					return;

				me.buttonDown = false;
				me.repaint();
			};

			/*
			 * onmouseout:
			 */
			this.canvas.onmouseout = function(evt) {

				if (me.buttonStatus === 'hidden')
					return;
				
				if (me.buttonInside) {
					me.buttonDown = false;
					me.buttonInside = false;
				}

				if (!me.isDragging) {
					me.repaint();
					return;
				}

				me.stopPointerDragging(true);
			};

			/*
			 * onmouseup:
			 */
			this.canvas.onmouseup = function(evt) {

				if (me.buttonStatus === 'hidden')
					return;

				me.buttonDown = false;

				if (!me.buttonInside)
					return;

				if (me.buttonStatus === 'annotate') {
					me.buttonStatus = 'dragging';
					me.canvas.style.cursor = 'url("images/drag-32.png"), default';

					me.cursorX = evt.offsetX;
					me.cursorY = evt.offsetY;

					me.startPointerDragging();
				} else {
					me.buttonStatus = 'annotate';
					me.canvas.style.cursor = 'default';

					me.stopPointerDragging(false);
				}

				me.repaint();
			};
		}
	};

	/* ================================================================	*
	 * XSlidePlugin.js 													*
	 * ================================================================	*/

	var XSlidePlugin = function(wt) {

		// checkTimebarVisibility --------------------------------------------
		XSlidePlugin.prototype.checkTimebarVisibility = function() {
			var l, t;

			t = new Date().getTime() / 1000;
			if ((t - me.lastEvent) < me.maxVisible)
				return;

			l = me.xslide.layout;
			l.setTimebarVisible(false);

			clearInterval(me.timer);
			me.timer = null;
			
			me.repaint();
		};

		// eventAction -------------------------------------------------------
		XSlidePlugin.prototype.eventAction = function(name, evt) {
			var l;

			l = this.xslide.layout;

			if (!l.hasVideoRecording())
				return;

			switch(name) {
			case 'check':
				break;
			case 'mousedown':
				l.setMouseDown();
				break;
			case 'mouseup':
				if (l.isButtonOver()) {
					if (l.isVideoPlaying())
						me.usession.eventAction('receiver-pause');
					else
						me.usession.eventAction('receiver-play');
				} else
				if (l.isCursorDragging())
					me.usession.eventAction('video-seek');

				l.setMouseUp();
				break;
			case 'mousemove':
				l.setMouseEvents(evt.offsetX, evt.offsetY, evt.buttons === 1);
				break;
			default:
				throw new Error('not yet implemented: ' + name);
			}

			this.lastEvent = new Date().getTime() / 1000;

			if (!l.isTimebarVisible()) {
				l.setTimebarVisible(true);
				this.timer = setInterval(this.checkTimebarVisibility, 1000);
			}

			this.repaint();
		};

		// eventVideoEnded ---------------------------------------------------
		XSlidePlugin.prototype.eventVideoEnded = function() {

			this.pointer.visible = false;
			this.repaint();
		};

		// repaint -----------------------------------------------------------
		XSlidePlugin.prototype.repaint = function(width, height) {
			var l, p, x, y, w, r, ctx, dragging, img, path, timebar, xcursor;
			var t1, t2, xtime;

			l = this.xslide.layout;

			ctx = this.canvas.getContext('2d');

			if (arguments.length === 2) {
				ctx.canvas.width = width;
				ctx.canvas.height = height;
			}

			ctx.fillStyle = this.canvas.style.backgroundColor;
			ctx.fillRect(0, 0, l.canvasw, l.canvash);

			if (this.img !== null) {
				r = l.getSlidePosition();
				ctx.drawImage(this.img, r.x, r.y, r.w, r.h);

				if (this.pointer.visible) {
					x = this.pointer.x * r.w / this.img.width;
					y = this.pointer.y * r.h / this.img.height;
					x += r.x;
					y += r.y;
					ctx.drawImage(this.pointerimg, x, y);
				}
			}

			if (!l.isTimebarVisible())
				return;

			l.update();

			ctx.font = this.font;
			ctx.textBaseline = 'top';
			
			/*
			 * frame...
	 		 */
			ctx.fillStyle = 'black';
			r = l.getFramePosition();
			ctx.fillRect(r.x, r.y, r.w, r.h);

			ctx.strokeStyle = 'gray';
			UILibN.drawBorder(ctx, r.x - 1, r.y - 1, r.w + 2, r.h + 2, 1);

			/*
			 * button...
			 */
			if (l.isButtonVisible()) {
				if (l.isButtonOver())
					if (l.isButtonPressed())
						ctx.fillStyle = '#004040';
					else
						ctx.fillStyle = 'teal';
				else
					ctx.fillStyle = '#006060';

				r = l.getButtonPosition();
				ctx.fillRect(r.x, r.y, r.w, r.h);

				p = l.getButtonImagePosition();

				if (l.isVideoPlaying())
					ctx.drawImage(this.pauseimg, p.x, p.y);
				else
					ctx.drawImage(this.playimg, p.x, p.y);
			}

			/*
			 * timebar...
			 */
			timebar = l.getTimebar();
			ctx.fillStyle = 'red';
			ctx.fillRect(timebar.x, timebar.y, timebar.w, timebar.h);

			/*
			 * barline...
			 */
			if ((r = l.getBarLine()) !== null) {
				ctx.strokeStyle = 'gray';
				ctx.beginPath();
				ctx.moveTo(r.x, r.y);
				ctx.lineTo(r.x + r.w, r.y);
				ctx.stroke();
			}

			/*
			 * events...
			 */
			ctx.fillStyle = 'orange';

			for(var d of l.getEventsPositions()) {
				if ((d -= 2) < 0)
					d = 0;

				ctx.fillRect(timebar.x + d, timebar.y, 4, timebar.h);
			}

			/*
			 * cursor... 
			 */
			ctx.fillStyle = 'white';

			if ((dragging = l.isCursorDragging()))
				xcursor = l.getCursorDraggingPosition();
			else
				xcursor = l.getCursorX();

			y = l.getCursorY();
			w = l.getCursorWidth() / 2;
			ctx.beginPath();
			ctx.arc(xcursor, y, w, 0, 2 * Math.PI);
			ctx.fill();

			/*
			 * total time...
			 */
			t1 = this.xslide.getDuration();
			w = ctx.measureText(t1).width;
			if ((xtime = timebar.x + timebar.w - w) < timebar.x)
				xtime = timebar.x;
			
			if (!dragging) {
				/*
				 * current time...
				 */
				t2 = this.xslide.getCurrentTime();
				w = ctx.measureText(t2).width;
				x = xcursor - w / 2;

				if (x + w + 10 > xtime) {
					x = xtime - w - 10;

					if (x < timebar.x) {
						xtime += timebar.x - x;
						x = timebar.x;
					}
				}

				ctx.fillText(t2, x, timebar.y + 14);
			}

			ctx.fillText(t1, xtime, timebar.y + 14);

			/*
			 * thumbnail...
			 */
			if (!l.isThumbnailVisible())
				return;

			if ((path = l.getThumbnailPath()) === null)
				return;

			if ((img = MW.HttpFiles.getImage(path)) === null)
				return;

			r = l.getThumbnailPosition();

			ctx.drawImage(img, r.x, r.y, r.w, r.h);
			ctx.strokeStyle = 'gray';
			UILibN.drawBorder(ctx, r.x, r.y, r.w, r.h, 1);

			t = l.getThumbnailTime();
			w = ctx.measureText(t).width;

			x = r.x + r.w - w - 15;
			y = r.y + r.h - 25;

			ctx.fillStyle = 'black';
			ctx.fillRect(x, y, w + 10, 19);
			ctx.fillStyle = 'white';
			ctx.fillText(t, x + 5, y + 3);
		};

		// setImage ----------------------------------------------------------
		XSlidePlugin.prototype.setImage = function(path) {

			this.img = MW.HttpFiles.getImage(path);
		};

		// setRoute ----------------------------------------------------------
		XSlidePlugin.prototype.setRoute = function(route) {
			var obj;

			if (route.length === 0) {
				this.pointer.visible = false;
				this.pointer.end = true;
				this.repaint();
				return;
			}

			this.pointer.route = route;
			this.pointer.end = false;
			this.pointer.idx = 0;

			this.setRoutePointer();

			if (route.length === 1)
				return;

			obj = {
				timer: null
			};

			obj.timer = setInterval(function(plugin, obj) {

				plugin.setRoutePointer(obj.timer);
			}, 100, this, obj);
		};

		// setRoutePointer ---------------------------------------------------
		XSlidePlugin.prototype.setRoutePointer = function(timer) {
			var p, idx, pointer, route;

			pointer = this.pointer;
			route = pointer.route;
			idx = pointer.idx;

			if (pointer.end || ++pointer.idx >= route.length) {
				clearInterval(timer);
				return;
			}

			p = route[idx];

			if (isNaN(p.x) || isNaN(p.y))
				return;
			
			if (p.x === -1)
				return;

			pointer.x = p.x;
			pointer.y = p.y;

			pointer.visible = true;
			this.repaint();
		};
		
		// Constructor -------------------------------------------------------
		{
			var me;

			this.xslide = wt;
			this.canvas = wt.uiwt.element;
			this.usession = wt.appl.uiappl.rplugin.usession;

			this.img = null;
			this.timer = null;
			this.lastEvent = 0;
			this.maxVisible = 10;

			this.pointer = {
				visible: false,
				end: false,
				route: null,
				idx: 0,
				x: 0,
				y: 0
			};

			this.pauseimg = new Image();
			this.pauseimg.src = 'images/pause-24.png';
			this.playimg = new Image();
			this.playimg.src = 'images/play-24.png'; 
			this.pointerimg = new Image();
			this.pointerimg.src = 'images/drag-32.png';

			wt.layout.update();

			/*
			 * events...
			 */
			me = this;
			
			/*
			 * onmouseenter...
			 */
			this.canvas.onmouseenter = function(evt) {

				try {
					me.eventAction('check', null);
				}
				catch(err) {
					HttpError.send(err);
					MW.xappl.runTask('#http-jserror');
				}
			};

			/*
			 * onmousedown...
			 */
			this.canvas.onmousedown = function(evt) {

				try {
					me.eventAction('mousedown', evt);
				}
				catch(err) {
					HttpError.send(err);
					MW.xappl.runTask('#http-jserror');
				}
			};

			/*
			 * onmouseup... 
			 */
			this.canvas.onmouseup = function(evt) {

				try {
					me.eventAction('mouseup', null);
				}
				catch(err) {
					HttpError.send(err);
					MW.xappl.runTask('#http-jserror');
				}
			};

			/*
			 * onmousemove...
			 */
			this.canvas.onmousemove = function(evt) {

				try {
					me.eventAction('mousemove', evt);
				}
				catch(err) {
					HttpError.send(err);
					MW.xappl.runTask('#http-jserror');
				}
			};

			/*
			 * video: onmousemove...
			 */
			video = DocLib.getElement('receiver-video');
			video.onmouseenter = this.canvas.onmouseenter;
			video.onmouseleave = this.canvas.onmouseleave;

			/*
			 * onmousemove...
			 */
			video.onmousemove = function(evt) {

				try {
					me.eventAction('check', null);
				}
				catch(err) {
					HttpError.send(err);
					MW.xappl.runTask('#http-jserror');
				}
			};
		}
	};
};

var MW = new MagicW();
